__afterburn_cpu = 0x1000
CPU_CR_OFFSET = 0
CPU_KR_OFFSET = 1024
CPU_RR_OFFSET = 1088
CPU_BANKED_OFFSET = 1152
CPU_PSR_OFFSET = 1408
CPU_IRQ_PENDING_OFFSET = 1416
CPU_IN_SERVICE_IRQ_OFFSET = 1420
CPU_CURRENT_BANK_OFFSET = 1424
CPU_CFLE_OFFSET = 1428

/*
.macro __afterburn_call func
	mov r7=ar.pfs
	brl.call.sptk.many b2=\func
	;;
	mov ar.pfs=r7
.endm
*/

.macro __afterburn_call func
	movl r7=\func
	;;
	mov b2=r7
	mov r7=ar.pfs
	;;
	br.call.sptk.many b2=b2
	;;
	mov ar.pfs=r7
.endm

.macro __afterburn_break_0_0 num
	break \num
	add r4=0x10,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endm

.macro __afterburn_break_x_0 num
	break \num
	add r4=0x20,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endm

.macro __afterburn_break_0_x num
	break \num
	add r7=0x10,r4
	;;
	mov b2=r7
	br.sptk.many b2
	;;
.endm

.macro __afterburn_break_x_x num
	break \num
	add r7=0x20,r4
	;;
	mov b2=r7
	br.sptk.many b2
	;;
.endm

.proc burn_deliver_interrupt
burn_deliver_interrupt:
	/* check psr.i */
	mov r5=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r5]
	;;
	tbit.nz p63,p0=r5,14
	;;

	/* check cr.ivr > in_service_irq */
(p63)	mov r5=__afterburn_cpu+CPU_CR_OFFSET+8*65
(p63)	mov r6=__afterburn_cpu+CPU_IN_SERVICE_IRQ_OFFSET
	;;
(p63)	ld8 r5=[r5]
(p63)	ld4 r6=[r6]
	;;
(p63)	cmp.gt p63,p0=r5,r6	
	;;

	/* check (cr.ivr >> 4) > (tpr >> 4) - includes mmi bit */
(p63)	mov r6=__afterburn_cpu+CPU_CR_OFFSET+8*66
	;;
(p63)	ld8 r6=[r6]
(p63)	shr.u r5=r5,4
	;;
(p63)	shr.u r6=r6,4
	;;
(p63)	cmp.gt p63,p0=r5,r6
	;;

	/* deliver interrupt via hypervisor */
(p63)	break 0x80021
	mov pr=r7,-1
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endp burn_deliver_interrupt	
	
.macro __afterburn_check_interrupts
	mov r5=__afterburn_cpu+CPU_IRQ_PENDING_OFFSET
	;;
	ld4 r5=[r5]
	mov r7=pr
	;;
	cmp.ne p63,p0=r5,r0
(p63)	br.spnt.many burn_deliver_interrupt
	;;
	mov pr=r7,-1
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endm

.global burn_enable_interrupts
.proc burn_enable_interrupts
burn_enable_interrupts:
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r6]
	;;
	dep r5=-1,r5,14,1
	;;
	st8 [r6]=r5
	add r4=0x10,r4
	__afterburn_check_interrupts
.endp burn_enable_interrupts

.macro __afterburn_bank_switch bank
	mov r5=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r5]
	mov r7=pr
	;;
.if \bank == 1
	tbit.z p63,p0=r5,44
.else
	tbit.nz p63,p0=r5,44
.endif
	;;
(p63)	mov r5=__afterburn_cpu+CPU_BANKED_OFFSET+(1-\bank)*8*16
(p63)	mov r6=__afterburn_cpu+CPU_BANKED_OFFSET+(1-\bank)*8*16+8
	;;
(p63)	st8 [r5] = r16, 16
(p63)	st8 [r6] = r17, 16
	;;
(p63)	st8 [r5] = r18, 16
(p63)	st8 [r6] = r19, 16
	;;
(p63)	st8 [r5] = r20, 16
(p63)	st8 [r6] = r21, 16
	;;
(p63)	st8 [r5] = r22, 16
(p63)	st8 [r6] = r23, 16
	;;
(p63)	st8 [r5] = r24, 16
(p63)	st8 [r6] = r25, 16
	;;
(p63)	st8 [r5] = r26, 16
(p63)	st8 [r6] = r27, 16
	;;
(p63)	st8 [r5] = r28, 16
(p63)	st8 [r6] = r29, 16
	;;
(p63)	st8 [r5] = r30
(p63)	st8 [r6] = r31
(p63)	mov r5=__afterburn_cpu+CPU_BANKED_OFFSET+\bank*8*16
(p63)	mov r6=__afterburn_cpu+CPU_BANKED_OFFSET+\bank*8*16+8
	;;
(p63)	ld8 r16 = [r5], 16
(p63)	ld8 r17 = [r6], 16
	;;
(p63)	ld8 r18 = [r5], 16
(p63)	ld8 r19 = [r6], 16
	;;
(p63)	ld8 r20 = [r5], 16
(p63)	ld8 r21 = [r6], 16
	;;
(p63)	ld8 r22 = [r5], 16
(p63)	ld8 r23 = [r6], 16
	;;
(p63)	ld8 r24 = [r5], 16
(p63)	ld8 r25 = [r6], 16
	;;
(p63)	ld8 r26 = [r5], 16
(p63)	ld8 r27 = [r6], 16
	;;
(p63)	ld8 r28 = [r5], 16
(p63)	ld8 r29 = [r6], 16
	;;
(p63)	ld8 r30 = [r5]
(p63)	ld8 r31 = [r6]
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r6]
	mov pr=r7,-1
	;;
	dep r5=-\bank,r5,44,1
	;;
	st8 [r6]=r5
	add r4=0x10,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endm

.global burn_rfi
.proc burn_rfi
burn_rfi:
	mov r5=__afterburn_cpu+CPU_CR_OFFSET+23*8
	;;
	ld8 r5=[r5]
	mov r7=pr
	;;
	tbit.nz p63,p0=r5,63
	;;
(p63)	br.spnt.many 1f
	;;
	mov pr=r7,-1
	break 0x80020
	;;
1:
	alloc r5=ar.pfs,0,0,0,0
	mov pr=r7,-1
	break 0x80020
	;;
.endp burn_rfi

.proc burn_read_cfm
burn_read_cfm:
	mov r5=ar.pfs
	br.ret.sptk.many b2
	;;
.endp burn_read_cfm

.global burn_cover
.proc burn_cover
burn_cover:
	mov r7=ar.pfs
	br.call.sptk.few b2=burn_read_cfm
	;;
	dep.z r5=r5,0,38
	mov ar.pfs=r7
	mov r6=__afterburn_cpu+CPU_CFLE_OFFSET
	;;
	ld4 r6=[r6]
	mov r7=pr
	;;
	cmp.eq p62,p63=r6,r0
	;;
(p63)	mov r6=__afterburn_cpu+CPU_CR_OFFSET+23*8
(p63)	dep r5=-1,r5,63,1
(p62)	mov r6=__afterburn_cpu+CPU_CFLE_OFFSET
(p62)	mov r5=1
	;;
(p63)	st8 [r6]=r5
(p62)	st4 [r6]=r5
	mov pr=r7,-1
	/* beware this can't be reexecuted */
	cover
	add r4=0x10,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
	/* __afterburn_break_noarg 0x80002 */
.endp burn_cover

.global burn_bsw_0
.proc burn_bsw_0
burn_bsw_0:
	__afterburn_bank_switch 0
	/* __afterburn_break_noarg 0x80010 */
.endp burn_bsw_0

.global burn_bsw_1
.proc burn_bsw_1
burn_bsw_1:
	__afterburn_bank_switch 1
	/* __afterburn_break_noarg 0x80011 */
.endp burn_bsw_1

.global burn_read_ivr
.proc burn_read_ivr
burn_read_ivr:
	__afterburn_break_0_x 0x80012
.endp burn_read_ivr

.global burn_write_itm
.proc burn_write_itm
burn_write_itm:
	__afterburn_call vmm_write_itm
	add r4=0x20,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;

	/* __afterburn_break 0x80013 */
.endp burn_write_itm

.global burn_write_itc
.proc burn_write_itc
burn_write_itc:
	__afterburn_call vmm_write_itc
	add r4=0x20,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endp burn_write_itc

.global burn_read_pmd
.proc burn_read_pmd
burn_read_pmd:
	__afterburn_call vmm_read_pmd
	add r4=0x20,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
.endp burn_read_pmd

/* N.B. dfh check not restartable! */
.global burn_write_psr
.proc burn_write_psr
burn_write_psr:
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r7=[r6]
	shl r5=r5,32
	;;
	mix4.l r5=r7,r5
	;;
	st8 [r6]=r5
	mov r6=pr
	/* psr.dfh 0 -> 1 ? */
	tbit.nz p63,p0=r5,19
	;;
(p63)	tbit.z p63,p0=r7,19
	;;
(p63)	break 0x80008
	mov pr=r6,-1
	add r4=0x20,r4
	__afterburn_check_interrupts
.endp burn_write_psr

.global burn_set_dfh
.proc burn_set_dfh
burn_set_dfh:
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r6]
	;;
	dep r5=-1,r5,14,1
	;;
	st8 [r6]=r5
	__afterburn_call vmm_set_dfh
	add r4=0x10,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;
	/* __afterburn_break_noarg 0x80008 */
.endp burn_set_dfh

.global burn_thash
.proc burn_thash
burn_thash:
	/* r6=cr.pta */
	mov r6=__afterburn_cpu+CPU_CR_OFFSET+8*8
	;;
	ld8 r6=[r6]

	/* r7=region */
	shr.u r7=r5,61
	;;

	/* deposit region into r6 */
	dep r6=r7,r6,61,3

	/* r7=rr[r7].ps */
	shl r7=r7,3
	;;
	add r7=__afterburn_cpu+CPU_RR_OFFSET,r7
	/* clear bottom 15 bits of r6 */
	dep r6=0,r6,0,15
	;;
	ld8 r7=[r7]
	/* clear region bits in r5 */
	dep r5=0,r5,61,3
	;;
	extr.u r7=r7,2,6
	;;
	/* r5>>=ps */
	shr.u r5=r5,r7
	;;
	/* res=r6+(r5<<3) */
	shladd r5=r5,3,r6
	add r7=0x20,r4
	;;
	mov b2=r7
	br.sptk.many b2
	;;
.endp burn_thash

.global burn_tpa
.proc burn_tpa
burn_tpa:
	__afterburn_call vmm_tpa
	add r7=0x20,r4
	;;
	mov b2=r7
	br.sptk.many b2
	;;
.endp burn_tpa

.global burn_write_rr
.proc burn_write_rr
burn_write_rr:
	mov r7=__afterburn_cpu+CPU_RR_OFFSET
	shr.u r5=r5,61
	;;
	shladd r7=r5,3,r7
	;;
	st8 [r7]=r6

	// translate_rr
	mux1 r7=r6,@rev
	;;
	shr r7=r7,40
	;;
	dep r6=r7,r6,8,16

	mov r7=(12<<2)|1
	;;
	dep r6=r7,r6,0,8
	__afterburn_call vmm_write_rr
	add r4=0x20,r4
	;;
	mov b2=r4
	br.sptk.many b2
	;;

	/* __afterburn_break 0x80007 */
.endp burn_write_rr

.global burn_tlbop
.proc burn_tlbop
burn_tlbop:
	__afterburn_break_x_0 0x80006
.endp burn_tlbop
