__afterburn_cpu = 0x1000
CPU_CR_OFFSET = 0
CPU_KR_OFFSET = 1024
CPU_RR_OFFSET = 1088
CPU_BANKED_OFFSET = 1152
CPU_PSR_OFFSET = 1408
CPU_IRQ_PENDING_OFFSET = 1416
CPU_IN_SERVICE_IRQ_OFFSET = 1420
CPU_CURRENT_BANK_OFFSET = 1424
CPU_CFLE_OFFSET = 1428
CPU_R4_OFFSET = 1432

.macro __afterburn_call_0_0 pr,func
{.mib
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov r4=r0
	;;
.endm

.macro __afterburn_call_1_0 pr,func,src
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=\src
}
{.mmb
(\pr)	nop.m 0x0
(\pr)	nop.m 0x0
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov r4=r0
	;;
.endm

.macro __afterburn_call_0_1 pr,func,dst
{.mib
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov \dst=r5
(\pr)	mov r4=r0
	;;
.endm

.macro __afterburn_call_1_1 pr,func,src,dst
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=\src
}
{.mmb
(\pr)	nop.m 0x0
(\pr)	nop.m 0x0
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov \dst=r5
(\pr)	mov r4=r0
	;;
.endm

.macro __afterburn_call_2_0 pr,func,src1,src2
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=\src1
}
{.mmb
(\pr)	mov r6=\src2
(\pr)	nop.m 0x0
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov r4=r0
	;;
.endm

.macro __afterburn_call_3_0 pr,func,src1,src2,src3
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=\src1
}
{.mmb
(\pr)	mov r6=\src2
(\pr)	mov r7=\src3
(\pr)	br.sptk.many \func
	;;
}
(\pr)	mov r4=r0
	;;
.endm

/* don't need to come back! */
.macro emul_rfi
	br.sptk.many burn_rfi
	;;
.endm

.macro emul_fc pr=p0,src
(\pr)	mov msr[\src]=r0
.endm

.macro emul_bsw_0
	__afterburn_call_0_0 p0,burn_bsw_0
.endm

.macro emul_bsw_1
	__afterburn_call_0_0 p0,burn_bsw_1
.endm

.macro emul_cover
	__afterburn_call_0_0 p0,burn_cover
.endm

.macro emul_thash pr=p0,dst,src
	__afterburn_call_1_1 \pr,burn_thash,\src,\dst
.endm

.macro emul_ttag pr=p0,dst,src
	__afterburn_call_1_1 \pr,burn_ttag,\src,\dst
.endm

.macro emul_tpa pr=p0,dst,src
	__afterburn_call_1_1 \pr,burn_tpa,\src,\dst
.endm

.macro emul_write_psr pr=p0,src
	__afterburn_call_1_0 \pr,burn_write_psr,\src
.endm

.macro emul_write_rr pr=p0,rr,src
	__afterburn_call_2_0 \pr,burn_write_rr,\rr,\src
.endm


/* TLB management operations are multiplexed over burn_tlbop
   since they are all routed to the hypervisor MM subsystem */

.macro emul_itc_d pr=p0,src
	__afterburn_call_3_0 \pr,burn_tlbop,0,\src,-1
.endm

.macro emul_itr_d pr=p0,dtr,src
	__afterburn_call_3_0 \pr,burn_tlbop,0,\src,\dtr
.endm

.macro emul_itc_i pr=p0,src
	__afterburn_call_3_0 \pr,burn_tlbop,1,\src,-1
.endm

.macro emul_itr_i pr=p0,itr,src
	__afterburn_call_3_0 \pr,burn_tlbop,1,\src,\itr
.endm

.macro emul_ptc_l pr=p0,va,sz
	__afterburn_call_3_0 \pr,burn_tlbop,2,\va,\sz
.endm

.macro emul_ptc_g pr=p0,va,sz
	__afterburn_call_3_0 \pr,burn_tlbop,3,\va,\sz
.endm

.macro emul_ptc_ga pr=p0,va,sz
	__afterburn_call_3_0 \pr,burn_tlbop,3,\va,\sz
.endm

.macro emul_ptr_d pr=p0,va,sz
	__afterburn_call_3_0 \pr,burn_tlbop,4,\va,\sz
.endm

.macro emul_ptr_i pr=p0,va,sz
	__afterburn_call_3_0 \pr,burn_tlbop,5,\va,\sz
.endm

.macro emul_ptc_e pr=p0,src
	__afterburn_call_1_0 \pr,burn_tlbop,6
.endm


/* remaining ones are trivial enough to be implemented inline */

/* XXX assumes r27-r31 are scratch */
.macro emul_epc
	mov r27=r4
	mov r28=r5
	mov r29=r6
	mov r30=r7
	mov r31=b2
{
	.mii
	nop.m 0x0
	mov r4=ip
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
}
	ld8 r7=[r6]
	mov r5=__afterburn_cpu+CPU_R4_OFFSET
	;;
	dep r7=0,r7,32,2
	;;
	st8 [r6]=r7
	st8 [r5]=r27,16
	mov r6=__afterburn_cpu+CPU_R4_OFFSET+8
	;;
	st8 [r6]=r28,16
	st8 [r5]=r29,16
	;;
	st8 [r6]=r30
	st8 [r5]=r31
	mov r4=r0
.endm

.macro unepc
{.mii
	nop.m 0x0
	mov r4=ip
	mov r5=__afterburn_cpu+CPU_R4_OFFSET
	;;
}
	ld8 r27=[r5],16
	mov r6=__afterburn_cpu+CPU_R4_OFFSET+8
	;;
	ld8 r28=[r6],16
	ld8 r29=[r5],16
	;;
	ld8 r30=[r6]
	ld8 r31=[r5]
	mov r6=__afterburn_cpu+CPU_PSR_OFFSET
	;;
	ld8 r5=[r6]
	;;
	dep r5=-1,r5,32,2
	;;
	st8 [r6]=r5
	mov r4=r27
	mov r5=r28
	mov r6=r29
	mov r7=r30
	mov b2=r31
.endm

.macro emul_read_cr pr=p0,dst,cr
.if \cr == 65
	__afterburn_call_0_1 \pr,burn_read_ivr,\dst
.else
(\pr)	mov \dst=__afterburn_cpu+CPU_CR_OFFSET+8*\cr
	;;
(\pr)	ld8 \dst=[\dst]
.endif
.endm

.macro emul_write_cr pr=p0,cr,src
.if \cr == 1
	__afterburn_call_1_0 \pr,burn_write_itm,\src
.elseif \cr == 67
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_IN_SERVICE_IRQ_OFFSET
	;;
}
(\pr)	st4 [r5]=r0
(\pr)	mov r4=r0
	;;
.else
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_CR_OFFSET+8*\cr
	;;
}
(\pr)	st8 [r5]=\src
(\pr)	mov r4=r0
	;;
.endif
.endm

.macro emul_read_kr pr=p0,dst,kr
(\pr)	mov \dst=__afterburn_cpu+CPU_KR_OFFSET+8*\kr
	;;
(\pr)	ld8 \dst=[\dst]
.endm

.macro emul_write_kr pr=p0,kr,src
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_KR_OFFSET+8*\kr
	;;
}
(\pr)	st8 [r5]=\src
(\pr)	mov r4=r0
	;;
.endm

.macro emul_write_itc pr=p0,src
	__afterburn_call_1_0 \pr,burn_write_itc,\src
.endm

.macro emul_read_pmd pr=p0,dst,pmd
	__afterburn_call_1_1 \pr,burn_read_pmd,\pmd,\dst
.endm

.macro emul_read_rr pr=p0,dst,rr
(\pr)	shr.u \dst=\rr,61
	;;
(\pr)	shl \dst=\dst,3
	;;
(\pr)	add \dst=__afterburn_cpu+CPU_RR_OFFSET,\dst
	;;
(\pr)	ld8 \dst=[\dst]
.endm

.macro emul_read_psr pr=p0,dst
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov \dst=__afterburn_cpu+CPU_PSR_OFFSET
}
(\pr)	mov r5=psr.um
	;;
(\pr)	ld8 \dst=[\dst]
	;;
(\pr)	dep \dst=r5,\dst,0,5
(\pr)	mov r4=r0
	;;
.endm

# ignore pk, dfl, sp, pp, di, si
.macro emul_ssm pr=p0,mask
.if \mask & ((1<<13)|(1<<17))
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_PSR_OFFSET
}
	;;
(\pr)	ld8 r6=[r5]
	;;
.endif
.if \mask & (1<<13)
(\pr)	dep r6=-1,r6,13,1 /* ic */
	;;
.endif
.if \mask & (1<<17)
(\pr)	dep r6=-1,r6,17,1 /* dt */
	;;
.endif
.if \mask & ((1<<13)|(1<<17))
(\pr)	st8 [r5]=r6
(\pr)	mov r4=r0
	;;
.endif
.if \mask & (1<<14)
	__afterburn_call_0_0 \pr,burn_enable_interrupts
.endif
.if \mask & (1<<19)
	__afterburn_call_0_0 \pr,burn_set_dfh
.endif
.if \mask & 0x3f
(\pr)	sum (\mask&0x3f)
.endif
.endm

# ignore pk, dfl, sp, pp, di, si
.macro emul_rsm pr=p0,mask
.if \mask & ((1<<13)|(1<<14)|(1<<17)|(1<<19))
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_PSR_OFFSET
}
	;;
(\pr)	ld8 r6=[r5]
	;;
.endif
.if \mask & (1<<13)
(\pr)	dep r6=0,r6,13,1 /* i */
	;;
.endif
.if \mask & (1<<14)
(\pr)	dep r6=0,r6,14,1 /* ic */
	;;
.endif
.if \mask & (1<<17)
(\pr)	dep r6=0,r6,17,1 /* dt */
	;;
.endif
.if \mask & (1<<19)
(\pr)	dep r6=0,r6,19,1 /* dfh */
	;;
.endif
.if \mask & ((1<<13)|(1<<14)|(1<<17)|(1<<19))
(\pr)	st8 [r5]=r6
(\pr)	mov r4=r0
	;;
.endif
.if \mask & 0x3f
(\pr)	rum (\mask&0x3f)
.endif
.endm

.macro emul_srlz_d pr=p0
.endm

.macro emul_srlz_i pr=p0
.endm

.macro real_cover
	cover
.endm

.macro emul_st8_spill_r pr=p0,idx,incr,reg
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_R4_OFFSET+((\reg-4)*8)
	;;
}
(\pr)	ld8 r5=[r5]
	;;
(\pr)	st8.spill [\idx]=r5
(\pr)	mov r4=r0
(\pr)	add \idx=\incr,\idx
	;;
.endm

.macro emul_ld8_fill_r pr=p0,idx,incr,reg
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_R4_OFFSET+((\reg-4)*8)
}
(\pr)	ld8.fill r6=[\idx]
	;;
(\pr)	st8 [r5]=r6
(\pr)	mov r4=r0
(\pr)	add \idx=\incr,\idx
	;;
.endm

.macro emul_read_b2 pr=p0,dst
(\pr)	mov \dst=__afterburn_cpu+CPU_R4_OFFSET+32
	;;
(\pr)	ld8 \dst=[\dst]
.endm

.macro emul_write_b2 pr=p0,src
{.mii
(\pr)	nop.m 0x0
(\pr)	mov r4=ip
(\pr)	mov r5=__afterburn_cpu+CPU_R4_OFFSET+32
	;;
}
(\pr)	st8 [r5]=\src
(\pr)	mov r4=r0
	;;
.endm
