/*
 @LICENSE@
*/

#ifndef _IDE_PCI_H_
#define _IDE_PCI_H_

#include <driver_interface.h>
#include <lock.h>

#define MAX_BUS 2 /* We only ever have two buses on the host */
#define MAX_DRIVES_BUS 2
#define MAX_DRIVES (MAX_BUS * MAX_DRIVES_BUS)

enum ide_command_types {
	EXECUTE_DEVICE_DIAGNOSTIC  = 0x90,
	IDENTIFY_DEVICE = 0xEC,
	READ_DMA = 0xC8,
	READ_DMA_EXT = 0x25,
	SET_FEATURES = 0xEF,
	WRITE_DMA = 0xCA,
	WRITE_DMA_EXT = 0x35
};

struct ide_command_names { 
	uint16_t cmd; 
	char *name;
};

#define MAX_EDD_TRIES 10

enum protocol { EDD, PIO_IN, DMA, NONDATA };


/* XXX: This is the PRD entry as described in: [1] (pg2) */
struct dma_descriptor {
	uint32_t address_pointer;
	uint32_t size_eot;
};

/* Device driver data */
typedef struct dide {
	struct driver *classp;
	bus_space_t data_port[MAX_BUS];
	bus_space_t control_port[MAX_BUS];
	bus_space_t bmaster_port;

	int wfi[MAX_BUS]; /* waiting for interrupt */
	uint32_t *ident[MAX_DRIVES];
	uint64_t capacity[MAX_DRIVES];
	int disks;
	int ready[MAX_DRIVES];
	int large_support[MAX_DRIVES];

	lock_t	cmd_lock[MAX_BUS];
	struct command *head[MAX_BUS];
	struct command *tail[MAX_BUS];

	bus_space_t pciconf;
	dma_handle_t dma;

	mem_handle_t dma_table_handle;
        dmabuf_handle_t dma_table_dmabuf;
	dma_addr_t	dma_table_addr;
        struct dma_descriptor *dma_table;
} dide_t;


struct command {
	enum ide_command_types cmd;
	enum protocol protocol;

	int device;
	mem_handle_t data;
	uint64_t block_number;
	uint16_t blocks;

	int done;
	struct command *next;

	bool (*cont_fn)(struct dide *, int);

	uintptr_t key;
	dmabuf_handle_t dmabuf;
	int teardown_dma;

	callback_t callback;
	callback_data_t callback_data;
};


struct driver *ide_init(void);

/* Exported functions */
#if 0
dide_t * bennoide_init (bs_seq spaces, dma_handle_t dma, bus_space_t pciconf);
void bennoide_enable(dide_t *dthis);
int bennoide_read (dide_t *dthis, int diskid, callback_t callback, 
	      callback_data_t callback_data, block_seq data);
int bennoide_write (dide_t *dthis, int diskid, callback_t callback, 
	       callback_data_t callback_data, block_seq data);
int bennoide_read_sg (dide_t *dthis, int diskid, callback_t callback, 
	      callback_data_t callback_data, dmabuf_handle_t data);
int bennoide_write_sg (dide_t *dthis, int diskid, callback_t callback, 
	       callback_data_t callback_data, dmabuf_handle_t data);
int bennoide_interrupt (dide_t *dthis);
int bennoide_get_disk_count (dide_t *dthis);
uint64_t bennoide_get_block_count (dide_t *dthis, int diskid);
#endif

#endif				/* _IDE_PCI_H_ */
