#ifndef _IDE_DATA_H_
#define _IDE_DATA_H_

#include "ide_data_types.reg.h"

#include <io/io.h>

static inline union ide_data read_ide_data(bus_space_t bus) {
	union ide_data foo;
	foo.data = io_read_32(bus, 0);
	return foo;
}

static inline void write_ide_data(bus_space_t bus, union ide_data foo) {
	return io_write_32 (bus, (0), foo.data);
}

static inline union ide_error read_ide_error(bus_space_t bus) {
	union ide_error foo;
	foo.data = io_read_8(bus, 1);
	return foo;
}

static inline void write_ide_feature(bus_space_t bus, union ide_feature foo) {
	return io_write_8 (bus, (1), foo.data);
}

static inline union ide_sectors read_ide_sectors(bus_space_t bus) {
	union ide_sectors foo;
	foo.data = io_read_8(bus, 2);
	return foo;
}

static inline void write_ide_sectors(bus_space_t bus, union ide_sectors foo) {
	return io_write_8 (bus, (2), foo.data);
}

static inline union ide_lba_low read_ide_lba_low(bus_space_t bus) {
	union ide_lba_low foo;
	foo.data = io_read_8(bus, 3);
	return foo;
}

static inline void write_ide_lba_low(bus_space_t bus, union ide_lba_low foo) {
	return io_write_8 (bus, (3), foo.data);
}

static inline union ide_lba_mid read_ide_lba_mid(bus_space_t bus) {
	union ide_lba_mid foo;
	foo.data = io_read_8(bus, 4);
	return foo;
}

static inline void write_ide_lba_mid(bus_space_t bus, union ide_lba_mid foo) {
	return io_write_8 (bus, (4), foo.data);
}

static inline union ide_lba_hi read_ide_lba_hi(bus_space_t bus) {
	union ide_lba_hi foo;
	foo.data = io_read_8(bus, 5);
	return foo;
}

static inline void write_ide_lba_hi(bus_space_t bus, union ide_lba_hi foo) {
	return io_write_8 (bus, (5), foo.data);
}

static inline union ide_dev read_ide_dev(bus_space_t bus) {
	union ide_dev foo;
	foo.data = io_read_8(bus, 6);
	return foo;
}

static inline void write_ide_dev(bus_space_t bus, union ide_dev foo) {
	return io_write_8 (bus, (6), foo.data);
}

static inline void write_ide_command(bus_space_t bus, union ide_command foo) {
	return io_write_8 (bus, (7), foo.data);
}

static inline union ide_status read_ide_status(bus_space_t bus) {
	union ide_status foo;
	foo.data = io_read_8(bus, 7);
	return foo;
}

static inline void print_ide_data ( union ide_data reg ) {	UserPrint("Debug ide_data - 0x%x\n", reg.data);}

static inline void print_ide_error ( union ide_error reg ) {	UserPrint("Debug ide_error - 0x%x\n", reg.data);}

static inline void print_ide_feature ( union ide_feature reg ) {	UserPrint("Debug ide_feature - 0x%x\n", reg.data);}

static inline void print_ide_sectors ( union ide_sectors reg ) {	UserPrint("Debug ide_sectors - 0x%x\n", reg.data);}

static inline void print_ide_lba_low ( union ide_lba_low reg ) {	UserPrint("Debug ide_lba_low - 0x%x\n", reg.data);}

static inline void print_ide_lba_mid ( union ide_lba_mid reg ) {	UserPrint("Debug ide_lba_mid - 0x%x\n", reg.data);}

static inline void print_ide_lba_hi ( union ide_lba_hi reg ) {	UserPrint("Debug ide_lba_hi - 0x%x\n", reg.data);}

static inline void print_ide_dev ( union ide_dev reg ) {	UserPrint("Debug ide_dev - 0x%x\n", reg.data);	if (reg.r.sector_hi != 0) {
		UserPrint("\tsector_hi: %x\n", reg.r.sector_hi);
	}
	if (reg.r.dev != 0) {
		UserPrint("\tdev: %x\n", reg.r.dev);
	}
	if (reg.r.lba != 0) {
		UserPrint("\tlba: %x\n", reg.r.lba);
	}
}

static inline void print_ide_command ( union ide_command reg ) {	UserPrint("Debug ide_command - 0x%x\n", reg.data);	if (reg.r.cmd != 0) {
		UserPrint("\tcmd: %x\n", reg.r.cmd);
	}
}

static inline void print_ide_status ( union ide_status reg ) {	UserPrint("Debug ide_status - 0x%x\n", reg.data);	if (reg.r.err != 0) {
		UserPrint("\terr: %x\n", reg.r.err);
	}
	if (reg.r.drq != 0) {
		UserPrint("\tdrq: %x\n", reg.r.drq);
	}
	if (reg.r.df != 0) {
		UserPrint("\tdf: %x\n", reg.r.df);
	}
	if (reg.r.drdy != 0) {
		UserPrint("\tdrdy: %x\n", reg.r.drdy);
	}
	if (reg.r.bsy != 0) {
		UserPrint("\tbsy: %x\n", reg.r.bsy);
	}
}

#endif /* _IDE_DATA_H_ */
