#include "mpt_kenge.h"
#include "mpt_interface.h"
#include <pci_dma.h>
#include <malloc.h>

static struct driver mpt_driver_info =
{
	.name = "mpt",
	/* .class = netif_device */
};

#define NSGL_FIRST 6
#define NSGL 12

static void
mpt_build_sgl(request_t *req, MSG_SCSI_IO_REQUEST *mpt_req, SGE_SIMPLE32 *se, struct scsi_sge *sge, uint32_t sge_count, enum scsi_io_direction direction)
{
	SGE_CHAIN32 *ce;
	uint32_t flags = MPI_SGE_FLAGS_SIMPLE_ELEMENT;
	uint32_t ntodo;
	if (direction == SCSI_IO_DIR_IN)
		flags |= MPI_SGE_FLAGS_IOC_TO_HOST;
	else
		flags |= MPI_SGE_FLAGS_HOST_TO_IOC;

	ntodo = (sge_count > NSGL_FIRST) ? NSGL_FIRST-1 : sge_count;
	for (; ntodo > 0; ntodo--, sge_count--, se++, sge++)
	{
		if (sge_count == 1)
			flags |= MPI_SGE_FLAGS_LAST_ELEMENT | MPI_SGE_FLAGS_END_OF_BUFFER | MPI_SGE_FLAGS_END_OF_LIST;
		else if (ntodo == 1)
			flags |= MPI_SGE_FLAGS_LAST_ELEMENT;

		//printf("%lx:%x:%x\n", sge->addr, sge->len, flags);
		se->Address = sge->addr;
		MPI_pSGE_SET_FLAGS_LENGTH(se, flags, sge->len);
		flags &= ~MPI_SGE_FLAGS_LAST_ELEMENT;
	}

	while (sge_count > 0)
	{
		//printf("%d SGEs remaining\n", sge_count);
		ce = (SGE_CHAIN32 *)se++;
		if (sge_count > NSGL) {
			ce->NextChainOffset = ((NSGL-1) * sizeof(SGE_SIMPLE32)) >> 2;
			ce->Length = (NSGL-1) * sizeof(SGE_SIMPLE32);
			ntodo = NSGL-1;
		} else {
			ce->NextChainOffset = 0;
			ce->Length = sge_count * sizeof(SGE_SIMPLE32);
			ntodo = sge_count;
		}
		ce->Flags = MPI_SGE_FLAGS_CHAIN_ELEMENT;
		ce->Address = req->req_pbuf + ((char *)se - (char *)mpt_req);
		//printf("%x:%x:%x:%x\n", ce->Address, ce->Length, ce->Flags, ce->NextChainOffset);

		for (; ntodo > 0; ntodo--, sge_count--, se++, sge++)
		{
			if (sge_count == 1)
				flags |= MPI_SGE_FLAGS_LAST_ELEMENT | MPI_SGE_FLAGS_END_OF_BUFFER | MPI_SGE_FLAGS_END_OF_LIST;
			else if (ntodo == 1)
				flags |= MPI_SGE_FLAGS_LAST_ELEMENT;

			se->Address = sge->addr;
			MPI_pSGE_SET_FLAGS_LENGTH(se, flags, sge->len);
			//printf("%x:%x:%x\n", se->Address, sge->len, flags);
			flags &= ~MPI_SGE_FLAGS_LAST_ELEMENT;
		}
	}
}

/* based on mpt_start/mpt_execute_req in mpt_freebsd.c */
int
mpt_scsi_io_request(void *device, uint8_t target_id, uint8_t target_lun, uint8_t *cdb, uint8_t cdb_len,
		    struct scsi_sge *sge, uint32_t sge_count, uint32_t dxfer_len, enum scsi_io_direction direction,
		    callback_t callback, callback_data_t callback_data)
{
	request_t *req;
	mpt_softc_t *mpt = device;
	MSG_SCSI_IO_REQUEST *mpt_req;
	SGE_SIMPLE32 *se;
#if 0
	mem_handle_t mem_handle;
#endif
	int i;

	/* Get a request structure off the free list */
	while ((req = mpt_get_request(mpt)) == NULL) {
		for (i = 0; i < 1000000; i++);
		mpt_intr(mpt);
		printf("Y");
	}

	/* Now we build the command for the IOC */
	mpt_req = req->req_vbuf;
	bzero(mpt_req, sizeof *mpt_req);

	mpt_req->Function = MPI_FUNCTION_SCSI_IO_REQUEST;
	mpt_req->SenseBufferLength = MPT_SENSE_SIZE;

	/* We use the message context to find the request structure when we */
	/* Get the command competion interrupt from the FC IOC.				*/
	mpt_req->MsgContext = req->index;

	/* Which physical device to do the I/O on */
	mpt_req->TargetID = target_id;
	mpt_req->LUN[1] = target_lun;

	/* Set the direction of the transfer */
	if (direction == SCSI_IO_DIR_IN)
		mpt_req->Control = MPI_SCSIIO_CONTROL_READ;
	else if (direction == SCSI_IO_DIR_OUT)
		mpt_req->Control = MPI_SCSIIO_CONTROL_WRITE;
	else
		mpt_req->Control = MPI_SCSIIO_CONTROL_NODATATRANSFER;

	/* POSSIBLE ENHANCEMENT: do tagged transfers if device supports it */
	if (1 || mpt->is_fc)
		mpt_req->Control |= MPI_SCSIIO_CONTROL_SIMPLEQ;
	else
		mpt_req->Control |= MPI_SCSIIO_CONTROL_UNTAGGED;

	/* Copy the scsi command block into place */
	bcopy(cdb, mpt_req->CDB, cdb_len);

	mpt_req->CDBLength = cdb_len;
	mpt_req->DataLength = dxfer_len;
	mpt_req->SenseBufferLowAddr = req->sense_pbuf;
	mpt_req->ChainOffset = (sge_count > NSGL_FIRST) ? 88 : 0;

	/*
	 * If we have any data to send with this command,
	 * map it into bus space.
	 */

	if (direction != SCSI_IO_DIR_NONE) {
		/* POSSIBLE ENHANCEMENT: scatter-gather support */
		se = (SGE_SIMPLE32 *) &mpt_req->SGL;
		mpt_build_sgl(req, mpt_req, se, sge, sge_count, direction);

#if 0
		mem_handle = new_mem_handle(data, dxfer_len);
		setup_dma(mpt->dma, mem_handle, dxfer_len, &req->dmap, (direction == SCSI_IO_DIR_IN) ? bus_to_mem : mem_to_bus);
		se->Address = GLUE_DMABUF_ENT_DMAADDR(req->dmap, 0);
		MPI_pSGE_SET_FLAGS_LENGTH(se,
		    (MPI_SGE_FLAGS_LAST_ELEMENT | MPI_SGE_FLAGS_END_OF_BUFFER |
		    MPI_SGE_FLAGS_SIMPLE_ELEMENT | MPI_SGE_FLAGS_END_OF_LIST |
		    ((direction == SCSI_IO_DIR_IN) ? MPI_SGE_FLAGS_IOC_TO_HOST : MPI_SGE_FLAGS_HOST_TO_IOC)),
		    dxfer_len);
#endif
	} else {
		se = (SGE_SIMPLE32 *) &mpt_req->SGL;
		/*
		 * No data to transfer so we just make a single simple SGL
		 * with zero length.
		 */
		MPI_pSGE_SET_FLAGS(se,
		    (MPI_SGE_FLAGS_LAST_ELEMENT | MPI_SGE_FLAGS_END_OF_BUFFER |
		    MPI_SGE_FLAGS_SIMPLE_ELEMENT | MPI_SGE_FLAGS_END_OF_LIST));
	}

	req->callback = callback;
	req->callback_data = callback_data;

	/* ENHANCEMENT: schedule timeout */
	if (mpt->verbose > 1)
		mpt_print_scsi_io_request(mpt_req);
	mpt_send_cmd(mpt, req);
	return 0;
}

/* from mpt_freebsd.c (unmodified) */
static void
mpt_event_notify_reply(mpt_softc_t *mpt, MSG_EVENT_NOTIFY_REPLY *msg)
{
	switch(msg->Event) {
	case MPI_EVENT_LOG_DATA:
		/* Some error occured that LSI wants logged */
		printf("\tEvtLogData: IOCLogInfo: 0x%08x\n", msg->IOCLogInfo);
		printf("\tEvtLogData: Event Data:");
		{
			int i;
			for (i = 0; i < msg->EventDataLength; i++) {
				printf("  %08x", msg->Data[i]);
			}
		}
		printf("\n");
		break;

	case MPI_EVENT_UNIT_ATTENTION:
		mpt_prt(mpt, "Bus: 0x%02x TargetID: 0x%02x",
		    (msg->Data[0] >> 8) & 0xff, msg->Data[0] & 0xff);
		break;

	case MPI_EVENT_IOC_BUS_RESET:
		/* We generated a bus reset */
		mpt_prt(mpt, "IOC Bus Reset Port: %d",
		    (msg->Data[0] >> 8) & 0xff);
		break;

	case MPI_EVENT_EXT_BUS_RESET:
		/* Someone else generated a bus reset */
		mpt_prt(mpt, "Ext Bus Reset");
		/*
		 * These replies don't return EventData like the MPI
		 * spec says they do
		 */	
/*		xpt_async(AC_BUS_RESET, path, NULL);  */
		break;

	case MPI_EVENT_RESCAN:
		/*
		 * In general this means a device has been added
		 * to the loop.
		 */
		mpt_prt(mpt, "Rescan Port: %d", (msg->Data[0] >> 8) & 0xff);
/*		xpt_async(AC_FOUND_DEVICE, path, NULL);  */
		break;

	case MPI_EVENT_LINK_STATUS_CHANGE:
		mpt_prt(mpt, "Port %d: LinkState: %s",
		    (msg->Data[1] >> 8) & 0xff,
		    ((msg->Data[0] & 0xff) == 0)?  "Failed" : "Active");
		break;

	case MPI_EVENT_LOOP_STATE_CHANGE:
		switch ((msg->Data[0] >> 16) & 0xff) {
		case 0x01:
			mpt_prt(mpt,
			    "Port 0x%x: FC LinkEvent: LIP(%02x,%02x) (Loop Initialization)\n",
			    (msg->Data[1] >> 8) & 0xff,
			    (msg->Data[0] >> 8) & 0xff,
			    (msg->Data[0]     ) & 0xff);
			switch ((msg->Data[0] >> 8) & 0xff) {
			case 0xF7:
				if ((msg->Data[0] & 0xff) == 0xF7) {
					printf("Device needs AL_PA\n");
				} else {
					printf("Device %02x doesn't like FC performance\n", 
									msg->Data[0] & 0xFF);
				}
				break;
			case 0xF8:
				if ((msg->Data[0] & 0xff) == 0xF7) {
					printf("Device had loop failure at its receiver prior to acquiring AL_PA\n");
				} else {
					printf("Device %02x detected loop failure at its receiver\n", 
									msg->Data[0] & 0xFF);
				}
				break;
			default:
				printf("Device %02x requests that device %02x reset itself\n", 
					msg->Data[0] & 0xFF,
					(msg->Data[0] >> 8) & 0xFF);
				break;
			}
			break;
		case 0x02:
			mpt_prt(mpt, "Port 0x%x: FC LinkEvent: LPE(%02x,%02x) (Loop Port Enable)",
				(msg->Data[1] >> 8) & 0xff, /* Port */
				(msg->Data[0] >>  8) & 0xff, /* Character 3 */
				(msg->Data[0]      ) & 0xff  /* Character 4 */
				);
			break;
		case 0x03:
			mpt_prt(mpt, "Port 0x%x: FC LinkEvent: LPB(%02x,%02x) (Loop Port Bypass)",
				(msg->Data[1] >> 8) & 0xff, /* Port */
			(msg->Data[0] >> 8) & 0xff, /* Character 3 */
			(msg->Data[0]     ) & 0xff  /* Character 4 */
				);
			break;
		default:
			mpt_prt(mpt, "Port 0x%x: FC LinkEvent: Unknown FC event (%02x %02x %02x)",
				(msg->Data[1] >> 8) & 0xff, /* Port */
				(msg->Data[0] >> 16) & 0xff, /* Event */
				(msg->Data[0] >>  8) & 0xff, /* Character 3 */
				(msg->Data[0]      ) & 0xff  /* Character 4 */
				);
		}
		break;

	case MPI_EVENT_LOGOUT:
		mpt_prt(mpt, "FC Logout Port: %d N_PortID: %02x",
		    (msg->Data[1] >> 8) & 0xff, msg->Data[0]);
		break;
	case MPI_EVENT_EVENT_CHANGE:
		/* This is just an acknowledgement of our 
		   mpt_send_event_request */
		break;
	default:
		mpt_prt(mpt, "Unknown event 0x%x\n", msg->Event);
	}
	if (msg->AckRequired) {
		MSG_EVENT_ACK *ackp;
		request_t *req;
		if ((req = mpt_get_request(mpt)) == NULL) {
			panic("unable to get request to acknowledge notify");
		}
		ackp = (MSG_EVENT_ACK *) req->req_vbuf;
		bzero(ackp, sizeof *ackp);
		ackp->Function = MPI_FUNCTION_EVENT_ACK;
		ackp->Event = msg->Event;
		ackp->EventContext = msg->EventContext;
		ackp->MsgContext = req->index | 0x80000000;
		mpt_check_doorbell(mpt);
		mpt_send_cmd(mpt, req);
	}
}

/* from mpt_freebsd.c (unmodified) */
static void
mpt_ctlop(mpt_softc_t *mpt, void *vmsg, u_int32_t reply)
{
	MSG_DEFAULT_REPLY *dmsg = vmsg;

	if (dmsg->Function == MPI_FUNCTION_EVENT_NOTIFICATION) {
		mpt_event_notify_reply(mpt, vmsg);
		mpt_free_reply(mpt, (reply << 1));
	} else if (dmsg->Function == MPI_FUNCTION_EVENT_ACK) {
		mpt_free_reply(mpt, (reply << 1));
	} else if (dmsg->Function == MPI_FUNCTION_PORT_ENABLE) {
		MSG_PORT_ENABLE_REPLY *msg = vmsg;
		int index = msg->MsgContext & ~0x80000000;
		if (mpt->verbose > 1) {
			mpt_prt(mpt, "enable port reply idx %d", index);
		}
		if (index >= 0 && index < MPT_MAX_REQUESTS(mpt)) {
			request_t *req = &mpt->request_pool[index];
			req->debug = REQ_DONE;
		}
		mpt_free_reply(mpt, (reply << 1));
	} else if (dmsg->Function == MPI_FUNCTION_CONFIG) {
		MSG_CONFIG_REPLY *msg = vmsg;
		int index = msg->MsgContext & ~0x80000000;
		if (index >= 0 && index < MPT_MAX_REQUESTS(mpt)) {
			request_t *req = &mpt->request_pool[index];
			req->debug = REQ_DONE;
			req->sequence = reply;
		} else {
			mpt_free_reply(mpt, (reply << 1));
		}
	} else {
		mpt_prt(mpt, "unknown mpt_ctlop: %x", dmsg->Function);
	}
}

/* from mpt_freebsd.c (modified) */
void
mpt_done(mpt_softc_t *mpt, u_int32_t reply)
{
	int index;
	request_t *req;
	MSG_REQUEST_HEADER *mpt_req;
	MSG_SCSI_IO_REPLY *mpt_reply;
	int error = 0;

	if ((reply & MPT_CONTEXT_REPLY) == 0) {
		/* context reply */
		mpt_reply = NULL;
		index = reply & MPT_CONTEXT_MASK;
	} else {
		unsigned *pReply;

		/* address reply (Error) */
		mpt_reply = MPT_REPLY_PTOV(mpt, reply);
		if (mpt->verbose > 1) {
			pReply = (unsigned *) mpt_reply;
			mpt_prt(mpt, "Address Reply (index %u)",
			    mpt_reply->MsgContext & 0xffff);
			printf("%08x %08x %08x %08x\n",
			    pReply[0], pReply[1], pReply[2], pReply[3]);
			printf("%08x %08x %08x %08x\n",
			    pReply[4], pReply[5], pReply[6], pReply[7]);
			printf("%08x %08x %08x %08x\n\n",
			    pReply[8], pReply[9], pReply[10], pReply[11]);
		}
		index = mpt_reply->MsgContext;
	}

	/*
	 * Address reply with MessageContext high bit set
	 * This is most likely a notify message so we try
	 * to process it then free it
	 */
	if ((index & 0x80000000) != 0) {
		if (mpt_reply != NULL) {
			mpt_ctlop(mpt, mpt_reply, reply);
		} else {
			mpt_prt(mpt, "mpt_done: index 0x%x, NULL reply", index);
		}
		return;
	}

	/* Did we end up with a valid index into the table? */
	if (index < 0 || index >= MPT_MAX_REQUESTS(mpt)) {
		mpt_prt(mpt, "mpt_done: invalid index (%x) in reply", index);
		return;
	}

	req = &mpt->request_pool[index];

	/* Make sure memory hasn't been trashed */
	if (req->index != index) {
		printf("mpt_done: corrupted request struct");
		return;
	}

	/* Short cut for task management replys; nothing more for us to do */
	mpt_req = req->req_vbuf;
	if (mpt_req->Function == MPI_FUNCTION_SCSI_TASK_MGMT) {
		if (mpt->verbose > 1) {
			mpt_prt(mpt, "mpt_done: TASK MGMT");
		}
		goto done;
	}

	if (mpt_req->Function == MPI_FUNCTION_PORT_ENABLE) {
		goto done;
	}

	/*
	 * At this point it better be a SCSI IO command, but don't
	 * crash if it isn't
	 */
	if (mpt_req->Function != MPI_FUNCTION_SCSI_IO_REQUEST)  {
		goto done;
	}

	if (mpt_reply != NULL)
	{
		/* POSSIBLE ENHANCEMENT: better SCSI error reporting */
		if (mpt_reply->IOCStatus != MPI_IOCSTATUS_SUCCESS)
		{
			mpt_prt(mpt, "IOC error %lx", mpt_reply->IOCStatus);
			error = 1;
		}
		else if (mpt_reply->SCSIStatus != SCSI_STATUS_OK)
		{
			mpt_prt(mpt, "SCSI error %lx %lx", mpt_reply->SCSIStatus, mpt_reply->SCSIState);
			error = 1;
		}
	}

#if 0
	teardown_dma(mpt->dma, &req->dmap);
#endif
	if (req->callback != NULL)
		req->callback(req->callback_data, error);

done:
	/* If IOC done with this request free it up */
	if (mpt_reply == NULL || (mpt_reply->MsgFlags & 0x80) == 0)
		mpt_free_request(mpt, req);

	/* If address reply; give the buffer back to the IOC */
	if (mpt_reply != NULL)
		mpt_free_reply(mpt, (reply << 1));
}

/* from mpt_pci.c (unmodified) */
int
mpt_intr(void *dummy)
{
	int nrepl = 0;
	u_int32_t reply;
	mpt_softc_t *mpt = (mpt_softc_t *)dummy;

	if ((mpt_read(mpt, MPT_OFFSET_INTR_STATUS) & MPT_INTR_REPLY_READY) == 0)
		return (0);
	reply = mpt_pop_reply_queue(mpt);
	while (reply != MPT_REPLY_EMPTY) {
		nrepl++;
		if (mpt->verbose > 1) {
			if ((reply & MPT_CONTEXT_REPLY) != 0)  {
				/* Address reply; IOC has something to say */
				mpt_print_reply(MPT_REPLY_PTOV(mpt, reply));
			} else {
				/* Context reply ; all went well */
				mpt_prt(mpt, "context %u reply OK", reply);
			}
		}
		mpt_done(mpt, reply);
		reply = mpt_pop_reply_queue(mpt);
	}
	return (nrepl != 0);
}

/* from mpt_pci.c, heavily modified */
static int
mpt_dma_mem_alloc(mpt_softc_t *mpt)
{
	int i;
	caddr_t vptr;
	bus_addr_t pptr, end;
	size_t len;
	dmabuf_handle_t dmabuf;
	mem_handle_t mem;

	/* FIXME this len calculation looks bogus - matthewc */
	len = sizeof (request_t *) * MPT_REQ_MEM_SIZE(mpt);
	mpt->request_pool = malloc(len);
	bzero(mpt->request_pool, len);

	/* Allocate some DMA accessible memory for replies */
	if (get_pci_consistent_memory(mpt->dma, PAGE_SIZE, &mem, &dmabuf) == -1) {
		printf("mpt: cannot allocate %u bytes of reply memory\n", PAGE_SIZE);
		return (-1);
	}
	mpt->reply = get_addr(mem);
	mpt->reply_phys = GLUE_DMABUF_ENT_DMAADDR(dmabuf, 0);

	/* Allocate some DMA accessible memory for requests */
	if (get_pci_consistent_memory(mpt->dma, MPT_REQ_MEM_SIZE(mpt), &mem, &dmabuf) == -1) {
		printf("mpt: cannot allocate %u bytes of request memory\n", MPT_REQ_MEM_SIZE(mpt));
		return (-1);
	}
	mpt->request = get_addr(mem);
	mpt->request_phys = GLUE_DMABUF_ENT_DMAADDR(dmabuf, 0);

	i = 0;
	pptr =  mpt->request_phys;
	vptr =  mpt->request;
	end = pptr + MPT_REQ_MEM_SIZE(mpt);
	while(pptr < end) {
		request_t *req = &mpt->request_pool[i];
		req->index = i++;

		/* Store location of Request Data */
		req->req_pbuf = pptr;
		req->req_vbuf = vptr;

		pptr += MPT_REQUEST_AREA;
		vptr += MPT_REQUEST_AREA;

		req->sense_pbuf = (pptr - MPT_SENSE_SIZE);
		req->sense_vbuf = (vptr - MPT_SENSE_SIZE);
	}
	return (0);
}

/* Deallocate memory that was allocated by mpt_dma_mem_alloc 
 */
static void
mpt_dma_mem_free(mpt_softc_t *mpt)
{
	/* FIXME: tear down DMA */
	free(mpt->request_pool);
}

void *
mpt_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{
	mpt_softc_t *mpt;

	mpt = malloc(sizeof(mpt_softc_t));
	mpt->classp = &mpt_driver_info;
	mpt->iospace = spaces.v[0];
	mpt->dma = dma;

	mpt->verbose = 1; /* 1 for some verbosity, 2 for full message decodes */
	mpt->is_fc = 0;

	mpt_disable_ints(mpt);

	if (mpt_dma_mem_alloc(mpt)) {
		printf("Could not allocate DMA memory\n");
		goto fail;
	}

	/* mpt_read_config_regs(mpt); */

	if (mpt_init(mpt, MPT_DB_INIT_HOST) != 0) {
		printf("Initialization failed\n");
		goto fail;
	}

	printf("Request frame size: %x\n", mpt->request_frame_size<<2);
	return mpt;

fail:
	mpt_dma_mem_free(mpt);
	free(mpt);
	return NULL;
}
