#include "driver_interface.h"
#include "atomic.h" /* mb() &c */
#include "bfrdriver.h"
#include <malloc.h>
#include <string.h> /* for memset */
#include "e1000_driver.h"
#include  "e1000_kenge.h"

#define EPRINT 
#define NAME "e1000"
static struct driver e1000 =
{
	.name = NAME,
	.class = netif_device,
};
static inline void e1000_irq_enable(struct e1000_adapter *dthis);
static inline void e1000_irq_disable(struct e1000_adapter *dthis);
void e1000_reset(struct e1000_adapter *dthis);
static void e1000_configure_tx(struct e1000_adapter *dthis);
static void e1000_setup_rctl(struct e1000_adapter *dthis);
static void e1000_configure_rx(struct e1000_adapter *dthis);
void e1000_update_stats(struct e1000_adapter *adapter);
static int e1000_init_desc_ring(struct e1000_adapter *dthis);
int e1000_setup_tx_resources(struct e1000_adapter *dthis);
int e1000_setup_rx_resources(struct e1000_adapter *dthis);
static void e1000_clean_tx_ring(struct e1000_adapter *dthis);
static void e1000_clean_rx_ring(struct e1000_adapter *dthis);
static int e1000_clean_tx_irq(struct e1000_adapter *dthis);
static int e1000_clean_rx_irq(struct e1000_adapter *dthis);
static int e1000_sw_init(struct e1000_adapter *dthis);
/* the transmit packet tag(type), used to set the tx_desc status when transmit a packet, syt*/
#define E1000_TX_FLAGS_CSUM		0x00000001
#define E1000_TX_FLAGS_VLAN		0x00000002
#define E1000_TX_FLAGS_TSO		0x00000004
#define E1000_TX_FLAGS_VLAN_MASK	0xffff0000
#define E1000_TX_FLAGS_VLAN_SHIFT	16
static void *
e1000_setup(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{	
	struct e1000_adapter *dthis;
	struct pci_dev * pdev;
	uint16_t eeprom_data;
	int i;
	uint16_t phy_ctrl, phy_stat;
	
	dthis = malloc(sizeof(struct e1000_adapter));
	UserPrint("e1000 setup(%p)!\n", dthis);
	if (dthis == 0)
	    goto out;
	memset(dthis, 0, sizeof (*dthis));
	dthis->classp = &e1000;
	dthis->dma = dma;
	pdev = dma;
	dthis->hw.vendor_id = pdev->vendor;
	dthis->hw.device_id = pdev->device;
	dthis->hw.subsystem_vendor_id = pdev->subsystem_vendor;
	dthis->hw.subsystem_id = pdev->subsystem_device;
	dthis->hw.mhandle = spaces.v[0];
	dthis->hw.pciconf = pciconf;
	dthis->hw.iohandle = spaces.v[1];
	UserPrint("iohandle: start: 0x%x, flags: 0x%x\n", 
	            dthis->hw.iohandle.u.normal.start, 
		    dthis->hw.iohandle.u.normal.flags);
	dthis->hw.revision_id = E1000_READ_REG_BYTE(&dthis->hw, PCI_REVISION_ID);
	dthis->hw.pci_cmd_word = E1000_READ_REG_WORD(&dthis->hw, PCI_COMMAND);
	dthis->hw.max_frame_size = ENET_HEADER_SIZE + ETHERNET_FCS_SIZE;
	dthis->hw.min_frame_size = MINIMUM_ETHERNET_FRAME_SIZE;
        dthis->hw.back = dthis;
	//dthis->tx_timeout = &e1000_tx_timeout;
        //dthis->watchdog_timeo = 5 * HZ;
	lock_init(&dthis->tx_lock);
	lock_init(&dthis->rx_lock);
	lock_init(&dthis->stats_lock);
	UserPrint("about to sw_init\n");
	if(e1000_sw_init(dthis))
		goto out_free;
        /* disable interrupts... syt  */
	UserPrint("about to reset hw\n");
	e1000_reset_hw(&dthis->hw);
	/* make sure the EEPROM is good */
	if(e1000_validate_eeprom_checksum(&dthis->hw) < 0) {
	  UserPrint("The EEPROM Checksum Is Not Valid\n");
	  goto out_free;
        }
       /* copy the MAC address out of the EEPROM an d set up the mac addr filter*/
	e1000_read_mac_addr(&dthis->hw);
	e1000_read_part_num(&dthis->hw, &(dthis->part_num));
	e1000_get_bus_info(&dthis->hw);
	/* load the default value for the parameters not changed*/
	
	e1000_check_options(dthis);

	/* Initial Wake on LAN setting
	 * If APM wake is enabled in the EEPROM,
	 * enable the ACPI Magic Packet filter
	 */

	switch(dthis->hw.mac_type) {
	case e1000_82542_rev2_0:
	case e1000_82542_rev2_1:
	case e1000_82543:
		break;
	case e1000_82546:
	case e1000_82546_rev_3:
		if((E1000_READ_REG(&dthis->hw, STATUS) & E1000_STATUS_FUNC_1)
		   && (dthis->hw.media_type == e1000_media_type_copper)) {
			e1000_read_eeprom(&dthis->hw,
				EEPROM_INIT_CONTROL3_PORT_B, 1, &eeprom_data);
			break;
		}
		/* Fall Through */
	default:
		e1000_read_eeprom(&dthis->hw,
			EEPROM_INIT_CONTROL3_PORT_A, 1, &eeprom_data);
		break;
	}
	if(eeprom_data & E1000_EEPROM_APME)
		dthis->wol |= E1000_WUFC_MAG;
	/* reset the hardware with the new settings, include set rx filter */
	e1000_reset(dthis);
	EPRINT("e1000_setup: returning dthis %p", dthis);
	e1000_read_phy_reg(&dthis->hw, PHY_CTRL, &phy_ctrl);
	e1000_read_phy_reg(&dthis->hw, PHY_STATUS, &phy_stat);
	EPRINT("\nthe phy ctl 0x%x, phy stat 0x%x\n", 
	           phy_ctrl, phy_stat);
	return dthis;
out_free:
	free(dthis);
out:
	return NULL;
}

static int
enable(void *dthis_v, struct bfr_drv *bfr_handle)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v;
	uint16_t phy_labi;
	
	EPRINT("Enable: rx(count: %d) tx(count: %d).\n",
	           dthis->rx_ring.count, dthis->tx_ring.count);
	e1000_init_desc_ring(dthis);
	e1000_configure_tx(dthis);
	e1000_setup_rctl(dthis);
	e1000_configure_rx(dthis);
	if (bfr_register_if(bfr_handle,
			    dthis,
			    dthis->rx_ring.count,
			    PAGESIZE,
			    dthis->tx_ring.count)) {
		UserPrint("Couldn't register interface with buffer manager!\n");
		return 1;
	}
	dthis->buf_man_handle = bfr_handle;
	e1000_irq_enable(dthis);
	uint32_t d_status = E1000_READ_REG(&dthis->hw,STATUS);
	EPRINT("device status 0x%lx\n", d_status);
	e1000_read_phy_reg(&dthis->hw, PHY_LP_ABILITY, &phy_labi);
	EPRINT("\nlink parter ability 0x%x\n", phy_labi);
	return 0;
}


static void
e1000_cleanup(void *dthis_v)
{
        struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v;
	e1000_irq_disable(dthis);
	dthis->link_speed = 0;
	dthis->link_duplex = 0;
	e1000_reset(dthis);
	e1000_clean_tx_ring(dthis);
	e1000_clean_rx_ring(dthis);
}

/* already enabled rx in e1000_enable(configure_rx), do nothing here, syt*/
static void
e1000_enable_receive(void *dthis_v)
{
	return;
}

/**
 * e1000_intr - Interrupt Handler
 * @irq: interrupt number
 * @data: pointer to a network interface device structure
 * @pt_regs: CPU registers structure
 **/
static void
e1000_interrupt(void *dthis_v)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v;
	struct e1000_hw *hw = &dthis->hw;
	uint32_t icr = E1000_READ_REG(&dthis->hw, ICR);
	int i;
	
	if(!icr)
		UserPrint("not our interrupt\n");

	if(icr & (E1000_ICR_RXSEQ | E1000_ICR_LSC)) {
		hw->get_link_status = 1;
		//printk("jiffies: %ld\n", jiffies);
		//mod_timer(&dthis->watchdog_timer, jiffies);
	}


	for(i = 0; i < E1000_MAX_INTR; i++)
		if(!e1000_clean_rx_irq(dthis) &
		   !e1000_clean_tx_irq(dthis))
			break;
#ifdef E1000_COUNT_ICR
	dthis->icr_txdw += icr & 0x01;
	icr >>= 1;
	dthis->icr_txqe += icr & 0x01;
	icr >>= 1;
	dthis->icr_lsc += icr & 0x01;
	icr >>= 1;
	dthis->icr_rxseq += icr & 0x01;
	icr >>= 1;
	dthis->icr_rxdmt += icr & 0x01;
	icr >>= 2;
	dthis->icr_rxo += icr & 0x01;
	icr >>= 1;
	dthis->icr_rxt += icr & 0x01;
	icr >>= 2;
	dthis->icr_mdac += icr & 0x01;
	icr >>= 1;
	dthis->icr_rxcfg += icr & 0x01;
	icr >>= 1;
	dthis->icr_gpi += icr & 0x01;
#endif
}

static int
drv_feedpacket(void *dthis_void, dma_addr_t buf_phys, size_t buf_length, 
	       int more, enum bfr_packet_type ptype, uintptr_t buf_key)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_void;
	struct e1000_desc_ring *tx_ring = &dthis->tx_ring;
	struct e1000_tx_desc *tx_desc = NULL;
	/* the E1000_TXD_STAT_DD bit is reset here for the card, syt*/
	uint32_t txd_upper = 0, txd_lower = E1000_TXD_CMD_IFCS;
	unsigned int i;
      
	lock_lock(&dthis->tx_lock);
	EPRINT("e1000 feed txpacket: buf_phys 0x%x, buf_keys %p\n", 
	           buf_phys, buf_key);
	/* the ptype passed is only one type: PACKET_NON, do nothing with this type*/
	/*  we may need to set transmit desc status according to different packet type, syt*/
	if(ptype & E1000_TX_FLAGS_TSO) {
		txd_lower |= E1000_TXD_CMD_DEXT | E1000_TXD_DTYP_D |
		             E1000_TXD_CMD_TSE;
		txd_upper |= (E1000_TXD_POPTS_IXSM | E1000_TXD_POPTS_TXSM) << 8;
	}
	if(ptype & E1000_TX_FLAGS_CSUM) {
		txd_lower |= E1000_TXD_CMD_DEXT | E1000_TXD_DTYP_D;
		txd_upper |= E1000_TXD_POPTS_TXSM << 8;
	}
	if(ptype & E1000_TX_FLAGS_VLAN) {
		txd_lower |= E1000_TXD_CMD_VLE;
		txd_upper |= (ptype & E1000_TX_FLAGS_VLAN_MASK);
	}

	if (buf_length > 0xffffU)
		UserPrint("TX length > 0xffff\n");
	i = tx_ring->next_to_use;
	tx_desc = E1000_TX_DESC(*tx_ring, i);
	tx_desc->buffer_addr = (uint64_t)buf_phys;
	tx_ring->desc_keys[i] = buf_key;  
	tx_desc->lower.data = txd_lower | buf_length;
	tx_desc->upper.data = txd_upper;
	/*what's the txd_cmd, syt*/
	tx_desc->lower.data |= dthis->txd_cmd;	
        if(++i == tx_ring->count) i = 0;
	tx_ring->next_to_use = i;
	wmb();
	E1000_WRITE_REG(&dthis->hw, TDT, i);
	lock_unlock(&dthis->tx_lock);
	
	return 0;
}
/*free only one desc one time, return 0: successfully; return 1: error. Could we free more one time?*/
/* should add desc inuse number and empty number statistic, but need to add this into desc structure. syt*/
static int
drv_freepacket(void *dthis_void, dma_addr_t buf_phys, size_t buf_size, uintptr_t key)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_void;
	struct e1000_desc_ring *rx_ring = &dthis->rx_ring;
	struct e1000_rx_desc *rx_desc;
	unsigned int i;
	EPRINT("e1000_driver: freepacket(dthis: %p)!!\n", dthis);
	
	lock_lock(&dthis->rx_lock);
	i = rx_ring->next_to_use;
	EPRINT("free %d rx desc\n", i);
	rx_desc = E1000_RX_DESC(*rx_ring, i);
	if(rx_desc->buffer_addr || rx_desc->status & E1000_RXD_STAT_DD)
	{
	  UserPrint("try to set a inused desc(buf_addr %p, desc_stat: 0x%x)\n!",
	             rx_desc->buffer_addr, rx_desc->status);
	  lock_unlock(&dthis->rx_lock);
	  return 1;
	};  
	 rx_desc->buffer_addr = (uint64_t)buf_phys;
	 dthis->rx_ring.desc_keys[i] = key;
	 EPRINT("e1000_driver: rdesc[%d], buff_addr(phy):0x%x \n", 
                     i, rx_desc->buffer_addr, key);
	wmb();
	E1000_WRITE_REG(&dthis->hw, RDT, i);
	/*  E1000_RXD_STAT_DD bit, 0 for card, 1 for driver, syt */ 
	if(++i == rx_ring->count) i = 0;
        rx_ring->next_to_use = i; 
	lock_unlock(&dthis->rx_lock);
	
	return 0;
}

static int
e1000_getmac(void *void_dthis, unsigned char *buf)
{
    int i;
    struct e1000_adapter *dthis = (struct e1000_adapter *) void_dthis;

    for(i = 0; i < 6; i++) {
	buf[i] = dthis->mac_address[i] = dthis->hw.mac_addr[i];
    }

    return 0;
}

static int setstatistics(void *dthis_v, struct netstats *nsp)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v;
	
	nsp = &dthis->net_stats;
	return 0;
}

/**
 * e1000_irq_disable - Mask off interrupt generation on the NIC
 * @adapter: board private structure
 **/

static inline void
e1000_irq_disable(struct e1000_adapter *dthis)
{
	E1000_WRITE_REG(&dthis->hw, IMC, ~0);
	E1000_WRITE_FLUSH(&dthis->hw);
}
/*
static void 
e1000_reset_phy(void *dthis_v)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v; 
	uint32_t manc;

	if(dthis->hw.mac_type >= e1000_82540 &&
	   dthis->hw.media_type == e1000_media_type_copper) {
		manc = E1000_READ_REG(&dthis->hw, MANC);
		if(manc & E1000_MANC_SMBUS_EN) {
			manc |= E1000_MANC_ARP_EN;
			E1000_WRITE_REG(&dthis->hw, MANC, manc);
		}
	}
	e1000_phy_hw_reset(&dthis->hw);
}*/
void
e1000_reset(struct e1000_adapter *dthis)
{	
	uint32_t manc;
	
	e1000_reset_hw(&dthis->hw);
	if(dthis->hw.mac_type >= e1000_82544)
		E1000_WRITE_REG(&dthis->hw, WUC, 0);
	if(e1000_init_hw(&dthis->hw))
		UserPrint("Hardware Error\n");
	e1000_reset_adaptive(&dthis->hw);
	e1000_phy_get_info(&dthis->hw, &dthis->phy_info);
	if(dthis->en_mng_pt) {
		manc = E1000_READ_REG(&dthis->hw, MANC);
		manc |= (E1000_MANC_ARP_EN | E1000_MANC_EN_MNG2HOST);
		E1000_WRITE_REG(&dthis->hw, MANC, manc);
	}
}
/**
 * e1000_setup_tx_resources - allocate Tx resources (Descriptors)
 * @adapter: board private structure
 *
 * Return 0 on success, negative on failure
 **/

int
e1000_setup_tx_resources(struct e1000_adapter *dthis)
{
	struct e1000_desc_ring *txdr = &dthis->tx_ring;
	struct e1000_tx_desc *tdesc;
	uint32_t pages_to_pin;
	int size, descsize = sizeof(struct e1000_tx_desc);
	int ndesc = txdr->count;
	int i;
	
	EPRINT("setup tx: tx_ring 0x%x, count: %d\n", txdr, txdr->count);
        /* desc_keys is to pass buffer info(structure) between interfaces, need to be*/
        /*allocated and zeroed, ndesc: the tx desc number, syt*/ 
	txdr->desc_keys = malloc(ndesc * sizeof (*(txdr->desc_keys)));
	size = sizeof(*(txdr->desc_keys)) * txdr->count;
	if(!txdr->desc_keys) {
		UserPrint(NAME ": Couldn't allocate tx ring memory!\n");
		return 1;
	}
	/* need to make sure the mem handle is right, syt*/
	memset(txdr->desc_keys, 0, size);
        /* allocated the tx ring, and zero out the space, syt */	
 	txdr->size = ndesc * sizeof(struct e1000_tx_desc);
	E1000_ROUNDUP(txdr->size, 4096);
	pages_to_pin = ((ndesc * descsize) + PAGESIZE -1)/PAGESIZE;
	if (pages_to_pin != 1)
	    UserPrint("e1000 tx ring desc page cnt != 1 unsupported.\n");
	txdr->desc = alloc_pci_consistent_memory(dthis->dma, txdr->size, 
	                                         &txdr->desc_phys);
	if(!txdr->desc) {
		UserPrint("couldn't alloc tx ring!\n");
		return 1;
	}	
	EPRINT("###setup tx: tx_desc: %p, desc_phys: %x", txdr->desc, txdr->desc_phys);
	memset(txdr->desc, 0, txdr->size);
	txdr->next_to_use = 0;
	txdr->next_to_clean = 0;

	return 0;
}
/**
 * e1000_configure_tx - Configure 8254x Transmit Unit after Reset
 * @adapter: board private structure
 *
 * Configure the Tx unit of the MAC after a reset.
 **/
static void
e1000_configure_tx(struct e1000_adapter *dthis)
{
	/*make sure the dma_addr_t dma type and meaning, if needs to be replaced */
	/*by mem_handle, syt*/
	EPRINT("tx_config: tx_count %d\n", dthis->tx_ring.count);
	uint64_t tdba = dthis->tx_ring.desc_phys;
	uint32_t tdlen = dthis->tx_ring.count * sizeof(struct e1000_tx_desc);
	uint32_t tctl, tipg;
	
	E1000_WRITE_REG(&dthis->hw, TDBAL, (tdba & 0x00000000ffffffffULL));
	E1000_WRITE_REG(&dthis->hw, TDBAH, (tdba >> 32));
	E1000_WRITE_REG(&dthis->hw, TDLEN, tdlen);
	/* Setup the HW Tx Head and Tail descriptor pointers */
	E1000_WRITE_REG(&dthis->hw, TDH, 0);
	E1000_WRITE_REG(&dthis->hw, TDT, 0);
	/* Set the default values for the Tx Inter Packet Gap timer */
	switch (dthis->hw.mac_type) {
	case e1000_82542_rev2_0:
	case e1000_82542_rev2_1:
		tipg = DEFAULT_82542_TIPG_IPGT;
		tipg |= DEFAULT_82542_TIPG_IPGR1 << E1000_TIPG_IPGR1_SHIFT;
		tipg |= DEFAULT_82542_TIPG_IPGR2 << E1000_TIPG_IPGR2_SHIFT;
		break;
	default:
		if(dthis->hw.media_type == e1000_media_type_fiber ||
		   dthis->hw.media_type == e1000_media_type_internal_serdes)
			tipg = DEFAULT_82543_TIPG_IPGT_FIBER;
		else
			tipg = DEFAULT_82543_TIPG_IPGT_COPPER;
		tipg |= DEFAULT_82543_TIPG_IPGR1 << E1000_TIPG_IPGR1_SHIFT;
		tipg |= DEFAULT_82543_TIPG_IPGR2 << E1000_TIPG_IPGR2_SHIFT;
	}
	E1000_WRITE_REG(&dthis->hw, TIPG, tipg);
	/* Set the Tx Interrupt Delay register */
	E1000_WRITE_REG(&dthis->hw, TIDV, dthis->tx_int_delay);
	if(dthis->hw.mac_type >= e1000_82540)
		E1000_WRITE_REG(&dthis->hw, TADV, dthis->tx_abs_int_delay);
	/* Program the Transmit Control Register */
	tctl = E1000_READ_REG(&dthis->hw, TCTL);
	tctl &= ~E1000_TCTL_CT;
	tctl |= E1000_TCTL_EN | E1000_TCTL_PSP |
		(E1000_COLLISION_THRESHOLD << E1000_CT_SHIFT);
	E1000_WRITE_REG(&dthis->hw, TCTL, tctl);
	e1000_config_collision_dist(&dthis->hw);
	/* Setup Transmit Descriptor Settings for eop descriptor */
	dthis->txd_cmd = E1000_TXD_CMD_IDE | E1000_TXD_CMD_EOP |
		E1000_TXD_CMD_IFCS;
	if(dthis->hw.mac_type < e1000_82543)
		dthis->txd_cmd |= E1000_TXD_CMD_RPS;
	else
		dthis->txd_cmd |= E1000_TXD_CMD_RS;
	/* Cache if we're 82544 running in PCI-X because we'll
	 * need this to apply a workaround later in the send path. */
	if(dthis->hw.mac_type == e1000_82544 &&
	   dthis->hw.bus_type == e1000_bus_type_pcix)
		dthis->pcix_82544 = 1;
}
/**
 * e1000_setup_rx_resources - allocate Rx resources (Descriptors)
 * @adapter: board private structure
 *
 * Returns 0 on success, negative on failure
 **/

int
e1000_setup_rx_resources(struct e1000_adapter *dthis)
{
	struct e1000_desc_ring *rxdr = &dthis->rx_ring;
	struct e1000_rx_desc *rdesc;
	uint32_t pages_to_pin;
	int size, descsize = sizeof(struct e1000_rx_desc);
	int ndesc = rxdr->count;
	int i, i_next;
	
	EPRINT("setup rx, rx_ring %p  count: %d\n", rxdr, ndesc);
       /* desc_keys is to pass buffer info(structure) between interfaces, need to be*/
        /*allocated and zeroed, ndesc: the rx desc number, syt*/ 
	rxdr->desc_keys = malloc(ndesc * sizeof (*(rxdr->desc_keys)));
	size = sizeof(*(rxdr->desc_keys)) * rxdr->count;
	if(!rxdr->desc_keys) {
		UserPrint(NAME ": Couldn't allocate rx ring memory!\n");
		return 1;
	}
	memset(rxdr->desc_keys, 0, size);
        /* allocated the rx ring, and zero out the space, syt */	
 	rxdr->size = ndesc * sizeof(struct e1000_rx_desc);
	E1000_ROUNDUP(rxdr->size, 4096);
	pages_to_pin = ((ndesc * descsize) + PAGESIZE -1)/PAGESIZE;
	if (pages_to_pin != 1)
	    UserPrint("e1000 rx ring desc page cnt != 1 unsupported.\n");
	rxdr->desc = alloc_pci_consistent_memory(dthis->dma, rxdr->size, 
	                                         &rxdr->desc_phys);
	if(!rxdr->desc) {
		UserPrint("couldn't alloc rx ring!\n");
		return 1;
	}	
	memset(rxdr->desc, 0, rxdr->size);
#if 0	
	for (i=0; i < ndesc; i++)
	{
		rdesc = E1000_RX_DESC(*rxdr, i);
		UserPrint("rdesc[%d].buff_addr: 0x%x, status: 0x%x", 
		           rdesc->buffer_addr, rdesc->status);
	}

#endif	
	rxdr->next_to_clean = 0;
	rxdr->next_to_use = 0;
	return 0;
}

/**
 * e1000_configure_rx - Configure 8254x Receive Unit after Reset
 * @adapter: board private structure
 *
 * Configure the Rx unit of the MAC after a reset.
 **/

static void
e1000_configure_rx(struct e1000_adapter *dthis)
{
        /*find the right type and member. syt*/
	uint64_t rdba = dthis->rx_ring.desc_phys;
	uint32_t rdlen = dthis->rx_ring.count * sizeof(struct e1000_rx_desc);
	uint32_t rctl;
	uint32_t rxcsum;

	/* disable receives while setting up the descriptors */
	rctl = E1000_READ_REG(&dthis->hw, RCTL);
	E1000_WRITE_REG(&dthis->hw, RCTL, rctl & ~E1000_RCTL_EN);
	/* set the Receive Delay Timer Register */
	E1000_WRITE_REG(&dthis->hw, RDTR, dthis->rx_int_delay);
	if(dthis->hw.mac_type >= e1000_82540) {
		E1000_WRITE_REG(&dthis->hw, RADV, dthis->rx_abs_int_delay);//dthis->rx_abs_int_delay);
		if(dthis->itr > 1)
			E1000_WRITE_REG(&dthis->hw, ITR,
				1000000000 / (dthis->itr * 256));
	}
	/* Setup the Base and Length of the Rx Descriptor Ring */
	E1000_WRITE_REG(&dthis->hw, RDBAL, (rdba & 0x00000000ffffffffULL));
	E1000_WRITE_REG(&dthis->hw, RDBAH, (rdba >> 32));
	E1000_WRITE_REG(&dthis->hw, RDLEN, rdlen);
	/* Setup the HW Rx Head and Tail Descriptor Pointers */
	E1000_WRITE_REG(&dthis->hw, RDH, 0);
	E1000_WRITE_REG(&dthis->hw, RDT, 0);
	/* Enable 82543 Receive Checksum Offload for TCP and UDP */
	if((dthis->hw.mac_type >= e1000_82543) &&
	   (dthis->rx_csum == TRUE)) {
		rxcsum = E1000_READ_REG(&dthis->hw, RXCSUM);
		rxcsum |= E1000_RXCSUM_TUOFL;
		E1000_WRITE_REG(&dthis->hw, RXCSUM, rxcsum);
	}
	/* Enable Receives */
	E1000_WRITE_REG(&dthis->hw, RCTL, rctl);

}

int
e1000_set_spd_dplx(void *dthis_v, uint16_t spddplx)
{
	struct e1000_adapter *dthis = (struct e1000_adapter *)dthis_v;
	
	dthis->hw.autoneg = 0;
	switch(spddplx) {
	case HALF_DUPLEX + SPEED_10:
		dthis->hw.forced_speed_duplex = e1000_10_half;
		break;
	case FULL_DUPLEX + SPEED_10:
		dthis->hw.forced_speed_duplex = e1000_10_full;
		break;
	case SPEED_100 + HALF_DUPLEX:
		dthis->hw.forced_speed_duplex = e1000_100_half;
		break;
	case FULL_DUPLEX + SPEED_100:
		dthis->hw.forced_speed_duplex = e1000_100_full;
		break;
	case SPEED_1000 + FULL_DUPLEX:
		dthis->hw.autoneg = 1;
		dthis->hw.autoneg_advertised = ADVERTISE_1000_FULL;
		break;
	case HALF_DUPLEX + SPEED_1000: /* not supported */
	default:
		UserPrint("Unsupported Speed/Duplexity configuration\n");
		return 1;
	}
	return 0;
}

/* for tx&rx descriptors ring allocation, syt*/
static int
e1000_init_desc_ring(struct e1000_adapter *dthis)
{
	int err;

	/* allocate transmit descriptors */
	if((err = e1000_setup_tx_resources(dthis)))
		goto err_setup_tx;
	/* allocate receive descriptors */
	if((err = e1000_setup_rx_resources(dthis)))
		goto err_setup_rx;

	return 0;
err_setup_rx:
	e1000_clean_rx_ring(dthis);
err_setup_tx:
	e1000_clean_tx_ring(dthis);

	return err;
}
/**
 * e1000_irq_enable - Enable default interrupt generation settings
 * @adapter: board private structure
 **/
static inline void
e1000_irq_enable(struct e1000_adapter *dthis)
{
	E1000_WRITE_REG(&dthis->hw, IMS, IMS_ENABLE_MASK);
	E1000_WRITE_FLUSH(&dthis->hw);
	
}
/**
 * e1000_clean_tx_ring 
 * @adapter: board private structure
 **/
static void
e1000_clean_tx_ring(struct e1000_adapter *dthis)
{
	struct e1000_desc_ring *tx_ring = &dthis->tx_ring;
	unsigned long size;
	unsigned int i;

        UserPrint("teardown tx dma\n");
	teardown_dma(dthis->dma, &tx_ring->dma_handle);
	/* Zero out the descriptor ring */
	memset(tx_ring->desc, 0, tx_ring->size);
	tx_ring->next_to_use = 0;
	tx_ring->next_to_clean = 0;
	dthis->tx_ring.desc = NULL;
	E1000_WRITE_REG(&dthis->hw, TDH, 0);
	E1000_WRITE_REG(&dthis->hw, TDT, 0);
}

/**
 * e1000_clean_rx_ring 
 * @adapter: board private structure
 **/

static void
e1000_clean_rx_ring(struct e1000_adapter *dthis)
{
	struct e1000_desc_ring *rx_ring = &dthis->rx_ring;
	struct e1000_buffer *buffer_info;
	unsigned long size;
	unsigned int i;

	UserPrint("teardown rx dma\n");
	teardown_dma(dthis->dma, &rx_ring->dma_handle);
        /* Zero out the descriptor ring */
	memset(rx_ring->desc, 0, rx_ring->size);
	rx_ring->next_to_clean = 0;
	rx_ring->next_to_use = 0;
	rx_ring->desc = NULL;
	E1000_WRITE_REG(&dthis->hw, RDH, 0);
	E1000_WRITE_REG(&dthis->hw, RDT, 0);
}
/**
 * e1000_setup_rctl - configure the receive control register
 * @adapter: Board private structure
 **/
static void
e1000_setup_rctl(struct e1000_adapter *dthis)
{
	uint32_t rctl;

	rctl = E1000_READ_REG(&dthis->hw, RCTL);

	rctl &= ~(3 << E1000_RCTL_MO_SHIFT);

	rctl |= E1000_RCTL_EN | E1000_RCTL_BAM  
	        | E1000_RCTL_LBM_NO | E1000_RCTL_RDMTS_HALF |
		(dthis->hw.mc_filter_type << E1000_RCTL_MO_SHIFT);

	if(dthis->hw.tbi_compatibility_on == 1)
		rctl |= E1000_RCTL_SBP;
	else
		rctl &= ~E1000_RCTL_SBP;

	/* Setup buffer sizes */
        rctl &= ~(E1000_RCTL_SZ_4096);
	switch (dthis->rx_buffer_len) {
	case E1000_RXBUFFER_2048:
	default:
		rctl |= E1000_RCTL_SZ_2048;
		rctl &= ~(E1000_RCTL_BSEX | E1000_RCTL_LPE);
		break;
	case E1000_RXBUFFER_4096:
		rctl |= E1000_RCTL_SZ_4096 | E1000_RCTL_BSEX | E1000_RCTL_LPE;
		break;
	case E1000_RXBUFFER_8192:
		rctl |= E1000_RCTL_SZ_8192;
		break;
	case E1000_RXBUFFER_16384:
		rctl |= E1000_RCTL_SZ_16384;
		break;
	}
        EPRINT("\nset rctl 0x%lx\n", rctl);
	E1000_WRITE_REG(&dthis->hw, RCTL, rctl);
}

#if 0
/*should move this function to module: e1000_mod.c, syt*/
/**
 * e1000_set_multi - Multicast and Promiscuous mode set
 * @netdev: network interface device structure
 *
 * The set_multi entry point is called whenever the multicast address
 * list or the network interface flags are updated.  This routine is
 * responsible for configuring the hardware for proper multicast,
 * promiscuous mode, and all-multi behavior.
 **/

static void
e1000_set_multi(struct net_device *netdev)
{
	struct e1000_adapter *dthis = netdev->priv;
	struct dev_mc_list *mc_ptr;
	uint32_t rctl;
	uint32_t hash_value;
	int i;

	/* Check for Promiscuous and All Multicast modes */

	rctl = E1000_READ_REG(&dthis->hw, RCTL);

	if(netdev->flags & IFF_PROMISC) {
		rctl |= (E1000_RCTL_UPE | E1000_RCTL_MPE);
	} else if(netdev->flags & IFF_ALLMULTI) {
		rctl |= E1000_RCTL_MPE;
		rctl &= ~E1000_RCTL_UPE;
	} else {
		rctl &= ~(E1000_RCTL_UPE | E1000_RCTL_MPE);
	}

	E1000_WRITE_REG(&dthis->hw, RCTL, rctl);

	/* 82542 2.0 needs to be in reset to write receive address registers */

	if(dthis->hw.mac_type == e1000_82542_rev2_0)
		e1000_enter_82542_rst(dthis);

	/* load the first 14 multicast address into the exact filters 1-14
	 * RAR 0 is used for the station MAC adddress
	 * if there are not 14 addresses, go ahead and clear the filters
	 */
	mc_ptr = netdev->mc_list;

	for(i = 1; i < E1000_RAR_ENTRIES; i++) {
		if(mc_ptr) {
			e1000_rar_set(&dthis->hw, mc_ptr->dmi_addr, i);
			mc_ptr = mc_ptr->next;
		} else {
			E1000_WRITE_REG_ARRAY(&dthis->hw, RA, i << 1, 0);
			E1000_WRITE_REG_ARRAY(&dthis->hw, RA, (i << 1) + 1, 0);
		}
	}

	/* clear the old settings from the multicast hash table */

	for(i = 0; i < E1000_NUM_MTA_REGISTERS; i++)
		E1000_WRITE_REG_ARRAY(&dthis->hw, MTA, i, 0);

	/* load any remaining addresses into the hash table */

	for(; mc_ptr; mc_ptr = mc_ptr->next) {
		hash_value = e1000_hash_mc_addr(&dthis->hw, mc_ptr->dmi_addr);
		e1000_mta_set(&dthis->hw, hash_value);
	}

	if(dthis->hw.mac_type == e1000_82542_rev2_0)
		e1000_leave_82542_rst(dthis);
}
#endif


/**
 * e1000_clean_rx_irq - Send received data up the network stack
 * @adapter: board private structure
 **/
/*syt, need to make clealy the machanism the desc use in e1000, then the desc
structure is also changed*/
static int
e1000_clean_rx_irq(struct e1000_adapter *dthis)
{       
        struct e1000_desc_ring *rx_ring = &dthis->rx_ring;
	struct e1000_rx_desc *rx_desc, *rx;
	unsigned long flags;
	uint32_t length;
	unsigned int i, j=0, n=0;
	
	lock_lock(&dthis->rx_lock);
	i = rx_ring->next_to_clean;
	EPRINT("###rx_int: ring %p desc%d\n", rx_ring, i);
	rx_desc = E1000_RX_DESC(*rx_ring, i);
	flags = rx_desc->status & E1000_RXD_STAT_DD;
	EPRINT("rx_int: status 0x%x\n", rx_desc->status);

	while(rx_desc->status & E1000_RXD_STAT_DD) {
	
#ifdef CONFIG_E1000_NAPI
		if(*work_done >= work_to_do)
			break;
		(*work_done)++;
#endif
		rx_desc->status = 0;
		rx_desc->buffer_addr = 0;
		length = rx_desc->length;
	       	bfr_feedpacket(dthis->buf_man_handle, rx_desc->buffer_addr, 
           		       length - ETHERNET_FCS_SIZE, dthis->rx_ring.desc_keys[i]);
		if(++i == rx_ring->count) i = 0;
		rx_desc = E1000_RX_DESC(*rx_ring, i);
	}
	rx_ring->next_to_clean = i;
	lock_unlock(&dthis->rx_lock);
	
	return 0;
}

/**
 * e1000_clean_tx_irq - Reclaim resources after transmit completes
 * @adapter: board private structure
 **/
/* use bfr_freepacket, make sure the machanism first*/
static int
e1000_clean_tx_irq(struct e1000_adapter *dthis)
{
	struct e1000_desc_ring *tx_ring = &dthis->tx_ring;
	struct e1000_tx_desc *tx_desc;
	unsigned int i;

	lock_lock(&dthis->tx_lock);
	i = tx_ring->next_to_clean;
	tx_desc = E1000_TX_DESC(*tx_ring, i);
	EPRINT("tx_int: ring %p desc%d, status: 0x%x\n", tx_ring, i, tx_desc->upper.data);
	while((tx_desc->upper.data & E1000_TXD_STAT_DD) && tx_desc->buffer_addr)
	{
		bfr_freepacket(dthis->buf_man_handle,
		               tx_desc->buffer_addr,
		               dthis->tx_ring.desc_keys[i]);
		tx_desc->buffer_addr = 0;
		tx_desc->lower.data = 0;
		tx_desc->upper.data = 0;
		if(++i == tx_ring->count) i = 0;
		tx_desc = E1000_TX_DESC(*tx_ring, i);
	}
        tx_ring->next_to_clean = i;
        lock_unlock(&dthis->tx_lock);
	
	return 0;
}


/**
 * e1000_sw_init - Initialize general software structures (struct e1000_adapter)
 * @adapter: board private structure to initialize
 *
 * e1000_sw_init initializes the Adapter private data structure.
 * Fields are initialized based on PCI device information and
 * OS network device settings (MTU size).
 **/

static int 
e1000_sw_init(struct e1000_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	
	/* identify the MAC */
        adapter->rx_buffer_len = E1000_RXBUFFER_2048;
	if (e1000_set_mac_type(hw)) {
		UserPrint("Unknown MAC Type\n");
		return 1;
	}
	/* initialize eeprom parameters */
	e1000_init_eeprom_params(hw);

	if((hw->mac_type == e1000_82541) ||
	   (hw->mac_type == e1000_82547) ||
	   (hw->mac_type == e1000_82541_rev_2) ||
	   (hw->mac_type == e1000_82547_rev_2))
		hw->phy_init_script = 1;
	EPRINT("set media type\n");
	e1000_set_media_type(hw);

	if(hw->mac_type < e1000_82543)
		hw->report_tx_early = 0;
	else
		hw->report_tx_early = 1;

	hw->wait_autoneg_complete = FALSE;
	hw->tbi_compatibility_en = TRUE;
	hw->adaptive_ifs = TRUE;

	/* Copper options */

	if(hw->media_type == e1000_media_type_copper) {
		hw->mdix = AUTO_ALL_MODES;
		hw->disable_polarity_correction = FALSE;
		hw->master_slave = E1000_MASTER_SLAVE;
	}

	return 0;
}

/**
 * e1000_smartspeed - Workaround for SmartSpeed on 82541 and 82547 controllers.
 * @adapter:
 **/

static void
e1000_smartspeed(void *dthis_v)
{
	struct e1000_adapter *adapter = (struct e1000_adapter *)dthis_v;
	uint16_t phy_status;
	uint16_t phy_ctrl;

	if((adapter->hw.phy_type != e1000_phy_igp) || !adapter->hw.autoneg ||
	   !(adapter->hw.autoneg_advertised & ADVERTISE_1000_FULL))
		return;

	if(adapter->smartspeed == 0) {
		/* If Master/Slave config fault is asserted twice,
		 * we assume back-to-back */
		e1000_read_phy_reg(&adapter->hw, PHY_1000T_STATUS, &phy_status);
		if(!(phy_status & SR_1000T_MS_CONFIG_FAULT)) return;
		e1000_read_phy_reg(&adapter->hw, PHY_1000T_STATUS, &phy_status);
		if(!(phy_status & SR_1000T_MS_CONFIG_FAULT)) return;
		e1000_read_phy_reg(&adapter->hw, PHY_1000T_CTRL, &phy_ctrl);
		if(phy_ctrl & CR_1000T_MS_ENABLE) {
			phy_ctrl &= ~CR_1000T_MS_ENABLE;
			e1000_write_phy_reg(&adapter->hw, PHY_1000T_CTRL,
					    phy_ctrl);
			adapter->smartspeed++;
			if(!e1000_phy_setup_autoneg(&adapter->hw) &&
			   !e1000_read_phy_reg(&adapter->hw, PHY_CTRL,
				   	       &phy_ctrl)) {
				phy_ctrl |= (MII_CR_AUTO_NEG_EN |
					     MII_CR_RESTART_AUTO_NEG);
				e1000_write_phy_reg(&adapter->hw, PHY_CTRL,
						    phy_ctrl);
			}
		}
		return;
	} else if(adapter->smartspeed == E1000_SMARTSPEED_DOWNSHIFT) {
		/* If still no link, perhaps using 2/3 pair cable */
		e1000_read_phy_reg(&adapter->hw, PHY_1000T_CTRL, &phy_ctrl);
		phy_ctrl |= CR_1000T_MS_ENABLE;
		e1000_write_phy_reg(&adapter->hw, PHY_1000T_CTRL, phy_ctrl);
		if(!e1000_phy_setup_autoneg(&adapter->hw) &&
		   !e1000_read_phy_reg(&adapter->hw, PHY_CTRL, &phy_ctrl)) {
			phy_ctrl |= (MII_CR_AUTO_NEG_EN |
				     MII_CR_RESTART_AUTO_NEG);
			e1000_write_phy_reg(&adapter->hw, PHY_CTRL, phy_ctrl);
		}
	}
	/* Restart process after E1000_SMARTSPEED_MAX iterations */
	if(adapter->smartspeed++ == E1000_SMARTSPEED_MAX)
		adapter->smartspeed = 0;
}

/**
 * e1000_update_stats - Update the board statistics counters
 * @adapter: board private structure
 **/

void
e1000_update_stats(struct e1000_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	unsigned long flags;
	uint16_t phy_tmp;

#define PHY_IDLE_ERROR_COUNT_MASK 0x00FF

	spin_lock_irqsave(&adapter->stats_lock, flags);

	/* these counters are modified from e1000_adjust_tbi_stats,
	 * called from the interrupt context, so they must only
	 * be written while holding adapter->stats_lock
	 */

	adapter->stats.crcerrs += E1000_READ_REG(hw, CRCERRS);
	adapter->stats.gprc += E1000_READ_REG(hw, GPRC);
	adapter->stats.gorcl += E1000_READ_REG(hw, GORCL);
	adapter->stats.gorch += E1000_READ_REG(hw, GORCH);
	adapter->stats.bprc += E1000_READ_REG(hw, BPRC);
	adapter->stats.mprc += E1000_READ_REG(hw, MPRC);
	adapter->stats.roc += E1000_READ_REG(hw, ROC);
	adapter->stats.prc64 += E1000_READ_REG(hw, PRC64);
	adapter->stats.prc127 += E1000_READ_REG(hw, PRC127);
	adapter->stats.prc255 += E1000_READ_REG(hw, PRC255);
	adapter->stats.prc511 += E1000_READ_REG(hw, PRC511);
	adapter->stats.prc1023 += E1000_READ_REG(hw, PRC1023);
	adapter->stats.prc1522 += E1000_READ_REG(hw, PRC1522);

	/* the rest of the counters are only modified here */

	adapter->stats.symerrs += E1000_READ_REG(hw, SYMERRS);
	adapter->stats.mpc += E1000_READ_REG(hw, MPC);
	adapter->stats.scc += E1000_READ_REG(hw, SCC);
	adapter->stats.ecol += E1000_READ_REG(hw, ECOL);
	adapter->stats.mcc += E1000_READ_REG(hw, MCC);
	adapter->stats.latecol += E1000_READ_REG(hw, LATECOL);
	adapter->stats.dc += E1000_READ_REG(hw, DC);
	adapter->stats.sec += E1000_READ_REG(hw, SEC);
	adapter->stats.rlec += E1000_READ_REG(hw, RLEC);
	adapter->stats.xonrxc += E1000_READ_REG(hw, XONRXC);
	adapter->stats.xontxc += E1000_READ_REG(hw, XONTXC);
	adapter->stats.xoffrxc += E1000_READ_REG(hw, XOFFRXC);
	adapter->stats.xofftxc += E1000_READ_REG(hw, XOFFTXC);
	adapter->stats.fcruc += E1000_READ_REG(hw, FCRUC);
	adapter->stats.gptc += E1000_READ_REG(hw, GPTC);
	adapter->stats.gotcl += E1000_READ_REG(hw, GOTCL);
	adapter->stats.gotch += E1000_READ_REG(hw, GOTCH);
	adapter->stats.rnbc += E1000_READ_REG(hw, RNBC);
	adapter->stats.ruc += E1000_READ_REG(hw, RUC);
	adapter->stats.rfc += E1000_READ_REG(hw, RFC);
	adapter->stats.rjc += E1000_READ_REG(hw, RJC);
	adapter->stats.torl += E1000_READ_REG(hw, TORL);
	adapter->stats.torh += E1000_READ_REG(hw, TORH);
	adapter->stats.totl += E1000_READ_REG(hw, TOTL);
	adapter->stats.toth += E1000_READ_REG(hw, TOTH);
	adapter->stats.tpr += E1000_READ_REG(hw, TPR);
	adapter->stats.ptc64 += E1000_READ_REG(hw, PTC64);
	adapter->stats.ptc127 += E1000_READ_REG(hw, PTC127);
	adapter->stats.ptc255 += E1000_READ_REG(hw, PTC255);
	adapter->stats.ptc511 += E1000_READ_REG(hw, PTC511);
	adapter->stats.ptc1023 += E1000_READ_REG(hw, PTC1023);
	adapter->stats.ptc1522 += E1000_READ_REG(hw, PTC1522);
	adapter->stats.mptc += E1000_READ_REG(hw, MPTC);
	adapter->stats.bptc += E1000_READ_REG(hw, BPTC);

	/* used for adaptive IFS */

	hw->tx_packet_delta = E1000_READ_REG(hw, TPT);
	adapter->stats.tpt += hw->tx_packet_delta;
	hw->collision_delta = E1000_READ_REG(hw, COLC);
	adapter->stats.colc += hw->collision_delta;

	if(hw->mac_type >= e1000_82543) {
		adapter->stats.algnerrc += E1000_READ_REG(hw, ALGNERRC);
		adapter->stats.rxerrc += E1000_READ_REG(hw, RXERRC);
		adapter->stats.tncrs += E1000_READ_REG(hw, TNCRS);
		adapter->stats.cexterr += E1000_READ_REG(hw, CEXTERR);
		adapter->stats.tsctc += E1000_READ_REG(hw, TSCTC);
		adapter->stats.tsctfc += E1000_READ_REG(hw, TSCTFC);
	}

	/* Fill out the OS statistics structure */

	adapter->net_stats.rx_packets = adapter->stats.gprc;
	adapter->net_stats.tx_packets = adapter->stats.gptc;
	adapter->net_stats.rx_bytes = adapter->stats.gorcl;
	adapter->net_stats.tx_bytes = adapter->stats.gotcl;
	adapter->net_stats.multicast = adapter->stats.mprc;
	adapter->net_stats.collisions = adapter->stats.colc;

	/* Rx Errors */

	adapter->net_stats.rx_errors = adapter->stats.rxerrc +
		adapter->stats.crcerrs + adapter->stats.algnerrc +
		adapter->stats.rlec + adapter->stats.rnbc +
		adapter->stats.mpc + adapter->stats.cexterr;
	adapter->net_stats.rx_dropped = adapter->stats.rnbc;
	adapter->net_stats.rx_length_errors = adapter->stats.rlec;
	adapter->net_stats.rx_crc_errors = adapter->stats.crcerrs;
	adapter->net_stats.rx_frame_errors = adapter->stats.algnerrc;
	adapter->net_stats.rx_fifo_errors = adapter->stats.mpc;
	adapter->net_stats.rx_missed_errors = adapter->stats.mpc;

	/* Tx Errors */

	adapter->net_stats.tx_errors = adapter->stats.ecol +
	                               adapter->stats.latecol;
	adapter->net_stats.tx_aborted_errors = adapter->stats.ecol;
	adapter->net_stats.tx_window_errors = adapter->stats.latecol;
	adapter->net_stats.tx_carrier_errors = adapter->stats.tncrs;

	/* Tx Dropped needs to be maintained elsewhere */

	/* Phy Stats */

	if(hw->media_type == e1000_media_type_copper) {
		if((adapter->link_speed == SPEED_1000) &&
		   (!e1000_read_phy_reg(hw, PHY_1000T_STATUS, &phy_tmp))) {
			phy_tmp &= PHY_IDLE_ERROR_COUNT_MASK;
			adapter->phy_stats.idle_errors += phy_tmp;
		}

		if((hw->mac_type <= e1000_82546) &&
		   (hw->phy_type == e1000_phy_m88) &&
		   !e1000_read_phy_reg(hw, M88E1000_RX_ERR_CNTR, &phy_tmp))
			adapter->phy_stats.receive_errors += phy_tmp;
	}

	spin_unlock_irqrestore(&adapter->stats_lock, flags);
}

static struct netif_ops nops =
{
	.setup = e1000_setup,
	.enable = enable,
	.cleanup = e1000_cleanup,
	.enable_receive = e1000_enable_receive,
	.interrupt = e1000_interrupt, 
	.feedpacket = drv_feedpacket,
	.freepacket = drv_freepacket,
	.macaddr = e1000_getmac,
	.setstatistics = setstatistics,
};

struct driver *e1000_init(void) {
	EPRINT("e1000 init\n");
	e1000.n_ops = &nops;
	return &e1000;
}

/* e1000_main.c */
