#ifndef _UART16550_H_
#define _UART16550_H_

#include "uart16550_types.reg.h"

#include <io/io.h>

static inline union data read_data(bus_space_t bus) {
	union data foo;
	foo.data = io_read_8(bus, 0);
	return foo;
}

#define READ_DATA() read_data(dthis->csr)

static inline void write_data(bus_space_t bus, union data foo) {
	return io_write_8 (bus, (0), foo.data);
}

#define WRITE_DATA(foo) write_data(dthis->csr, foo)

static inline union ier read_ier(bus_space_t bus) {
	union ier foo;
	foo.data = io_read_8(bus, 1);
	return foo;
}

#define READ_IER() read_ier(dthis->csr)

static inline void write_ier(bus_space_t bus, union ier foo) {
	return io_write_8 (bus, (1), foo.data);
}

#define WRITE_IER(foo) write_ier(dthis->csr, foo)

static inline void write_fcr(bus_space_t bus, union fcr foo) {
	return io_write_8 (bus, (2), foo.data);
}

#define WRITE_FCR(foo) write_fcr(dthis->csr, foo)

static inline union iir read_iir(bus_space_t bus) {
	union iir foo;
	foo.data = io_read_8(bus, 2);
	return foo;
}

#define READ_IIR() read_iir(dthis->csr)

static inline union lcr read_lcr(bus_space_t bus) {
	union lcr foo;
	foo.data = io_read_8(bus, 3);
	return foo;
}

#define READ_LCR() read_lcr(dthis->csr)

static inline void write_lcr(bus_space_t bus, union lcr foo) {
	return io_write_8 (bus, (3), foo.data);
}

#define WRITE_LCR(foo) write_lcr(dthis->csr, foo)

static inline union mcr read_mcr(bus_space_t bus) {
	union mcr foo;
	foo.data = io_read_8(bus, 4);
	return foo;
}

#define READ_MCR() read_mcr(dthis->csr)

static inline void write_mcr(bus_space_t bus, union mcr foo) {
	return io_write_8 (bus, (4), foo.data);
}

#define WRITE_MCR(foo) write_mcr(dthis->csr, foo)

static inline union lsr read_lsr(bus_space_t bus) {
	union lsr foo;
	foo.data = io_read_8(bus, 5);
	return foo;
}

#define READ_LSR() read_lsr(dthis->csr)

static inline void write_lsr(bus_space_t bus, union lsr foo) {
	return io_write_8 (bus, (5), foo.data);
}

#define WRITE_LSR(foo) write_lsr(dthis->csr, foo)

static inline union msr read_msr(bus_space_t bus) {
	union msr foo;
	foo.data = io_read_8(bus, 6);
	return foo;
}

#define READ_MSR() read_msr(dthis->csr)

static inline void write_msr(bus_space_t bus, union msr foo) {
	return io_write_8 (bus, (6), foo.data);
}

#define WRITE_MSR(foo) write_msr(dthis->csr, foo)

static inline union dll read_dll(bus_space_t bus) {
	union dll foo;
	foo.data = io_read_8(bus, 0);
	return foo;
}

#define READ_DLL() read_dll(dthis->csr)

static inline void write_dll(bus_space_t bus, union dll foo) {
	return io_write_8 (bus, (0), foo.data);
}

#define WRITE_DLL(foo) write_dll(dthis->csr, foo)

static inline union dlh read_dlh(bus_space_t bus) {
	union dlh foo;
	foo.data = io_read_8(bus, 1);
	return foo;
}

#define READ_DLH() read_dlh(dthis->csr)

static inline void write_dlh(bus_space_t bus, union dlh foo) {
	return io_write_8 (bus, (1), foo.data);
}

#define WRITE_DLH(foo) write_dlh(dthis->csr, foo)

static inline void print_data ( union data reg ) {	UserPrint("Debug data - 0x%x\n", reg.data);}

static inline void print_ier ( union ier reg ) {	UserPrint("Debug ier - 0x%x\n", reg.data);	if (reg.r.rx_int != 0) {
		UserPrint("\trx_int: %x\n", reg.r.rx_int);
	}
	if (reg.r.tx_int != 0) {
		UserPrint("\ttx_int: %x\n", reg.r.tx_int);
	}
	if (reg.r.line_status_int != 0) {
		UserPrint("\tline_status_int: %x\n", reg.r.line_status_int);
	}
	if (reg.r.modem_status_int != 0) {
		UserPrint("\tmodem_status_int: %x\n", reg.r.modem_status_int);
	}
}

static inline void print_fcr ( union fcr reg ) {	UserPrint("Debug fcr - 0x%x\n", reg.data);	if (reg.r.fifo_enable != 0) {
		UserPrint("\tfifo_enable: %x\n", reg.r.fifo_enable);
	}
	if (reg.r.rx_fifo_clear != 0) {
		UserPrint("\trx_fifo_clear: %x\n", reg.r.rx_fifo_clear);
	}
	if (reg.r.tx_fifo_clear != 0) {
		UserPrint("\ttx_fifo_clear: %x\n", reg.r.tx_fifo_clear);
	}
	if (reg.r.trigger_level != 0) {
		UserPrint("\ttrigger_level: %x\n", reg.r.trigger_level);
	}
}

static inline void print_iir ( union iir reg ) {	UserPrint("Debug iir - 0x%x\n", reg.data);	if (reg.r.npending != 0) {
		UserPrint("\tnpending: %x\n", reg.r.npending);
	}
	if (reg.r.source != 0) {
		UserPrint("\tsource: %x\n", reg.r.source);
	}
	if (reg.r.timeout != 0) {
		UserPrint("\ttimeout: %x\n", reg.r.timeout);
	}
	if (reg.r.trigger_level != 0) {
		UserPrint("\ttrigger_level: %x\n", reg.r.trigger_level);
	}
}

static inline void print_lcr ( union lcr reg ) {	UserPrint("Debug lcr - 0x%x\n", reg.data);	if (reg.r.data_bits != 0) {
		UserPrint("\tdata_bits: %x\n", reg.r.data_bits);
	}
	if (reg.r.stop_bits != 0) {
		UserPrint("\tstop_bits: %x\n", reg.r.stop_bits);
	}
	if (reg.r.parity_enable != 0) {
		UserPrint("\tparity_enable: %x\n", reg.r.parity_enable);
	}
	if (reg.r.even_parity != 0) {
		UserPrint("\teven_parity: %x\n", reg.r.even_parity);
	}
	if (reg.r.stick_parity != 0) {
		UserPrint("\tstick_parity: %x\n", reg.r.stick_parity);
	}
	if (reg.r.brk != 0) {
		UserPrint("\tbrk: %x\n", reg.r.brk);
	}
	if (reg.r.dlab != 0) {
		UserPrint("\tdlab: %x\n", reg.r.dlab);
	}
}

static inline void print_mcr ( union mcr reg ) {	UserPrint("Debug mcr - 0x%x\n", reg.data);	if (reg.r.ndtr != 0) {
		UserPrint("\tndtr: %x\n", reg.r.ndtr);
	}
	if (reg.r.nrts != 0) {
		UserPrint("\tnrts: %x\n", reg.r.nrts);
	}
	if (reg.r.out1 != 0) {
		UserPrint("\tout1: %x\n", reg.r.out1);
	}
	if (reg.r.out2 != 0) {
		UserPrint("\tout2: %x\n", reg.r.out2);
	}
	if (reg.r.test != 0) {
		UserPrint("\ttest: %x\n", reg.r.test);
	}
}

static inline void print_lsr ( union lsr reg ) {	UserPrint("Debug lsr - 0x%x\n", reg.data);	if (reg.r.rx_avail != 0) {
		UserPrint("\trx_avail: %x\n", reg.r.rx_avail);
	}
	if (reg.r.overrun != 0) {
		UserPrint("\toverrun: %x\n", reg.r.overrun);
	}
	if (reg.r.parity_err != 0) {
		UserPrint("\tparity_err: %x\n", reg.r.parity_err);
	}
	if (reg.r.framing_err != 0) {
		UserPrint("\tframing_err: %x\n", reg.r.framing_err);
	}
	if (reg.r.brk != 0) {
		UserPrint("\tbrk: %x\n", reg.r.brk);
	}
	if (reg.r.tx_empty != 0) {
		UserPrint("\ttx_empty: %x\n", reg.r.tx_empty);
	}
	if (reg.r.tx_done != 0) {
		UserPrint("\ttx_done: %x\n", reg.r.tx_done);
	}
	if (reg.r.err_in_fifo != 0) {
		UserPrint("\terr_in_fifo: %x\n", reg.r.err_in_fifo);
	}
}

static inline void print_msr ( union msr reg ) {	UserPrint("Debug msr - 0x%x\n", reg.data);	if (reg.r.dcts != 0) {
		UserPrint("\tdcts: %x\n", reg.r.dcts);
	}
	if (reg.r.ddsr != 0) {
		UserPrint("\tddsr: %x\n", reg.r.ddsr);
	}
	if (reg.r.dri != 0) {
		UserPrint("\tdri: %x\n", reg.r.dri);
	}
	if (reg.r.ddcd != 0) {
		UserPrint("\tddcd: %x\n", reg.r.ddcd);
	}
	if (reg.r.ncts != 0) {
		UserPrint("\tncts: %x\n", reg.r.ncts);
	}
	if (reg.r.ndsr != 0) {
		UserPrint("\tndsr: %x\n", reg.r.ndsr);
	}
	if (reg.r.nri != 0) {
		UserPrint("\tnri: %x\n", reg.r.nri);
	}
	if (reg.r.ndcd != 0) {
		UserPrint("\tndcd: %x\n", reg.r.ndcd);
	}
}

static inline void print_dll ( union dll reg ) {	UserPrint("Debug dll - 0x%x\n", reg.data);}

static inline void print_dlh ( union dlh reg ) {	UserPrint("Debug dlh - 0x%x\n", reg.data);}

#endif /* _UART16550_H_ */
