#ifndef IO_H_
#define IO_H_

static inline uint8_t io_read_8(bus_space_t handle, uintptr_t offset);
static inline void io_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t io_read_16(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_16(bus_space_t handle, uintptr_t offset,
					 uint16_t value);

static inline uint32_t io_read_32(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_32(bus_space_t handle, uintptr_t offset,
					 uint32_t value);

#define pci_conf_read_8   io_read_8
#define pci_conf_read_16  io_read_16
#define pci_conf_read_32  io_read_32
#define pci_conf_write_8  io_write_8
#define pci_conf_write_16 io_write_16
#define pci_conf_write_32 io_write_32

#define readb(address)		(*(volatile uint8_t *)(address))
#define writeb(address,value)	(*(volatile uint8_t *)(address) = (value))
#define readw(address)		(*(volatile uint16_t *)(address))
#define writew(address,value)	(*(volatile uint16_t *)(address) = (value))
#define readl(address)		(*(volatile uint32_t *)(address))
#define writel(address,value)	(*(volatile uint32_t *)(address) = (value))
#define readl_physical(address)		({ unsigned long value; asm volatile(";; rsm psr.dt ;; srlz.d ;; ld4 %0=[%1] ;; ssm psr.dt ;; srlz.d ;;" : "=r"(value) : "r"(address)); value; })
#define writel_physical(address,value)	({ asm volatile(";; rsm psr.dt ;; srlz.d ;; st4 [%0]=%1 ;; ssm psr.dt ;; srlz.d ;;" : : "r"(address), "r"(value)); })
#define inb(port)		readb(ia64_io_address(port))
#define outb(port,value)	writeb(ia64_io_address(port),value)
#define inw(port)		readw(ia64_io_address(port))
#define outw(port,value)	writew(ia64_io_address(port),value)
#define inl(port)		readl(ia64_io_address(port))
#define outl(port,value)	writel(ia64_io_address(port),value)

extern uintptr_t ia64_io_base;

static inline uintptr_t
ia64_io_address(uintptr_t port)
{
	return ia64_io_base + (((port >> 2) << 12) | (port & ((1 << 12) - 1)));
}

static inline uint8_t
io_read_8(bus_space_t handle, uintptr_t offset)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			return readb(address);
		case IORESOURCE_IO:
			return inb(address);
		case IORESOURCE_PCICONF:
			return pci_read_config_byte(address);
	}
	return 0;
}

static inline void
io_write_8(bus_space_t handle, uintptr_t offset,
		 uint8_t value)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			writeb(address, value);
			break;
		case IORESOURCE_IO:
			outb(address, value);
			break;
		case IORESOURCE_PCICONF:
			pci_write_config_byte(address, value);
			break;
	}
}


static inline uint16_t
io_read_16(bus_space_t handle, uintptr_t offset)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			return readw(address);
		case IORESOURCE_IO:
			return inw(address);
		case IORESOURCE_PCICONF:
			return pci_read_config_word(address);
	}
	return 0;
}

static inline void
io_write_16(bus_space_t handle, uintptr_t offset, uint16_t value)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			writew(address, value);
			break;
		case IORESOURCE_IO:
			outw(address, value);
			break;
		case IORESOURCE_PCICONF:
			pci_write_config_word(address, value);
			break;
	}
}

static inline uint32_t 
io_read_32(bus_space_t handle, uintptr_t offset)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			return readl(address);
		case IORESOURCE_IO:
			return inl(address);
		case IORESOURCE_PCICONF:
			return pci_read_config_dword(address);
	}
	return 0;
}

static inline void
io_write_32(bus_space_t handle, uintptr_t offset,
		 uint32_t value)
{
	uintptr_t address = handle.address + offset;
	switch (handle.type)
	{
		case IORESOURCE_MEM:
			writel(address, value);
			break;
		case IORESOURCE_IO:
			outl(address, value);
			break;
		case IORESOURCE_PCICONF:
			pci_write_config_dword(address, value);
			break;
	}
}

#endif
