struct request {
	uint8_t type;
	uint8_t from;
	uint8_t synchronous;
	uint8_t pad1;
	uint16_t request_id;
	uint16_t pad2;
	union {
		// 16 bytes
		struct {
			uint64_t va;
			uint32_t rr;
			uint32_t itir;
		} purge;

		// 12 bytes
		struct {
			uint32_t mpfn;
			uint8_t is_write;
			uint8_t access_type;
			uint8_t fragment;
			uint8_t reply_to;
			copyset_t copyset;
			uint16_t wu_version;
		} fetch;

		// 5 bytes
		struct {
			uint32_t mpfn;
			uint8_t reply_to;
		} invalidate;

		// 5 bytes
		struct {
			uint32_t mpfn;
			uint8_t npages;
		} invwhole;

		// 1 byte
		struct {
			uint8_t irq;
		} ipi;

		// 16 bytes
		struct {
			uint64_t ip;
			uint64_t gp;
		} startup;

		// 5 bytes
		struct {
			uint32_t type;
			uint8_t irq;
		} connectirq;

#ifdef GUESTIO_DISK
		// 32 bytes
		struct {
			uint64_t mpa;
			uint32_t len;
			uint8_t nsge;
			uint8_t is_write;
			uint8_t key_valid;
			uint64_t offset;
			uint64_t key;
		} diskio;
#endif

		// 1 byte
		struct {
			uint8_t c;
		} putchar;

		// 8 bytes
		struct {
			uint32_t nodes;
			uint32_t node_id;
		} configure;

		// 6 bytes
		struct {
			uint32_t nstores;
			uint16_t wu_version;
		} store;

		// 24 bytes
		struct {
			uint64_t mpa;
			uint64_t value;
			uint32_t size;
			uint16_t wu_version;
			uint16_t arrival_time;
		} store1;

		// 6 bytes
		struct {
			uint32_t mpfn;
			uint16_t count;
		} storehistory;

		// 10 bytes
		struct {
			uint64_t mpa;
			uint8_t size;
			uint8_t reply_to;
		} load;

		// 8 bytes
		struct {
			uint64_t value;
		} loadreply;

		// 18 bytes
		struct {
			uint64_t start;
			uint64_t end;
			uint16_t count;
		} profile;

#ifdef SPECIAL_CMPXCHG_FETCHADD
		struct {
			uint64_t mpa;
			uint64_t value;
			uint64_t cmpvalue;
			uint8_t size;
		} cmpxchg;

		struct {
			uint64_t mpa;
			uint8_t size;
			uint8_t inc;
		} fetchadd;

		struct {
			uint64_t oldvalue;
			uint64_t newvalue;
		} semreply;
#endif

#ifdef SPECIAL_LOCKS
		struct {
			uint64_t mpa;
		} lock;

		struct {
			uint32_t status;
		} lockreply;
#endif
	};
};

struct pkt {
	uint8_t dstaddr[6];
	uint8_t srcaddr[6];
	uint16_t proto;
	struct request request;
};

#define ETHER_PROTO_VMM 0x1234

#define IPC_FETCH		0
#define IPC_FETCH_REPLY		1
#define IPC_INVALIDATE		2
#define IPC_INVALIDATE_REPLY	3
#define IPC_PURGE		4
#define IPC_PURGE_REPLY		5
#define IPC_IPI			6
#define IPC_STARTUP		7
#define IPC_CONNECT_IRQ		8
#ifdef GUESTIO_DISK
#define IPC_DISK_IO_SIMPLE	9
#define IPC_DISK_IO_SGL		10
#endif
#define IPC_RESET		11
#define IPC_PUTCHAR		12
#define IPC_DISCOVER		13 /* broadcast */
#define IPC_DISCOVER_REPLY	14
#define IPC_CONFIGURE		15
#define IPC_CONFIGURE_REPLY	16
#define IPC_STORE		17
#define IPC_STORE_SINGLE	18
#define IPC_STORE_HISTORY	19
#define IPC_LOAD		20
#define IPC_LOAD_REPLY		21
#ifdef SPECIAL_CMPXCHG_FETCHADD
#define IPC_CMPXCHG		22
#define IPC_FETCHADD		23
#define IPC_SEMAPHORE_REPLY	24
#endif
#ifdef SPECIAL_LOCKS
#define IPC_LOCK_ACQUIRE	25
#define IPC_LOCK_GRANT		26
#define IPC_LOCK_READ		27
#define IPC_LOCK_READ_REPLY	28
#define IPC_LOCK_UNLOCK		29
#endif
#define IPC_INVALIDATE_WHOLE	30
#define IPC_PROFILE		31
#define IPC_STATS_START		32
#define IPC_STATS_STOP		33
#define IPC_PROBE		34
