#ifndef PCI_DMA_H
#define PCI_DMA_H
#include <stdint.h>

typedef enum direction { mem_to_bus, bus_to_mem, consistent } direction_t;

/*
 * Allocate sz bytes of PCI=consistent memory; i.e., memory that is
 * visible from the PCI bus (at busaddr) and in the driver (at vaddr).
 * Remember to use memory barriers where appropriate!
 * 
 * Returns: 0 for success; anything else means error.
 */
extern int get_pci_consistent_memory(dma_handle_t dma, 
				     size_t sz, 
				     mem_handle_t *vaddr, 
				     dmabuf_handle_t *busaddr);


extern int setup_dma(dma_handle_t dma, 
	      mem_handle_t handle, 
	      size_t sz, 
	      dmabuf_handle_t *dmabuf,
	      direction_t dir);

extern int teardown_dma(dma_handle_t dma, dmabuf_handle_t *x);

mem_handle_t    map_physical(/*host_handle_t host,*/ void *phys_addr,
			     uintptr_t size);



#endif /* PCI_DMA_H */
