/* head.S */
unsigned long tpa(unsigned long address);
void reset_system(void);
void sample_pmi(void);
struct pal_ret call_pal(unsigned long nr, unsigned long arg1, unsigned long arg2, unsigned long arg3);

/* driver-glue.c */
void device_init(void);
void device_enter_virtual();
void do_irq(unsigned int irq);
int disk_xfer(struct scsi_sglist *sgl, unsigned long offset, unsigned int length, int rw, void (*callback)(void *, uint64_t));
void early_serial_setup(void);
int getc(void);
uint8_t pci_read_config_byte(uintptr_t address);
void pci_write_config_byte(uintptr_t address, uint8_t value);
uint16_t pci_read_config_word(uintptr_t address);
void pci_write_config_word(uintptr_t address, uint16_t value);
uint32_t pci_read_config_dword(uintptr_t address);
void pci_write_config_dword(uintptr_t address, uint32_t value);

/* debug.c */
void putchar(int c);
int puts(const char *s);
int printf(const char *format, ...);
int logf(const char *format, ...);
int process_debug_char(char c);
void debug(void);

/* profile_dsm.c */
void profile_dsm_flush(void);
void profile_fetch(unsigned long start_time, unsigned int pending, mpfn_t mpfn, boolean is_write, uint8_t access_type, node_t forward_to, node_t fetch_rcvd_from, boolean fetch_rcvd_data);
void profile_invalidate(unsigned long start_time, unsigned int pending, mpfn_t mpfn, copyset_t copyset);

/* profile_tlb.c */
void profile_tlb(unsigned long ip, unsigned long va, unsigned int rr, unsigned char type);

/* emu.c */
void dtr_insert(unsigned long va, unsigned long pa, unsigned long sizebits, unsigned long attr);
void itr_insert(unsigned long va, unsigned long pa, unsigned long sizebits, unsigned long attr);
void emu_precover(void);
void deliver_exception(unsigned long vector, unsigned long isr);
void check_interrupts(void);
void deliver_interrupt(unsigned int irq);
unsigned long thash(unsigned long addr);
void stats_start(void);
void stats_stop(void);
void emu_store_PIB(unsigned long addr);
void emu_fwb(void);
void emu_fwb_if_hit(mpfn_t mpfn);
void emu_store_remote(unsigned long addr, unsigned long mpa, unsigned int size, int release);
int emu_store(unsigned long addr, unsigned long mpa);
int emu_load(unsigned long addr, unsigned long mpa);
void print_stats(void);
void startup_secondary(unsigned long ip, unsigned long gp);

/* ssc.c */
void ssc_connect_irq(unsigned int type, unsigned char irq);
unsigned long do_ssc(unsigned long arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3, unsigned long nr);

/* mm.c */
void mm_handle_fetch(node_t node, mpfn_t mpfn, boolean is_write, access_type_t access_type, node_t for_node, struct request *orig_req, boolean synchronous);
void mm_fetch(mpfn_t mpfn, boolean is_write, access_type_t access_type, unsigned long va, boolean wait);
void mm_handle_fetch_data(mpfn_t mpfn, void *data);
void mm_handle_fetch_data_partial(mpfn_t mpfn, void *data, uint32_t offset, uint32_t length);
void mm_handle_fetch_reply(mpfn_t mpfn, boolean is_write, copyset_t copyset, uint16_t wu_version, node_t from, boolean rcvd_data);
void mm_handle_invalidate(node_t node, mpfn_t mpfn, struct request *orig_req);
void mm_handle_invalidate_whole(mpfn_t mpfn, unsigned int npages, node_t node);
void mm_invalidate_whole(mpfn_t mpfn, unsigned int npages);
void mm_handle_invalidate_reply(node_t node, mpfn_t mpfn);
void mm_chmod(mpfn_t mpfn, unsigned int prot);
void mm_handle_tlb_purge(node_t node, unsigned long rr, unsigned long orig_va, unsigned long sz);
void mm_handle_purge_reply(void);
void tlb_insert(unsigned long ifa, unsigned long itir, unsigned long pte, boolean pinned, boolean is_exec);
void tlb_purge(unsigned long address, unsigned long sz, boolean global);
void tlb_purge_all(void);
void mm_vhpt_miss(unsigned long address, unsigned long isr);
void mm_tlb_miss(unsigned long address, unsigned long isr);
void mm_fault(unsigned long address, unsigned long isr);
uint64_t mm_handle_cmpxchg(uint64_t mpa, uint64_t value, uint64_t cmpvalue, uint8_t size, uint64_t *newvalue);
uint64_t mm_handle_fetchadd(uint64_t mpa, uint8_t size, uint8_t inc, uint64_t *newvalue);
void mm_handle_semaphore_reply(uint64_t oldvalue, uint64_t newvalue);
uint64_t mm_cmpxchg(uint64_t mpa, uint64_t value, uint64_t cmpvalue, uint8_t size, int release, uint64_t va);
uint64_t mm_fetchadd(uint64_t mpa, uint8_t size, uint8_t inc, int release);
boolean mm_load(uint64_t mpa, uint8_t size, uint64_t *value);
void mm_handle_load(node_t node, uint64_t mpa, uint8_t size, node_t for_node, struct request *orig_req);
void mm_handle_load_reply(uint64_t value);
void mm_handle_store_history(mpfn_t mpfn, uint16_t count, void *data);
void mm_new_generation();
void mm_handle_store(uint64_t mpa, uint64_t value, uint8_t size, node_t node, uint16_t wu_version, int replay);
void process_store_queue();
void process_store_queue_partial(node_t from, uint16_t page_wu_version);
void mm_check_mpa(unsigned long mpa, unsigned int prot);
void mm_copy_to_user(unsigned long dest_mpa, const void *src, unsigned int n);
void mm_copy_from_user(void *dest, unsigned long src_mpa, unsigned int n);
void mm_release_page(mpfn_t mpfn, unsigned long pa);
void mm_init(void);
void vhpt_init(void);

/* ipc.c */
void ipc_handle_async(struct request *request, unsigned char *ethaddr, int delayed);
unsigned int dequeue_packet(unsigned long buffer_mpa, unsigned int length);
struct driver *ipc_init(void);
void ipc_send(node_t node, struct request *request);
void ipc_send_page(node_t node, struct request *request, unsigned long pa, unsigned int length, uintptr_t data_area);
void ipc_send_raw(unsigned long buffer_mpa, unsigned int length);
void ipc_wait(node_t node, struct request *request, uint8_t *lock);
void ipc_send_reply(node_t node, unsigned char type);
void ipc_send_fetch_reply(node_t node, mpfn_t mpfn, unsigned long pa, boolean is_write, boolean send_data, unsigned int wu_version, copyset_t copyset);
void ipc_send_sequenced(node_t node, struct request *request);
unsigned int ipc_broadcast(unsigned long mask, struct request *request);
unsigned int ipc_broadcast_sequenced(unsigned long mask, struct request *request);
void ipc_discover(void);
void ipc_reset_system(void);

/* fw-emu.c */
struct ia64_boot_param *init_guest_fw(char *command_line);
void init_fw_interfaces(struct ia64_boot_param *bp);
void do_reset_system(void);

/* clear_page.S */
void clear_page(void *p);

/* copy_page.S */
void copy_page(void *dst, void *src);
void copy_area(void *dst, void *src, unsigned int length);

