#include <string.h>
#include <malloc.h>
#include <assert.h>
#include "types.h"
#include "ipc.h"
#include "proto.h"
#include "stats.h"

#define BITS(val,from,n)	((val >> from) & ((1UL << n) - 1))
#define BIT(val,from)		BITS(val,from,1)

#define CR_DCR				0
#define CR_ITM				1
#define CR_IVA				2
#define CR_PTA				8
#define CR_IPSR				16
#define CR_ISR				17
#define CR_IIP				19
#define CR_IFA				20
#define CR_ITIR				21
#define CR_IIPA				22
#define CR_IFS				23
#define CR_IIM				24
#define CR_IHA				25
#define CR_LID				64
#define CR_IVR				65
#define CR_TPR				66
#define CR_EOI				67
#define CR_IRR0				68
#define CR_IRR1				69
#define CR_IRR2				70
#define CR_IRR3				71
#define CR_ITV				72
#define CR_PMV				73
#define CR_CMCV				74
#define CR_LRR0				80
#define CR_LRR1				81

#define TPR_MMI				(1UL << 16)

#define PSR_BE_BIT	1
#define PSR_UP_BIT	2
#define PSR_AC_BIT	3
#define PSR_MFL_BIT	4
#define PSR_MFH_BIT	5
#define PSR_IC_BIT	13
#define PSR_I_BIT	14
#define PSR_PK_BIT	15
#define PSR_DT_BIT	17
#define PSR_DFL_BIT	18
#define PSR_DFH_BIT	19
#define PSR_SP_BIT	20
#define PSR_PP_BIT	21
#define PSR_DI_BIT	22
#define PSR_SI_BIT	23
#define PSR_DB_BIT	24
#define PSR_LP_BIT	25
#define PSR_TB_BIT	26
#define PSR_RT_BIT	27
#define PSR_CPL_BITS	32
#define PSR_IS_BIT	34
#define PSR_MC_BIT	35
#define PSR_IT_BIT	36
#define PSR_ID_BIT	37
#define PSR_DA_BIT	38
#define PSR_DD_BIT	39
#define PSR_SS_BIT	40
#define PSR_RI_BITS	41
#define PSR_ED_BIT	43
#define PSR_BN_BIT	44
#define PSR_IA_BIT	45

#define PSR_BE		(1UL << PSR_BE_BIT)
#define PSR_UP		(1UL << PSR_UP_BIT)
#define PSR_AC		(1UL << PSR_AC_BIT)
#define PSR_MFL		(1UL << PSR_MFL_BIT)
#define PSR_MFH		(1UL << PSR_MFH_BIT)
#define PSR_IC		(1UL << PSR_IC_BIT)
#define PSR_I		(1UL << PSR_I_BIT)
#define PSR_PK		(1UL << PSR_PK_BIT)
#define PSR_DT		(1UL << PSR_DT_BIT)
#define PSR_DFL		(1UL << PSR_DFL_BIT)
#define PSR_DFH		(1UL << PSR_DFH_BIT)
#define PSR_SP		(1UL << PSR_SP_BIT)
#define PSR_PP		(1UL << PSR_PP_BIT)
#define PSR_DI		(1UL << PSR_DI_BIT)
#define PSR_SI		(1UL << PSR_SI_BIT)
#define PSR_DB		(1UL << PSR_DB_BIT)
#define PSR_LP		(1UL << PSR_LP_BIT)
#define PSR_TB		(1UL << PSR_TB_BIT)
#define PSR_RT		(1UL << PSR_RT_BIT)
#define PSR_CPL_3	(3UL << PSR_CPL_BITS)
#define PSR_IS		(1UL << PSR_IS_BIT)
#define PSR_MC		(1UL << PSR_MC_BIT)
#define PSR_IT		(1UL << PSR_IT_BIT)
#define PSR_ID		(1UL << PSR_ID_BIT)
#define PSR_DA		(1UL << PSR_DA_BIT)
#define PSR_DD		(1UL << PSR_DD_BIT)
#define PSR_SS		(1UL << PSR_SS_BIT)
#define PSR_RI		(3UL << PSR_RI_BITS)
#define PSR_ED		(1UL << PSR_ED_BIT)
#define PSR_BN		(1UL << PSR_BN_BIT)
#define PSR_IA		(1UL << PSR_IA_BIT)

#define DCR_PP_BIT	0
#define DCR_LC_BIT	2
#define DCR_DM_BIT	8
#define DCR_DP_BIT	9
#define DCR_DK_BIT	10
#define DCR_DX_BIT	11
#define DCR_DR_BIT	12
#define DCR_DA_BIT	13
#define DCR_DD_BIT	14
#define DCR_PP		(1UL << DCR_PP_BIT)
#define DCR_LC		(1UL << DCR_LC_BIT)
#define DCR_DM		(1UL << DCR_DM_BIT)
#define DCR_DP		(1UL << DCR_DP_BIT)
#define DCR_DK		(1UL << DCR_DK_BIT)
#define DCR_DX		(1UL << DCR_DX_BIT)
#define DCR_DR		(1UL << DCR_DR_BIT)
#define DCR_DA		(1UL << DCR_DA_BIT)
#define DCR_DD		(1UL << DCR_DD_BIT)

#define PTE_P_BIT	0
#define PTE_VNUMA_PIB_BIT	1
#define PTE_MA_BITS	2
#define PTE_A_BIT	5
#define PTE_D_BIT	6
#define PTE_PL_BITS	7
#define PTE_AR_BITS	9
#define PTE_PPN_BITS	12
#define PTE_ED_BIT	52
#define PTE_VNUMA_A_BIT	53
#define PTE_VNUMA_D_BIT	54

#define PTE_P		(1UL << PTE_P_BIT)
#define PTE_VNUMA_PIB	(1UL << PTE_VNUMA_PIB_BIT)
#define PTE_MA_UC	(4UL << PTE_MA_BITS)
#define PTE_A		(1UL << PTE_A_BIT)
#define PTE_D		(1UL << PTE_D_BIT)
#define PTE_PL_3	(3UL << PTE_PL_BITS)
#define PTE_AR		(7UL << PTE_AR_BITS)
#define PTE_AR_RW	(2UL << PTE_AR_BITS)
#define PTE_AR_RWX	(3UL << PTE_AR_BITS)
#define PTE_AR_GATE	(7UL << PTE_AR_BITS)
#define PTE_PPN		(0x3fffffffffUL << PTE_PPN_BITS)
#define PTE_ED		(1UL << PTE_ED_BIT)

extern struct cpu {
	unsigned long cr[128];
	unsigned long kr[8];
	unsigned long rr[8];
	unsigned long banked[2][16];
	unsigned long psr;
	unsigned int irq_pending;
	unsigned int in_service_irq;
	unsigned int current_bank;
	unsigned int cfle;
	unsigned long saved_r4;
	unsigned long saved_r5;
	unsigned long saved_r6;
	unsigned long saved_r7;
	unsigned long saved_b2;
	unsigned long banked_unat[2];
	unsigned long temp[2];
} cpu;

#define read_psr() ((cpu.psr & ~(PSR_MFL|PSR_MFH)) | (read_ipsr() & (PSR_MFL|PSR_MFH)))
#define write_psr(value) { cpu.psr = value; write_ipsr((read_ipsr() & ~(PSR_DFL|PSR_DFH|PSR_MFL|PSR_MFH|PSR_ED)) | ((value) & (PSR_DFL|PSR_DFH|PSR_MFL|PSR_MFH|PSR_ED))); check_interrupts(); }
#define read_register(reg)	({ unsigned long ret; asm("mov %0 = " reg : "=r"(ret)); ret; })
#define read_register_volatile(reg)	({ unsigned long ret; asm volatile("mov %0 = " reg : "=r"(ret)); ret; })
#define write_register(reg,value)	asm volatile("mov " reg " = %0" : : "r"(value))
#define read_bank1_register(reg) ({ unsigned long ret; asm("bsw.1 ;; mov %0 = " reg " ;; bsw.0 ;;" : "=r"(ret)); ret; })
#define write_bank1_register(reg,value)	asm("bsw.1 ;; mov " reg " = %0 ;; bsw.0 ;;" : : "r"(value))
#define read_rr(address)		({ unsigned long ret; asm("mov %0 = rr[%1]" : "=r"(ret) : "r"(address)); ret; })
#define write_rr(address,value)		asm("mov rr[%0] = %1" : : "r"(address), "r"(value))
#define read_iip()  read_register("cr.iip")
#define write_iip(value)	write_register("cr.iip", value)
#define read_ipsr() read_register("cr.ipsr")
#define write_ipsr(value)	write_register("cr.ipsr", value)
#define read_ifs()  read_register("cr.ifs")
#define write_ifs(value)	write_register("cr.ifs", value)
#define read_ifa()  read_register("cr.ifa")
#define write_ifa(value)	write_register("cr.ifa", value)
#define read_isr()  read_register("cr.isr")
#define write_isr(value)	write_register("cr.isr", value)
#define read_iipa()  read_register("cr.iipa")
#define write_itm(value)	write_register("cr.itm", value)
#define read_itv(value)	read_register("cr.itv")
#define write_itv(value)	write_register("cr.itv", value)
#define write_pmv(value)	write_register("cr.pmv", value)
#define write_tpr(value)	write_register("cr.tpr", value)
#define read_ivr()		read_register_volatile("cr.ivr")
#define write_eoi()		write_register("cr.eoi", 0)
#define read_iim()		read_register("cr.iim")
#define write_pta(value)	write_register("cr.pta", value)
#define read_dcr()		read_register("cr.dcr")
#define write_dcr(value)	write_register("cr.dcr", value)
#define read_fpsr()		read_register("ar.fpsr")
#define write_fpsr(value)	write_register("ar.fpsr", value)
#define read_pr()		read_register_volatile("pr")
#define write_pr(value)		write_register("pr", value)
#define read_pmd(n)		({ unsigned long ret; asm("mov %0 = pmd[%1]" : "=r"(ret) : "r"(n)); ret; })
#define write_pmd(n,value)		asm("mov pmd[%0] = %1" : : "r"(n), "r"(value))
#define read_pmc(n)		({ unsigned long ret; asm("mov %0 = pmc[%1]" : "=r"(ret) : "r"(n)); ret; })
#define write_pmc(n,value)		asm("mov pmc[%0] = %1" : : "r"(n), "r"(value))
#define write_dbr(n,value)		asm("mov dbr[%0] = %1" : : "r"(n), "r"(value))
#define write_ibr(n,value)		asm("mov ibr[%0] = %1" : : "r"(n), "r"(value))
#define read_msr(n)		({ unsigned long ret; asm volatile("mov %0 = msr[%1]" : "=r"(ret) : "r"(n)); ret; })
#define write_msr(n,value)		asm volatile("mov msr[%0] = %1" : : "r"(n), "r"(value))
#define read_bsp()	read_register("r27")
#define read_itc()	read_register_volatile("ar.itc")

#define ia64_thash(addr)	({ unsigned long val; asm("thash %0=%1" : "=r"(val) : "r"(addr)); val; })
#define ia64_ttag(addr)		({ unsigned long val; asm("ttag  %0=%1" : "=r"(val) : "r"(addr)); val; })
#define ia64_ptc_l(addr,size)	asm volatile("ptc.l %0,%1" : : "r"(addr), "r"(size))
#define ia64_ptc_ga(addr,size)	asm volatile("ptc.ga %0,%1" : : "r"(addr), "r"(size))
#define ia64_ptc_e(x)		asm volatile("ptc.e %0" : : "r"(x))
#define ia64_itc_d(va,ps,pte)		asm volatile("mov cr.ifa=%0; mov cr.itir=%1;; itc.d %2" : : "r"(va), "r"(ps), "r"(pte))
#define ia64_itc_i(va,ps,pte)		asm volatile("mov cr.ifa=%0; mov cr.itir=%1;; itc.i %2" : : "r"(va), "r"(ps), "r"(pte))
#define ia64_itr_d(n,va,ps,pte)		asm volatile("mov cr.ifa=%0; mov cr.itir=%1;; itr.d dtr[%2]=%3;; srlz.d" : : "r"(va), "r"(ps), "r"(n), "r"(pte))
#define ia64_ptr_d(addr,size)	asm volatile("ptr.d %0,%1" : : "r"(addr), "r"(size))
#define ia64_srlz_d()		asm volatile("srlz.d")

#if 0
#define translate_rr(rr) (rr)
#elif 0
#define translate_rr(rr) ({unsigned long preserve = (rr) & 0xff0000ff, ridlo = ((rr) >> 8) & 0xff, ridmid = ((rr) >> 16) & 0xff; (preserve | (ridlo << 16) | (ridmid << 8)); })
#elif 0 /* WRONG */
#define translate_rr(rr) ({unsigned long preserve = (rr) & 0xf80000ff, ridlo = ((rr) >> 8) & 0x7ff, ridmid = ((rr) >> 19) & 0xff; (preserve | (ridlo << 16) | (ridmid << 8)); })
#elif 0 /* 8/11 */
#define translate_rr(rr) ({unsigned long preserve = (rr) & 0xf80000ff, ridlo = ((rr) >> 8) & 0xff, ridmid = ((rr) >> 16) & 0x7ff; (preserve | (ridlo << 19) | (ridmid << 8)); })
#elif 1 /* 3/5/11 */
#define translate_rr(rr) ({unsigned long preserve = (rr) & 0xf80000ff, ridlo = ((rr) >> 8) & 7, ridmid = ((rr) >> 11) & 0x1f, ridhi = ((rr) >> 16) & 0x7ff; (preserve | (ridlo << 24) | (ridmid << 19) | (ridhi << 8)); })
#else /* for short */
#define translate_rr(rr) ({unsigned long preserve = (rr) & 0xff, ridlo = ((rr) >> 8) & 7, ridmid = ((rr) >> 11) & 0x1f, ridhi = (rr) >> 16; (preserve | (ridlo << 29) | (ridmid << 24) | (ridhi << 8)); })
#endif

#define PROT_NONE	0
#define PROT_RX		1
#define PROT_RWX	3

#define PAGE_SHIFT	12
#define PAGE_SIZE	(1UL << PAGE_SHIFT)
#define PAGE_MASK	(PAGE_SIZE-1)
#define CACHE_LINE_SIZE	128

#define RGN_PAGE_SHIFT(va)	BITS(cpu.rr[va >> 61], 2, 6)
#define RGN_PAGE_SIZE(va)	(1UL << RGN_PAGE_SHIFT(va))

#define __pa(addr) ((unsigned long)(addr))

/* these values used to indicate source of data to ipc_send_page */
#define DATA_AREA_HEADER	(uintptr_t)-1
#define DATA_AREA_MALLOC	(uintptr_t)-2
#define DATA_AREA_STATIC	(uintptr_t)-3
#define DATA_AREA_GUEST(mfn)	(uintptr_t)mfn

#define PAGE_MODE_WRITE_INVALIDATE	0
#define PAGE_MODE_WRITE_UPDATE		1

#define MAX_NODES	8
//#define NODE_BITS	2
//#define NODE_MASK	0x03
#define NODE_BITS	3
#define NODE_MASK	0x07

#define likely(cond)	__builtin_expect(cond,1)
#define unlikely(cond)	__builtin_expect(cond,0)

extern unsigned int debug_flags;
extern char *last_interruption;

#define DEBUG_FLAG_PRINT_SHORT	1
#define DEBUG_FLAG_PRINT_LONG	2
#define DEBUG_FLAG_PROFILE	4

#ifdef DEBUG_VM
#define DEBUG_SHORT(c) if (debug_flags & DEBUG_FLAG_PRINT_SHORT) putchar(c)
#define DEBUG_EMU(args...) if (debug_flags & DEBUG_FLAG_PRINT_LONG) printf(args)
#define DEBUG_MM(args...) if (debug_flags & DEBUG_FLAG_PRINT_LONG) printf(args)
#else
#define DEBUG_SHORT(c)
#define DEBUG_EMU(args...)
#define DEBUG_MM(args...)
#endif

#ifndef NDEBUG
#define DEBUG_INTERRUPTION(format, args...) { if (read_isr() & (1UL << 39)) { printf("NESTED EXCEPTION isr=%lx: " format, read_isr(), args); printf("LAST INTERRUPTION: %s", last_interruption); debug(); } last_interruption = format; }
#else
#define DEBUG_INTERRUPTION(format...)
#endif

extern void *ether_drv;
#ifdef TIGON3
#define get_pending_tx() tigon3_get_pending(ether_drv);
#else
#define get_pending_tx() dp83820_get_pending(ether_drv);
#endif

void monitor_rel_start(void);
