/*
 @LICENSE@
*/

/*-
 * Copyright (c) 2003 Cristan Szmajda.  All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by Cristan Szmajda. 
 * 4. The name Cristan Szmajda may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY CRISTAN SZMAJDA ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL CRISTAN SZMAJDA BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * $Id: ctype.h,v 1.3 2004/03/02 05:50:40 nfd Exp $
 */

/*
 * XXX: ignores locale 
 */

#ifndef _CTYPE_H_
#define	_CTYPE_H_

#ifndef	SAVE_SPACE

/*
 * Look-up table implementation. 
 */

/*
 * Doesn't use the fact that _CTYPE_CNTRL, _CTYPE_PUNCT, _CTYPE_UPPER,
 * _CTYPE_LOWER, and _CTYPE_DIGIT are mutually exclusive. 
 */
#define	_CTYPE_CNTRL	0001
#define	_CTYPE_SPACE	0002
#define	_CTYPE_BLANK	0004
#define	_CTYPE_PUNCT	0010
#define	_CTYPE_LOWER	0020	/* MUST equal 'A' - 'a' >> 1 */
#define	_CTYPE_UPPER	0040	/* MUST equal 'A' - 'a' */
#define	_CTYPE_DIGIT	0100
#define	_CTYPE_XDIGIT	0200

#define INLINE static inline

extern const unsigned char __ctype[257];	/* lookup table: -1 to 255 
						 */
extern const unsigned char __digit[257];	/* lookup table: -1 to 255 
						 */

INLINE int
isalnum(int c)
{
	return __ctype[c +
		       1] & (_CTYPE_UPPER | _CTYPE_LOWER | _CTYPE_DIGIT);
}
INLINE int
isalpha(int c)
{
	return __ctype[c + 1] & (_CTYPE_UPPER | _CTYPE_LOWER);
}
INLINE int
isblank(int c)
{
	return __ctype[c + 1] & _CTYPE_BLANK;
}
INLINE int
iscntrl(int c)
{
	return __ctype[c + 1] & _CTYPE_CNTRL;
}
INLINE int
isdigit(int c)
{
	return __ctype[c + 1] & _CTYPE_DIGIT;
}
INLINE int
isgraph(int c)
{
	return __ctype[c +
		       1] & (_CTYPE_UPPER | _CTYPE_LOWER | _CTYPE_DIGIT |
			     _CTYPE_PUNCT);
}
INLINE int
islower(int c)
{
	return __ctype[c + 1] & _CTYPE_LOWER;
}
INLINE int
isprint(int c)
{
	return __ctype[c +
		       1] & (_CTYPE_UPPER | _CTYPE_LOWER | _CTYPE_DIGIT |
			     _CTYPE_PUNCT | _CTYPE_SPACE);
}
INLINE int
ispunct(int c)
{
	return __ctype[c + 1] & _CTYPE_PUNCT;
}
INLINE int
isspace(int c)
{
	return __ctype[c + 1] & _CTYPE_SPACE;
}
INLINE int
isupper(int c)
{
	return __ctype[c + 1] & _CTYPE_UPPER;
}
INLINE int
isxdigit(int c)
{
	return __ctype[c + 1] & _CTYPE_XDIGIT;
}
/*
 * _CTYPE_LOWER == 020 and _CTYPE_UPPER == 040 
 */
INLINE int
tolower(int c)
{
	return c + (__ctype[c + 1] & _CTYPE_UPPER);
}
INLINE int
toupper(int c)
{
	return c - ((__ctype[c + 1] & _CTYPE_LOWER) << 1);
}
#if !defined(_ANSI_SOURCE) && !defined(_POSIX_SOURCE)
INLINE int
isascii(int c)
{
	return 0 <= c && c <= '\177';
}
INLINE int
toascii(int c)
{
	return c & 0177;
}
INLINE int
digittoint(int c)
{
	return __digit[c + 1];
}
#endif

#else				/* SAVE_SPACE */

INLINE int
iscntrl(int c)
{
	return 0 <= c && c <= '\037' || c == '\177';
}
INLINE int
isgraph(int c)
{
	return ' ' < c && c < '\177';
}
INLINE int
isprint(int c)
{
	return ' ' <= c && c < '\177';
}
INLINE int
isdigit(int c)
{
	return '0' <= c && c <= '9';
}
INLINE int
islower(int c)
{
	return 'a' <= c && c <= 'z';
}
INLINE int
isspace(int c)
{
	return c == ' ' || '\011' <= c && c <= '\015';
}
INLINE int
isupper(int c)
{
	return 'A' <= c && c <= 'Z';
}
INLINE int
isalpha(int c)
{
	return isupper(c) || islower(c);
}
INLINE int
isalnum(int c)
{
	return isalpha(c) || isdigit(c);
}
INLINE int
ispunct(int c)
{
	return isgraph(c) && !isalnum(c);
}
INLINE int
isxdigit(int c)
{
	return isdigit(c) || 'A' <= c && c <= 'F' || 'a' <= c && c <= 'f';
}
INLINE int
tolower(int c)
{
	return isupper(c) ? c + ('a' - 'A') : c;
}
INLINE int
toupper(int c)
{
	return islower(c) ? c - ('a' - 'A') : c;
}
#if !defined(_ANSI_SOURCE) && !defined(_POSIX_SOURCE)
INLINE int
isascii(int c)
{
	return 0 <= c && c <= '\177';
}
INLINE int
toascii(int c)
{
	return c & 0177;
}
INLINE int
digittoint(int c)
{
	return isdigit(c) ? c - '0' : 'A' <= c
	    && c <= 'Z' ? c - 'A' : 'a' <= c && c <= 'z' ? c - 'a' : 99;
}
#endif

#endif				/* SAVE_SPACE */

#endif				/* !_CTYPE_H_ */
