/*
 @LICENSE@
*/

/*-
 * Copyright (c) 2003 Cristan Szmajda.	All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions, and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions, and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution. 
 * 3. All advertising materials mentioning features or use of this software
 *	must display the following acknowledgement:
 * 	This product includes software developed by Cristan Szmajda. 
 * 4. The name Cristan Szmajda may not be used to endorse or promote
 *	products derived from this software without specific prior written
 *	permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY CRISTAN SZMAJDA ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL CRISTAN SZMAJDA BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * $Id: stdint.h,v 1.2 2004/02/23 06:09:59 benjl Exp $
 */

#ifndef _STDINT_H_
#define	_STDINT_H_

/*
 * integers with exactly the specified width 
 */
typedef signed char int8_t;
typedef short   int16_t;
typedef int     int32_t;
typedef long long int64_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned long long uint64_t;

#define	INT8_MIN		(-128)
#define	INT16_MIN		(-32767-1)
#define	INT32_MIN		(-2147483647-1)
#define	INT64_MIN		(-9223372036854775807-1)
#define	INT8_MAX		127
#define	INT16_MAX		32767
#define	INT32_MAX		2147483647
#define	INT64_MAX		9223372036854775807
#define	UINT8_MAX		0xff
#define	UINT16_MAX		0xffff
#define	UINT32_MAX		0xffffffff
#define	UINT64_MAX		0xffffffffffffffff

/*
 * smallest integers with at least the specified width 
 */
typedef int8_t  int_least8_t;
typedef int16_t int_least16_t;
typedef int32_t int_least32_t;
typedef int64_t int_least64_t;
typedef uint8_t uint_least8_t;
typedef uint16_t uint_least16_t;
typedef uint32_t uint_least32_t;
typedef uint64_t uint_least64_t;
#define	INT_LEAST8_MIN		INT8_MIN
#define	INT_LEAST16_MIN		INT16_MIN
#define	INT_LEAST32_MIN		INT32_MIN
#define	INT_LEAST64_MIN		INT64_MIN
#define	INT_LEAST8_MAX		INT8_MAX
#define	INT_LEAST16_MAX		INT16_MAX
#define	INT_LEAST32_MAX		INT32_MAX
#define	INT_LEAST64_MAX		INT64_MAX
#define	UINT_LEAST8_MAX		UINT8_MAX
#define	UINT_LEAST16_MAX	UINT16_MAX
#define	UINT_LEAST32_MAX	UINT32_MAX
#define	UINT_LEAST64_MAX	UINT64_MAX

/*
 * `fastest' integers with at least the specified width 
 */
/*
 * XXX try wider ones, especially on Alpha 
 */
typedef int8_t  int_fast8_t;
typedef int16_t int_fast16_t;
typedef int32_t int_fast32_t;
typedef int64_t int_fast64_t;
typedef uint8_t uint_fast8_t;
typedef uint16_t uint_fast16_t;
typedef uint32_t uint_fast32_t;
typedef uint64_t uint_fast64_t;
#define	INT_FAST8_MIN		INT8_MIN
#define	INT_FAST16_MIN		INT16_MIN
#define	INT_FAST32_MIN		INT32_MIN
#define	INT_FAST64_MIN		INT64_MIN
#define	INT_FAST8_MAX		INT8_MAX
#define	INT_FAST16_MAX		INT16_MAX
#define	INT_FAST32_MAX		INT32_MAX
#define	INT_FAST64_MAX		INT64_MAX
#define	UINT_FAST8_MAX		UINT8_MAX
#define	UINT_FAST16_MAX		UINT16_MAX
#define	UINT_FAST32_MAX		UINT32_MAX
#define	UINT_FAST64_MAX		UINT64_MAX

/*
 * integer that can hold a void * 
 */
typedef long    intptr_t;	/* XXX */
typedef unsigned long uintptr_t;	/* XXX */
#define	INTPTR_MIN		LONG_MIN
#define	INTPTR_MAX		LONG_MAX
#define	UINTPTR_MAX		ULONG_MAX

/*
 * largest integer supported by implementation 
 */
#ifndef	NO_LONG_LONG
typedef long long intmax_t;
typedef unsigned long long uintmax_t;
#define	INTMAX_MIN		LLONG_MIN
#define	INTMAX_MAX		LLONG_MAX
#define	UINTMAX_MAX		ULLONG_MAX
#else
typedef long    intmax_t;
typedef unsigned long uintmax_t;
#define	INTMAX_MIN		LONG_MIN
#define	INTMAX_MAX		LONG_MAX
#define	UINTMAX_MAX		ULONG_MAX
#endif				/* NO_LONG_LONG */

/*
 * miscellany 
 */
#define	PTRDIFF_MIN		INTPTR_MIN
#define	PTRDIFF_MAX		INTPTR_MAX
#define	SIG_ATOMIC_MIN		INT_MIN
#define	SIG_ATOMIC_MAX		INT_MAX
#define	SIZE_MAX		UINTPTR_MAX
#define	WCHAR_MIN		0
#define	WCHAR_MAX		UINT16_MAX
#define	WINT_MIN		0
#define	WINT_MAX		UINT16_MAX

#ifdef	__STDC_LIMIT_MACROS
#define	INT8_C(x)		(int8_t)(x)
#define	INT16_C(x)		(int16_t)(x)
#define	INT32_C(x)		(int32_t)(x)
#define	INT64_C(x)		(int64_t)(x)
#define	UINT8_C(x)		(uint8_t)(x)
#define	UINT16_C(x)		(uint16_t)(x)
#define	UINT32_C(x)		(uint32_t)(x)
#define	UINT64_C(x)		(uint64_t)(x)
#define	INT_FAST8_C(x)		(int_fast8_t)(x)
#define	INT_FAST16_C(x)		(int_fast16_t)(x)
#define	INT_FAST32_C(x)		(int_fast32_t)(x)
#define	INT_FAST64_C(x)		(int_fast64_t)(x)
#define	UINT_FAST8_C(x)		(uint_fast8_t)(x)
#define	UINT_FAST16_C(x)	(uint_fast16_t)(x)
#define	UINT_FAST32_C(x)	(uint_fast32_t)(x)
#define	UINT_FAST64_C(x)	(uint_fast64_t)(x)
#define	INT_LEAST8_C(x)		(int_least8_t)(x)
#define	INT_LEAST16_C(x)	(int_least16_t)(x)
#define	INT_LEAST32_C(x)	(int_least32_t)(x)
#define	INT_LEAST64_C(x)	(int_least64_t)(x)
#define	UINT_LEAST8_C(x)	(uint_least8_t)(x)
#define	UINT_LEAST16_C(x)	(uint_least16_t)(x)
#define	UINT_LEAST32_C(x)	(uint_least32_t)(x)
#define	UINT_LEAST64_C(x)	(uint_least64_t)(x)
#define	INTPTR_C(x)		(intptr_t)(x)
#define	UINTPTR_C(x)		(uintptr_t)(x)
#define	INTMAX_C(x)		(intmax_t)(x)
#define	UINTMAX_C(x)		(uintmax_t)(x)
#endif				/* __STDC_LIMIT_MACROS */

#endif				/* _STDINT_H_ */
