/*
 @LICENSE@
*/

/*-
 * Copyright (c) 2003 Cristan Szmajda.  All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by Cristan Szmajda. 
 * 4. The name Cristan Szmajda may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY CRISTAN SZMAJDA ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL CRISTAN SZMAJDA BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * $Id: stdlib.h,v 1.3 2004/03/23 06:09:08 benjl Exp $
 */

#ifndef	_STDLIB_H_
#define	_STDLIB_H_

//#include <types.h>
#include <stdint.h>
#include <stddef.h>

#define	EXIT_FAILURE	1
#define	EXIT_SUCCESS	0

#define INLINE static __inline__

typedef struct {
	int             quot,
	                rem;
} div_t;
typedef struct {
	long            quot,
	                rem;
} ldiv_t;
typedef struct {
	long long       quot,
	                rem;
} lldiv_t;

// INLINE void abort(void) { assert(0); } /* XXX */
// INLINE void *alloca(size_t n) { return malloc(n); } /* XXX */

INLINE int
abs(int x)
{
	return x < 0 ? -x : x;
}
INLINE long
labs(long x)
{
	return x < 0 ? -x : x;
}
INLINE long long
llabs(long long x)
{
	return x < 0 ? -x : x;
}

// #include <div.c>
// #include <ldiv.c>
// #include <lldiv.c>

/*
 * unimplemented system functions 
 */
#if 0
INLINE int
atexit(void     (*func) (void))
{
	return 0;
}
INLINE void
exit(int status)
{
	for (;;);
}
INLINE int
system(const char *cmd)
{
	return -1;
}
char           *getenv(const char *);
int             putenv(const char *);
int             setenv(const char *, const char *, int);
void            unsetenv(const char *);

double          strtod(const char *s, const char **endp);
long            strtol(const char *s, const char **endp, int base);
unsigned long
                strtoul(const char *s, const char **endp, int base);
long long
                strtoll(const char *s, const char **endp, int base);
unsigned long long
                strtoull(const char *s, const char **endp, int base);
#endif

// #include <fac2.c>
// #include <ilog2.c>

/*
 * deprecated 
 */
// INLINE int ffs(int x) { return x != 0 ? fac2(x) + 1 : 0; }

// #include <strtoull.c>

/*
 * deprecated 
 */
#if 0
INLINE int
atoi(const char *s)
{
	return strtol(s, NULL, 10);
}
INLINE long
atol(const char *s)
{
	return strtol(s, NULL, 10);
}
INLINE double
atof(const char *s)
{
	return strtod(s, NULL);
}
#endif

/*
 * allocate n bytes of memory 
 */
void           *malloc(size_t n);
void            __malloc_init(void *bss_p, void *top_p);

#if 0
INLINE void    *
calloc(size_t n, size_t size)
{
	void           *p;

	if ((p = malloc(n * size)) != NULL)
		memset(p, '\0', n * size);
	return p;
}

/*
 * resize buffer allocated by malloc, with contents preserved 
 */
void           *realloc(void *p, size_t n);
#endif

/*
 * free buffer allocated by malloc (or strdup) 
 */
void            free(void *p);


#define	RAND_MAX	INT_MAX
// #include <rand.c>

// #include <bsearch.c>
// #include <qsort.c>

#endif				/* _STDLIB_H_ */
