/*
 @LICENSE@
*/

/*-
 * Copyright (c) 2003 Cristan Szmajda.  All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by Cristan Szmajda. 
 * 4. The name Cristan Szmajda may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY CRISTAN SZMAJDA ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL CRISTAN SZMAJDA BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * $Id: ctype.c,v 1.2 2004/01/27 22:23:56 benjl Exp $
 */

#include <ctype.h>

#ifndef	SAVE_SPACE

/*
 * lookup table of character attributes, -1 to 255 
 */
const unsigned char __ctype[257] = {
	0,			/* EOF */
	_CTYPE_CNTRL,		/* NUL */
	_CTYPE_CNTRL,		/* SOH */
	_CTYPE_CNTRL,		/* STX */
	_CTYPE_CNTRL,		/* ETX */
	_CTYPE_CNTRL,		/* EOT */
	_CTYPE_CNTRL,		/* ENQ */
	_CTYPE_CNTRL,		/* ACK */
	_CTYPE_CNTRL,		/* BEL */
	_CTYPE_CNTRL,		/* BS */
	_CTYPE_CNTRL | _CTYPE_SPACE | _CTYPE_BLANK,	/* HT */
	_CTYPE_CNTRL | _CTYPE_SPACE,	/* LF */
	_CTYPE_CNTRL | _CTYPE_SPACE,	/* VT */
	_CTYPE_CNTRL | _CTYPE_SPACE,	/* FF */
	_CTYPE_CNTRL | _CTYPE_SPACE,	/* CR */
	_CTYPE_CNTRL,		/* SO */
	_CTYPE_CNTRL,		/* SI */
	_CTYPE_CNTRL,		/* DLE */
	_CTYPE_CNTRL,		/* DC1 */
	_CTYPE_CNTRL,		/* DC2 */
	_CTYPE_CNTRL,		/* DC3 */
	_CTYPE_CNTRL,		/* DC4 */
	_CTYPE_CNTRL,		/* NAK */
	_CTYPE_CNTRL,		/* SYN */
	_CTYPE_CNTRL,		/* ETB */
	_CTYPE_CNTRL,		/* CAN */
	_CTYPE_CNTRL,		/* EM */
	_CTYPE_CNTRL,		/* SUB */
	_CTYPE_CNTRL,		/* ESC */
	_CTYPE_CNTRL,		/* FS */
	_CTYPE_CNTRL,		/* GS */
	_CTYPE_CNTRL,		/* RS */
	_CTYPE_CNTRL,		/* US */
	_CTYPE_SPACE | _CTYPE_BLANK,	/* SP */
	_CTYPE_PUNCT,		/* '!' */
	_CTYPE_PUNCT,		/* '"' */
	_CTYPE_PUNCT,		/* '#' */
	_CTYPE_PUNCT,		/* '$' */
	_CTYPE_PUNCT,		/* '%' */
	_CTYPE_PUNCT,		/* '&' */
	_CTYPE_PUNCT,		/* '\'' */
	_CTYPE_PUNCT,		/* '(' */
	_CTYPE_PUNCT,		/* ')' */
	_CTYPE_PUNCT,		/* '*' */
	_CTYPE_PUNCT,		/* '+' */
	_CTYPE_PUNCT,		/* ',' */
	_CTYPE_PUNCT,		/* '-' */
	_CTYPE_PUNCT,		/* '.' */
	_CTYPE_PUNCT,		/* '/' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '0' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '1' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '2' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '3' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '4' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '5' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '6' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '7' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '8' */
	_CTYPE_DIGIT | _CTYPE_XDIGIT,	/* '9' */
	_CTYPE_PUNCT,		/* ':' */
	_CTYPE_PUNCT,		/* ';' */
	_CTYPE_PUNCT,		/* '<' */
	_CTYPE_PUNCT,		/* '=' */
	_CTYPE_PUNCT,		/* '>' */
	_CTYPE_PUNCT,		/* '?' */
	_CTYPE_PUNCT,		/* '@' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'A' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'B' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'C' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'D' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'E' */
	_CTYPE_UPPER | _CTYPE_XDIGIT,	/* 'F' */
	_CTYPE_UPPER,		/* 'G' */
	_CTYPE_UPPER,		/* 'H' */
	_CTYPE_UPPER,		/* 'I' */
	_CTYPE_UPPER,		/* 'J' */
	_CTYPE_UPPER,		/* 'K' */
	_CTYPE_UPPER,		/* 'L' */
	_CTYPE_UPPER,		/* 'M' */
	_CTYPE_UPPER,		/* 'N' */
	_CTYPE_UPPER,		/* 'O' */
	_CTYPE_UPPER,		/* 'P' */
	_CTYPE_UPPER,		/* 'Q' */
	_CTYPE_UPPER,		/* 'R' */
	_CTYPE_UPPER,		/* 'S' */
	_CTYPE_UPPER,		/* 'T' */
	_CTYPE_UPPER,		/* 'U' */
	_CTYPE_UPPER,		/* 'V' */
	_CTYPE_UPPER,		/* 'W' */
	_CTYPE_UPPER,		/* 'X' */
	_CTYPE_UPPER,		/* 'Y' */
	_CTYPE_UPPER,		/* 'Z' */
	_CTYPE_PUNCT,		/* '[' */
	_CTYPE_PUNCT,		/* '\\' */
	_CTYPE_PUNCT,		/* ']' */
	_CTYPE_PUNCT,		/* '^' */
	_CTYPE_PUNCT,		/* '_' */
	_CTYPE_PUNCT,		/* '`' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'a' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'b' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'c' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'd' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'e' */
	_CTYPE_LOWER | _CTYPE_XDIGIT,	/* 'f' */
	_CTYPE_LOWER,		/* 'g' */
	_CTYPE_LOWER,		/* 'h' */
	_CTYPE_LOWER,		/* 'i' */
	_CTYPE_LOWER,		/* 'j' */
	_CTYPE_LOWER,		/* 'k' */
	_CTYPE_LOWER,		/* 'l' */
	_CTYPE_LOWER,		/* 'm' */
	_CTYPE_LOWER,		/* 'n' */
	_CTYPE_LOWER,		/* 'o' */
	_CTYPE_LOWER,		/* 'p' */
	_CTYPE_LOWER,		/* 'q' */
	_CTYPE_LOWER,		/* 'r' */
	_CTYPE_LOWER,		/* 's' */
	_CTYPE_LOWER,		/* 't' */
	_CTYPE_LOWER,		/* 'u' */
	_CTYPE_LOWER,		/* 'v' */
	_CTYPE_LOWER,		/* 'w' */
	_CTYPE_LOWER,		/* 'x' */
	_CTYPE_LOWER,		/* 'y' */
	_CTYPE_LOWER,		/* 'z' */
	_CTYPE_PUNCT,		/* '{' */
	_CTYPE_PUNCT,		/* '|' */
	_CTYPE_PUNCT,		/* '}' */
	_CTYPE_PUNCT,		/* '~' */
	_CTYPE_CNTRL,		/* DEL */

	/*
	 * no locale stuff yet 
	 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/*
 * lookup table of numeral values, -1 to 255 
 */
const unsigned char __digit[257] = {
	99,

	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 99, 99, 99, 99, 99, 99,
	99, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
	25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 99, 99, 99, 99, 99,
	99, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
	25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 99, 99, 99, 99, 99,

	/*
	 * no locale stuff yet 
	 */
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
};

#endif				/* SAVE_SPACE */
