/*
 @LICENSE@
*/

/*
 * K&R Malloc 
 */

#define NALLOC  1024		/* minimum #units to request */

#define NULL ((void*)0)

typedef long long Align;	/* for alignment to long long boundary */

union header {			/* block header */
	struct {
		union header   *ptr;	/* next block if on free list */
		unsigned        size;	/* size of this block */
	} s;
	Align           x;	/* force alignment of blocks */
};

typedef union header Header;

static Header  *morecore(unsigned nu);

static Header   base;		/* empty list to get started */
static void    *bss = (void *) 0x20000000;	/* `sbrk` point */
static void    *top = (void *) 0x30000000;	/* `sbrk` point */
static Header  *freep = NULL;	/* start of free list */

void
__malloc_init(void *bss_p, void *top_p)
{
	bss = bss_p;
	top = top_p;
}


/*
 * malloc: general-purpose storage allocator 
 */
void           *
malloc(unsigned nbytes)
{
	Header         *p,
	               *prevp;
	unsigned        nunits;
	nunits = (nbytes + sizeof(Header) - 1) / sizeof(Header) + 1;
	if ((prevp = freep) == NULL) {	/* no free list yet */
		base.s.ptr = freep = prevp = &base;
		base.s.size = 0;
	}
	for (p = prevp->s.ptr;; prevp = p, p = p->s.ptr) {
		if (p->s.size >= nunits) {	/* big enough */
			if (p->s.size == nunits)	/* exactly */
				prevp->s.ptr = p->s.ptr;
			else {	/* allocate tail end */
				p->s.size -= nunits;
				p += p->s.size;
				p->s.size = nunits;
			}
			freep = prevp;
			return (void *) (p + 1);
		}
		if (p == freep) {	/* wrapped around free list */
			if ((p = (Header *) morecore(nunits)) == NULL) {
				return NULL;	/* none left */
			} else {

			}
		}
	}
}

/*
 * free: put block ap in free list 
 */
void
free(void *ap)
{
	Header         *bp,
	               *p;

	bp = (Header *) ap - 1;	/* point to block header */
	for (p = freep; !(bp > p && bp < p->s.ptr); p = p->s.ptr)
		if (p >= p->s.ptr && (bp > p || bp < p->s.ptr))
			break;	/* freed block at start or end of arena */

	if (bp + bp->s.size == p->s.ptr) {	/* join to upper nbr */
		bp->s.size += p->s.ptr->s.size;
		bp->s.ptr = p->s.ptr->s.ptr;
	} else
		bp->s.ptr = p->s.ptr;
	if (p + p->s.size == bp) {	/* join to lower nbr */
		p->s.size += bp->s.size;
		p->s.ptr = bp->s.ptr;
	} else
		p->s.ptr = bp;
	freep = p;
}


/*
 * morecore: this is sbrk on unix systems 
 */
static Header  *
morecore(unsigned nu)
{
	void           *cp;
	Header         *up;
	cp = bss;
	if (nu < NALLOC)
		nu = NALLOC;
	if (bss + (nu * sizeof(Header)) > top)
		return NULL;
	bss += (nu * sizeof(Header));

	up = (Header *) cp;
	up->s.size = nu;
	free((void *) (up + 1));
	return freep;
}
