/*
 @LICENSE@
*/

#ifndef ALPHA_DENSE_IO_H
#define ALPHA_DENSE_IO_H
#include <dev.h>

#define mb() __asm__ __volatile__("mb": : :"memory")

static inline uint8_t dense_mem_read_8(bus_space_t handle,
				       uintptr_t offset);
static inline void dense_mem_write_8(bus_space_t handle, uintptr_t offset,
				     uint8_t value);

static inline uint16_t dense_mem_read_16(bus_space_t handle,
					 uintptr_t offset);
static inline void dense_mem_write_16(bus_space_t handle, uintptr_t offset,
				      uint16_t value);

static inline uint32_t dense_mem_read_32(bus_space_t handle,
					 uintptr_t offset);
static inline void dense_mem_write_32(bus_space_t handle, uintptr_t offset,
				      uint32_t value);

static inline uint64_t dense_mem_read_64(bus_space_t handle,
					 uintptr_t offset);
static inline void dense_mem_write_64(bus_space_t handle, uintptr_t offset,
				      uint64_t value);


static inline   uint8_t
dense_mem_read_8(bus_space_t handle, uintptr_t offset)
{
	uint8_t         result;
	result = *(volatile uint8_t *)
	    (offset + handle);
	mb();
	return result;
}

static inline void
dense_mem_write_8(bus_space_t handle, uintptr_t offset, uint8_t value)
{
	*(volatile uint8_t *) (offset + handle)
	    = value;
	mb();
}


static inline   uint16_t
dense_mem_read_16(bus_space_t handle, uintptr_t offset)
{
	uint16_t        result;
	result = *(volatile uint16_t *)
	    (offset + handle);
	mb();
	return result;
}

static inline void
dense_mem_write_16(bus_space_t handle, uintptr_t offset, uint16_t value)
{
	*(volatile uint16_t *) (offset + handle)
	    = value;
	mb();
}


static inline   uint32_t
dense_mem_read_32(bus_space_t handle, uintptr_t offset)
{
	uint32_t        result;
	result = *(volatile uint32_t *)
	    (offset + handle);
	mb();
	return result;
}

static inline void
dense_mem_write_32(bus_space_t handle, uintptr_t offset, uint32_t value)
{
	*(volatile uint32_t *) (offset + handle)
	    = value;
	mb();
}


static inline   uint64_t
dense_mem_read_64(bus_space_t handle, uintptr_t offset)
{
	uint64_t        result;
	result = *(volatile uint64_t *)
	    (offset + handle);
	mb();
	return result;
}

static inline void
dense_mem_write_64(bus_space_t handle, uintptr_t offset, uint64_t value)
{
	*(volatile uint64_t *) (offset + handle)
	    = value;
	mb();
}


#endif
