/*
 @LICENSE@
*/

#ifndef MUNGI_ALPHA_IO_IO_H_
#define MUNGI_ALPHA_IO_IO_H_

#include <dev.h>
#include <assert.h>

#ifdef ALPHA_TSUNAMI

/*
 * Don't need sparse I/O on Tsunami 
 */
#include <io/dense_mem.h>
#define sparse_io_read_8	dense_mem_read_8
#define sparse_io_write_8	dense_mem_write_8
#define sparse_io_read_16	dense_mem_read_16
#define sparse_io_write_16	dense_mem_write_16
#define sparse_io_read_32	dense_mem_read_32
#define sparse_io_write_32	dense_mem_write_32

#else

static inline uint8_t sparse_io_read_8(bus_space_t handle,
				       uintptr_t offset);
static inline void sparse_io_write_8(bus_space_t handle, uintptr_t offset,
				     uint8_t value);

static inline uint16_t sparse_io_read_16(bus_space_t handle,
					 uintptr_t offset);
static inline void sparse_io_write_16(bus_space_t handle, uintptr_t offset,
				      uint16_t value);

static inline uint32_t sparse_io_read_32(bus_space_t handle,
					 uintptr_t offset);
static inline void sparse_io_write_32(bus_space_t handle, uintptr_t offset,
				      uint32_t value);

#define mb() \
__asm__ __volatile__("mb": : :"memory")

#define __kernel_extbl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("extbl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })

#define __kernel_insbl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("insbl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })

#define __kernel_extwl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("extwl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })

#define __kernel_inswl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("inswl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })

#define __kernel_extbl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("extbl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })

#define __kernel_insbl(val, shift)					\
  ({ unsigned long __kir;						\
     __asm__("insbl %2,%1,%0" : "=r"(__kir) : "rI"(shift), "r"(val));	\
     __kir; })


static inline   uint8_t
sparse_io_read_8(bus_space_t handle, uintptr_t offset)
{
	/*
	 * XXX 
	 */
	long            temp;
#if 0
	UserPrint("address is: %p\n", (volatile int *) (((offset)
							 << 5) + handle +
							0x00));
	UserPrint("offset is: %p\n", handle);
	UserPrint("blah: %lx\n", ((offset) << 5));
#endif
	temp = *(volatile int *) (((offset) << 5)
				  + handle + 0x00);
	mb();
	return __kernel_extbl(temp, (offset) & 3);
}

static inline void
sparse_io_write_8(bus_space_t handle, uintptr_t offset, uint8_t value)
{
	/*
	 * XXX 
	 */
	unsigned long   w = __kernel_insbl(value, (offset) & 3);
	*(volatile unsigned int *) (((offset) << 5)
				    + handle + 0x00) = w;
	mb();
}

static inline   uint16_t
sparse_io_read_16(bus_space_t handle, uintptr_t offset)
{
	long            temp;
	temp = *(volatile uint32_t *) (((offset) << 5)
				       + handle + 0x08);
	mb();
	return __kernel_extwl(temp, (offset) & 3);
}

static inline void
sparse_io_write_16(bus_space_t handle, uintptr_t offset, uint16_t value)
{
	unsigned long   w = __kernel_inswl(value, (offset) & 3);
	*(volatile uint32_t *) (((offset) << 5)
				+ handle + 0x08) = w;
	mb();
}

static inline   uint32_t
sparse_io_read_32(bus_space_t handle, uintptr_t offset)
{
	uint32_t        result;

#if 0
	UserPrint("sparse_io_read_32([%d <%p %lx> %lx %lx %lx], %lx)\n",
		  handle.identifier, handle,
		  handle.address.passwd, handle.flags, handle.data,
		  handle.size, offset);
	UserPrint("sparse_io_read_32: about to read 0x%p\n",
		  ((offset) << 5) + handle + 0x18);
#endif

	result = *(volatile uint32_t *) (((offset) << 5)
					 + handle + 0x18);
	mb();

	return result;
}

static inline void
sparse_io_write_32(bus_space_t handle, uintptr_t offset, uint32_t value)
{
	*(volatile uint32_t *) (((offset) << 5)
				+ handle + 0x18) = value;
	mb();
}
#endif
#endif
