/*****************************************************************************
 * block_driver.h
 * Define the interface to a block device
 */

#ifndef BLOCK_DRIVER_H
#define BLOCK_DRIVER_H

#include "driver_interface.h"

struct block_ops {
	void *(*setup)(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf);
	void  (*enable)(void *device);
	int   (*read)(void *device, int diskid, 
		      callback_t callback, callback_data_t callback_data,
		      block_seq blocks); 
	int   (*write)(void *device, int diskid, callback_t callback, 
		       callback_data_t callback_data,
		       block_seq blocks);
	int (*interrupt)(void *device);

	int  (*maxdisc)(void *device);
	uint64_t (*size)(void *device, int diskid);
	int (*blksz)(void *device, int diskid);
};

static inline void blkdev_enable(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	(*dp->classp->b_ops->enable)(device);
}

static inline int blkdev_read(void *device, int diskid, 
		      callback_t callback, callback_data_t callback_data,
		      block_seq blocks)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return (*dp->classp->b_ops->read)(device, diskid, callback, callback_data, blocks);
}

static inline int blkdev_write(void *device, int diskid, 
		      callback_t callback, callback_data_t callback_data,
		      block_seq blocks)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return (*dp->classp->b_ops->write)(device, diskid, callback, callback_data, blocks);
}

static inline int blkdev_interrupt(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->b_ops->interrupt(device);
}

static inline int blkdev_maxdisc(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->b_ops->maxdisc(device);
}

static inline uint64_t blkdev_size(void *device, int diskid)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->b_ops->size(device, diskid);
}

static inline int blkdev_blksz(void *device, int diskid)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->b_ops->blksz(device, diskid);
}

#endif /* BLOCK_DRIVER_H */
