#ifndef _DRIVER_INTERFACE_H
#define _DRIVER_INTERFACE_H

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <driver_env_glue.h>

#include "bfrdriver_types.h"


/*
 * A `sequence' of bus spaces, to be passed as argument to the driver.
 */
typedef struct {
	unsigned int c;
	bus_space_t *v;
} bs_seq;

/* CEG HACK */
#define NULL_MEM_HANDLE ((mem_handle_t){NULL,0})

/*
 * A descriptor of an I/O request for the IDE driver.
 */
typedef struct {
	mem_handle_t handle;	/* where in memory to transfer to/from */
	uint64_t blockid;	/* Where on disc to transfer to/from */
	uint64_t block_count;	/* How many sectors to transfer */
	uintptr_t key;		/* Tag assigned by caller */
} block_t;

/*
 * Sequence of blocks to be read in a single request to the driver.
 */
typedef struct {
	unsigned int c;
	block_t *v;
} block_seq;

struct pci_id { uint16_t vendor; uint16_t device; };

typedef void (*callback_t)(callback_data_t, uint64_t);
extern callback_t null_callback;

struct block_ops;
struct network_ops;
struct character_ops;
struct buffer_ops;

enum driver_class { block_device, netif_device, character_device, buffer_manager };

struct driver {
	const char *name;
	enum driver_class class;
	union {
		struct block_ops *b_ops;
		struct netif_ops *n_ops;
		struct character_ops *c_ops;
		struct buffer_ops *bfr_ops;
	};
};

struct driver_instance {
	struct driver *classp;
};

/* Interface functions */
int get_pinnable_memory(size_t size, mem_handle_t * handle);

int             register_cpu_interrupt_handler(/*host_handle_t host,*/ int num,
					       void *fn, void *data);

#endif
