#ifndef _DRIVER_ENV_GLUE_H
#define _DRIVER_ENV_GLUE_H

#include <config.h>
#include <pagebits.h>

#define DMA_TO_DEVICE 1
#define DMA_FROM_DEVICE 2
#define DMA_BIDIRECTIONAL 3

#define GLUE_DMABUF_NENTS(x) 0
#define GLUE_DMABUF_ENT_LEN(x, i) 0
#define GLUE_DMABUF_ENT_DMAADDR(x, i) 0

typedef int dma_handle_t;
typedef unsigned long dma_addr_t;
typedef void *mem_handle_t;
typedef void   *bus_addr_t;

typedef struct {
	void            (*function) (void *);
	void           *data;
} exec_handle_t;

#define EXECUTE_HANDLE(x) (x.function(x.data))

#define NULL_EXEC_HANDLE { NULL, NULL }

/*
 * Kind of bus space
 */
enum bus_access {
	IOPORT,
	PCI_CONFIG,
	PCI_MEM,
};

//struct pci_dev;

/*
 * Describe access to part of a device's bus space
 * -- currently IO space and PCI Config space; could add MMIO space later.
 */
typedef union {
	struct bus_ioport {
		enum bus_access accesstype; /* IOPORT */
		/*unsigned short base;
		  unsigned short len;*/
		mem_handle_t address;
	} ioports;
	struct bus_pciconfig {
		enum bus_access accesstype; /* PCI_CONFIG */
		//struct pci_dev *devp;
	} pciconfig;
} bus_space_t;

struct pci_sglist {
	uintptr_t address;
};
/*
 * A DMA scatter-gather  list in addresses as seen from the device.
 */
typedef struct dmabuf_handle {
	unsigned	nentries;
	struct pci_sglist   sglist[];
} *dmabuf_handle_t;

#define UserPrint printf

static inline void *
get_addr(mem_handle_t handle) 
{
	return (void*)handle;
}

static inline mem_handle_t 
new_mem_handle(void *addr, size_t sz) 
{
	(void)sz;
	return (mem_handle_t)addr;
}

static inline void
free_mem_handle(mem_handle_t *h)
{
	(void)h;
}

static inline mem_handle_t
handle_increment(mem_handle_t *handle, size_t sz)
{
	return (mem_handle_t)((char *)handle + sz);
}

//extern int pagesize;
#define pagesize PAGESIZE

#endif
