#ifndef ALPHA_DENSE_IO_H
#define ALPHA_DENSE_IO_H
#include <bus.h>

#define mb() __asm__ __volatile__("mb": : :"memory")

static inline uint8_t dense_mem_read_8(bus_space_t handle,
					   uintptr_t offset);
static inline void dense_mem_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t dense_mem_read_16(bus_space_t handle,
					   uintptr_t offset);
static inline void dense_mem_write_16(bus_space_t handle, uintptr_t offset,
					uint16_t value);

static inline uint32_t dense_mem_read_32(bus_space_t handle,
					   uintptr_t offset);
static inline void dense_mem_write_32(bus_space_t handle, uintptr_t offset,
					uint32_t value);

static inline uint64_t dense_mem_read_64(bus_space_t handle,
					   uintptr_t offset);
static inline void dense_mem_write_64(bus_space_t handle, uintptr_t offset,
					uint64_t value);


static inline uint8_t
dense_mem_read_8(bus_space_t handle, uintptr_t offset)
{
	uint8_t result;
#if 0
	UserPrint("8 -- Reeading from: %p\n", (volatile uint32_t *) 
		  (offset + handle.data + handle.address.address));
	UserPrint("Reading from .addr: %p\n", handle.address.address);
	UserPrint("Reading from .data: %p\n", handle.data);
#endif
	result = *(volatile uint8_t *)
		(offset + handle.data + handle.address.address);
	mb();
	return result;
}

static inline void
dense_mem_write_8(bus_space_t handle, uintptr_t offset, uint8_t value)
{
#if 0
	UserPrint("8 -- Writing to: %p\n", (volatile uint32_t *) 
		  (offset + handle.data + handle.address.address));
	UserPrint("Writing to .addr: %p\n", handle.address.address);
	UserPrint("Writing to .data: %p\n", handle.data);
#endif
	*(volatile uint8_t *) (offset + handle.data + handle.address.address)
		= value;
	mb();
}


static inline uint16_t
dense_mem_read_16(bus_space_t handle, uintptr_t offset)
{
	uint16_t result;
	result = *(volatile uint16_t *)
		(offset + handle.data + handle.address.address);
	mb();
	return result;
}

static inline void
dense_mem_write_16(bus_space_t handle, uintptr_t offset, uint16_t value)
{
	*(volatile uint16_t *) (offset + handle.data + handle.address.address)
		= value;
	mb();
}


static inline uint32_t
dense_mem_read_32(bus_space_t handle, uintptr_t offset)
{
	uint32_t result;
	result = *(volatile uint32_t *)
		(offset + handle.data + handle.address.address);
	mb();
	return result;
}

static inline void
dense_mem_write_32(bus_space_t handle, uintptr_t offset, uint32_t value)
{
	UserPrint("32 -- Writing to: %p\n", (volatile uint32_t *) 
		  (offset + handle.data + handle.address.address));
	*(volatile uint32_t *) (offset + handle.data + handle.address.address)
		= value;
	mb();
}


static inline uint64_t
dense_mem_read_64(bus_space_t handle, uintptr_t offset)
{
	uint64_t result;
	result = *(volatile uint64_t *)
		(offset + handle.data + handle.address.address);
	mb();
	return result;
}

static inline void
dense_mem_write_64(bus_space_t handle, uintptr_t offset, uint64_t value)
{
	*(volatile uint64_t *) (offset + handle.data + handle.address.address)
		= value;
	mb();
}

#define io_read_8 dense_mem_read_8
#define io_write_8   dense_mem_write_8
#define io_read_16   dense_mem_read_16
#define io_write_16  dense_mem_write_16
#define io_read_32   dense_mem_read_32
#define io_write_32  dense_mem_write_32

#endif
