#ifndef IO_H_
#define IO_H_

#include <linux/pci.h>
#include <asm/io.h>
#include "driver_env_glue.h"   /* for bus_space_t */

static inline uint8_t io_read_8(bus_space_t handle, uintptr_t offset);
static inline void io_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t io_read_16(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_16(bus_space_t handle, uintptr_t offset,
					 uint16_t value);

static inline uint32_t io_read_32(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_32(bus_space_t handle, uintptr_t offset,
					 uint32_t value);

#define DEBUG(blah...)

static inline uint8_t 
io_read_8(bus_space_t handle, uintptr_t offset)
{
	DEBUG("io_read_8: reading from %d\n", offset);
	if (handle.is_pciconfig) {
		uint8_t result;
		pci_read_config_byte(handle.u.pciconfig.pcidev,
			offset, &result);
		return result;
	}
	else if (handle.u.normal.flags & IORESOURCE_IO)
	        return inb(offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
	        return readb(handle.u.normal.start + offset);
	else {
	        printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
		        handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
		return 0;
	}
}

static inline void
io_write_8(bus_space_t handle, uintptr_t offset,
		 uint8_t value)
{	 
	DEBUG("io_write_8: writing %x to %d\n", value, offset);
	if (handle.is_pciconfig)
		pci_write_config_byte(handle.u.pciconfig.pcidev,
			offset, value);
	else if (handle.u.normal.flags & IORESOURCE_IO)
	        outb(value, offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
	        writeb(value, handle.u.normal.start + offset);
	else
	       printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
			handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
}


static inline uint16_t 
io_read_16(bus_space_t handle, uintptr_t offset)
{
	DEBUG("io_read_16: reading from %d\n", offset);
	if (handle.is_pciconfig) {
		uint16_t result;
		pci_read_config_word(handle.u.pciconfig.pcidev,
			offset, &result);
		return result;
	}
	else if (handle.u.normal.flags & IORESOURCE_IO)
	        return inw(offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
		return readw(handle.u.normal.start + offset);
	else {
	        printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
			handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
		return 0;
	}
}

static inline void
io_write_16(bus_space_t handle, uintptr_t offset, uint16_t value)
{	
	DEBUG("io_write_16: writing %x to %d\n", value, offset);
	if (handle.is_pciconfig)
		pci_write_config_word(handle.u.pciconfig.pcidev,
			offset, value);
	else if (handle.u.normal.flags & IORESOURCE_IO)
	        outw(value, offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
	        writew(value, handle.u.normal.start + offset);
	else
	        printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
			handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
}

static inline uint32_t 
io_read_32(bus_space_t handle, uintptr_t offset)
{
	DEBUG("io_read_32: reading from %d\n", offset);
	if (handle.is_pciconfig) {
		uint32_t result;
		pci_read_config_dword(handle.u.pciconfig.pcidev,
			offset, &result);
		return result;
	}
	else if (handle.u.normal.flags & IORESOURCE_IO)
		        return inl(offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
		return readl(handle.u.normal.start + offset);
	else {
	       printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
			handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
		return 0;
	}
}

static inline void
io_write_32(bus_space_t handle, uintptr_t offset,
		 uint32_t value)
{
	DEBUG("io_write_32: writing %x to %d\n", value, offset);
	if (handle.is_pciconfig)
		pci_write_config_dword(handle.u.pciconfig.pcidev,
			offset, value);
	else if (handle.u.normal.flags & IORESOURCE_IO)
	        outl(value, offset);
	else if (handle.u.normal.flags & IORESOURCE_MEM)
	        writel(value, handle.u.normal.start + offset);
	else
	        printk("bad bus_space_t: {start:0x%lx,end:0x%lx,flags:0x%lx}\n",
			handle.u.normal.start, handle.u.normal.end,
			handle.u.normal.flags);
}

#endif
