#ifndef PCI_DMA_H
#define PCI_DMA_H
#include "assert.h"

/*
 * This must match whatever the kernel was compiled with.
 */
typedef enum direction 
{ 
	mem_to_bus = DMA_TO_DEVICE, 
	bus_to_mem = DMA_FROM_DEVICE, 
	consistent = DMA_BIDIRECTIONAL 
} direction_t;


#define GLUE_DMABUF_NENTS(x) ((x).nents)
#define GLUE_DMABUF_ENT_LEN(x,i) (sg_dma_len( &(x).list[i] ))
#define GLUE_DMABUF_ENT_DMAADDR(x,i) (sg_dma_address( &(x).list[i] ))
#define GLUE_DMABUF_ENT_VADDR(x,i) \
	(page_address((x).list[i].page) + (x).list[i].offset)

typedef struct pci_dev *dma_handle_t;


typedef struct {
	struct scatterlist *list, *origlist;
	size_t size;
	unsigned nents;
} mem_handle_t;



mem_handle_t new_mem_handle(void *addr, size_t sz);
void free_mem_handle(mem_handle_t *handle);

typedef struct {
	struct scatterlist *list;
	unsigned nents;
	int direction;
} dmabuf_handle_t;

static inline void *get_addr(mem_handle_t handle)
{
	assert(handle.nents >= 1);
	return GLUE_DMABUF_ENT_VADDR(handle, 0);
}



/*
 * setup_dma
 *   For the device referred to by dma_handle,
 *   map the area of memory in the driver's address space
 *   from handle to handle + size - 1
 *   into a scatter-gather list, and return the scatter-gather list 
 *   and its length in *sgp  and *sglen.
 *
 *   This function allocates the sglist.
 */
extern int setup_dma(dma_handle_t dma, 
	      mem_handle_t handle, 
	      size_t sz, 
	      dmabuf_handle_t *dmabuf,
	      direction_t dir);

extern int teardown_dma(dma_handle_t dma, dmabuf_handle_t *x);

/*
 * Allocate sz bytes of PCI=consistent memory; i.e., memory that is
 * visible from the PCI bus (at busaddr) and in the driver (at vaddr).
 * Remember to use memory barriers where appropriate!
 * 
 * Returns: 0 for success; anything else means error.
 */
extern int get_pci_consistent_memory(dma_handle_t dma, 
				     size_t sz, 
				     mem_handle_t *vaddr, 
				     dmabuf_handle_t *busaddr);


#endif /* PCI_DMA_H */
