/************************************************************************
 * glue.h -- os-specific routines/methods for the IDE driver.
 * This file for linux in user mode.
 *
 * $Id: driver_env_glue.h,v 1.4 2004/05/06 03:03:24 ianw Exp $
 *
 */

#ifndef _DRIVER_ENV_GLUE_H
#define _DRIVER_ENV_GLUE_H

#include <linux/types.h>
#include <linux/usrdrv.h>
#include <stdio.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>

#include <UserPrint.h>

typedef int dma_handle_t;
typedef unsigned long dma_addr_t;

/*
 * Kind of bus space
 */
enum bus_access {
	IOPORT,
	PCI_CONFIG,
	PCI_MEM,
};

struct pci_dev;

/*
 * Describe access to part of a device's bus space
 * -- currently IO space and PCI Config space; could add MMIO space later.
 */
typedef union {
	struct bus_ioport {
		enum bus_access accesstype; /* IOPORT */
		unsigned long base;
		unsigned long len;
	} ioports;
	struct bus_pciconfig {
		enum bus_access accesstype; /* PCI_CONFIG */
		struct pci_dev *devp;
	} pciconfig;
} *bus_space_t;


/*
 * A DMA scatter-gather  list in addresses as seen from the device.
 */

#include <asm/scatterlist.h>
#include <linux/usrblk.h>

typedef struct usr_io_req *dmabuf_handle_t;

#define GLUE_DMABUF_NENTS(x) ((x)->nentries)
#define GLUE_DMABUF_ENT_LEN(x, i) ((x)->sglist[i].dma_length)
#define GLUE_DMABUF_ENT_DMAADDR(x, i) ((x)->sglist[i].dma_address)


/*
 * This must match whatever the kernel was compiled with.
 */
typedef enum direction 
{ 
	mem_to_bus = DMA_TO_DEVICE, 
	bus_to_mem = DMA_FROM_DEVICE, 
	consistent = DMA_BIDIRECTIONAL 
} direction_t;

enum mem_type { mem_virt, mem_bus };
typedef union {
	struct {
		enum mem_type t; /* == mem_virt */
		void *vaddr;
	} v;
	struct {
		enum mem_type t; /* == mem_bus */
		struct usr_io_req *iorp;
	} p;
} mem_handle_t;

/*
 * Data to be placed in callback from driver.
 */
typedef struct {
	int	notify; /* master file fd */
	char   *cthis; /* buffer that was being used (may now be released */
} callback_data_t;

extern callback_data_t null_callback_data;

static inline void *
get_addr(mem_handle_t handle) 
{
	assert(handle.v.t == mem_virt);
	return handle.v.vaddr;
}

static inline mem_handle_t 
new_mem_handle(void *addr, size_t sz) 
{
	mem_handle_t mp;
	mp.v.t = mem_virt;
	mp.v.vaddr = addr;
	return mp;
}

static inline void
free_mem_handle(mem_handle_t *h)
{
	(void)h;
}

static inline mem_handle_t
handle_increment(mem_handle_t *handle, size_t sz)
{
	mem_handle_t *mp = handle;

	if (mp->v.t == mem_virt){
		mp->v.vaddr = ((char *)mp->v.vaddr) + sz;
		return *mp;
	}

	assert(sz/512 == mp->p.iorp->nsectors);
	return *mp;
}

extern int linux_init(int busid, int slot, int function, char *name);


extern int create_interrupt_thread(int irq, void *device, int (*handler)(void *));

extern void destroy_interrupt_thread(void);

static void inline 
udelay(unsigned long usec) {
	struct timespec ts;

	ts.tv_sec = 0;
	ts.tv_nsec = usec * 1000;
	//nanosleep(&ts, NULL);
}

#define PAGESIZE pagesize
extern int pagesize;

extern int linux_loop_init(char *name, int dma, void *cookie);

int
__setup_dma(dma_handle_t dma, 
	  mem_handle_t handle, 
	  size_t sz, 
	  dmabuf_handle_t *dmabuf,
	    direction_t dir);
int 
__teardown_dma(dma_handle_t dma, dmabuf_handle_t *x);


#endif /* _DRIVER_ENV_GLUE_H_ */
