#ifndef PCI_CONF_PCI_CONF_H_
#define PCI_CONF_PCI_CONF_H_
#include <pci/pci.h>

#include <assert.h>
#ifdef DEBUG_PCI
# undef PRINT_DEBUG
# define PRINT_DEBUG UserPrint
#else
# define PRINT_DEBUG(X, ...) /* Nothing */
#endif


static inline uint8_t pci_conf_read_8(bus_space_t handle, uintptr_t offset);
static inline void pci_conf_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t pci_conf_read_16(bus_space_t handle,
					    uintptr_t offset);
static inline void pci_conf_write_16(bus_space_t handle, uintptr_t offset,
					 uint16_t value);

static inline uint32_t pci_conf_read_32(bus_space_t handle,
					    uintptr_t offset);
static inline void pci_conf_write_32(bus_space_t handle, uintptr_t offset,
					 uint32_t value);





static inline uint8_t 
pci_conf_read_8(bus_space_t handle, uintptr_t offset)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_read_8: reading from %p %d\n", bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	return pci_read_byte(bp->devp, offset);

}

static inline void
pci_conf_write_8(bus_space_t handle, uintptr_t offset,
		 uint8_t value)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_write_8: writing %x to %p %d\n", value, bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	pci_write_byte(bp->devp, offset, value);
}


static inline uint16_t 
pci_conf_read_16(bus_space_t handle, uintptr_t offset)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_read_16: reading from %p %d\n", bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	return pci_read_word(bp->devp, offset);
}

static inline void
pci_conf_write_16(bus_space_t handle, uintptr_t offset,
		 uint16_t value)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_write_16: writing %x to %p %d\n", value, bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	pci_write_word(bp->devp, offset, value);
}

static inline uint32_t 
pci_conf_read_32(bus_space_t handle, uintptr_t offset)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_read_32: reading from %p %d\n", bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	return pci_read_long(bp->devp, offset);
}

static inline void
pci_conf_write_32(bus_space_t handle, uintptr_t offset,
		 uint32_t value)
{
	struct bus_pciconfig *bp = &handle->pciconfig;
	PRINT_DEBUG("pci_conf_write_32: writing %x to %p %d\n", value, bp, offset);
	assert(bp->accesstype == PCI_CONFIG);
	pci_write_long(bp->devp, offset, value);
}
#endif
