#include <asm/types.h>
#ifdef HAS_MSA
#include <linux/msa.h>
#endif
#include <asm/unistd.h>
#include <sys/resource.h>
#include <errno.h>
#include <unistd.h>

extern int has_msa;
#ifndef HAS_MSA
typedef unsigned long clk_t;
enum { ONCPU_USER, ONCPU_SYS, NR_MICRO_STATES } msa_states;
#define MSA_SELF 0
#endif

#define TO_NANO 1000000000L
#define UTONANO 1000UL


static inline long msa(int ntimers, int which, clk_t *timers)
{
#ifdef HAS_MSA
	int ret;
	if (has_msa) {
		ret = syscall(__NR_msa, ntimers, which, timers);
		if (ret == -1 && errno == ENOSYS) {
			has_msa = 0;
			goto no_msa;
		}
		return ret;
	}
no_msa: 
#endif
	{
		struct rusage r;
		getrusage(RUSAGE_SELF, &r);
		timers[ONCPU_USER] = r.ru_utime.tv_sec * TO_NANO + r.ru_utime.tv_usec * UTONANO;
		timers[ONCPU_SYS] = r.ru_stime.tv_sec * TO_NANO + r.ru_stime.tv_usec * UTONANO;
		return 0;
	}
	
}

struct msa_name {
	int index;
	char *name;
};

extern struct msa_name msa_names[];

extern void msa_refresh_irqs(void);
extern clk_t irqtime(int);
extern unsigned long irqcount(int);
extern unsigned long max_irq(void);

