#ifndef _DRIVER_ENV_GLUE_H
#define _DRIVER_ENV_GLUE_H

#include <linux/usrdrv.h>
#include <stdio.h>

#define GLUE_DMABUF_NENTS(x) ((x)->nentries)
#define GLUE_DMABUF_ENT_LEN(x, i) ((x)->sglist[i].len)
#define GLUE_DMABUF_ENT_DMAADDR(x, i) ((x)->sglist[i].dmaaddr)

typedef int dma_handle_t;
typedef unsigned long dma_addr_t;
typedef void *mem_handle_t;

/*
 * Data to be placed in callback from driver.
 */
typedef struct {
	int notify;
	void *cthis;
} callback_data_t;

extern callback_data_t null_callback_data;

/*
 * This must match whatever the kernel was compiled with.
 */
typedef enum direction 
{ 
	mem_to_bus = DMA_TO_DEVICE, 
	bus_to_mem = DMA_FROM_DEVICE, 
	consistent = DMA_BIDIRECTIONAL 
} direction_t;


/*
 * Kind of bus space
 */
enum bus_access {
	IOPORT,
	PCI_CONFIG,
	PCI_MEM,
};

struct pci_dev;

/*
 * Describe access to part of a device's bus space
 * -- currently IO space and PCI Config space; could add MMIO space later.
 */
typedef union {
	struct bus_ioport {
		enum bus_access accesstype; /* IOPORT */
		unsigned long base;
		unsigned long len;
	} ioports;
	struct bus_pciconfig {
		enum bus_access accesstype; /* PCI_CONFIG */
		struct pci_dev *devp;
	} pciconfig;
} *bus_space_t;

/*
 * A DMA scatter-gather  list in addresses as seen from the device.
 */
typedef struct dmabuf_handle {
	unsigned	nentries;
	struct usr_pci_sglist   sglist[];
} *dmabuf_handle_t;

#define UserPrint printf

static inline void *
get_addr(mem_handle_t handle) 
{
	return (void*)handle;
}

static inline mem_handle_t 
new_mem_handle(void *addr, size_t sz) 
{
	(void)sz;
	return (mem_handle_t)addr;
}

static inline void
free_mem_handle(mem_handle_t *h)
{
	(void)h;
}

static inline mem_handle_t
handle_increment(mem_handle_t *handle, size_t sz)
{
	return (mem_handle_t)((char *)handle + sz);
}

#define get_addr(x) (x)

extern int pagesize;
#define PAGESIZE pagesize

extern unsigned long interrupts;

#endif
