#ifndef IO_IO_H_
#define IO_IO_H_

#include <assert.h>
#include <sys/io.h>

static inline uint8_t io_read_8(bus_space_t handle, uintptr_t offset);
static inline void io_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t io_read_16(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_16(bus_space_t handle, uintptr_t offset,
					 uint16_t value);

static inline uint32_t io_read_32(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_32(bus_space_t handle, uintptr_t offset,
					 uint32_t value);

//#define DEBUG_IO
#ifdef DEBUG_IO
# undef PRINT_DEBUG
# define PRINT_DEBUG UserPrint
#else
# define PRINT_DEBUG(X, ...) /* Nothing */
#endif

static inline uint8_t 
io_read_8(bus_space_t handle, uintptr_t offset)
{
	struct bus_ioport *bp = &handle->ioports;

	PRINT_DEBUG("io_read_8: reading from %p %ju (%lx))\n", handle, (uintmax_t)offset, (unsigned long)bp->base + offset);
	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset);
	return inb(bp->base + offset);
}

static inline void
io_write_8(bus_space_t handle, uintptr_t offset,
		 uint8_t value)
{
	struct bus_ioport *bp = &handle->ioports;
	PRINT_DEBUG("io_write_8: writing %x to %p %ju (%lx)\n", value, handle, (uintmax_t)offset, (unsigned long)(bp->base + offset));

	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset);

	outb(value, bp->base + offset);
}


static inline uint16_t 
io_read_16(bus_space_t handle, uintptr_t offset)
{
	struct bus_ioport *bp = &handle->ioports;
	PRINT_DEBUG("io_read_16: reading from %p %ju (%lx)\n", handle, (uintmax_t)offset, (unsigned long)(bp->base + offset));
	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset + 1);

	return inw(bp->base + offset);
}

static inline void
io_write_16(bus_space_t handle, uintptr_t offset,
		 uint16_t value)
{
	struct bus_ioport *bp = &handle->ioports;
	PRINT_DEBUG("io_write_16: writing %x to %p %ju (%lx)\n", value, handle, (uintmax_t)offset, (unsigned long)(bp->base + offset));

	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset + 1);


	outw(value, bp->base +offset);
}

static inline uint32_t 
io_read_32(bus_space_t handle, uintptr_t offset)
{
	struct bus_ioport *bp = &handle->ioports;
	PRINT_DEBUG("io_read_32: reading from %p %ju(%lx)\n", handle, (uintmax_t)offset, (unsigned long)(bp->base + offset));

	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset + 3);

	return inl(bp->base + offset);
}

static inline void
io_write_32(bus_space_t handle, uintptr_t offset,
		 uint32_t value)
{
	struct bus_ioport *bp = &handle->ioports;

	PRINT_DEBUG("io_write_32: writing %x to %p %ju (%lx)\n", value, handle, (uintmax_t)offset, (unsigned long)(bp->base + offset));
	assert(bp->accesstype == IOPORT);
	assert(bp->len > offset + 3);

	outl(value, bp->base + offset);
}
#undef PRINT_DEBUG
#endif
