/*
 * lock.h -- simple wrappers for the driver to give locking, 
 * using pthreads primitives
 */

#ifndef _LOCK_H
#define _LOCK_H

#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <string.h>

typedef pthread_mutex_t lock_t;

static inline void
lock_init(lock_t *lp) {
	pthread_mutexattr_t attr;

	pthread_mutexattr_init(&attr);
#ifdef LOCK_DEBUG
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK_NP);
#endif
	pthread_mutex_init(lp, &attr);
	pthread_mutexattr_destroy(&attr);
}

static inline int
lock_lock(lock_t *lp) {
	if (pthread_mutex_lock(lp) == -1) {
		fprintf(stderr, "Error: Lock %p gave error code %s\n", lp, strerror(errno));
		return 1;
	}
	return 0;
}

static inline int 
lock_unlock(lock_t *lp) {
	return pthread_mutex_unlock(lp);
}

static inline int 
lock_destroy(lock_t *lp) {
	return pthread_mutex_destroy(lp);
}

static inline int
lock_maybelock(lock_t *lp)
{
	return pthread_mutex_trylock(lp);
}

#endif
