#ifndef MUNGI_MIPS_IO_IO_H_
#define MUNGI_MIPS_IO_IO_H_

#include <bus.h>
#include <assert.h>

static inline uint8_t io_read_8(bus_space_t handle, uintptr_t offset);
static inline void io_write_8(bus_space_t handle, uintptr_t offset,
					uint8_t value);

static inline uint16_t io_read_16(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_16(bus_space_t handle, uintptr_t offset,
					 uint16_t value);

static inline uint32_t io_read_32(bus_space_t handle,
					    uintptr_t offset);
static inline void io_write_32(bus_space_t handle, uintptr_t offset,
					 uint32_t value);

#define mb() __asm__ __volatile__("sync": : :"memory")

/* TEMPORARY hack by CEG to make PCI drivers work on MIPS */

#define HACK_NUM (3UL)

static uintptr_t
endian_fix( uintptr_t off )
{
	uintptr_t end = 0;

	end = off & HACK_NUM;
	off = off & ~HACK_NUM;

	end = HACK_NUM - end;
	
	off |= end;

	return off;
}

static inline 
uint8_t io_read_8( bus_space_t handle, uintptr_t offset)
{
	volatile uint8_t *c;

	c = handle.address.address + offset;
	c = (void*) endian_fix( (uintptr_t) c);

	UserPrint( "asi: doing read8 from %p\n", c );
	return *c;
}

static inline void 
io_write_8( bus_space_t handle, uintptr_t offset, uint8_t value )
{
	volatile unsigned char *c;

	c = handle.address.address + offset;
	c = (void*) endian_fix( (uintptr_t) c);

	// UserPrint( "asi: doing write8 to %p\n", c );
	*c = value;
}

static inline uint16_t 
io_read_16( bus_space_t handle, uintptr_t offset )
{
	assert( !"asr_16 NYI" );
	return 0;
}

static inline void 
io_write_16( bus_space_t handle, uintptr_t offset, 
					  uint16_t value )
{
	assert( !"asw_16 NYI" );
}

static inline uint32_t 
io_read_32( bus_space_t handle, uintptr_t offset )
{
	volatile uint32_t *p = handle.address.address + offset;
	uint32_t i;
	UserPrint( "sparse read from %p\n", p );
	i = *p;
	return i;
}

static inline void 
io_write_32( bus_space_t handle, uintptr_t offset, uint32_t value )
{
	volatile uint32_t *p = handle.address.address + offset;
	// UserPrint( "sparse write to %p\n", p );

	*p = value;
}

static inline void
wbflush(void)
{
    asm volatile ("sync");
}

#endif

