/*
 * Netdriver.h -- generic network driver interface definitions
 *
 * @LICENCE@
 *
 */
#ifndef NETDRIVER_H_
#define NETDRIVER_H_

#include "bfrdriver.h"

//struct bfr_drv;



struct netstats {
	uint64_t    rx_packets; /* number of received packets */
	uint64_t	rx_dropped; /* number of dropped packets */
	uint64_t	rx_bytes;   /* number of received bytes */
	uint64_t	rx_sovr;    /* number of status FIFO overruns */
	uint64_t	rx_dorn;    /* number of data FIFO overruns */
	uint64_t	rx_idle;    /* number of times receiver idle */
	uint64_t	tx_packets; /* number of sent packets */
	uint64_t	tx_bytes;   /* number of sent bytes */
	uint64_t	tx_dropped; /* number of dropped transmit packets */
	uint64_t	interrupts; /* number of interrupts received */
};

struct netif_ops {
	void *(*setup)(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf);
	int   (*enable)(void *device, struct bfr_drv *bthis);
	void  (*cleanup)(void *device);
	void  (*enable_receive)(void *device);
	void  (*interrupt)(void *device);
	int   (*feedpacket)(void *device, dma_addr_t buf_phys, 
			    size_t buf_length, int more, 
			    enum bfr_packet_type ptype, uintptr_t buf_key);
	int   (*freepacket)(void *device, dma_addr_t buf_phys, size_t buf_size, uintptr_t key);
	int   (*macaddr)(void *device, unsigned char buf[6]);
	int   (*setstatistics)(void *device, struct netstats *ns);
	int   (*setholdoff)(void *device, uint32_t nsec);
};

static inline int netif_enable(void *device, void *bthis)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->enable(device, bthis);
}

static inline void netif_cleanup(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	dp->classp->n_ops->cleanup(device);
}

static inline void netif_interrupt(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	dp->classp->n_ops->interrupt(device);
}

static inline void netif_enable_receive(void *device)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	dp->classp->n_ops->enable_receive(device);
}


static inline int netif_feedpacket(void *device, dma_addr_t buf_phys, 
			    size_t buf_length, int more, 
			    enum bfr_packet_type ptype, uintptr_t buf_key)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->feedpacket(device, buf_phys, buf_length, more,
				      ptype, buf_key);
}

static inline int netif_freepacket(void *device, dma_addr_t buf_phys, size_t buf_size, uintptr_t key)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->freepacket(device, buf_phys, buf_size, key);

}

static inline int netif_macaddr(void *device, unsigned char buf[6])
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->macaddr(device, buf);
}

static inline int netif_setstatistics(void *device, struct netstats *ns)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->setstatistics(device, ns);
}

static inline int netif_setholdoff(void *device, uint32_t nsec)
{
	struct driver_instance *dp = (struct driver_instance *)device;
	return dp->classp->n_ops->setholdoff(device, nsec);
}

#endif /* NETDRIVER_H_ */
