#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "msa.h"
#include <pthread.h>

int has_msa = 1;

static void
__msa_refresh_interrupts(clk_t **irqs, int *nirq);


static void
__msa_refresh_interrupts(clk_t **irqs, int *nirq)
{
	static FILE *irqf;
	char *p;
	int ncpu;
	unsigned irq;
	int i;
	char buf[128];

	if (has_msa == 0)
		return;
	if (irqf == 0) {
		irqf = fopen("/proc/interrupts", "r");
		if (irqf == NULL)
			return;
	}

	rewind(irqf);
	/*
	 * Get the header line; how many processors?
	 */
	for (;;) {
		fgets(buf, sizeof buf, irqf);

		for (p = buf, ncpu = 0; *p && *p != '\n'; p++)
			if (*p == 'C')
				ncpu++;
		if (*p == '\n')
			break;
	} 

	memset(*irqs, 0, *nirq * sizeof(*irqs)[0]);
	/*
	 * And scan the irq times
	 */
	while (!feof(irqf)) {
		unsigned long long count;
		unsigned long long itime;

		if (fscanf(irqf, "%u:", &irq) != 1) {
			fgets(buf, sizeof buf, irqf);
			continue;
		}
			
		if (irq >= *nirq) {
			*nirq = irq+1;
			*irqs = realloc(*irqs, 2*(irq+1)*sizeof(clk_t));
			(*irqs)[2*irq] = 0;
			(*irqs)[2*irq + 1] = 0;
		}
		for (i = 0; i < ncpu; i++) {
			fscanf(irqf, "%llu%llu", &count, &itime);
                        (*irqs)[2*irq] = count;
			(*irqs)[2*irq + 1] = itime;
		}
		fgets(buf, sizeof buf, irqf);
	}
}

static clk_t *irqs;
static int nirqs;
pthread_mutex_t irqmutex = PTHREAD_MUTEX_INITIALIZER;

void msa_refresh_irqs(void)
{
	if (has_msa == 0)
		return;

	pthread_mutex_lock(&irqmutex);
	
	__msa_refresh_interrupts(&irqs, &nirqs);

	pthread_mutex_unlock(&irqmutex);
}

clk_t irqtime(int irq)
{
	if (has_msa == 0)
		return 0;
	if (irq >= nirqs)
		return (clk_t)0;
	return irqs[2*irq + 1];
}

unsigned long irqcount(int irq)
{
	if (has_msa == 0)
		return 0;
	if (irq >= nirqs)
		return (clk_t)0;
	return (unsigned long)irqs[2*irq];
}

unsigned long max_irq(void)
{
	return nirqs;
}

#ifdef HAS_MSA
struct msa_name msa_names[] = {
#define X(a) {a, #a}
	X(ONCPU_USER),
	X(ONCPU_SYS),
	X(INTERRUPTIBLE_SLEEP),
	X(UNINTERRUPTIBLE_SLEEP),
	X(ONACTIVEQUEUE),
	X(ONEXPIREDQUEUE),
	X(ZOMBIE),
	X(STOPPED),
	X(INTERRUPTED),
	X(PAGING_SLEEP),
	X(FUTEX_SLEEP),
	X(POLL_SLEEP),
	{-1, 0},
};
#endif
