#include <stddef.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>
#include "driver_env_glue.h"
#include "usr_pci_map.h"
#include "xmalloc.h"
#include "asm/unistd.h"

#include "linux_usr_pci.h"
#include "error.h"
#include "pci_dma.h"




int
__setup_dma(dma_handle_t dma, 
	  mem_handle_t handle, 
	  size_t sz, 
	  dmabuf_handle_t *dmabuf,
	  direction_t dir)
{
	struct mapping_info m;
	dmabuf_handle_t y;
	int nents = sz / pagesize + 2;

	y = xmalloc(sizeof *y +sizeof(*(y->sglist)) * nents);
	m.virtaddr = get_addr(handle);
	m.size = sz;
	m.direction = dir;
	m.nents = nents;
	m.sglist = y->sglist;

	if (usr_pci_map(dma, USR_MAP, &m) != 0) {
		syserror("usr_pci_map");
		return -1;
	}

	y->nentries = m.nents;
	*dmabuf = y;

	return 0;
}

int setup_dma(dma_handle_t dma, 
	  mem_handle_t handle, 
	  size_t sz, 
	  dmabuf_handle_t *dmabuf,
	      direction_t dir) __attribute__ ((weak, alias ("__setup_dma")));



int 
__teardown_dma(dma_handle_t dma, dmabuf_handle_t *x)
{
	struct mapping_info m;
	int error_code;
	dmabuf_handle_t y = *x;

	m.dmaaddr = y->sglist[0].dmaaddr;
	m.sglist = y->sglist;

	if ((error_code = usr_pci_map(dma, USR_UNMAP, &m)) != 0) {
		syserror("USR_UNMAP");
		return error_code;
	}

	free(y);
	*x = 0;
	return 0;
}

int teardown_dma(dma_handle_t dma, dmabuf_handle_t *x) __attribute__ ((weak, alias ("__teardown_dma")));

