#include <driver_interface.h>
#include <pci_dma.h>
#include <assert.h>
#include <paxlib.h>
#include <stdlib.h>
#include <string.h>


#include <si.h>

#include "io/io.h"
#include <mungi/clock.h>
#include <heap_alloc.h>

#define MAX_PINNED 100

#define PROFILE
//#define PROFILE_PAGEPIN
//#define PROFILE_PAGEUNPIN
//#define PROFILE_INOVERHEAD
#define PROFILE_EXTERN
#ifdef PROFILE
#include "cycles.h"
#define TIMES 100

PROFILE_EXTERN int profile_count = 0, profile_i;
PROFILE_EXTERN cycle_t profile_total[TIMES];
PROFILE_EXTERN cycle_t profile_x1, profile_x2;
#endif

volatile uint64_t force_write = 0 ;


callback_t null_callback = NULL;
callback_data_t null_callback_data = {{NULL, 0}, NULL};
dmabuf_handle_t null_dmabuf = NULL;
//int pagesize = BASE_PAGESIZE;

//#define NO_DMA_MALLOC
//#define NO_SETUP_DMA
//#define NO_RELEASE

#define DMA_HANDLE_PB_SIZE (8*PAGESIZE)

dma_handle_t
get_dma_handle(cicap_t dma_cicap)
{
	environment_t env;
	dma_handle_t dma_handle;
	IDMA_t *dma;
	void **_pb;
	void *rp;
	cap_t pb;
	int r;

	dma_handle = malloc(sizeof(struct dma_handle));
	if (dma_handle == NULL) 
		return NULL;

	dma_handle->numpinned = 0;
	rp = object_create(DMA_HANDLE_PB_SIZE, &pb);
	assert(rp != NULL);
	dma = idma_cicap_constructor(dma_cicap, pb, DMA_HANDLE_PB_SIZE, &env );

	_pb = (void**)dma->pb.address;
	*((cicap_t*)_pb) = dma->ecicap;



	r = pax_cl_get_entry_pt(dma->cid, IID_IDMA, MID_IDMA_MODIFY,
				&dma->epts[MID_IDMA_MODIFY - MID_IDMA_START]
				);
	if( r != 0 ) {
		return NULL;
	}

	dma_handle->dma = dma;

	_pb += 2;
	dma_handle->pin = ((pin_seq *)_pb);
	_pb += ((sizeof(pin_seq)/sizeof(void*)) + 1);
	dma_handle->unpin = ((cap_seq *)_pb);
	_pb += ((sizeof(cap_seq)/sizeof(void*)) + 1);

	dma_handle->pin->c = 1;
	dma_handle->pin->v = (void*) _pb;
	
	dma_handle->unpin->c = 0;
	dma_handle->unpin->v = (void*) ((uintptr_t) pb.address) + 
		DMA_HANDLE_PB_SIZE - sizeof(void*);

	_pb += ((MAX_PINNED * sizeof(pin_t))/sizeof(void*)) + 1;
	//dma_handle->sg_area = _pb;
	
	dma_handle->max_sg = (void*)dma->pb.address + DMA_HANDLE_PB_SIZE;

	/* sg_area is our working heap.*/ 
	dma_handle->sg_area = heap_alloc_init(_pb, dma_handle->max_sg);
	assert(dma_handle->sg_area);
	return dma_handle ;
}

bus_space_t
get_pciconf_bus(cicap_t pci_conf, cap_t pb, int size)
{
	environment_t env;
	bus_space_t bus;
	bus.address.address = ipci_conf_cicap_constructor(pci_conf, 
							  pb, PAGESIZE, &env );
	return bus;
}

int
get_pci_consistent_memory(dma_handle_t dma, size_t size, mem_handle_t * handle,
			  dmabuf_handle_t *busaddr)
{
	int r, i;
	char *obj;
	struct objinfo info;

	/* setup the objinfo */
	bzero( &info, sizeof( info ) );
	info.special |= O_DMACONST;

	/* create the object */
	obj = object_create_info(size, handle, &info);

	if (obj == NULL) {
		return 1;
	} 

	/* Hack for now.. lets touch it so its in the page table later */
	for (i = 0; i < size; i++) {
		obj[i] = 0;
	}

	/* telk to the chipset */
	r = setup_dma(dma, *handle, size, busaddr, DMA_BIDIRECTIONAL);
	assert( r == 0 );

	return r;
}

int
setup_dma(dma_handle_t dma, mem_handle_t handle, size_t size, 
	  dmabuf_handle_t *dmabuf, direction_t dir)
{
	cap_t ret;
	cap_t param = dma->dma->pb;

	dma->pin->c = 1;
	dma->unpin->c = 0;

	/* Allocate space for the sglist */
	dma->pin->v[0].handle = handle;
	dma->pin->v[0].length = size;
	/* +2: One for round-up, one because we may cross a pg boundary 
	 * (but can't tell)
	 * FIXME: why can't we tell?!
	 */
	dma->pin->v[0].sg.c = (size / PAGESIZE) + 2;
	*dmabuf = heap_alloc(dma->sg_area, sizeof(struct dmabuf_handle) + 
		(dma->pin->v[0].sg.c * sizeof(sg_t)) );
	assert(*dmabuf!=NULL);
	if ((void *)*dmabuf >= dma->max_sg) {
		UserPrint("pinned: %d\n", dma->numpinned);
	}
	assert ((void *)*dmabuf < dma->max_sg);
	//(*dmabuf)->prev = NULL; 
	dma->pin->v[0].sg.v = (*dmabuf)->v;

	PdxCall(dma->dma->epts[MID_IDMA_MODIFY - MID_IDMA_START], 
		param, &ret, PD_EMPTY);
	
	(*dmabuf)->c = dma->pin->v[0].sg.c;
	(*dmabuf)->handle = handle;
	dma->numpinned += 1;


	return 0;
}

/* FIXME: move. -nfd */
int setup_dma_pktseq(dma_handle_t dma, pkt_seq *pktseq, dmabuf_handle_t *dmabufs,
	direction_t dir);

/* Same thing but for a packet sequence. */
/* Pin each fragment in each packet separately */
int
setup_dma_pktseq(dma_handle_t dma, pkt_seq *pktseq, dmabuf_handle_t *dmabuffers,
	direction_t dir)
{
	cap_t ret;
	packet_t *currentpacket;
	buf_t *currentfragment;
	cap_t param = dma->dma->pb;
	int frames = 0;
	int currentpin = 0, i, j;
	dmabuf_handle_t *currentbuf = dmabuffers;
	
	dma->pin->c = 0;
	dma->unpin->c = 0;
	for (i = 0; i < pktseq->c; i++) {
		currentpacket = &pktseq->v[i];
		for (j = 0; j < currentpacket->c; j++) {
			currentfragment = &currentpacket->v[j];
			frames = (currentfragment->length / PAGESIZE) + 2;
			(*currentbuf) = heap_alloc(dma->sg_area,
				sizeof(struct dmabuf_handle) + (frames * sizeof(sg_t)));
			assert(*currentbuf != NULL);
			(*currentbuf)->handle = currentfragment->handle;
			(*currentbuf)->client_ref = currentfragment;
			assert((void *)&dma->pin->v[currentpin] < (void *)dma->unpin->v);
			dma->pin->v[currentpin].handle = currentfragment->handle;
			assert(currentpin < MAX_PINNED);
			dma->pin->v[currentpin].length = currentfragment->length;
			dma->pin->v[currentpin].sg.c = frames;
			dma->pin->v[currentpin].sg.v = (*currentbuf)->v;
			dma->pin->c += 1;
			currentpin ++;
			currentbuf ++;
		}
	}
	PdxCall(dma->dma->epts[MID_IDMA_MODIFY - MID_IDMA_START], 
		param, &ret, PD_EMPTY);

	currentbuf = dmabuffers;
	for (i=0;i<currentpin;i++) {
		(*currentbuf)->c = dma->pin->v[i].sg.c;
		currentbuf ++;
	}
	return currentpin;
}

int
teardown_dma(dma_handle_t dma, dmabuf_handle_t *dmabuf)
{
	//dma->sg_area = (*dmabuf)->prev;
	heap_free(dma->sg_area, *dmabuf);
	//UserPrint("Teardown to %p (dma is %p, dmabuf is %p)\n", dma->sg_area, dma, *dmabuf);
	dma->numpinned -= 1;
	return 0;
}

