/*
 @LICENSE@
*/

/*
 * K&R Malloc 
 */

#include <heap_alloc.h>

#define NALLOC  1024		/* minimum #units to request */

#define NULL ((void*)0)

typedef long long Align;	/* for alignment to long long boundary */

union header {			/* block header */
	struct {
		union header   *ptr;	/* next block if on free list */
		unsigned        size;	/* size of this block */
	} s;
	Align           x;	/* force alignment of blocks */
};

typedef union header Header;

struct mallocinfo {
	Header   base;		/* empty list to get started */
	void    *bss;	/* `sbrk` point */
	void    *top;	/* `sbrk` point */
	Header  *freep;	/* start of free list */
	void	*align;
};

static Header  *morecore(struct mallocinfo *info, unsigned nu);


void *
heap_alloc_init(void *bss_p, void *top_p)
{
	struct mallocinfo *info = bss_p;
	info->base.s.ptr = NULL;
	info->base.s.size = 0;
	info->bss = bss_p + sizeof (struct mallocinfo);
	info->top = top_p;
	info->freep = NULL;
	
	return info;
}

/*
 * malloc: general-purpose storage allocator 
 */
void *
heap_alloc(void *heap, unsigned nbytes)
{
	Header         *p,
	               *prevp;
	unsigned        nunits;
	struct mallocinfo *info = heap;
	nunits = (nbytes + sizeof(Header) - 1) / sizeof(Header) + 1;
	if ((prevp = info->freep) == NULL) {	/* no free list yet */
		info->base.s.ptr = info->freep = prevp = &(info->base);
		info->base.s.size = 0;
	}
	for (p = prevp->s.ptr;; prevp = p, p = p->s.ptr) {
		if (p->s.size >= nunits) {	/* big enough */
			if (p->s.size == nunits)	/* exactly */
				prevp->s.ptr = p->s.ptr;
			else {	/* allocate tail end */
				p->s.size -= nunits;
				p += p->s.size;
				p->s.size = nunits;
			}
			info->freep = prevp;
			return (void *) (p + 1);
		}
		if (p == info->freep) {	/* wrapped around free list */
			if ((p = (Header *) morecore(info, nunits)) == NULL) {
				out_of_memory();
				return NULL;	/* none left */
			} else {

			}
		}
	}
}

/*
 * free: put block ap in free list 
 */
void
heap_free(void *heap, void *ap)
{
	Header         *bp,
	               *p;
	struct mallocinfo *info = heap;

	bp = (Header *) ap - 1;	/* point to block header */
	for (p = info->freep; !(bp > p && bp < p->s.ptr); p = p->s.ptr)
		if (p >= p->s.ptr && (bp > p || bp < p->s.ptr))
			break;	/* freed block at start or end of arena */

	if (bp + bp->s.size == p->s.ptr) {	/* join to upper nbr */
		bp->s.size += p->s.ptr->s.size;
		bp->s.ptr = p->s.ptr->s.ptr;
	} else
		bp->s.ptr = p->s.ptr;
	if (p + p->s.size == bp) {	/* join to lower nbr */
		p->s.size += bp->s.size;
		p->s.ptr = bp->s.ptr;
	} else
		p->s.ptr = bp;
	info->freep = p;
}


/*
 * morecore: this is sbrk on unix systems 
 */
static Header  *
morecore(struct mallocinfo *info, unsigned nu)
{
	void           *cp;
	Header         *up;
	cp = info->bss;
	if (nu < NALLOC)
		nu = NALLOC;
	if (info->bss + (nu * sizeof(Header)) > info->top)
		return NULL;
	info->bss += (nu * sizeof(Header));

	up = (Header *) cp;
	up->s.size = nu;
	heap_free(info, (void *) (up + 1));
	return info->freep;
}
