/* Check that heap_alloc isn't crazy. nfd */

#include <heap_alloc.h>

#include <assert.h>

#define TEST_SIZE 16384
#define CANARY1 0
#define CANARY2 (TEST_SIZE+2)
#define HEAP_START 2

typedef unsigned char byte;

static void setcanary(byte *val)
{
	val[0]=0xbe; val[1]=0xef; /* In the outlet by the light switch */
}

static void checkcanaries(byte *val)
{
	assert (val[CANARY1]==0xbe && val[CANARY1+1]==0xef); 
	assert (val[CANARY2]==0xbe && val[CANARY2+1]==0xef); 
}

int main(int argc, char **argv)
{
	byte scratch[TEST_SIZE+4];
	void *heap;
	byte *smallobj, *largeobj, *objects[TEST_SIZE/10];
	int maxsize, i;
	
	setcanary(scratch); setcanary(&scratch[2+TEST_SIZE]);
	heap_alloc_init(&scratch[HEAP_START], &scratch[CANARY2]);
	//assert (heap == &scratch[HEAP_START]); /* good so far... */
	checkcanaries(scratch);
	
	smallobj = heap_alloc(37);
	checkcanaries(scratch);
	assert(smallobj);
	assert(!((unsigned long)smallobj & 3));
	heap_free (smallobj);
	checkcanaries(scratch);
	
	for (maxsize=TEST_SIZE; maxsize; maxsize--) {
		largeobj = heap_alloc(maxsize);
		checkcanaries(scratch);
		if (largeobj) break;
	}
	assert(largeobj);
	assert(!((unsigned long)largeobj & 7));
	printf("Heap of %d results in largest alloc of %d\n", TEST_SIZE, maxsize);
	heap_free(largeobj);
	checkcanaries(scratch);
	
	for (i=0; i<9; i++) {
		objects[i] = (byte *)heap_alloc(maxsize/10);
		assert(objects[i]);
		assert(!((unsigned long)objects[i] & 7));
		checkcanaries(scratch);
	}
	
	for (i=0; i<9; i++) {
		heap_free(objects[i]);
		checkcanaries(scratch);
	}
	printf("All tests passed.\n");
}


