/* Check that heap_alloc isn't crazy. nfd */

#include <heap_alloc.h>

#include <assert.h>

#define TEST_SIZE 32768
#define CANARY1 0
#define CANARY2 (TEST_SIZE)
#define HEAP_START 8

typedef unsigned char byte;

static void setcanary(byte *val)
{
	val[0]=0xbe; val[1]=0xef; /* In the outlet by the light switch */
}

static void checkcanaries(byte *val)
{
	assert (val[CANARY1]==0xbe && val[CANARY1+1]==0xef); 
	assert (val[CANARY2]==0xbe && val[CANARY2+1]==0xef); 
}

int main(int argc, char **argv)
{
	byte scratch[TEST_SIZE+2];
	void *heap;
	byte *smallobj, *largeobj, *objects[TEST_SIZE/10];
	int maxsize, i;
	
	printf("scratch = %p - %p\n", scratch, &scratch[CANARY2]);
	assert(!((unsigned long)scratch & 7));
	setcanary(scratch); setcanary(&scratch[TEST_SIZE]);
	heap = heap_alloc_init(&scratch[HEAP_START], &scratch[CANARY2]);
	assert (heap == &scratch[HEAP_START]); /* good so far... */
	checkcanaries(scratch);
	
	smallobj = heap_alloc(heap, 37);
	printf("smallobj = %p\n", smallobj);
	checkcanaries(scratch);
	assert(smallobj);
	assert(!((unsigned long)smallobj & 7));
	heap_free (heap, smallobj);
	checkcanaries(scratch);
	
	for (maxsize=TEST_SIZE; maxsize; maxsize--) {
		largeobj = heap_alloc(heap, maxsize);
		checkcanaries(scratch);
		if (largeobj) break;
	}
	printf("largeobj = %p\n", smallobj);
	assert(largeobj);
	printf("Heap of %d results in largest alloc of %d\n", TEST_SIZE, maxsize);
	heap_free(heap, largeobj);
	assert(!((unsigned long)largeobj & 7));
	checkcanaries(scratch);
	
	for (i=0; i<9; i++) {
		objects[i] = (byte *)heap_alloc(heap, maxsize/10);
		printf("objects[%d] = %p\n", i, objects[i]);
		assert(objects[i]);
		assert(!((unsigned long)objects[i] & 7));
		checkcanaries(scratch);
	}
	
	for (i=0; i<9; i++) {
		heap_free(heap, objects[i]);
		checkcanaries(scratch);
	}
	printf("All tests passed.\n");
}


