#include "vm.h"
#include "config.h"
#include <asm/rse.h>
#include "pal.h"
#include "sal.h"
#include "fpswa.h"

#define M (0 << 4)
#define I (1 << 4)
#define F (2 << 4)
#define L (3 << 4)
#define B (4 << 4)
#define R (5 << 4)

static const char templates[16][4] = {
	{ M, I, I }, { M, I, I }, { M, L, R }, { R, R, R },
	{ M, M, I }, { M, M, I }, { M, F, I }, { M, M, F },
	{ M, I, B }, { M, B, B }, { R, R, R }, { B, B, B },
	{ M, M, B }, { R, R, R }, { M, F, B }, { R, R, R }
};

extern unsigned int node_id;
extern unsigned int write_update_threshold;
extern unsigned long extra_latency; /* for latency sensitivity analysis */

unsigned int debug_flags;
char *last_interruption;

unsigned int dtr_index = 1, itr_index = 1;

void
dtr_insert(unsigned long va, unsigned long pa, unsigned long sizebits, unsigned long attr)
{
	unsigned long pte = (pa & 0x3fffffffff000UL) | PTE_D | PTE_A | PTE_P | attr;
	asm volatile("mov cr.ifa = %0; mov cr.itir = %1 ;; itr.d dtr[%2] = %3" : : "r"(va), "r"(sizebits << 2), "r"(dtr_index), "r"(pte));
	dtr_index++;
}

void
itr_insert(unsigned long va, unsigned long pa, unsigned long sizebits, unsigned long attr)
{
	unsigned long pte = (pa & 0x3fffffffff000UL) | PTE_A | PTE_P | attr;
	asm volatile("mov cr.ifa = %0; mov cr.itir = %1 ;; itr.i itr[%2] = %3" : : "r"(va), "r"(sizebits << 2), "r"(itr_index), "r"(pte));
	itr_index++;
}

#define read_fr_const(regno, type) ({ unsigned long value; asm("getf."#type " %0 = f"#regno : "=r"(value)); value; })

#define read_fr(regno, type) \
	switch (regno) { \
		case 0: return 0; \
		case 1: return read_fr_const(1, type); /* s:0x3f800000, d:0x3ff0000000000000, sig:0x8000000000000000 */ \
		case 2: return read_fr_const(2, type); \
		case 3: return read_fr_const(3, type); \
		case 4: return read_fr_const(4, type); \
		case 5: return read_fr_const(5, type); \
		case 6: return read_fr_const(6, type); \
		case 7: return read_fr_const(7, type); \
		case 8: return read_fr_const(8, type); \
		case 9: return read_fr_const(9, type); \
		case 10: return read_fr_const(10, type); \
		case 11: return read_fr_const(11, type); \
		case 12: return read_fr_const(12, type); \
		case 13: return read_fr_const(13, type); \
		case 14: return read_fr_const(14, type); \
		case 15: return read_fr_const(15, type); \
		case 16: return read_fr_const(16, type); \
		case 17: return read_fr_const(17, type); \
		case 18: return read_fr_const(18, type); \
		case 19: return read_fr_const(19, type); \
		case 20: return read_fr_const(20, type); \
		case 21: return read_fr_const(21, type); \
		case 22: return read_fr_const(22, type); \
		case 23: return read_fr_const(23, type); \
		case 24: return read_fr_const(24, type); \
		case 25: return read_fr_const(25, type); \
		case 26: return read_fr_const(26, type); \
		case 27: return read_fr_const(27, type); \
		case 28: return read_fr_const(28, type); \
		case 29: return read_fr_const(29, type); \
		case 30: return read_fr_const(30, type); \
		case 31: return read_fr_const(31, type); \
		case 32: return read_fr_const(32, type); \
		case 33: return read_fr_const(33, type); \
		case 34: return read_fr_const(34, type); \
		case 35: return read_fr_const(35, type); \
		case 36: return read_fr_const(36, type); \
		case 37: return read_fr_const(37, type); \
		case 38: return read_fr_const(38, type); \
		case 39: return read_fr_const(39, type); \
		case 40: return read_fr_const(40, type); \
		case 41: return read_fr_const(41, type); \
		case 42: return read_fr_const(42, type); \
		case 43: return read_fr_const(43, type); \
		case 44: return read_fr_const(44, type); \
		case 45: return read_fr_const(45, type); \
		case 46: return read_fr_const(46, type); \
		case 47: return read_fr_const(47, type); \
		case 48: return read_fr_const(48, type); \
		case 49: return read_fr_const(49, type); \
		case 50: return read_fr_const(50, type); \
		case 51: return read_fr_const(51, type); \
		case 52: return read_fr_const(52, type); \
		case 53: return read_fr_const(53, type); \
		case 54: return read_fr_const(54, type); \
		case 55: return read_fr_const(55, type); \
		case 56: return read_fr_const(56, type); \
		case 57: return read_fr_const(57, type); \
		case 58: return read_fr_const(58, type); \
		case 59: return read_fr_const(59, type); \
		case 60: return read_fr_const(60, type); \
		case 61: return read_fr_const(61, type); \
		case 62: return read_fr_const(62, type); \
		case 63: return read_fr_const(63, type); \
		case 64: return read_fr_const(64, type); \
		case 65: return read_fr_const(65, type); \
		case 66: return read_fr_const(66, type); \
		case 67: return read_fr_const(67, type); \
		case 68: return read_fr_const(68, type); \
		case 69: return read_fr_const(69, type); \
		case 70: return read_fr_const(70, type); \
		case 71: return read_fr_const(71, type); \
		case 72: return read_fr_const(72, type); \
		case 73: return read_fr_const(73, type); \
		case 74: return read_fr_const(74, type); \
		case 75: return read_fr_const(75, type); \
		case 76: return read_fr_const(76, type); \
		case 77: return read_fr_const(77, type); \
		case 78: return read_fr_const(78, type); \
		case 79: return read_fr_const(79, type); \
		case 80: return read_fr_const(80, type); \
		case 81: return read_fr_const(81, type); \
		case 82: return read_fr_const(82, type); \
		case 83: return read_fr_const(83, type); \
		case 84: return read_fr_const(84, type); \
		case 85: return read_fr_const(85, type); \
		case 86: return read_fr_const(86, type); \
		case 87: return read_fr_const(87, type); \
		case 88: return read_fr_const(88, type); \
		case 89: return read_fr_const(89, type); \
		case 90: return read_fr_const(90, type); \
		case 91: return read_fr_const(91, type); \
		case 92: return read_fr_const(92, type); \
		case 93: return read_fr_const(93, type); \
		case 94: return read_fr_const(94, type); \
		case 95: return read_fr_const(95, type); \
		case 96: return read_fr_const(96, type); \
		case 97: return read_fr_const(97, type); \
		case 98: return read_fr_const(98, type); \
		case 99: return read_fr_const(99, type); \
		case 100: return read_fr_const(100, type); \
		case 101: return read_fr_const(101, type); \
		case 102: return read_fr_const(102, type); \
		case 103: return read_fr_const(103, type); \
		case 104: return read_fr_const(104, type); \
		case 105: return read_fr_const(105, type); \
		case 106: return read_fr_const(106, type); \
		case 107: return read_fr_const(107, type); \
		case 108: return read_fr_const(108, type); \
		case 109: return read_fr_const(109, type); \
		case 110: return read_fr_const(110, type); \
		case 111: return read_fr_const(111, type); \
		case 112: return read_fr_const(112, type); \
		case 113: return read_fr_const(113, type); \
		case 114: return read_fr_const(114, type); \
		case 115: return read_fr_const(115, type); \
		case 116: return read_fr_const(116, type); \
		case 117: return read_fr_const(117, type); \
		case 118: return read_fr_const(118, type); \
		case 119: return read_fr_const(119, type); \
		case 120: return read_fr_const(120, type); \
		case 121: return read_fr_const(121, type); \
		case 122: return read_fr_const(122, type); \
		case 123: return read_fr_const(123, type); \
		case 124: return read_fr_const(124, type); \
		case 125: return read_fr_const(125, type); \
		case 126: return read_fr_const(126, type); \
		case 127: return read_fr_const(127, type); \
	} \
	return 0;

#define write_fr_const(regno, value, type) asm("setf."#type " f"#regno " = %0" :: "r"(value));

#define write_fr(regno, value, type) \
	switch (regno) { \
		case 0: case 1: printf("attempt to write f0/f1\n"); return; \
		case 2: write_fr_const(2, value, type); return; \
		case 3: write_fr_const(3, value, type); return; \
		case 4: write_fr_const(4, value, type); return; \
		case 5: write_fr_const(5, value, type); return; \
		case 6: write_fr_const(6, value, type); return; \
		case 7: write_fr_const(7, value, type); return; \
		case 8: write_fr_const(8, value, type); return; \
		case 9: write_fr_const(9, value, type); return; \
		case 10: write_fr_const(10, value, type); return; \
		case 11: write_fr_const(11, value, type); return; \
		case 12: write_fr_const(12, value, type); return; \
		case 13: write_fr_const(13, value, type); return; \
		case 14: write_fr_const(14, value, type); return; \
		case 15: write_fr_const(15, value, type); return; \
		case 16: write_fr_const(16, value, type); return; \
		case 17: write_fr_const(17, value, type); return; \
		case 18: write_fr_const(18, value, type); return; \
		case 19: write_fr_const(19, value, type); return; \
		case 20: write_fr_const(20, value, type); return; \
		case 21: write_fr_const(21, value, type); return; \
		case 22: write_fr_const(22, value, type); return; \
		case 23: write_fr_const(23, value, type); return; \
		case 24: write_fr_const(24, value, type); return; \
		case 25: write_fr_const(25, value, type); return; \
		case 26: write_fr_const(26, value, type); return; \
		case 27: write_fr_const(27, value, type); return; \
		case 28: write_fr_const(28, value, type); return; \
		case 29: write_fr_const(29, value, type); return; \
		case 30: write_fr_const(30, value, type); return; \
		case 31: write_fr_const(31, value, type); return; \
		case 32: write_fr_const(32, value, type); return; \
		case 33: write_fr_const(33, value, type); return; \
		case 34: write_fr_const(34, value, type); return; \
		case 35: write_fr_const(35, value, type); return; \
		case 36: write_fr_const(36, value, type); return; \
		case 37: write_fr_const(37, value, type); return; \
		case 38: write_fr_const(38, value, type); return; \
		case 39: write_fr_const(39, value, type); return; \
		case 40: write_fr_const(40, value, type); return; \
		case 41: write_fr_const(41, value, type); return; \
		case 42: write_fr_const(42, value, type); return; \
		case 43: write_fr_const(43, value, type); return; \
		case 44: write_fr_const(44, value, type); return; \
		case 45: write_fr_const(45, value, type); return; \
		case 46: write_fr_const(46, value, type); return; \
		case 47: write_fr_const(47, value, type); return; \
		case 48: write_fr_const(48, value, type); return; \
		case 49: write_fr_const(49, value, type); return; \
		case 50: write_fr_const(50, value, type); return; \
		case 51: write_fr_const(51, value, type); return; \
		case 52: write_fr_const(52, value, type); return; \
		case 53: write_fr_const(53, value, type); return; \
		case 54: write_fr_const(54, value, type); return; \
		case 55: write_fr_const(55, value, type); return; \
		case 56: write_fr_const(56, value, type); return; \
		case 57: write_fr_const(57, value, type); return; \
		case 58: write_fr_const(58, value, type); return; \
		case 59: write_fr_const(59, value, type); return; \
		case 60: write_fr_const(60, value, type); return; \
		case 61: write_fr_const(61, value, type); return; \
		case 62: write_fr_const(62, value, type); return; \
		case 63: write_fr_const(63, value, type); return; \
		case 64: write_fr_const(64, value, type); return; \
		case 65: write_fr_const(65, value, type); return; \
		case 66: write_fr_const(66, value, type); return; \
		case 67: write_fr_const(67, value, type); return; \
		case 68: write_fr_const(68, value, type); return; \
		case 69: write_fr_const(69, value, type); return; \
		case 70: write_fr_const(70, value, type); return; \
		case 71: write_fr_const(71, value, type); return; \
		case 72: write_fr_const(72, value, type); return; \
		case 73: write_fr_const(73, value, type); return; \
		case 74: write_fr_const(74, value, type); return; \
		case 75: write_fr_const(75, value, type); return; \
		case 76: write_fr_const(76, value, type); return; \
		case 77: write_fr_const(77, value, type); return; \
		case 78: write_fr_const(78, value, type); return; \
		case 79: write_fr_const(79, value, type); return; \
		case 80: write_fr_const(80, value, type); return; \
		case 81: write_fr_const(81, value, type); return; \
		case 82: write_fr_const(82, value, type); return; \
		case 83: write_fr_const(83, value, type); return; \
		case 84: write_fr_const(84, value, type); return; \
		case 85: write_fr_const(85, value, type); return; \
		case 86: write_fr_const(86, value, type); return; \
		case 87: write_fr_const(87, value, type); return; \
		case 88: write_fr_const(88, value, type); return; \
		case 89: write_fr_const(89, value, type); return; \
		case 90: write_fr_const(90, value, type); return; \
		case 91: write_fr_const(91, value, type); return; \
		case 92: write_fr_const(92, value, type); return; \
		case 93: write_fr_const(93, value, type); return; \
		case 94: write_fr_const(94, value, type); return; \
		case 95: write_fr_const(95, value, type); return; \
		case 96: write_fr_const(96, value, type); return; \
		case 97: write_fr_const(97, value, type); return; \
		case 98: write_fr_const(98, value, type); return; \
		case 99: write_fr_const(99, value, type); return; \
		case 100: write_fr_const(100, value, type); return; \
		case 101: write_fr_const(101, value, type); return; \
		case 102: write_fr_const(102, value, type); return; \
		case 103: write_fr_const(103, value, type); return; \
		case 104: write_fr_const(104, value, type); return; \
		case 105: write_fr_const(105, value, type); return; \
		case 106: write_fr_const(106, value, type); return; \
		case 107: write_fr_const(107, value, type); return; \
		case 108: write_fr_const(108, value, type); return; \
		case 109: write_fr_const(109, value, type); return; \
		case 110: write_fr_const(110, value, type); return; \
		case 111: write_fr_const(111, value, type); return; \
		case 112: write_fr_const(112, value, type); return; \
		case 113: write_fr_const(113, value, type); return; \
		case 114: write_fr_const(114, value, type); return; \
		case 115: write_fr_const(115, value, type); return; \
		case 116: write_fr_const(116, value, type); return; \
		case 117: write_fr_const(117, value, type); return; \
		case 118: write_fr_const(118, value, type); return; \
		case 119: write_fr_const(119, value, type); return; \
		case 120: write_fr_const(120, value, type); return; \
		case 121: write_fr_const(121, value, type); return; \
		case 122: write_fr_const(122, value, type); return; \
		case 123: write_fr_const(123, value, type); return; \
		case 124: write_fr_const(124, value, type); return; \
		case 125: write_fr_const(125, value, type); return; \
		case 126: write_fr_const(126, value, type); return; \
		case 127: write_fr_const(127, value, type); return; \
	}

unsigned long
read_fr_sig(unsigned int regno)
{
	read_fr(regno, sig);
}

unsigned long
read_fr_s(unsigned int regno)
{
	read_fr(regno, s);
}

unsigned long
read_fr_d(unsigned int regno)
{
	read_fr(regno, d);
}

void
write_fr_sig(unsigned int regno, unsigned long value)
{
	write_fr(regno, value, sig);
}

void
write_fr_s(unsigned int regno, unsigned long value)
{
	write_fr(regno, value, s);
}

void
write_fr_d(unsigned int regno, unsigned long value)
{
	write_fr(regno, value, d);
}

unsigned long
read_gr(unsigned int regno)
{
	long double f6, f7;
	unsigned long value;
	unsigned long *bsp;

	switch (regno)
	{
	case 0:  return 0;
	case 1:  return read_register("r28");
	case 2:  return read_register("r29");
	case 3:  return read_register("r3");
	case 4:  return read_register("r4");
	case 5:  return read_register("r5");
	case 6:  return read_register("r6");
	case 7:  return read_register("r7");
	case 8:  return read_register("r30");
	case 9:  return read_register("r9");
	case 10: return read_register("r10");
	case 11: return read_register("r11");
	case 12: return read_register("r31");
	case 13: return read_register("r13");
	case 14: return read_register("r14");
	case 15: return read_register("r15");
	case 16: return read_bank1_register("r16");
	case 17: return read_bank1_register("r17");
	case 18: return read_bank1_register("r18");
	case 19: return read_bank1_register("r19");
	case 20: return read_bank1_register("r20");
	case 21: return read_bank1_register("r21");
	case 22: return read_bank1_register("r22");
	case 23: return read_bank1_register("r23");
	case 24: return read_bank1_register("r24");
	case 25: return read_bank1_register("r25");
	case 26: return read_bank1_register("r26");
	case 27: return read_bank1_register("r27");
	case 28: return read_bank1_register("r28");
	case 29: return read_bank1_register("r29");
	case 30: return read_bank1_register("r30");
	case 31: return read_bank1_register("r31");
	default:
		asm volatile("stf.spill [%0]=f6 ; stf.spill [%1]=f7 ;;" : : "r"(&f6), "r"(&f7));
		asm volatile("flushrs ;;");
		bsp = ia64_rse_skip_regs((unsigned long *)read_bsp(), -(read_ifs() & 0x7f) + (regno - 32));		
		value = *bsp;
		asm volatile("ldf.fill f6=[%0] ; ldf.fill f7=[%1] ;;" : : "r"(&f6), "r"(&f7));
		return value;
	}
}

static void
write_gr(unsigned int regno, unsigned long value)
{
	long double f6, f7;
	unsigned long *bsp;

	switch (regno)
	{
	case 0:  printf("attempt to write GR0\n"); return;
	case 1:  write_register("r28", value); return;
	case 2:  write_register("r29", value); return;
	case 3:  write_register("r3", value);  return;
	case 4:  write_register("r4", value);  return;
	case 5:  write_register("r5", value);  return;
	case 6:  write_register("r6", value);  return;
	case 7:  write_register("r7", value);  return;
	case 8:  write_register("r30", value); return;
	case 9:  write_register("r9", value);  return;
	case 10: write_register("r10", value); return;
	case 11: write_register("r11", value); return;
	case 12: write_register("r31", value); return;
	case 13: write_register("r13", value); return;
	case 14: write_register("r14", value); return;
	case 15: write_register("r15", value); return;
	case 16: write_bank1_register("r16", value); return;
	case 17: write_bank1_register("r17", value); return;
	case 18: write_bank1_register("r18", value); return;
	case 19: write_bank1_register("r19", value); return;
	case 20: write_bank1_register("r20", value); return;
	case 21: write_bank1_register("r21", value); return;
	case 22: write_bank1_register("r22", value); return;
	case 23: write_bank1_register("r23", value); return;
	case 24: write_bank1_register("r24", value); return;
	case 25: write_bank1_register("r25", value); return;
	case 26: write_bank1_register("r26", value); return;
	case 27: write_bank1_register("r27", value); return;
	case 28: write_bank1_register("r28", value); return;
	case 29: write_bank1_register("r29", value); return;
	case 30: write_bank1_register("r30", value); return;
	case 31: write_bank1_register("r31", value); return;
	default:
		asm volatile("stf.spill [%0]=f6 ; stf.spill [%1]=f7 ;;" : : "r"(&f6), "r"(&f7));
		asm volatile("mov ar.rsc = 0 ;; flushrs ;; loadrs ;;");
		bsp = ia64_rse_skip_regs((unsigned long *)read_bsp(), -(read_ifs() & 0x7f) + (regno - 32));
		*bsp = value;
		asm volatile(";; mov ar.rsc = 3");
		asm volatile("ldf.fill f6=[%0] ; ldf.fill f7=[%1] ;;" : : "r"(&f6), "r"(&f7));
	}
}

extern unsigned long save_bank1(unsigned long *bank);
extern void load_bank1(unsigned long *bank, unsigned long unat);

static void
write_ar(unsigned int regno, unsigned long value)
{
	switch (regno)
	{
	case 0:  write_register("ar.k0", value); return;
	case 1:  write_register("ar.k1", value); return;
	case 2:  write_register("ar.k2", value); return;
	case 3:  write_register("ar.k3", value); return;
	case 4:  write_register("ar.k4", value); return;
	case 5:  write_register("ar.k5", value); return;
	case 6:  write_register("ar.k6", value); return;
	case 7:  write_register("ar.k7", value); return;
	case 44: write_register("ar.itc", value); return;
	default: printf("attempt to write AR %d\n", regno);
	}
}

void
rewind_ip(void)
{
	unsigned long cr_iip;
	unsigned long cr_ipsr;
	unsigned int slot;

	cr_iip = read_iip();
	cr_ipsr = read_ipsr();
	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);
	switch (slot)
	{
		case 0:
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			write_iip(cr_iip - 0x10);
			break;
		case 1:
			write_ipsr(cr_ipsr &~ PSR_RI);
			break;
		case 2:
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		default:
			printf("bad slot %d\n", slot);
	}
}

static void
emu_bsw(unsigned int target_bank)
{
	unsigned int bank;
	bank = BITS(cpu.psr, PSR_BN_BIT, 1);
	if (target_bank != bank)
	{
		cpu.banked_unat[1-target_bank] = save_bank1(cpu.banked[1-target_bank]);
		load_bank1(cpu.banked[target_bank], cpu.banked_unat[target_bank]);
		cpu.psr ^= PSR_BN;
	}
}

void
emu_precover(void)
{
	DEBUG_EMU("[%lx] precovering register frame\n", read_iip());
	cpu.cr[CR_IFS] = read_ifs();
	write_ifs(1UL << 63);
	cpu.cfle = 0;
}

static void
emu_cover(void)
{
	if (cpu.cfle)
		cpu.cr[CR_IFS] = read_ifs();
	else
		cpu.cfle = 1;

	write_ifs(1UL << 63);
}

void check_interrupts(void);
static void
emu_rfi(void)
{
	if (cpu.psr & PSR_CPL_3)
		printf("woah! rfi from CPL3 (iip=%lx psr=%lx)\n", read_iip(), cpu.psr);

	if ((cpu.cr[CR_IFS] & (1UL << 63)) && ((read_ifs() & 0x7f) != 0))
	{
		extern void shrink_frame;
		/* printf("This should not happen anymore\n"); */
		write_iip((unsigned long)&shrink_frame);
		write_ipsr(read_ipsr() & ~(PSR_RI|PSR_I));
		return;
	}
	write_iip(cpu.cr[CR_IIP]);
	write_ipsr((read_ipsr() & ~PSR_RI) | (BITS(cpu.cr[CR_IPSR], PSR_RI_BITS, 2) << PSR_RI_BITS) | PSR_I);
	if (cpu.cr[CR_IFS] & (1UL << 63))
		write_ifs(cpu.cr[CR_IFS]);
	cpu.cfle = 1;
	emu_bsw(BITS(cpu.cr[CR_IPSR], PSR_BN_BIT, 1));

#ifdef AFTERBURN
	if (cpu.cr[CR_IPSR] & PSR_CPL_3)
	{
		write_register("r4", cpu.saved_r4);
		write_register("r5", cpu.saved_r5);
		write_register("r6", cpu.saved_r6);
		write_register("r7", cpu.saved_r7);
		write_register("b2", cpu.saved_b2);
	}
#endif

	/* this write_psr needs to be at the end because it can call deliver_interrupt */
	write_psr(cpu.cr[CR_IPSR] &~ PSR_RI);
}

int via_deliver_interrupt;

void
deliver_exception(unsigned long vector, unsigned long isr)
{
	unsigned long iip = read_iip();
	unsigned long ri;

	//printf("exception iip = %lx vector = %lx ifa = %lx\n", iip, vector, read_ifa());
#if 0
	extern void vmm_tpa_faulted;
	if ((iip >= 0x100000) && (iip < 0x101000))
	{
		printf("hmm, need to be very careful here!\n");
		write_register("r5", vector);
		write_iip(&vmm_tpa_faulted);
		write_ipsr(read_ipsr() &~ PSR_RI);
		return;
	}
#endif

	if (cpu.psr & PSR_IC)
	{
		ri = read_ipsr() & PSR_RI;
		cpu.cr[CR_IIP] = iip;
		cpu.cr[CR_IIPA] = cpu.cr[CR_IIP];
		cpu.cr[CR_IFS] = 0;
		cpu.cr[CR_ISR] = (isr & ~PSR_RI) | ri;
		cpu.cr[CR_IPSR] = (read_psr() & ~PSR_RI) | ri;

#ifdef AFTERBURN
		/* Afterburner hack */
		if (cpu.psr & PSR_CPL_3)
		{
			cpu.saved_r4 = read_register("r4");
			cpu.saved_r5 = read_register("r5");
			cpu.saved_r6 = read_register("r6");
			cpu.saved_r7 = read_register("r7");
			cpu.saved_b2 = read_register("b2");
			write_register("r4", 0xfff);
		}
#if AFTERBURN != 2
		else
		{
			unsigned long restart_ip;
			restart_ip = read_register("r4");
			if (restart_ip != 0)
			{
				cpu.cr[CR_IIP] = restart_ip;
				cpu.cr[CR_IPSR] &= ~PSR_RI;
			}
		}
#endif
#endif
	}
	else if (vector != 0x1400)
	{
		printf("[%lx] Exception with psr.ic off (%lx)\n", read_iip(), vector);
		debug();
		return;
	}

	DEBUG_EMU("exception %lx isr=%lx iip=%lx ipsr=%lx\n", vector, cpu.cr[CR_ISR], cpu.cr[CR_IIP], cpu.cr[CR_IPSR]);

	if (isr & (1UL << 38)) /* incomplete register frame */
		emu_precover();

	write_iip(cpu.cr[CR_IVA] + vector);
	write_ipsr(read_ipsr() &~ PSR_RI);
	emu_bsw(0);
	cpu.psr &= ~(PSR_BN | PSR_CPL_3 | PSR_I | PSR_IC);

#if AFTERBURN != 2
	write_register("r4", 0);
#endif
}

void
check_interrupts(void)
{
	if (cpu.irq_pending && (cpu.psr & PSR_I)
		&& !(cpu.cr[CR_TPR] & TPR_MMI)
		&& (cpu.cr[CR_IVR] > cpu.in_service_irq)
		&& ((cpu.cr[CR_IVR] >> 4) > (cpu.cr[CR_TPR] >> 4)))
	{
#if AFTERBURN == 2
		if (!(cpu.psr & PSR_CPL_3) && !(read_register("r4") & 1))
		{
			/* setup deferred interrupt */
			//printf("deferring irq @ %lx\n", read_iip());
			return;
		}

#endif
#if 0
		if (cpu.cr[CR_IVR] == 0x30)
			printf("%");
#endif
		deliver_exception(0x3000, 0);
	}
}

void
deliver_interrupt(unsigned int irq)
{
	cpu.cr[CR_IRR0+(irq/64)] |= (1UL << (irq % 64));
	if (irq > cpu.cr[CR_IVR])
		cpu.cr[CR_IVR] = irq;
	cpu.irq_pending = 1;
}

static unsigned long
ia64_fls (unsigned long x)
{
	unsigned long ret = 63;

	while (!(x & (1UL << 63)))
	{
		x <<= 1;
		ret--;
	}

	return ret;
}

static void
clear_interrupt(unsigned int irq)
{
	int i;

	cpu.cr[CR_IRR0+(irq/64)] &= ~(1UL << (irq % 64));
	for (i = 3; i >= 0; i--)
	{
		if (cpu.cr[CR_IRR0+i] != 0)
		{
			cpu.cr[CR_IVR] = ia64_fls(cpu.cr[CR_IRR0+i]) + (i*64);
			return;
		}
	}

	cpu.cr[CR_IVR] = 0xf;
	cpu.irq_pending = 0;
}

unsigned long
thash(unsigned long addr)
{
	unsigned long vhpt_mask, vhpt_base, vhpt_offset, region;

	region = addr >> 61;
	vhpt_mask = (1UL << BITS(cpu.cr[CR_PTA], 2, 6)) - 1;
	vhpt_base = BITS(cpu.cr[CR_PTA], 15, 49) << 15;
	vhpt_offset = ((addr << 3) >> 3) >> BITS(cpu.rr[region], 2, 6);
	if (BITS(cpu.cr[CR_PTA], 8, 1)) /* long format */
		return (vhpt_base | ((vhpt_offset << 5) & vhpt_mask));
	else
		return ((region << 61) | ((vhpt_base << 3) >> 3) | ((vhpt_offset << 3) & vhpt_mask));
}

void
handle_disabled_fp(void)
{
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("disabled_fp %lx\n", read_iip());
	DEBUG_SHORT('f');
	STATS_INC(disabled_fp);

	/* we handle rsm psr.df[hl] lazily, so here we sync the psr bits */
	write_ipsr((read_ipsr() & ~(PSR_DFL|PSR_DFH)) | (cpu.psr & (PSR_DFL|PSR_DFH)));

	if (((isr & 1) && (cpu.psr & PSR_DFL))
	    || ((isr & 2) && (cpu.psr & PSR_DFH)))
	{
		deliver_exception(0x5500, isr);
	}
}

#define PMC_PL0	(1UL << 0)
#define PMC_PL3	(1UL << 3)
#define PMC_PL_ANY (0xFUL)
#define PMC_OI	(1UL << 5)
#define PMC_PM	(1UL << 6)
#define PMC4_ENABLE (1UL << 23)
#define PMC_EVENT_IA64_TAGGED_INST_RETIRED	(0x08UL << 8)
#define PMC_EVENT_CPU_CYCLES			(0x12UL << 8)
#if 0 /* Itanium 1?? */
#define PMC_EVENT_ITLB_MISSES_FETCH		(0x27UL << 8)
#define PMC_EVENT_ITLB_INSERTS_HPW		(0x28UL << 8)
#define PMC_EVENT_DTLB_MISSES			(0x61UL << 8)
#define PMC_EVENT_DTLB_INSERTS_HPW		(0x62UL << 8)
#endif
#define PMC_EVENT_L2DTLB_MISSES			(0xc1UL << 8)
#define PMC_EVENT_DTLB_INSERTS_HPW		(0xc9UL << 8)
#define PMC_EVENT_DTLB_INSERTS_HPW_RETIRED	(0x2cUL << 8)
#define PMC_EVENT_L2_OZQ_RELEASE		(0xa3UL << 8)

#define CORE_STATS_DEF(var) #var,
#ifndef NSTATS
#define STATS_DEF(var) CORE_STATS_DEF(var)
#else
#define STATS_DEF(var)
#endif

unsigned long stats[stat_max];
char *stats_names[] = {
#include "stats_def.h"
};

#undef STATS_DEF
#undef CORE_STATS_DEF

void
stats_start(void)
{
	memset(stats, 0, sizeof(stats));
	write_pmc(0, 0);
#if 1
	//write_pmc(5, PMC_PL_ANY | PMC_PM | PMC_EVENT_CPU_CYCLES);
	write_pmc(6, PMC_PL3 | PMC_PM | PMC_EVENT_CPU_CYCLES);
	write_pmc(7, PMC_PL0 | PMC_PM | PMC_EVENT_CPU_CYCLES);
#endif
#if 0
	//write_pmc(5, PMC_PL_ANY | PMC_PM | PMC_EVENT_L2DTLB_MISSES);
	write_pmc(6, PMC_PL_ANY | PMC_PM | PMC_EVENT_DTLB_INSERTS_HPW);
	write_pmc(7, PMC_PL_ANY | PMC_PM | PMC_EVENT_DTLB_INSERTS_HPW_RETIRED);
#endif
	//write_pmc(5, PMC_PL0 | PMC_PL3 | PMC_PM | PMC_EVENT_DTLB_MISSES);
	//write_pmc(6, PMC_PL0 | PMC_PL3 | PMC_PM | PMC_EVENT_DTLB_INSERTS_HPW);
	//write_pmc(7, PMC_PL0 | PMC_PL3 | PMC_PM | PMC_EVENT_ITLB_MISSES_FETCH);
	//write_pmc(7, PMC_PL0 | PMC_PL3 | PMC_PM | PMC_EVENT_ITLB_INSERTS_HPW);
	//write_pmd(5, 0);
	write_pmd(6, 0);
	write_pmd(7, 0);
	//pp already enabled
	//write_ipsr(read_ipsr() | PSR_PP);
	//write_dcr(read_dcr() | DCR_PP);

	debug_flags |= DEBUG_FLAG_PROFILE;
}

void
stats_stop(void)
{
	unsigned int i;

	asm volatile("rsm psr.pp;; srlz.d");
	//write_ipsr(read_ipsr() & ~PSR_PP);
	//write_dcr(read_dcr() & ~DCR_PP);
	//write_gr(8, read_pmd(4));
	printf("PMD6: %lx\nPMD7: %lx\n", read_pmd(6), read_pmd(7));
	for (i = 0; i < sizeof(stats)/sizeof(unsigned long); i++)
		printf("%s: %lx\n", stats_names[i], stats[i]);

	debug_flags &= ~DEBUG_FLAG_PROFILE;
	profile_dsm_flush();
}

#if 0
void
monitor_mf_start(void)
{
	// mf    00000000100010 0000000000000
	// mask  00001000000000 1111111111111
	write_pmc(8, 0x800880000200ffffUL); /* program opcode matcher */
	/* write_pmc(8, 0x800000003fffffffUL); */ /* program opcode matcher */
	write_pmc(15, 0xfffffffeUL); /* enable opcode matching */
	write_pmc(4, PMC_PL_ANY | PMC_PM | PMC_EVENT_IA64_TAGGED_INST_RETIRED | PMC4_ENABLE);
	write_pmd(4, 0); /* reset PMD */
	write_pmc(0, 0); /* unfreeze PMU */
}
#endif

void
monitor_rel_start(void)
{
	/* release detection */
	write_pmc(4, PMC_PL3 | PMC_PM | PMC_EVENT_L2_OZQ_RELEASE | PMC4_ENABLE);

	/* thrash prevention */
	write_pmc(5, PMC_PL3 | PMC_PM | PMC_EVENT_IA64_TAGGED_INST_RETIRED);

#if 0	/* delayed fwb */
	write_pmc(5, PMC_PL3 | PMC_OI | PMC_PM | PMC_EVENT_CPU_CYCLES);
#endif

#if 0	/* mf detection - won't work */
	write_pmc(8, 0x800880000200ffffUL); /* program opcode matcher */
	write_pmc(15, 0xfffffffeUL); /* enable opcode matching */
	write_pmc(5, PMC_PL3 | PMC_OI | PMC_PM | PMC_EVENT_IA64_TAGGED_INST_RETIRED);
	write_pmd(5, (1UL<<47)-1);
#endif

	write_pmd(4, 0); /* reset PMDs */
	write_pmd(5, 0);
	write_pmc(0, 0); /* unfreeze PMU */
}

#ifdef DELAYED_FWB
unsigned long fwb_delay; /* = 100000; */

void
delayed_fwb_start(void)
{
	write_pmd(5, (1UL<<47)-fwb_delay);
	//write_pmd(5, 0);
}

void
delayed_fwb_stop(void)
{
	write_pmd(5, 0);
}
#endif

#if 0
void
tlbop(unsigned int type, unsigned long arg1, unsigned long arg2)
{
	switch (type)
	{
	case 0: /* ITC.D/ITR.D */
		tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], arg1, (arg2 != -1), 0);
		break;
	case 1: /* ITC.I/ITR.I */
		tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], arg1, (arg2 != -1), 1);
		break;
	default:
		printf("tlbop %d\n");
		break;
	}
}
#endif

static void
emu_M0(unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;
	unsigned long mask = BITS(instr, 6, 21) | ((x6 >> 4) << 21) | (BITS(instr, 36, 1) << 23);

	//vstats[x6]++;
	switch (x6)
	{
		case 0x6:
		case 0x16:
		case 0x26:
		case 0x36:
			DEBUG_EMU("[%lx] SSM 0x%lx\n", read_iip(), mask);
			write_psr(read_psr() | mask);
			break;

		case 0x7:
		case 0x17:
		case 0x27:
		case 0x37:
			DEBUG_EMU("[%lx] RSM 0x%lx\n", read_iip(), mask);
			write_psr(read_psr() & ~mask);
			break;

		default:
			printf("[%lx] M0: %lx\n", read_iip(), x6);
			debug();
	}

}

static void
emu_M1(unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;
	unsigned int r1 = (instr >> 6) & 0x7f;
	unsigned int r2 = (instr >> 13) & 0x7f;
	unsigned int r3 = (instr >> 20) & 0x7f;
	unsigned long address, value;
	unsigned int region;

	//vstats[x6]++;
	switch (x6)
	{
		case 0x00:
			address = read_gr(r3);
			region = address >> 61;
			value = read_gr(r2);
			DEBUG_MM("[%lx] MOVE to RR %lx = %lx\n", read_iip(), address, value);
			cpu.rr[region] = value;
			if (region != 0)
				write_rr(address, translate_rr((value & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1));
			break;

		case 0x01: /* MOV to DBR */
#ifdef PMU_DIRECT
			address = read_gr(r3);
			value = read_gr(r2);
			write_dbr(address, value);
#endif
			break;

		case 0x02: /* MOV to IBR */
#ifdef PMU_DIRECT
			address = read_gr(r3);
			value = read_gr(r2);
			write_ibr(address, value);
#endif
			break;

		case 0x04: /* MOV to PMC */
			//printf("[%lx] MOVE to PMC\n", read_iip());
#ifdef PMU_DIRECT
			address = read_gr(r3);
			value = read_gr(r2);
			write_pmc(address, value);
#endif
			break;

		case 0x05: /* MOV to PMD */
#ifdef PMU_DIRECT
			address = read_gr(r3);
			value = read_gr(r2);
			write_pmd(address, value);
#endif
			break;

		case 0x06: /* MOV to MSR */
			DEBUG_EMU("[%lx] FC\n", read_iip());
			address = read_gr(r3);
			value = tpa(address);
			if (value == 0)
			{
				//printf("FC on unmapped address %lx\n", address);
				rewind_ip();
				mm_tlb_miss(address, 0);
				return;
			}
			asm volatile("fc %0" : : "r"(address));
			break;

		case 0x09: /* PTC.L */
			DEBUG_MM("[%lx] PTC.L\n", read_iip());
			tlb_purge(read_gr(r3), read_gr(r2), 0);
			break;

		case 0x0a: /* PTC.G */
		case 0x0b: /* PTC.GA */
			DEBUG_MM("[%lx] PTC.G\n", read_iip());
			tlb_purge(read_gr(r3), read_gr(r2), 1);
			break;

		case 0x0e:
			address = read_gr(r3);
			value = read_gr(r2);
			DEBUG_MM("[%lx] ITR.D DTR[%ld] = %lx (IFA = %lx, ITIR = %lx)\n", read_iip(), address, value,
									cpu.cr[CR_IFA], cpu.cr[CR_ITIR]);
			tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], value, 1, 0);
			break;

		case 0x0f:
			address = read_gr(r3);
			value = read_gr(r2);
			DEBUG_MM("[%lx] ITR.I ITR[%ld] = %lx (IFA = %lx, ITIR = %lx)\n", read_iip(), address, value,
									cpu.cr[CR_IFA], cpu.cr[CR_ITIR]);
			tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], value, 1, 1);
			break;

		case 0x10:
			address = read_gr(r3);
			region = address >> 61;
			DEBUG_MM("[%lx] MOVE from RR %lx = %lx\n", read_iip(), address, cpu.rr[region]);
			write_gr(r1, cpu.rr[region]);
			break;

		case 0x14: /* MOVE from PMC */
#ifdef PMU_DIRECT
			address = read_gr(r3);
			value = read_pmc(address);
			write_gr(r1, value);
			//logf("[%lx] rd PMC%lx = %lx\n", read_iip(), address, value);
#endif
			break;

		case 0x16: /* MOVE from MSR */
			address = read_gr(r3);
			value = thash(address);
			DEBUG_MM("[%lx] THASH\n", read_iip());
			write_gr(r1, value);
			break;

		case 0x1e: /* TPA */
			address = read_gr(r3);
			DEBUG_MM("[%lx] TPA %lx\n", read_iip(), address);
			value = tpa(address);
			if (value == 0)
			{
				//printf("TPA on unmapped address %lx\n", address);
				rewind_ip();
				mm_tlb_miss(address, 0);
				return;
			}
			write_gr(r1, value);
			break;

		case 0x24:
			DEBUG_EMU("[%lx] MOVE from CR %d = %lx\n", read_iip(), r3, cpu.cr[r3]);
			write_gr(r1, cpu.cr[r3]);
			switch (r3)
			{
				case CR_IVR:
					cpu.in_service_irq = cpu.cr[CR_IVR];
					clear_interrupt(cpu.cr[CR_IVR]);
					break;
			}
			break;

		case 0x25:
			DEBUG_EMU("[%lx] MOVE from PSR = %lx\n", read_iip(), read_psr());
			write_gr(r1, read_psr());
			break;

		case 0x2a:
			value = read_gr(r2);
			DEBUG_EMU("[%lx] MOVE to AR %d = 0x%lx\n", read_iip(), r3, value);
			write_ar(r3, value);
			break;

		case 0x2c:
			cpu.cr[r3] = read_gr(r2);
			DEBUG_EMU("[%lx] MOVE to CR %d = 0x%lx\n", read_iip(), r3, cpu.cr[r3]);
			switch (r3)
			{
				case CR_EOI:
					cpu.in_service_irq = 0;
					break;
				case CR_ITM:
					write_tpr(0);
					write_itv(0xff);
					write_itm(cpu.cr[CR_ITM]);
					break;
			}
			break;

		case 0x2d:
			value = read_gr(r2);
			DEBUG_EMU("[%lx] MOVE to PSR.l = %lx\n", read_iip(), value);
			write_psr((value & 0xffffffffUL) | (read_psr() >> 32 << 32));
			break;

		case 0x2e:
			value = read_gr(r2);
			DEBUG_MM("[%lx] ITC.D %lx (IFA = %lx, ITIR = %lx)\n", read_iip(), value,
					cpu.cr[CR_IFA], cpu.cr[CR_ITIR]);
			tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], value, 0, 0);
			break;

		case 0x2f:
			value = read_gr(r2);
			DEBUG_MM("[%lx] ITC.I %lx (IFA = %lx, ITIR = %lx)\n", read_iip(), value,
					cpu.cr[CR_IFA], cpu.cr[CR_ITIR]);
			tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], value, 0, 1);
			break;

		case 0x34:
			DEBUG_MM("[%lx] PTC.E\n", read_iip());
			tlb_purge_all();
			break;

		default:
			printf("[%lx] M1: %lx\n", read_iip(), x6);
			debug();
	}
}

static void
emu_B0(unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;

	//vstats[x6]++;
	switch (x6)
	{
		case 0x8:
			DEBUG_EMU("[%lx] RFI to %lx ipsr=%lx ifs=%lx\n", read_iip(), cpu.cr[CR_IIP], cpu.cr[CR_IPSR], cpu.cr[CR_IFS]);
			emu_rfi();
			break;

		case 0xc:
		case 0xd:
			DEBUG_EMU("[%lx] BSW.%d\n", read_iip(), (unsigned int)(x6 & 1));
			emu_bsw(x6 & 1);
			break;

		default:
			printf("[%lx] B0: %lx\n", read_iip(), x6);
			debug();
	}
}

void
handle_general_exception(void)
{
	unsigned long cr_iip, cr_ipsr;
	unsigned long w1, w2, instr;
	unsigned int slot, type, opcode;

	cr_iip = read_iip();
	cr_ipsr = read_ipsr();

	DEBUG_INTERRUPTION("general exception %lx", read_iip());
	DEBUG_SHORT('e');

	if (tpa(cr_iip) == 0)
	{
		mm_tlb_miss(cr_iip, read_isr());
		return;
	}
	//asm volatile("ssm psr.ic;; srlz.d;;");
	w1 = *(unsigned long *)cr_iip;
	w2 = *(unsigned long *)(cr_iip+8);
	//asm volatile("rsm psr.ic;; srlz.d;;");

	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			instr = (w2 >> 23);
			write_ipsr((cr_ipsr &~ PSR_RI));
			write_iip(cr_iip + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			return;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 37) & 0xf;

	if ((type == M) && (opcode == 0))
		emu_M0(instr);
	else if ((type == M) && (opcode == 1))
		emu_M1(instr);
	else if ((type == B) && (opcode == 0))
		emu_B0(instr);
	else
	  {
		printf("[%lx] unexpected type (data = %08lx %08lx isr = %lx iipa = %lx)\n", read_iip(), w2, w1, read_isr(), read_iipa());
		debug();
	  }
}

void
emu_store_PIB(unsigned long addr)
{
	unsigned long cr_iip, cr_ipsr;
	unsigned long w1, w2, instr;
	unsigned int slot, type, opcode;
	unsigned int r2, r3, target;
	unsigned long staddr, value;
	struct request request;

	cr_iip = read_iip();
	cr_ipsr = read_ipsr();

	if (tpa(cr_iip) == 0)
	{
		mm_tlb_miss(cr_iip, read_isr());
		return;
	}
	//asm volatile("ssm psr.ic;; srlz.d;;");
	w1 = *(unsigned long *)cr_iip;
	w2 = *(unsigned long *)(cr_iip+8);
	//asm volatile("rsm psr.ic;; srlz.d;;");

	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			instr = (w2 >> 23);
			write_ipsr((cr_ipsr &~ PSR_RI));
			write_iip(cr_iip + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			return;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 33) & 0xff; // includes bits 40:37,36,35:33

	if ((type != M) || (opcode != 0x46))
	{
		printf("emu_store_PIB: ip %lx, type %x, opcode %x\n", cr_iip, type, opcode);
		return;
	}

	r2 = (instr >> 13) & 0x7f;
	r3 = (instr >> 20) & 0x7f;
	staddr = read_gr(r3);
	value = read_gr(r2);
	if (staddr != addr)
	{
		printf("emu_store_PIB: staddr %lx, addr %lx\n", staddr, addr);
		return;
	}

	if ((addr - 0xc0000000fee00000UL) > 2UL*1024*1024)
	{
		printf("emu_store_PIB: ip %lx, addr %lx not in PIB\n", cr_iip, addr);
		return;
	}
	target = (addr >> 4) & 63;

#ifdef DSM
	if (value == 0xff)
	{
		extern unsigned long smp_rendezvous_ip, smp_rendezvous_gp;
		request.type = IPC_STARTUP;
		request.startup.ip = smp_rendezvous_ip;
		request.startup.gp = smp_rendezvous_gp;
	}
	else
	{
		request.type = IPC_IPI;
		request.ipi.irq = value;
	}

	ipc_send(target, &request);
#endif
	//printf("IPI to CPU%d %x\n", target, request.ipi.irq);
}

#define WRITE_BUFFER_SIZE 32
static struct store *write_buffer;
static unsigned int write_buffer_idx;
int seen_release;
extern uint16_t wu_applied;
extern uint16_t current_generation;

void emu_wbinit(void)
{
	write_buffer = malloc(WRITE_BUFFER_SIZE * sizeof(struct store));
}

void emu_fwb(void)
{
	struct request request;

#ifdef DELAYED_FWB
	delayed_fwb_stop();
#endif
	write_pmd(4, 0);
	seen_release = 0;

	if (!write_buffer_idx)
		return;

	//printf("[%lx] fwb %x\n", read_iip(), write_buffer_idx);

	request.type = IPC_STORE;
	request.store.nstores = write_buffer_idx;
	request.store.wu_version = wu_applied;
	ipc_send_page(63 /* !node_id */, &request, __pa(write_buffer), write_buffer_idx*sizeof(struct store), DATA_AREA_MALLOC);
	wu_applied += 1<<NODE_BITS;
	if ((wu_applied >> 14) != current_generation)
		mm_new_generation();
	write_buffer = malloc(WRITE_BUFFER_SIZE * sizeof(struct store));
	write_buffer_idx = 0;
}

void
emu_fwb_if_hit(mpfn_t mpfn)
{
	unsigned int i;
	for (i = 0; i < write_buffer_idx; i++)
	{
		if ((write_buffer[i].mpa >> PAGE_SHIFT) == mpfn)
		{
			STATS_INC(fwb_hit);
			emu_fwb();
			return;
		}
	}

	if (seen_release || read_pmd(4))
	{
		//write_pmd(4, 0);
		//printf("release seen, forcing release!\n");
		STATS_INC(fwb_pmu_release);
		emu_fwb();
	}
}

void
emu_store_remote(unsigned long mpa, unsigned long value, unsigned int size,
		   int release)
{
	/* queue the store to the write buffer */
	/* the store will become remotely visible when the write buffer is flushed */
	/* certain stores may become remotely visible earlier if a remote node requests the page */
	write_buffer[write_buffer_idx].mpa = mpa;
	write_buffer[write_buffer_idx].value = value;
	write_buffer[write_buffer_idx].wu_version = wu_applied;
	write_buffer[write_buffer_idx].size = size;
	write_buffer_idx++;

#if 0
	if (!release && (write_buffer_idx > 1) && read_pmd(4))
	{
		//write_pmd(4, 0);
		//printf("release seen, forcing release!\n");
		release = 1;
	}
#endif

#if 1
	if (write_buffer_idx == WRITE_BUFFER_SIZE)
	{
		STATS_INC(fwb_full);
		emu_fwb();
	}
	else if (release)
	{
		STATS_INC(fwb_st_release);
#ifdef DELAYED_FWB
		if (fwb_delay)
		{
			seen_release = 1;
			delayed_fwb_start();
		}
		else
#endif
			emu_fwb();
	}
#elif 0
	if (release || (write_buffer_idx == WRITE_BUFFER_SIZE))
		emu_fwb();
#else
	if (write_buffer_idx == WRITE_BUFFER_SIZE)
		emu_fwb();
	else if (release)
		seen_release = 1;
#endif
}

int
emu_store(unsigned long addr, unsigned long mpa)
{
	unsigned long cr_iip, cr_ipsr;
	unsigned long w1, w2, instr;
	unsigned int slot, type, opcode;
	unsigned int r1, r2, size, release;
	unsigned long value;
#ifndef NDEBUG
	unsigned long staddr;
	unsigned int r3;
#endif

	cr_iip = read_iip();
	cr_ipsr = read_ipsr();

	if (tpa(cr_iip) == 0)
	{
		//printf("emu_store missed on instruction page\n");
		mm_tlb_miss(cr_iip, read_isr());
		return 1;
	}
	//asm volatile("ssm psr.ic;; srlz.d;;");
	w1 = *(unsigned long *)cr_iip;
	w2 = *(unsigned long *)(cr_iip+8);
	//asm volatile("rsm psr.ic;; srlz.d;;");

	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			instr = (w2 >> 23);
			write_ipsr((cr_ipsr &~ PSR_RI));
			write_iip(cr_iip + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			return 0;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 33) & 0xff; // includes bits 40:37,36,35:33
#if 0
	opcode = (instr >> 32) & 0x1ff; // includes bits 40:37,36,35:32
#endif
	release = (instr >> 32) & 1;
	size = (instr >> 30) & 3;
	r2 = (instr >> 13) & 0x7f;
#ifndef NDEBUG
	r3 = (instr >> 20) & 0x7f;
	staddr = read_gr(r3);
#endif

	/* FIXME how to handle cmpxchg? */
	/* deal with ,imm form (0x56, 0x76) as well? may not be a good optimisation as then there'll most likely be another store */
	/* the size != 0 disqualifies stfe */
	if (type != M)
		goto badop;
	if (opcode == 0x46)
		value = read_gr(r2);
	else if (opcode == 0x66)
	{
		if (size == 1)
		{
			value = read_fr_sig(r2);
			size = 3;
		}
		else if (size == 2)
			value = read_fr_s(r2);
		else if (size == 3)
			value = read_fr_d(r2);
		else goto badop;
	}
	else goto notstore;

#ifndef NDEBUG
	if (staddr != addr)
	{
		printf("emu_store: staddr %lx, addr %lx\n", staddr, addr);
		rewind_ip();
		return 0;
	}

	//printf("emu_store: ip %lx, type %x, opcode %x, staddr %lx, mpa %lx, value %lx size %x\n", cr_iip, type, opcode, staddr, mpa, value, size);
#endif

	mm_handle_store(mpa, value, size, node_id, wu_applied, 0); /* apply locally */
	emu_store_remote(mpa, value, size, release);
	return 1;

notstore:
#ifdef SPECIAL_CMPXCHG_FETCHADD
	if (opcode == 0x40) /* cmpxchg */
	{
		value = mm_cmpxchg(mpa, read_gr(r2), read_register("ar.ccv"), size, release, addr);
		r1 = (instr >> 6) & 0x7f;
		write_gr(r1, value);
		return 1;
	}
	else if (opcode == 0x42) /* fetchadd */
	{
		value = mm_fetchadd(mpa, size, r2, release);
		r1 = (instr >> 6) & 0x7f;
		write_gr(r1, value);
		return 1;
	}
#endif

badop:
	//printf("emu_store: ip %lx, type %x, opcode %x\n", cr_iip, type, opcode);
	rewind_ip();
	return 0;
}

int
emu_load(unsigned long addr, unsigned long mpa)
{
	unsigned long cr_iip, cr_ipsr;
	unsigned long w1, w2, instr;
	unsigned int slot, type, opcode;
	unsigned int r1, r2, size, release, loadtype;
	uint64_t value;
#ifndef NDEBUG
	unsigned long ldaddr;
	unsigned int r3;
#endif

	cr_iip = read_iip();
	cr_ipsr = read_ipsr();

	if (tpa(cr_iip) == 0)
	{
		//printf("emu_store missed on instruction page\n");
		mm_tlb_miss(cr_iip, read_isr());
		return 1;
	}
	//asm volatile("ssm psr.ic;; srlz.d;;");
	w1 = *(unsigned long *)cr_iip;
	w2 = *(unsigned long *)(cr_iip+8);
	//asm volatile("rsm psr.ic;; srlz.d;;");

	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			instr = (w2 >> 23);
			write_ipsr((cr_ipsr &~ PSR_RI));
			write_iip(cr_iip + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			return 0;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 33) & 0xff; // includes bits 40:37,36,35:33
#if 0
	opcode = (instr >> 32) & 0x1ff; // includes bits 40:37,36,35:32
	release = (instr >> 32) & 1;
#endif
	size = (instr >> 30) & 3;
	r1 = (instr >> 6) & 0x7f;
#ifndef NDEBUG
	r3 = (instr >> 20) & 0x7f;
	ldaddr = read_gr(r3);
#endif

	/* 0x40 = ld, ld.s, 0x42 = ld.acq, ld.bias */
	/* deal with ,imm form (0x56, 0x76) as well? may not be a good optimisation as then there'll most likely be another load */
	if (type != M)
		goto badop;
	if ((opcode == 0x40) || (opcode == 0x42))
		loadtype = 0; /* integer */
	else if (opcode == 0x60)
	{
		loadtype = size;
		if (size == 1)
			size = 3;
		else if (size == 0)
			goto badop;
	}
	else
		goto badop;

#ifndef NDEBUG
	if (ldaddr != addr)
	{
		printf("emu_load: ldaddr %lx, addr %lx\n", ldaddr, addr);
		rewind_ip();
		return 0;
	}

	//printf("emu_store: ip %lx, type %x, opcode %x, staddr %lx, mpa %lx, value %lx size %x\n", cr_iip, type, opcode, staddr, mpa, value, size);
#endif

	//printf("do_load: ip %lx, mpa %lx, size %x\n", cr_iip, mpa, size);
	if (!mm_load(mpa, size, &value))
	{
		/* need to wait */
		rewind_ip();
		return 1;
	}
	//printf("do_load: r%x value %lx\n", r1, value);
	switch (loadtype)
	{
		case 0: write_gr(r1, value); break;
		case 1: write_fr_sig(r1, value); break;
		case 2: write_fr_s(r1, value); break;
		case 3: write_fr_d(r1, value); break;
	}
	return 1;

#if 0
notload:
	if (opcode == lfetch)
	{
	}
#endif
badop:
	//printf("emu_load: ip %lx, type %x, opcode %x\n", cr_iip, type, opcode);
	rewind_ip();
	return 0;
}

#if 0
void
emu_diff(unsigned long mpa, void *page, void *oldpage)
{
	void *end = page + PAGE_SIZE;
	uint64_t *p = page;
	uint64_t *oldp = oldpage;

	if (!write_buffer)
		write_buffer = malloc(WRITE_BUFFER_SIZE * sizeof(struct store));

	while (; p != end; p++, oldp++)
	{
		if (*p != *old)
		{
			write_buffer[write_buffer_idx].mpa = (p - page) + mpa;
			write_buffer[write_buffer_idx].value = *p;
			write_buffer[write_buffer_idx].size = 3; /* 8 */
			write_buffer_idx++;

			if (write_buffer_idx == WRITE_BUFFER_SIZE)
			{
				emu_fwb();
				write_buffer = malloc(WRITE_BUFFER_SIZE * sizeof(struct store));
			}
		}
	}

	free(oldpage);
}
#endif

void
handle_break(void)
{
	unsigned long ipsr = read_ipsr();
	unsigned long iim = read_iim();
	unsigned long fn, arg1, arg2, res, start;
	unsigned int slot;

	DEBUG_INTERRUPTION("break %lx %lx\n", read_iip(), iim);
	DEBUG_SHORT('b');

	slot = BITS(ipsr, PSR_RI_BITS, 2);

	switch (slot)
	{
		case 0:
			write_ipsr((ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			write_ipsr((ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			write_ipsr(ipsr &~ PSR_RI);
			write_iip(read_iip() + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			debug();
			return;
	}

	switch (iim)
	{
#if 0
		case 0:
			printf("oops, break 0\n");
			break;
#endif
		case 0x80001:
			//DEBUG_EMU("[%lx] SSC %ld\n", read_iip(), read_gr(15));
			write_gr(8, do_ssc(read_gr(32), read_gr(33), read_gr(34), read_gr(35), read_gr(15)));
			break;
		case 0x80002:
			//DEBUG_EMU("[%lx] COVER\n", read_iip());
			/* printf("BAD: COVER should not be hit\n"); */
			emu_cover();
#if 0
			/* Afterburner hack */
			{
				unsigned long restart_ip = read_register("r4");
				if (!(cpu.psr & PSR_CPL_3) && (restart_ip != 0))
				{
					write_iip(restart_ip + 0x10);
					write_ipsr(read_ipsr() &~ PSR_RI);
					write_register("r4", 0);
				}
			}
#endif
			break;
		case 0x80003:
			//printf("PAL assist fault\n");
			fn = read_gr(28);
			switch (fn)
			{
#ifdef DO_PAL_HALT
				case PAL_HALT_LIGHT:
					//write_gr(8, 0);
					start = read_itc();
					res = call_pal_halt();
					if (res != 0)
						printf("PAL_HALT_LIGHT: %d\n", res);
					handle_interrupt();
					CORE_STATS_ADD(idle_time, read_itc()-start);
					break;
#endif
				case PAL_FREQ_RATIOS:
					write_gr(8,  0);
					write_gr(9,  0x100000001UL);
					write_gr(10, 0x100000001UL);
					write_gr(11, 0x100000001UL);
					break;
				case PAL_PTCE_INFO:
					write_gr(8,  0);
					write_gr(9,  0);
					write_gr(10, 0x100000001UL);
					write_gr(11, 0);
					break;
				default:
					write_gr(8, -1);
					break;
			}
			break;
		case 0x80004:
			//printf("SAL assist fault\n");
			fn = read_gr(32);
			switch (fn)
			{
				case SAL_FREQ_BASE:
					write_gr(8, 0);
#ifdef SKI
#error CHECKME
					write_gr(9,  40000000);
#else
					write_gr(9, /* 1500000000 */ 900000000);
#endif
					break;
				default:
					write_gr(8, -1);
					break;
			}
			break;
#ifdef DELAYED_FWB
		case 0x80005:
			fwb_delay = read_register("r5");
			printf("fwb_delay = %lx\n", fwb_delay);
			break;
#endif
		case 0x80006:
			{
			STATS_IF(unsigned long start_time = read_itc());
			STATS_IF(unsigned long time_taken);
			fn = read_register("r5");
			arg1 = read_register("r6");
			arg2 = read_register("r7");
			//DEBUG_EMU("[%lx] TLBOP\n", read_iip());
			switch (fn) {
			case 0: /* ITC.D/ITR.D */
				tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], arg1, arg2 != (unsigned long)-1, 0);
				break;
			case 1: /* ITC.I/ITR.I */
				tlb_insert(cpu.cr[CR_IFA], cpu.cr[CR_ITIR], arg1, arg2 != (unsigned long)-1, 1);
				break;
			case 2: /* PTC.L */
				tlb_purge(arg1, arg2, 0);
				break;
			case 3: /* PTC.G/PTC.GA */
				tlb_purge(arg1, arg2, 1);
				break;
			case 4: /* PTR.D */
				printf("PTR.D not implemented\n");
				break;
			case 5: /* PTR.I */
				printf("PTR.I not implemented\n");
				break;
			case 6: /* PTC.E */
				tlb_purge_all();
				break;
			}
			STATS_IF(time_taken = read_itc() - start_time);
			if (fn >= 2)
				STATS_ADD(tlb_purge_time, time_taken);
			else
				STATS_ADD(tlb_insert_time, time_taken);
			}
			break;
		case 0x80007: /* EPC */
			cpu.psr &= ~PSR_CPL_3;
			break;
#if 0
		case 0x80007:
			arg1 = read_register("r5");
			arg2 = read_register("r6");
			arg1 >>= 61;
			//DEBUG_EMU("[%lx] MOV RR=\n", read_iip());
			printf("BAD: MOV RR= should not be hit\n");
			cpu.rr[arg1] = arg2;
			if (arg1 != 0)
				write_rr(arg1 << 61, translate_rr((arg2 & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1));
			break;
#endif
		case 0x80008:
			//DEBUG_EMU("[%lx] SSM PSR.DFH\n", read_iip());
			printf("BAD: SSM PSR.DFH should not be hit\n");
			write_psr(read_psr() | PSR_DFH);
			break;
		case 0x80009:
			ipc_reset_system();
			reset_system();
			break;
		case 0x8000a:
			printf("get_time called!\n");
			write_gr(8, 0x8000000000000003UL);
			break;
		case 0x8000b:
			write_update_threshold = read_register("r5");
			printf("write_update_threshold = %d\n", write_update_threshold);
			break;
		case 0x8000c:
			extra_latency = read_register("r5");
			printf("extra_latency = %lx\n", extra_latency);
			break;
		case 0x8000d: /* halt */
			start = read_itc();
			while ((res = read_ivr()) == 0xf);
			do_irq(res);
			check_interrupts();
			CORE_STATS_ADD(idle_time, read_itc()-start);
			break;
		case 0x8000e: /* clear_page */
			arg1 = read_register("r5");
			arg2 = read_register("r6");
			//printf("clear_page %lx\n", arg1);
			arg1 >>= PAGE_SHIFT;
#if 0
			arg2 = arg1 + (1UL << (arg2-PAGE_SHIFT));
			for (; arg1 < arg2; arg1++)
			{
				mm_check_mpa(arg1 << PAGE_SHIFT, PROT_RWX);
				clear_page(arg1 << PAGE_SHIFT);
			}
#else
			arg2 = (1UL << (arg2-PAGE_SHIFT));
			mm_invalidate_whole(arg1, arg2);
#endif
			break;
#ifdef SPECIAL_LOCKS
		case 0x8000c:
			arg1 = read_register("r30"); // r8
			res = tpa(arg1);
			if (res == 0)
			{
				rewind_ip();
				mm_tlb_miss(arg1, 0);
				return;
			}
			lock_acquire(res);
			break;
		case 0x8000d:
			arg1 = read_register("r30");
			res = tpa(arg1);
			if (res == 0)
			{
				rewind_ip();
				mm_tlb_miss(arg1, 0);
				return;
			}
			lock_release(res);
			break;
		case 0x8000e:
			arg1 = read_register("r30");
			res = tpa(arg1);
			if (res == 0)
			{
				rewind_ip();
				mm_tlb_miss(arg1, 0);
				return;
			}
			write_register("r30", lock_is_locked(res));
			break;
		case 0x8000f:
			arg1 = read_register("r30");
			res = tpa(arg1);
			if (res == 0)
			{
				rewind_ip();
				mm_tlb_miss(arg1, 0);
				return;
			}
			lock_init(res);
			break;
#endif
#if 0
		case 0x80009:
			//DEBUG_EMU("[%lx] SSM PSR.I\n", read_iip());
			printf("BAD: SSM PSR.I should not be hit\n");
			write_psr(read_psr() | PSR_I);
			break;
		case 0x80010:
			//DEBUG_EMU("[%lx] BSW.0\n", read_iip());
			printf("BAD: BSW.0 should not be hit\n");
			emu_bsw(0);
			break;
		case 0x80011:
			//DEBUG_EMU("[%lx] BSW.1\n", read_iip());
			printf("BAD: BSW.1 should not be hit\n");
			emu_bsw(1);
			break;
#endif
		case 0x80010:
			//profile((unsigned long)-1, read_register("r5"), 0, 0, 0, 0);
			ipc_send_reply(63, IPC_STATS_START);
			stats_start();
			break;
		case 0x80011:
			//profile((unsigned long)-1, read_register("r5"), 0, 0, 0, 0);
			ipc_send_reply(63, IPC_STATS_STOP);
			stats_stop();
			break;
		case 0x80012:
			//DEBUG_EMU("[%lx] MOV =CR.IVR\n", read_iip());
#if 0
			if (cpu.cr[CR_IVR] == 0x30)
				printf("^");
#endif
			write_register("r5", cpu.cr[CR_IVR]);
			cpu.in_service_irq = cpu.cr[CR_IVR];
			clear_interrupt(cpu.cr[CR_IVR]);
			break;
		case 0x80013:
			//DEBUG_EMU("[%lx] MOV CR.ITM=\n", read_iip());
			printf("BAD: MOV CR.ITM= should not be hit\n");
			cpu.cr[CR_ITM] = read_register("r5");
			write_tpr(0);
			write_itv(0xff);
			write_itm(cpu.cr[CR_ITM]);
			break;
		case 0x80014:
			//DEBUG_EMU("[%lx] THASH\n", read_iip());
			printf("BAD: THASH should not be hit\n");
			write_register("r5", thash(read_register("r5")));
			break;
		case 0x80015:
			//DEBUG_EMU("[%lx] TPA\n", read_iip());
			printf("BAD: TPA should not be hit\n");
			arg1 = read_register("r5");
			res = tpa(arg1);
			if (res == 0)
			{
				//printf("TPA on unmapped address %lx\n", address);
				rewind_ip();
				mm_tlb_miss(arg1, 0);
				return;
			}
			write_register("r5", res);
			break;
		case 0x80016:
			//DEBUG_EMU("[%lx] MOV PSR.L=\n", read_iip());
			printf("BAD: MOV PSR.L= should not be hit\n");
			write_psr((read_register("r5") & 0xffffffffUL) | (read_psr() >> 32 << 32));
			break;
		case 0x8001d:
			logf("=");
			break;
		case 0x8001e:
			logf("|");
			break;
		case 0x8001f:
			process_debug_char('l');
			break;
		case 0x80020:
			//DEBUG_EMU("RFI\n");
			emu_rfi();
			break;
		case 0x80021:
			via_deliver_interrupt = 1;
			check_interrupts();
			via_deliver_interrupt = 0;
			break;
#if 0
		case 0x80022:
			arg1 = read_register("r4");
			arg2 = read_register("r5");
			printf("delivering exception (%lx %lx)\n", arg1, arg2);
			write_iip(arg1);
			write_register("r4", 0xfff);
			deliver_exception(arg2, cpu.cr[CR_ISR]);
			break;
#endif
		case 0x80023:
			/* tpa assist vector */
			//printf("tpa assist %lx %lx\n", cpu.temp[0], cpu.temp[1]);
			write_iip(cpu.temp[0]);
			write_register("r4", 0xfff);
			mm_tlb_miss(cpu.temp[1], 0);
			break;
		default:
			//DEBUG_EMU("[%lx] BREAK %lx %ld\n", read_iip(), iim, read_gr(15));
			rewind_ip();
			cpu.cr[CR_IIM] = iim;
			deliver_exception(0x2c00, 0);
			return;
	}
}

void
handle_unaligned(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("unaligned %lx %lx\n", read_iip(), address);
	DEBUG_SHORT('u');
	cpu.cr[CR_IFA] = address;
	cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
	deliver_exception(0x5a00, isr);
}

void
handle_speculation(void)
{
	long offset = (((long)read_iim() << 43) >> 39);

	printf("spec offset=%ld\n", offset);
	write_iip(read_iip() + offset);
	write_ipsr(read_ipsr() &~ PSR_RI);

#ifdef THRASH_PROTECTION
	/* On Itanium 2 we only get here when psr.ss=1 on a resteer.
	 * So handle the single-step trap now */
	handle_single_step();
#endif
}

void
handle_unhandled(void)
{
	unsigned long vector = read_register("r3");
	DEBUG_INTERRUPTION("unhandled %lx\n", vector);
	printf("unhandled exception %lx\n", vector);
	debug();
}

void
startup_secondary(unsigned long ip, unsigned long gp)
{
	write_gr(1, gp);
	write_iip(ip);
}

#ifdef DO_FPSWA
extern efi_fpswa_t fpswa;

void
handle_fpswa(int is_fault)
{
	unsigned long pr = read_pr();
	unsigned long cr_iip = read_iip();
	unsigned long cr_isr = read_isr();
	unsigned long cr_ipsr = read_ipsr();
	unsigned long cr_ifs = read_ifs();
	unsigned long fpsr = read_fpsr();
	unsigned long bundle[2];
	unsigned int slot;
	fp_state_t fp_state;
	fpswa_ret_t ret;
	unsigned int i;

	DEBUG_SHORT('f');
	printf("fpswa2 @ %lx\n", cr_iip);
	printf("calling interface @ %lx\n", (unsigned long)fpswa);

	slot = BITS(cr_ipsr, PSR_RI_BITS, 2);
	if (!is_fault && (slot == 0))
		cr_iip -= 16;

	if (tpa(cr_iip) == 0)
	{
		mm_tlb_miss(cr_iip, 0);
		return;
	}
	//asm volatile("ssm psr.ic;; srlz.d;;");
	bundle[0] = *(unsigned long *)cr_iip;
	bundle[1] = *(unsigned long *)(cr_iip+8);
	//asm volatile("rsm psr.ic;; srlz.d;;");

	memset(&fp_state, 0, sizeof(fp_state));
	ret = call_fpswa(is_fault, bundle, &cr_ipsr, &fpsr, &cr_isr, &pr, &cr_ifs, &fp_state);
	if (ret != 0)
		printf("fpswa failed (%x)!\n", ret);

	if (is_fault)
	{
		switch (slot)
		{
		case 0:
			write_ipsr((cr_ipsr &~ PSR_RI) | (1UL << PSR_RI_BITS));
			break;
		case 1:
			write_ipsr((cr_ipsr &~ PSR_RI) | (2UL << PSR_RI_BITS));
			break;
		case 2:
			write_ipsr(cr_ipsr &~ PSR_RI);
			write_iip(cr_iip + 0x10);
			break;
		default:
			printf("bad slot %d\n", slot);
			return;
		}
	}

	write_fpsr(fpsr); /* write new fp status values */
	write_pr(pr);
}

void
handle_fp_fault(void)
{
	handle_fpswa(1);
}

void
handle_fp_trap(void)
{
	handle_fpswa(0);
}
#endif

