/*
 * PAL & SAL emulation.
 *
 * Copyright (C) 1998-2001 Hewlett-Packard Co
 *	David Mosberger-Tang <davidm@hpl.hp.com>
 */

#include "vm.h"
#include "bootparam.h"
#include "efi.h"
#include "sal.h"
#include "fpswa.h"
#include "ssc.h"

#define MB	(1024*1024UL)

#define NUM_MEM_DESCS	2

extern void palproc, palproc_end, salproc, salproc_end, efiproc, efiproc_end, efigettimeproc, efigettimeproc_end, efiresetproc, efiresetproc_end;

struct firmware {
	efi_system_table_t efi_systab;
	efi_runtime_services_t efi_runtime;
	efi_config_table_t efi_tables;
	efi_memory_desc_t efi_memmap[NUM_MEM_DESCS];
	struct ia64_sal_systab sal_systab;
	struct ia64_boot_param bp;
	char cmd_line[1024];
	char pal_proc[64];
	char sal_proc[64];
	char efi_proc[64];
	char efi_get_time_proc[64];
	char efi_reset_proc[64];
	uint64_t efi_unimplemented[2];
	uint64_t efi_get_time[2];
	uint64_t efi_reset_system[2];
	uint16_t fw_vendor[5];
} *vm_firmware = (void *)(16*MB-PAGE_SIZE);

struct ia64_boot_param *
init_guest_fw (char *command_line)
{
	efi_memory_desc_t *md;
	unsigned char checksum;
	unsigned char *p;
	unsigned int i;

#	define MAKE_MD(n, typ, attr, start, end)	\
	do {						\
		md = &vm_firmware->efi_memmap[n];	\
		md->type = typ;				\
		md->pad = 0;				\
		md->phys_addr = start;			\
		md->virt_addr = 0;			\
		md->num_pages = (end - start) >> 12;	\
		md->attribute = attr;			\
	} while (0)

	memset(vm_firmware, 0, sizeof(struct firmware));

	/* Set up EFI tables */
	vm_firmware->efi_systab.hdr.signature = EFI_SYSTEM_TABLE_SIGNATURE;
	vm_firmware->efi_systab.hdr.revision  = EFI_SYSTEM_TABLE_REVISION;
	vm_firmware->efi_systab.hdr.headersize = sizeof(vm_firmware->efi_systab.hdr);
	vm_firmware->efi_systab.fw_vendor = (uint64_t)(vm_firmware->fw_vendor);
	vm_firmware->efi_systab.fw_revision = 1;
	vm_firmware->efi_systab.runtime = (uint64_t)&vm_firmware->efi_runtime;
	vm_firmware->efi_systab.nr_tables = 1;
	vm_firmware->efi_systab.tables = (uint64_t)&vm_firmware->efi_tables;

	vm_firmware->efi_runtime.hdr.signature = EFI_RUNTIME_SERVICES_SIGNATURE;
	vm_firmware->efi_runtime.hdr.revision = EFI_RUNTIME_SERVICES_REVISION;
	vm_firmware->efi_runtime.hdr.headersize = sizeof(vm_firmware->efi_runtime.hdr);
	vm_firmware->efi_runtime.get_time = (uint64_t)vm_firmware->efi_get_time;
	vm_firmware->efi_runtime.set_time = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.get_wakeup_time = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.set_wakeup_time = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.set_virtual_address_map = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.get_variable = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.get_next_variable = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.set_variable = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.get_next_high_mono_count = (uint64_t)vm_firmware->efi_unimplemented;
	vm_firmware->efi_runtime.reset_system = (uint64_t)vm_firmware->efi_reset_system;

	vm_firmware->efi_tables.guid = SAL_SYSTEM_TABLE_GUID;
	vm_firmware->efi_tables.table = (uint64_t)&vm_firmware->sal_systab;

	MAKE_MD(0, EFI_RUNTIME_SERVICES_DATA,		EFI_MEMORY_WB,    0*MB,    16*MB);
	MAKE_MD(1, EFI_CONVENTIONAL_MEMORY,	EFI_MEMORY_WB,    16*MB,  896*MB-16*MB);

	/* Set up SAL tables */
	memcpy(vm_firmware->sal_systab.signature, "SST_", 4);
	vm_firmware->sal_systab.size = sizeof(vm_firmware->sal_systab);
	vm_firmware->sal_systab.sal_rev_minor = 1;
	vm_firmware->sal_systab.sal_rev_major = 0;
	vm_firmware->sal_systab.entry_count = 1;
	memcpy(vm_firmware->sal_systab.oem_id, "UNSW", sizeof("UNSW"));
	memcpy(vm_firmware->sal_systab.product_id, "vNUMA", sizeof("vNUMA"));
	vm_firmware->sal_systab.ep.type = SAL_DESC_ENTRY_POINT;
	vm_firmware->sal_systab.ep.pal_proc = (uint64_t)vm_firmware->pal_proc;
	vm_firmware->sal_systab.ep.sal_proc = (uint64_t)vm_firmware->sal_proc;
	vm_firmware->sal_systab.ep.gp = 0;
#if 0
	vm_firmware->sal_systab.pf.type = SAL_DESC_PLATFORM_FEATURE;
	vm_firmware->sal_systab.pf.feature_mask = IA64_SAL_PLATFORM_FEATURE_ITC_DRIFT;
#endif

	checksum = 0;
	p = (unsigned char *)&vm_firmware->sal_systab;
	for (i = 0; i < sizeof(vm_firmware->sal_systab); i++)
		checksum += p[i];

	vm_firmware->sal_systab.checksum = -checksum;

	memcpy(vm_firmware->cmd_line, command_line, strlen(command_line));
	memcpy(vm_firmware->pal_proc, &palproc, &palproc_end - &palproc);
	memcpy(vm_firmware->sal_proc, &salproc, &salproc_end - &salproc);
	memcpy(vm_firmware->efi_proc, &efiproc, &efiproc_end - &efiproc);
	memcpy(vm_firmware->efi_get_time_proc, &efigettimeproc, &efigettimeproc_end - &efigettimeproc);
	memcpy(vm_firmware->efi_reset_proc, &efiresetproc, &efiresetproc_end - &efiresetproc);
	vm_firmware->efi_unimplemented[0] = (uint64_t)vm_firmware->efi_proc;
	vm_firmware->efi_unimplemented[1] = 0;
	vm_firmware->efi_get_time[0] = (uint64_t)vm_firmware->efi_get_time_proc;
	vm_firmware->efi_get_time[1] = 0;
	vm_firmware->efi_reset_system[0] = (uint64_t)vm_firmware->efi_reset_proc;
	vm_firmware->efi_reset_system[1] = 0;
	vm_firmware->fw_vendor[0] = 'U';
	vm_firmware->fw_vendor[1] = 'N';
	vm_firmware->fw_vendor[2] = 'S';
	vm_firmware->fw_vendor[3] = 'W';

	vm_firmware->bp.efi_systab = (uint64_t)&vm_firmware->efi_systab;
	vm_firmware->bp.efi_memmap = (uint64_t)&vm_firmware->efi_memmap;
	vm_firmware->bp.efi_memmap_size = sizeof(vm_firmware->efi_memmap);
	vm_firmware->bp.efi_memdesc_size = sizeof(efi_memory_desc_t);
	vm_firmware->bp.efi_memdesc_version = 1;
	vm_firmware->bp.command_line = (uint64_t)&vm_firmware->cmd_line;
	vm_firmware->bp.console_info.num_cols = 80;
	vm_firmware->bp.console_info.num_rows = 25;
	vm_firmware->bp.console_info.orig_x = 0;
	vm_firmware->bp.console_info.orig_y = 24;
	vm_firmware->bp.fpswa = 0;

	return &vm_firmware->bp;
}

ia64_sal_handler sal;
unsigned long sal_desc[2];
unsigned long pal;
unsigned long acpi_root;
unsigned long hcdp;
efi_reset_system_t *efi_reset_system;
efi_fpswa_t fpswa;

void
init_fw_interfaces (struct ia64_boot_param *bp)
{
	efi_system_table_t *efisystab;
	efi_config_table_t *tables;
	struct ia64_sal_systab *systab = NULL;
	fpswa_interface_t *fpswa_interface;
	unsigned int i;

	efisystab = (efi_system_table_t *)(bp->efi_systab);
	tables = (efi_config_table_t *)(efisystab->tables);
	for (i = 0; i < efisystab->nr_tables; i++)
	{
		if (memcmp(&tables[i].guid, &SAL_SYSTEM_TABLE_GUID, sizeof(efi_guid_t)) == 0)
			systab = (struct ia64_sal_systab *)(tables[i].table);
		else if (memcmp(&tables[i].guid, &ACPI_20_TABLE_GUID, sizeof(efi_guid_t)) == 0)
			acpi_root = tables[i].table;
		else if (memcmp(&tables[i].guid, &HCDP_TABLE_GUID, sizeof(efi_guid_t)) == 0)
			hcdp = tables[i].table;
	}

	if (systab == NULL)
		return;

	if (systab->ep.type != SAL_DESC_ENTRY_POINT)
		return;

	efi_reset_system = (efi_reset_system_t *)(((efi_runtime_services_t *)(efisystab->runtime))->reset_system);
	sal_desc[0] = systab->ep.sal_proc;
	sal_desc[1] = systab->ep.gp;
	sal = (ia64_sal_handler)sal_desc;
	pal = systab->ep.pal_proc;

	fpswa_interface = (fpswa_interface_t *)(bp->fpswa);
	if (fpswa_interface != NULL)
		fpswa = fpswa_interface->fpswa;
}

void
do_reset_system(void)
{
	efi_reset_system(EFI_RESET_WARM, EFI_SUCCESS, 0, NULL);
	printf("Reset failed\n");
	while (1);
}
