#include "vm.h"
#include "config.h"
#include <netdriver.h>
#include <bfrdriver.h>

#if 1
#define ETHERMTU 1408
/* #define ETHERMTU (1500-64) */
#define FRAGMENTS 3
#else
#define ETHERMTU 512
#define FRAGMENTS (PAGE_SIZE/ETHERMTU)
#endif
// -8 because request is larger than it needs to be over the wire
#define DATA_PAYLOAD(request) ((void *)(request+1)-8)

node_t node_id, nodes = 1;

#ifdef DSM
unsigned char ether_table[64][6];
unsigned char my_ethaddr[6];
//unsigned char profile_ethaddr[6] = { 0x00,0x05,0x5D,0x6B,0xC2,0x72 };
//unsigned char profile_ethaddr[6] = { 0x00,0x30,0x6E,0x4A,0x92,0xE1 };
//unsigned char profile_ethaddr[6] = { 0x00,0x30,0x6E,0x4A,0xC2,0x0B };
unsigned char profile_ethaddr[6] = { 0x00,0x30,0x6E,0x4A,0xC2,0x92 };
void *ether_drv;
uint8_t ipc_configure_lock;

unsigned short my_request_id[MAX_NODES], remote_request_id[MAX_NODES];
unsigned long extra_latency;
extern unsigned int dtr_index;

extern mpfn_t fetch_waiting_for_mpfn;
extern uint16_t fetch_wu_version;
extern node_t fetch_rcvd_from;
extern boolean fetch_rcvd_data;

extern uint16_t wu_applied;
extern uint16_t current_generation;
extern uint16_t remote_wu_version[MAX_NODES];
//extern uint8_t updates_from[MAX_NODES];

static void ipc_print_nodes(void);

#if 0
#define IPC_DEBUG logf
static char *messages[] = {
	"IPC_FETCH",
	"IPC_FETCH_REPLY",
	"IPC_INVALIDATE",
	"IPC_INVALIDATE_REPLY",
	"IPC_PURGE",
	"IPC_PURGE_REPLY",
	"IPC_IPI",
	"IPC_STARTUP",
	"IPC_CONNECT_IRQ",
	"IPC_DISK_IO_SIMPLE",
	"IPC_DISK_IO_SGL",
	"IPC_RESET",
	"IPC_STATS_START",
	"IPC_STATS_STOP",
	"IPC_PUTCHAR",
	"IPC_DISCOVER",
	"IPC_DISCOVER_REPLY",
	"IPC_CONFIGURE",
	"IPC_CONFIGURE_REPLY",
	"IPC_PROFILE",
	"IPC_STORE",
	"IPC_LOAD",
	"IPC_LOAD_REPLY",
	"IPC_CMPXCHG",
	"IPC_FETCHADD",
	"IPC_SEMAPHORE_REPLY",
	"IPC_LOCK_ACQUIRE",
	"IPC_LOCK_GRANT",
	"IPC_LOCK_READ",
	"IPC_LOCK_READ_REPLY",
	"IPC_LOCK_UNLOCK",
	"IPC_INVALIDATE_WHOLE"
};
#else
#define IPC_DEBUG(...) /* nothing */
#endif

void startup_secondary(unsigned long ip, unsigned long gp);

static void
recover(int from, struct request *request)
{
	struct request resp;

	switch (request->type)
	{
#if 0
		case IPC_FETCH:
			/* must already have been invalidated, reply lost */
			ipc_send_fetch_reply(from, request->pa, request->is_write, request->access_type, 1, 0);
			return;
#endif
		case IPC_INVALIDATE:
			/* must already have been invalidated, reply lost */
			resp.type = IPC_INVALIDATE_REPLY;
			resp.invalidate.mpfn = request->invalidate.mpfn;
			ipc_send(request->invalidate.reply_to, &resp);
			return;
	}
	printf("recovery failed\n");
}
#define CHECK_REQUEST_ID() \
	if (ethaddr) { \
		if (request->request_id != remote_request_id[from]) { \
			printf("request out of sequence (expect %d got %d)\n", remote_request_id[from], request->request_id); \
			process_debug_char('l'); \
			/* recover(from, request); */ \
			break; \
		} \
		remote_request_id[from] = (request->request_id + 1) & 0xffff; \
	}

void
ipc_handle_async(struct request *request, unsigned char *ethaddr, int delayed)
{
	node_t from = request->from;
	IPC_DEBUG("request %d %d (%s)\n", request->request_id, request->type, messages[request->type]);
	switch (request->type)
	{
		case IPC_FETCH:
			CHECK_REQUEST_ID();
			mm_handle_fetch(request->fetch.reply_to, request->fetch.mpfn, request->fetch.is_write, request->fetch.access_type, request->fetch.reply_to, request, 1);
#ifdef SPECIAL_LOCKS
/* this is here to deal with a corner case in lock_acquire */
#warning FIXME: Fetch shouldn't really need to break out of wait loop
#endif
			break;
		case IPC_FETCH_REPLY:
			if (request->fetch.fragment != 0xff)
			{
#ifdef JUMBO_FRAMES
				mm_handle_fetch_data(request->fetch.mpfn, DATA_PAYLOAD(request));
#else
				mm_handle_fetch_data_partial(request->fetch.mpfn, DATA_PAYLOAD(request), request->fetch.fragment * ETHERMTU, (request->fetch.fragment == (FRAGMENTS-1)) ? (PAGE_SIZE-(FRAGMENTS-1)*ETHERMTU) : ETHERMTU);
				if (request->fetch.fragment != (FRAGMENTS-1))
					break;
#endif
			}
			mm_handle_fetch_reply(request->fetch.mpfn, request->fetch.is_write, request->fetch.copyset, request->fetch.wu_version, from, request->fetch.fragment != 0xff);
			break;
		case IPC_INVALIDATE:
			//assert(node_id != 0);
			CHECK_REQUEST_ID();
			mm_handle_invalidate(request->invalidate.reply_to, request->invalidate.mpfn, request);
			break;
		case IPC_INVALIDATE_REPLY:
			//assert(node_id == 0);
			mm_handle_invalidate_reply(from, request->invalidate.mpfn);
			break;
		case IPC_INVALIDATE_WHOLE:
			mm_handle_invalidate_whole(request->invwhole.mpfn, request->invwhole.npages, from);
			break;
		case IPC_PURGE:
			//CHECK_REQUEST_ID();
			mm_handle_tlb_purge(from, request->purge.rr, request->purge.va, request->purge.itir);
			break;
#if 0
		case IPC_PURGE_REPLY:
			mm_handle_purge_reply();
			break;
#endif
		case IPC_IPI:
			//printf("IPI %x\n", request->ipi.irq);
			deliver_interrupt(request->ipi.irq);
			break;
		case IPC_STARTUP:
			printf("startup: %lx %lx\n", request->startup.ip, request->startup.gp);
			startup_secondary(request->startup.ip, request->startup.gp);
			break;
		case IPC_CONNECT_IRQ:
			ssc_connect_irq(request->connectirq.type, request->connectirq.irq);
			break;
		case IPC_DISCOVER:
			/* avoid badness when starting multiple single-node instances */
			if (node_id == 0)
				break;
			memcpy(ether_table[0], ethaddr, 6);
			ipc_send_reply(0, IPC_DISCOVER_REPLY);
			break;
		case IPC_DISCOVER_REPLY:
			memcpy(ether_table[nodes], ethaddr, 6);
			nodes++;
			break;
		case IPC_CONFIGURE:
			nodes = request->configure.nodes;
			node_id = request->configure.node_id;
			printf("I am node %d of %d\n", node_id, nodes);
			cpu.cr[CR_LID] = node_id;
			memcpy(ether_table, DATA_PAYLOAD(request), 6*nodes);
			ipc_send_reply(from, IPC_CONFIGURE_REPLY);
			ipc_print_nodes();
			wu_applied = node_id;
			break;
		case IPC_CONFIGURE_REPLY:
			ipc_configure_lock = 0;
			break;
#ifdef GUESTIO_DISK
		case IPC_DISK_IO_SIMPLE:
			mm_handle_disk_io((struct scsi_sge *)(&request->diskio), 1, request->diskio.offset, request->diskio.is_write, request->diskio.key_valid, request->diskio.key);
			break;
		case IPC_DISK_IO_SGL:
			mm_handle_disk_io((struct scsi_sge *)DATA_PAYLOAD(request), request->diskio.nsge, request->diskio.offset, request->diskio.is_write, request->diskio.key_valid, request->diskio.key);
			break;
#endif
		case IPC_RESET:
			printf("resetting system\n");
			reset_system();
			break;
		case IPC_STATS_START:
			stats_start();
			break;
		case IPC_STATS_STOP:
			stats_stop();
			break;
		case IPC_STORE:
			//CHECK_REQUEST_ID();
			{
				unsigned int i;
				struct store *stores = (struct store *)DATA_PAYLOAD(request);
				for (i = 0; i < request->store.nstores; i++)
					mm_handle_store(stores[i].mpa, stores[i].value, stores[i].size, from, /* request->store.wu_version */ stores[i].wu_version, 0);
			}
			remote_wu_version[from] = request->store.wu_version;
			//updates_from[from]++;
			wu_applied += 1<<NODE_BITS;
			if ((wu_applied >> 14) != current_generation)
				mm_new_generation();
			break;
#if 0
		case IPC_STORE_SINGLE:
			//logf("%x/%x/%x:%x %lx=%lx\n", fetch_wu_version, request->store1.wu_version, waiting_for_mpfn, !waiting_for_mpfn || ((int16_t)(request->store1.wu_version - fetch_wu_version) >= 0), request->store1.mpa, request->store1.value);
			//if (!waiting_for_mpfn)
				//printf("NOT IN FETCH REPLY!\n");
			if (waiting_for_mpfn)
				assert((request->store1.mpa >> 12) == waiting_for_mpfn);

			if (!waiting_for_mpfn || ((int16_t)((request->store1.arrival_time|NODE_MASK) - (fetch_wu_version|NODE_MASK)) >= 0)) {
				//printf("APPLYING wu_applied=%x fetch_wu_version=%x store_version=%x waiting_for_mpfn=%x!\n", wu_applied, fetch_wu_version, request->store1.wu_version, waiting_for_mpfn);
#warning Is this the right version to use?
#if 1
				if ((request->store1.mpa >> 12) == 0x4001)
					logf("A[%x/%x/%x]", request->store1.arrival_time, fetch_wu_version, waiting_for_mpfn);
#endif
				mm_handle_store(request->store1.mpa, request->store1.value, request->store1.size, from, request->store1.wu_version);
			}
#if 1
			else if ((request->store1.mpa >> 12) == 0x4001) {
				logf("D[%x/%x]^%lx", request->store1.arrival_time, fetch_wu_version, request->store1.value);
			}
#endif
			break;
#endif
		case IPC_STORE_HISTORY:
			mm_handle_store_history(request->storehistory.mpfn, request->storehistory.count, DATA_PAYLOAD(request));
			break;
#ifdef SPECIAL_CMPXCHG_FETCHADD
		case IPC_CMPXCHG:
			{
			struct request reply;
			reply.type = IPC_SEMAPHORE_REPLY;
			reply.semreply.oldvalue = mm_handle_cmpxchg(request->cmpxchg.mpa, request->cmpxchg.value, request->cmpxchg.cmpvalue, request->cmpxchg.size, &reply.semreply.newvalue);
			ipc_send(from, &reply);
			}
			break;
		case IPC_FETCHADD:
			{
			struct request reply;
			reply.type = IPC_SEMAPHORE_REPLY;
			reply.semreply.oldvalue = mm_handle_fetchadd(request->fetchadd.mpa, request->fetchadd.size, request->fetchadd.inc, &reply.semreply.newvalue);
			ipc_send(from, &reply);
			}
			break;
		case IPC_SEMAPHORE_REPLY:
			mm_handle_semaphore_reply(request->semreply.oldvalue, request->semreply.newvalue);
			break;
#endif
		case IPC_LOAD:
			mm_handle_load(from, request->load.mpa, request->load.size, request->load.reply_to, request);
			break;
		case IPC_LOAD_REPLY:
			mm_handle_load_reply(request->loadreply.value);
			break;
#ifdef SPECIAL_LOCKS
		case IPC_LOCK_ACQUIRE:
			lock_handle_acquire(from, request->lock.mpa);
			break;
		case IPC_LOCK_GRANT:
			lock_handle_grant(request->lock.mpa);
			break;
		case IPC_LOCK_READ:
			{
			struct request reply;
			reply.type = IPC_LOCK_READ_REPLY;
			reply.lockreply.status = lock_handle_read(request->lock.mpa);
			ipc_send(from, &reply);
			break;
			}
		case IPC_LOCK_READ_REPLY:
			lock_handle_read_reply(request->lockreply.status);
			break;
		case IPC_LOCK_UNLOCK:
			lock_handle_unlock(request->lock.mpa);
			break;
#endif
		case IPC_PROFILE:
			break; /* ignore */
		case IPC_PROBE:
			printf("IPC_PROBE from %d\n", from);
			break;
		default:
			printf("Unexpected message %d\n", request->type);
	}
}

struct bfr_drv {
	struct driver *classp;
};

static struct driver bfr_drv =
{
	.name = "ipcbuffer",
	.class = buffer_manager
};

static struct bfr_drv *bfr_init(dma_handle_t dma)
{
	struct bfr_drv *buf_mgr;
	buf_mgr = malloc(sizeof(struct bfr_drv));
	buf_mgr->classp = &bfr_drv;
	return buf_mgr;
}

static int register_if(struct bfr_drv *buf_mgr, void *etherif,
	uint32_t num_recv_buffers, uint32_t recv_buffer_size, uint32_t num_tx_buffers)
{
	void *buf;

	ether_drv = etherif;
	netif_macaddr(ether_drv, my_ethaddr);
	while (num_recv_buffers-- > 0)
	{
		buf = malloc(RX_BUF_SIZE);
		netif_freepacket(ether_drv, (dma_addr_t)(uintptr_t)buf, RX_BUF_SIZE, (uintptr_t)buf);
	}
	return 0;
}

#define MAX_TX_HDRS 256
struct pkt tx_hdrs[MAX_TX_HDRS];
unsigned int tx_hdr_lo, tx_hdr_hi;

static inline struct pkt *alloc_tx_hdr()
{
	unsigned int tx_hdr_this = tx_hdr_hi;
	unsigned int tx_hdr_next = (tx_hdr_hi+1) & (MAX_TX_HDRS-1);
	if (unlikely(tx_hdr_next == tx_hdr_lo))
	{
#ifndef TIGON3
		extern void *ether_drv;
		dp83820_reap_tx_descs(ether_drv);
		if (tx_hdr_next == tx_hdr_lo)
		{
			printf("pending: %d\n", dp83820_get_pending(ether_drv));
			return NULL;
		}
#else
		return NULL;
#endif
	}
	tx_hdr_hi = tx_hdr_next;
	return &tx_hdrs[tx_hdr_this];
}

static inline void free_tx_hdr(struct pkt *p)
{
	assert(p == &tx_hdrs[tx_hdr_lo]);
	tx_hdr_lo = (tx_hdr_lo+1) & (MAX_TX_HDRS-1);
}

struct packet {
	dma_addr_t buffer_addr;
	size_t size;
	uintptr_t internal_key;
};

// 128
#define MAX_PACKETS 256
struct packet guest_packet_queue[MAX_PACKETS];
unsigned int guest_packet_lo, guest_packet_hi;
extern int guest_network_irq;

static int
queue_packet(dma_addr_t buffer_addr, size_t size, uintptr_t internal_key)
{
	struct packet *packet;
	unsigned int next_hi = (guest_packet_hi + 1) % MAX_PACKETS;

	if (next_hi == guest_packet_lo)
	{
		/* if this is hit, try increasing MAX_PACKETS */
		printf("*");
		return 0;
	}

	packet = &guest_packet_queue[guest_packet_hi];
	packet->buffer_addr = buffer_addr;
	packet->size = size;
	packet->internal_key = internal_key;
	guest_packet_hi = next_hi;
	return 1;
}

unsigned int
dequeue_packet(unsigned long buffer_mpa, unsigned int length)
{
	struct packet *packet;

	if (guest_packet_hi == guest_packet_lo)
		return 0;

	packet = &guest_packet_queue[guest_packet_lo];
	if (packet->size < length)
		length = packet->size;
	mm_copy_to_user(buffer_mpa, (void *)(packet->buffer_addr), length);
	netif_freepacket(ether_drv, packet->buffer_addr, RX_BUF_SIZE, packet->internal_key);
	guest_packet_lo = (guest_packet_lo + 1) % MAX_PACKETS;
	return length;
}

static void
feedpacket(struct bfr_drv *buf_mgr, dma_addr_t buffer_addr, size_t size, uintptr_t internal_key)
{
	struct pkt *pkt = (struct pkt *)buffer_addr;

	if (pkt->proto == ETHER_PROTO_VMM)
	{
		STATS_INC(vmm_packets_in);
		ipc_handle_async(&pkt->request, pkt->srcaddr, 0);
	}
	else if (guest_network_irq && (node_id == 0))
	{
		STATS_INC(guest_packets_in);
		if (queue_packet(buffer_addr, size, internal_key))
		{
			deliver_interrupt(guest_network_irq);
			return;
		}
#if 0
		//printf("guest packet!\n");
		if (!guest_packet_size)
		{
			if (size > sizeof(guest_packet))
				size = sizeof(guest_packet);
			memcpy(guest_packet, pkt, size);
			guest_packet_size = size;
			deliver_interrupt(guest_network_irq);
		}
		else
		{
			printf("X");
		}
#endif
	}
	else
	{
		//printf("unhandled packet, proto = %x\n", pkt->proto);
	}
	netif_freepacket(ether_drv, buffer_addr, RX_BUF_SIZE, internal_key);
}

static int
freepacket(struct bfr_drv *buf_mgr,
		dma_addr_t addr,
		uintptr_t key)
{
	if (key == DATA_AREA_HEADER)
		free_tx_hdr((struct pkt *)addr);
	else if (key == DATA_AREA_MALLOC)
		free((void *)addr);
	else if (key != DATA_AREA_STATIC)
		mm_release_page(key, addr);
	return 0;
}

static struct buffer_ops buffer_ops = {
	.bfr_init = bfr_init,
	.register_if = register_if,
	.freepacket = freepacket,
	.feedpacket = feedpacket
};

struct driver *
ipc_init(void)
{
	printf("sizeof(struct request) = %x\n", sizeof(struct request));

	memcpy(ether_table[62], profile_ethaddr, 6);
	memset(ether_table[63], 0xff, 6);

	bfr_drv.bfr_ops = &buffer_ops;
	return &bfr_drv;
}

void
ipc_send(node_t node, struct request *request)
{
	struct pkt *data;

	STATS_INC(short_packets_out);
	IPC_DEBUG("send request %d %d (%s)\n", request->request_id, request->type, messages[request->type]);
	request->from = node_id;
	data = alloc_tx_hdr();
	if (unlikely(data == NULL))
	{
		printf("ipc_send: alloc_tx_hdr failed\n");
		return;
	}
#ifdef BROADCAST_ALL
	memset(data->dstaddr, 0xff, 6);
#else
	memcpy(data->dstaddr, ether_table[node], 6);
#endif
	memcpy(data->srcaddr, my_ethaddr, 6);
	data->proto = ETHER_PROTO_VMM;
	data->request = *request;

	// -8 because request is larger than it needs to be over the wire
	if (netif_feedpacket(ether_drv, (dma_addr_t)(uintptr_t)data, sizeof(*data)-8, 0, PACKET_NONE, DATA_AREA_HEADER) == -1)
		printf("netif_feedpacket failed\n");
}

void
ipc_send_page(node_t node, struct request *request, unsigned long pa, unsigned int length, uintptr_t data_area)
{
	struct pkt *data;

	if (length == PAGE_SIZE)
		STATS_INC(page_packets_out);
	else
		STATS_INC(long_packets_out);

	IPC_DEBUG("send request %d %d (%s)\n", request->request_id, request->type, messages[request->type]);
	request->from = node_id;
	data = alloc_tx_hdr();
	if (unlikely(data == NULL))
	{
		printf("ipc_send: alloc_tx_hdr failed\n");
		return;
	}
#ifdef BROADCAST_ALL
	memset(data->dstaddr, 0xff, 6);
#else
	memcpy(data->dstaddr, ether_table[node], 6);
#endif
	memcpy(data->srcaddr, my_ethaddr, 6);
	data->proto = ETHER_PROTO_VMM;
	data->request = *request;

	// -8 because request is larger than it needs to be over the wire
	if (netif_feedpacket(ether_drv, (dma_addr_t)(uintptr_t)data, sizeof(*data)-8, 1, PACKET_NONE, DATA_AREA_HEADER) == -1)
		printf("netif_feedpacket failed\n");

	if (netif_feedpacket(ether_drv, (dma_addr_t)pa, length, 0, PACKET_NONE, data_area) == -1)
		printf("netif_feedpacket failed\n");
}

void
ipc_send_raw(unsigned long buffer_mpa, unsigned int length)
{
	/* make sure to use our local srcaddr so we don't confuse the switch */
	char *newbuffer;
	struct pkt *data;

	STATS_INC(guest_packets_out);
	newbuffer = malloc(length);
	mm_copy_from_user(newbuffer, buffer_mpa, length);
	data = (struct pkt *)newbuffer;
	memcpy(data->srcaddr, my_ethaddr, 6);

	if (netif_feedpacket(ether_drv, (dma_addr_t)newbuffer, length, 0, PACKET_NONE, DATA_AREA_MALLOC) == -1)
		printf("netif_feedpacket failed\n");
}

void
ipc_wait(node_t node, struct request *request, uint8_t *lock)
{
	unsigned int irq, count = 0;

	assert(*lock);
	while (1)
	{
                if ((++count & 0xfffffff) == 0)
		{
			printf("wait timed out!\n");
			printf("node=%d id=%x type=%x %lx %x\n", node, request->request_id, request->type, request->fetch.mpfn, request->fetch.is_write);
		}

#if 0
		if (dp83820_rx_poll(ether_drv) && !(*lock))
			break;
#endif

		irq = read_ivr();
		if (irq == 0xf)
			continue;

		do_irq(irq);
		if (!(*lock))
			break;
	}
#if 0 /* LATENCY_TEST */
	{
		unsigned long wait_until = end + extra_latency;
		while ((end = read_itc()) < wait_until)
			;
	}
#endif
}

void
ipc_send_reply(node_t node, unsigned char type)
{
	struct request request;

	request.type = type;
	ipc_send(node, &request);
}

void
ipc_send_fetch_reply(node_t node, mpfn_t mpfn, unsigned long pa, boolean is_write, boolean send_data, unsigned int wu_version, copyset_t copyset)
{
	struct request request;
	unsigned int offset, fragment, thislength;

	request.type = IPC_FETCH_REPLY;
	request.fetch.mpfn = mpfn;
	request.fetch.is_write = is_write;
	request.fetch.copyset = copyset;
	request.fetch.wu_version = wu_version;
	if (send_data)
	{
#ifdef JUMBO_FRAMES
		request.fetch.fragment = 0;
		ipc_send_page(node, &request, pa, PAGE_SIZE, DATA_AREA_GUEST(mpfn));
#else
		for (offset = 0, fragment = 0; offset < PAGE_SIZE; offset += ETHERMTU, fragment++)
		{
			request.fetch.fragment = fragment;
			thislength = ((PAGE_SIZE - offset) >= ETHERMTU) ? ETHERMTU : (PAGE_SIZE - offset);
			ipc_send_page(node, &request, pa+offset, thislength, ((offset + thislength) == PAGE_SIZE) ? DATA_AREA_GUEST(mpfn) : DATA_AREA_STATIC);
		}
#endif
	}
	else
	{
		request.fetch.fragment = 0xff;
		ipc_send(node, &request);
	}
}

void
ipc_send_sequenced(node_t node, struct request *request)
{
	request->request_id = my_request_id[node];
	my_request_id[node] = (my_request_id[node] + 1) & 0xffff;
	ipc_send(node, request);
}

unsigned int
ipc_broadcast(unsigned long mask, struct request *request)
{
	node_t node;
	unsigned int sent = 0;

	for (node = 0; node < nodes; node++)
	{
		if ((node != node_id) && (mask & (1 << node)))
		{
			ipc_send(node, request);
			sent++;
		}
	}

	return sent;
}

unsigned int
ipc_broadcast_sequenced(unsigned long mask, struct request *request)
{
	node_t node;
	unsigned int sent = 0;

	for (node = 0; node < nodes; node++)
	{
		if ((node != node_id) && (mask & (1 << node)))
		{
			request->request_id = my_request_id[node];
			my_request_id[node] = (my_request_id[node] + 1) & 0xffff;
			ipc_send(node, request);
			sent++;
		}
	}

	return sent;
}

static void
ipc_print_nodes(void)
{
	unsigned char *ethaddr;
	unsigned int i;

	printf("ipc: %d node(s)\n", nodes);
	for (i = 0; i < nodes; i++)
	{
		ethaddr = ether_table[i];
		printf("ipc: node %d: %02x:%02x:%02x:%02x:%02x:%02x\n", i,
				ethaddr[0], ethaddr[1], ethaddr[2], ethaddr[3], ethaddr[4], ethaddr[5]);
	}
}

void
ipc_discover(void)
{
	struct request request;
	unsigned long time_start, time_end;
	node_t node;
	unsigned int i;

	if (!ether_drv)
	{
		printf("ipc: no known network cards detected\n");
		return;
	}

	printf("ipc: discovering nodes...\n");
	time_start = read_itc();
	for (i = 0; i < 5; i++) {
		netif_interrupt(ether_drv);
		do {
		} while (read_itc() - time_start < 10000000000UL);
	}

	memcpy(ether_table[0], my_ethaddr, 6);
	ipc_send_reply(63, IPC_DISCOVER);

	time_start = read_itc();
	for (i = 0; i < 5; i++) {
		netif_interrupt(ether_drv);
		do {
		} while (read_itc() - time_start < 10000000000UL);
	}

	ipc_print_nodes();

	for (node = 1; node < nodes; node++)
	{
		ipc_configure_lock = 1;
		request.type = IPC_CONFIGURE;
		request.configure.node_id = node;
		request.configure.nodes = nodes;
		ipc_send_page(node, &request, (unsigned long)ether_table, 6*nodes, DATA_AREA_STATIC);
		time_start = read_itc();
		ipc_wait(-1, &request, &ipc_configure_lock);
		time_end = read_itc();
		printf("Node %d: ~%lx cycles RTT\n", node, time_end-time_start);
	}

	printf("ipc: %d node(s) configured\n", nodes);
}

void
ipc_reset_system(void)
{
	printf("resetting system\n");
	if (ether_drv && (nodes > 1))
	{
		//memset(ether_table[0], 0xff, 6);
		ipc_send_reply(63, IPC_RESET);
	}
	reset_system();
}
#else
struct driver *
ipc_init(void)
{
	return NULL;
}

void
ipc_reset_system(void)
{
	printf("resetting system\n");
}

void
ipc_handle_async(struct request *request, unsigned char *ethaddr)
{
	assert(0);
}

void
ipc_send(node_t node, struct request *request)
{
	assert(0);
}

int
ipc_wait(node_t node, struct request *request)
{
	assert(0);
	return 0;
}

void
ipc_send_reply(node_t node, unsigned char type)
{
	assert(0);
}

void
ipc_send_fetch_reply(node_t node, mpfn_t mpfn, unsigned long pa, boolean is_write, boolean send_data, copyset_t copyset)
{
	assert(0);
}

void
ipc_send_sequenced(node_t node, struct request *request)
{
	assert(0);
}

unsigned int
ipc_broadcast(unsigned long mask, struct request *request)
{
	assert(0);
	return 0;
}

#endif
