#include "vm.h"
#include "config.h"
#include <elf.h>
#include <heap_alloc.h>
#include "bootparam.h"
#include "pal.h"

void *heap;
char heap_mem[8192*1024] __attribute__((aligned(16)));
extern unsigned int node_id;
extern struct cpu cpu;
extern unsigned long pal;
extern char *build_time;

struct boot_info {
	unsigned long ip;
	struct ia64_boot_param *guest_param;
};

struct boot_info
load_kernel(struct ia64_boot_param *param)
{
	void *initrd_start = (void *)(param->initrd_start);
	Elf64_Ehdr *elf_ehdr;
	Elf64_Phdr *elf_phdrs, *elf_phdr;
	struct boot_info boot_info;
	unsigned int i;
	extern void io_wait;
	struct pal_ret ret;

	assert(((unsigned long)heap_mem & 15) == 0);
	heap = heap_alloc_init(heap_mem, heap_mem+sizeof(heap_mem));
	init_fw_interfaces(param);

	early_serial_setup();

	printf("*** vNUMA: built at %s\n", build_time);
	printf("*** vNUMA: command line: %s\n", (void *)(param->command_line));

	printf("*** vNUMA: enabling McKinley Return Stack Buffer\n");
	ret = call_pal(PAL_PROC_GET_FEATURES, 0, 16, 0);
	printf("Before: %lx %lx %lx %lx\n", ret.status, ret.v0, ret.v1, ret.v2);
	if (ret.status == 0)
		call_pal(PAL_PROC_SET_FEATURES, ret.v1|0x80, 16, 0);
	ret = call_pal(PAL_PROC_GET_FEATURES, 0, 16, 0);
	printf("After : %lx %lx %lx %lx\n", ret.status, ret.v0, ret.v1, ret.v2);

	printf("*** vNUMA: bringing up devices\n");
	device_init();

	printf("*** vNUMA: initialising paging\n");
	node_id = (initrd_start != NULL) ? 0 : 1;
	mm_init();
	vhpt_init();

	printf("*** vNUMA: initialising CPU\n");
	dtr_insert(0x1000, 0x1000, PAGE_SHIFT, PTE_AR_RW|PTE_PL_3);
#ifdef AFTERBURN
	printf("*** vNUMA: initialising gate page\n");
	itr_insert(0x100000, 0x100000, PAGE_SHIFT, PTE_AR_GATE);
	write_register("r4", 0xfff);
#endif
	memset(&cpu, 0, sizeof(cpu));
	cpu.psr = PSR_BN;
	cpu.cfle = 1;
	cpu.cr[CR_ITV] = 1UL << 16; /* timer interrupt masked */
	if (initrd_start == NULL)
	{
		cpu.cr[CR_LID] = 1;
		node_id = 1;
	}
	write_dcr(DCR_PP|DCR_DP|DCR_DK|DCR_DX|DCR_DR|DCR_DA|DCR_DD|DCR_LC|DCR_PP);

	printf("*** vNUMA: setting up PMU to monitor releases\n");
	monitor_rel_start();

#ifdef DO_PAL_HALT
	printf("*** vNUMA: setting up PAL mapping (%lx)\n", pal);
	dtr_insert(pal, pal, 26, PTE_AR_RWX);
	itr_insert(pal, pal, 26, PTE_AR_RWX);
#endif

	if (initrd_start != NULL)
	{
		printf("*** vNUMA: locating nodes\n");
#ifdef DSM
		ipc_discover();
#endif

		printf("*** vNUMA: relocating kernel\n");
		elf_ehdr = initrd_start;
		elf_phdrs = initrd_start + elf_ehdr->e_phoff;
		for (i = 0; i < elf_ehdr->e_phnum; i++)
		{
			elf_phdr = &elf_phdrs[i];
			if (elf_phdr->p_type != PT_LOAD)
				continue;

			memcpy((void *)(elf_phdr->p_paddr), initrd_start + elf_phdr->p_offset, elf_phdr->p_filesz);
			memset((void *)(elf_phdr->p_paddr) + elf_phdr->p_filesz, 0, elf_phdr->p_memsz - elf_phdr->p_filesz);
		}

		printf("*** vNUMA: initialising guest firmware\n");
		boot_info.ip = elf_ehdr->e_entry;
		boot_info.guest_param = init_guest_fw((char *)param->command_line);
		printf("*** vNUMA: starting kernel\n");
	}
	else
	{
		printf("*** vNUMA: waiting for master\n");
		boot_info.ip = (unsigned long)&io_wait;
		boot_info.guest_param = NULL;
	}
	device_enter_virtual();
	return boot_info;
}

void
out_of_memory(void)
{
	printf("OUT OF MEMORY!!\n");
	mm_dump_write_log_stats();
	debug();
}

