#include "vm.h"
#include "config.h"
//#include "profile_tlb.h"

#if 0
#define DEBUG_DSM(mpfn, args...) if (is_mpfn_monitored(mpfn)) logf(args)
#else
#define DEBUG_DSM(...) /* nothing */
#endif

#if 0
#define DEBUG_LOCK(mpfn, args...) if (is_mpfn_monitored(mpfn)) logf(args)
#else
#define DEBUG_LOCK(...) /* nothing */
#endif

extern unsigned int dtr_index;
extern node_t node_id, nodes;
#ifdef THRASH_PROTECTION
mpfn_t last_exec_fault_mpfn, last_data_fault_mpfn;
#if THRASH_PROTECTION == 2
//#define THRASH_PROTECTION_NO_PROGRESS() ({unsigned long pmd5=read_pmd(5); unsigned long ip = read_iip() | BITS(read_ipsr(), PSR_RI_BITS, 2); boolean noprogress=!pmd5; if (noprogress) logf("!%lx", ip); else logf("=%lx[%lx]", ip, pmd5); noprogress; })
#define THRASH_PROTECTION_NO_PROGRESS() (!read_pmd(5))
#else
unsigned long last_fault_ip;
#define THRASH_PROTECTION_NO_PROGRESS() ((read_iip() | BITS(read_ipsr(), PSR_RI_BITS, 2)) == last_fault_ip)
#endif
#endif

mpfn_t fetch_waiting_for_mpfn;
access_type_t fetch_access_type;
uint16_t fetch_wu_version;
node_t fetch_rcvd_from;
boolean fetch_rcvd_data;

unsigned int write_update_threshold;
uint16_t wu_applied;
uint16_t current_generation;
uint16_t remote_wu_version[MAX_NODES];
//uint8_t updates_from[MAX_NODES];

#ifndef NDEBUG
#define MONITOR
#endif
#ifdef MONITOR
//#define MONITOR_ADDRESS 0x45d9580
//#define MONITOR_ADDRESS 0x47140c0
//#define MONITOR_ADDRESS 0x45d1680
//#define MONITOR_ADDRESS 0x4001000
#define MONITOR_ADDRESS 0x11d0000
#if 1
#define is_mpa_monitored(mpa) 1
#define is_mpfn_monitored(mpfn) 1
#elif 0
#define is_mpa_monitored(mpa) is_mpfn_monitored(mpa_to_mpfn(mpa)) /* ((mpa >= 0x4015000) && (mpa < 0x4016000)) */
#define is_mpfn_monitored(mpfn) ((mpfn == 0x4001) || (mpfn == 0x4013) || (mpfn == 0x4015))
#else
#define is_mpa_monitored(mpa) ((mpa) == MONITOR_ADDRESS)
#define is_mpfn_monitored(mpfn) ((mpfn) == mpa_to_mpfn(MONITOR_ADDRESS))
#endif
#endif

#if 0
#define PROFILE_TLB(address, type) profile_tlb(read_iip(), address, cpu.rr[address>>61], type)
#else
#define PROFILE_TLB(address, type)
#endif

#define MAX_MEM_SIZE	(1024*1024*1024)
#define MAX_MEM_SHIFT	30
#ifdef DISTRIBUTED_MANAGERS
#ifdef STRIPED_MANAGERS
#define mpfn_to_manager(mpfn)	((mpfn) & ((1<<DM_NODE_BITS)-1))
#else 
#define mpfn_to_manager(mpfn)	((mpfn) >> (MAX_MEM_SHIFT-PAGE_SHIFT-DM_NODE_BITS))
#endif
#else
#define mpfn_to_manager(mpfn)	0
#endif

#ifdef LONG_VHPT
// *********
// long VHPT
// *********
// should be CPU local

struct vhpt_entry
{
	uint64_t pte;
	uint64_t ps;
	uint64_t tag;
	uint64_t ig;
};

#define vhpt_lookup(va)		(struct vhpt_entry *)ia64_thash(va)
#define vhpt_tag(va)		ia64_ttag(va)
#define vhpt_entry_valid(e,va)	((e)->tag == vhpt_tag(va))

static void
vhpt_insert(unsigned long va, unsigned long pte, mpfn_t mpfn, boolean is_exec)
{
	struct vhpt_entry *entry = vhpt_lookup(va);

#if 0
	extern void _vhpt;
	profile_tlb(entry->tag, va, entry-(struct vhpt_entry *)&_vhpt, PROFILE_TLB_INSERT);
#endif

	DEBUG_MM("vhpt_insert %lx %lx %lx\n", va, (unsigned long)entry, read_rr(va));
	entry->pte = pte;
	entry->ps = PAGE_SHIFT << 2;
	entry->tag = vhpt_tag(va);
	entry->ig = mpfn;

#ifdef PRE_INSERT
	if (is_exec)
		ia64_itc_i(va, PAGE_SHIFT << 2, pte);
	else
		ia64_itc_d(va, PAGE_SHIFT << 2, pte);
#else
#warning pre-insert disabled
#endif
}

static void
vhpt_purge(unsigned long va)
{
	struct vhpt_entry *entry = vhpt_lookup(va);

#if 0
	extern void _vhpt;

	unsigned long iip = read_iip();
	if (iip == 0xa00000010000b320UL) /* burn_tlbop */
		iip = read_register("r4");
	profile_tlb(iip, va, entry-(struct vhpt_entry *)&_vhpt, PROFILE_TLB_PURGE);
#endif

	if (entry->tag == vhpt_tag(va))
	{
		entry->tag = 1UL << 63;
		entry->pte = 0;
	}
}

static void
vhpt_chmod(unsigned long va, mpfn_t mpfn, unsigned int prot)
{
	struct vhpt_entry *entry = vhpt_lookup(va);
	unsigned int da;

	/* check mpfn instead of tag, in case we have stale entries in
	 * the mapping DB */
	/* tpa = BITS(entry->pte, 12, 38) << 12; */
	if (entry->ig == mpfn)
	{
		da = (entry->pte >> 53) & 3;
		prot &= da;
		entry->pte = (entry->pte & ~(PTE_A|PTE_D)) | (prot << PTE_A_BIT);
		//DEBUG_MM("vhpt_chmod %lx %lx\n", prev, entry->pte);
	}
}

#ifdef ENABLE_REMAP
static void
vhpt_remap(unsigned long va, mpfn_t mpfn, pfn_t pfn)
{
	struct vhpt_entry *entry = vhpt_lookup(va);

	/* check mpfn instead of tag, in case we have stale entries in
	 * the mapping DB */
	/* tpa = BITS(entry->pte, 12, 38) << 12; */
	if (entry->ig == mpfn)
	{
		entry->pte = (entry->pte & ~0x3fffffffff000UL) | ((uint64_t)pfn << PAGE_SHIFT);
		//DEBUG_MM("vhpt_remap %lx %lx\n", prev, entry->pte);
	}
}
#endif
#else
/* SHORT VHPT */
#define VHPT_TWO_WAY

struct vhpt_entry
{
	uint64_t pte;
};

#define vhpt_lookup(va)		(struct vhpt_entry *)ia64_thash(va)
#define vhpt_entry_valid(e,va)	(tpa((unsigned long)e) != 0)

#define PGD_HASH_TABLE_SIZE (16*1024*1024/PAGE_SIZE) /* 16MB VHPT */
#define PGD_HASH_TABLE_BITS (24-PAGE_SHIFT)
#define rid_mangle(rr)	(((rr >> 16) & 0xf) | (((rr >> 11) & 0x1f) << 4) | ((rr >> 8) << 9))

struct pgd_info
{
	int32_t pgdi;
	uint32_t rr;
};

struct pgd_line
{
	struct pgd_info pgd[2];
};

struct pgd_line pgd_hash_table[PGD_HASH_TABLE_SIZE];
uint8_t next_way[PGD_HASH_TABLE_SIZE];
#ifndef NDEBUG
unsigned long last_replace_time[PGD_HASH_TABLE_SIZE*2];
#endif

/* FIXME */
#define pte_index(va)	(((va) >> PAGE_SHIFT) & ( (1UL<<(PAGE_SHIFT-3)) - 1))
#define pgd_index(va)	(((va) << 3) >> (PAGE_SHIFT+PAGE_SHIFT))
#if 1
#define pgd_hash_index(pgdi, rr)  (((pgdi) ^ ((rr)>>20) ^ ((uint32_t)(pgdi)>>31)) & (PGD_HASH_TABLE_SIZE-1))
#else
#define pgd_hash_index(pgdi, rr)  (((pgdi) ^ ((rr)>>20) ^ ((uint32_t)(pgdi)>>31<<11)) & (PGD_HASH_TABLE_SIZE-1))
#endif

static void
flush_va_novhpt(unsigned long va, unsigned long rr)
{
	unsigned long prev_rr = read_rr(1UL << 61);
	unsigned long shadow_va = va | (1UL << 61);

	asm volatile("mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(rr));
	ia64_ptc_l(shadow_va, PAGE_SHIFT<<2);
	asm volatile(";; srlz.d ;; mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(prev_rr));
}

#if 0 /* now in assembly code */
void
mm_vhpt_miss(unsigned long va, unsigned long isr)
{
	int32_t pgdi = pgd_index(va); /* sufficient to cover 53 bits */
	uint32_t rr = read_rr(va);
	unsigned int hashi = pgd_hash_index(pgdi, rr);
	struct pgd_line *hash_line = &pgd_hash_table[hashi];
	extern void _vhpt;
	void *pte_page = &_vhpt + (hashi<<(PAGE_SHIFT+1));
	unsigned long pte;
	struct vhpt_entry *entry = vhpt_lookup(va);

	if ((hash_line->pgd[0].pgdi == pgdi) && (hash_line->pgd[0].rr == rr))
		;
	else if ((hash_line->pgd[1].pgdi == pgdi) && (hash_line->pgd[1].rr == rr))
		pte_page += PAGE_SIZE;
	else
	{
		mm_tlb_miss(va, isr);
		return;
	}

	pte = ((unsigned long)pte_page)|PTE_AR_RW|PTE_PL_3|PTE_D|PTE_A|PTE_P;
	//printf("mapping %p @ %p\n", pte_page, entry);
	ia64_itc_d(entry, PAGE_SHIFT << 2, pte);
	//ia64_srlz_d();
}
#endif

static void
vhpt_insert(unsigned long va, unsigned long pte, mpfn_t mpfn, boolean is_exec)
{
	struct vhpt_entry *entry = vhpt_lookup(va);
	//printf("vhpt_insert %lx (%p)\n", va, entry);
	if (unlikely(!vhpt_entry_valid(entry,va)))
	{
		int32_t pgdi = pgd_index(va); /* sufficient to cover 53 bits */
		uint32_t rr = read_rr(va);
		unsigned int hashi = pgd_hash_index(pgdi, rr);
		struct pgd_line *hash_line = &pgd_hash_table[hashi];
		extern void _vhpt;
		void *pte_page = &_vhpt + (hashi<<(PAGE_SHIFT+1));
		unsigned long pte;

#ifdef VHPT_TWO_WAY
		if ((hash_line->pgd[0].pgdi == pgdi) && (hash_line->pgd[0].rr == rr))
			;
		else if ((hash_line->pgd[1].pgdi == pgdi) && (hash_line->pgd[1].rr == rr))
			pte_page += PAGE_SIZE;
		else
		{
			uint8_t way = next_way[hashi];
			struct pgd_info *hash_entry = &hash_line->pgd[way];
#ifndef NDEBUG
			unsigned long last_time = last_replace_time[hashi*2+way];
			unsigned long now = read_itc();
			if (unlikely(now - last_time < 90000000)) /* 0.1s */
			{
				logf("%lx %lx %x\n", va, rr, hashi);
				STATS_INC(upper_vhpt_conflicts);
			}
			last_replace_time[hashi*2+way] = now;
#endif
			pte_page += way<<PAGE_SHIFT;
			next_way[hashi] ^= 1;
#else
		if ((hash_entry->pgdi != pgdi) || (hash_entry->rr != rr))
		{
#endif
			//logf("%lx %lx %lx %lx\n", hash_entry->pgdi, pgdi, hash_entry->rr, rr);
			if (hash_entry->rr)
			{
				unsigned long old_hash_address = (((unsigned long)hash_entry->pgdi << PAGE_SHIFT) & ((1UL<<52) - 1)) | (1UL << 60);
				flush_va_novhpt(old_hash_address, hash_entry->rr);
				//printf("old=%lx new=%p oldrr=%x newrr=%x hashi=%x\n", old_hash_address, entry, hash_entry->rr, rr, hashi);
			}

			/* XX use faster clear_page?? */
			clear_page(pte_page);
			hash_entry->pgdi = pgdi;
			hash_entry->rr = rr;
			STATS_INC(upper_vhpt_clearpage);
		}

		pte = ((unsigned long)pte_page)|PTE_AR_RW|PTE_PL_3|PTE_D|PTE_A|PTE_P;
		//printf("mapping %p @ %p\n", pte_page, entry);
		ia64_itc_d(entry, PAGE_SHIFT << 2, pte);
		ia64_srlz_d();
	}

	//printf("pte = %lx\n", pte);
	entry->pte = pte;
#ifdef PRE_INSERT
	if (is_exec)
		ia64_itc_i(va, PAGE_SHIFT << 2, pte);
	else
		ia64_itc_d(va, PAGE_SHIFT << 2, pte);
#else
#warning pre-insert disabled
	// FIXME
	ia64_ptc_l(va, PAGE_SHIFT<<2);
#endif
}

static void
vhpt_purge(unsigned long va)
{
	struct vhpt_entry *entry = vhpt_lookup(va);
	if (unlikely(!vhpt_entry_valid(entry,va)))
	{
		int32_t pgdi = pgd_index(va); /* sufficient to cover 53 bits */
		uint32_t rr = read_rr(va);
		unsigned int hashi = pgd_hash_index(pgdi, rr);
		struct pgd_line *hash_line = &pgd_hash_table[hashi];
		extern void _vhpt;
		void *pte_page = &_vhpt + (hashi<<(PAGE_SHIFT+1));

#ifdef VHPT_TWO_WAY
		if ((hash_line->pgd[0].pgdi == pgdi) && (hash_line->pgd[0].rr == rr))
			;
		else if ((hash_line->pgd[1].pgdi == pgdi) && (hash_line->pgd[1].rr == rr))
			pte_page += PAGE_SIZE;
		else
#else
		if ((hash_entry->pgdi != pgdi) || (hash_entry->rr != rr))
#endif
			return;

		entry = (struct vhpt_entry *)((unsigned long *)pte_page + pte_index(va));
	}

	entry->pte = 0;
}

static void
vhpt_chmod(unsigned long va, mpfn_t mpfn, unsigned int prot)
{
	struct vhpt_entry *entry = vhpt_lookup(va);
	unsigned int da;

	if (!vhpt_entry_valid(entry,va))
		return;

	da = (entry->pte >> 53) & 3;
	prot &= da;
	entry->pte = (entry->pte & ~(PTE_A|PTE_D)) | (prot << PTE_A_BIT);
}
#endif

static void
flush_va(unsigned long va, unsigned long rr)
{
	if (rr == cpu.rr[va >> 61])
	{
		vhpt_purge(va);
		ia64_ptc_l(va, PAGE_SHIFT<<2);
	}
	else
	{
		unsigned long shadow_va = ((va << 3) >> 3) | (1UL << 61);

		asm volatile("mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((rr & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
		vhpt_purge(shadow_va);
		ia64_ptc_l(shadow_va, PAGE_SHIFT<<2);
		asm volatile(";; srlz.d ;; mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((cpu.rr[1] & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
	}
}

static void
chmod_va(unsigned long va, unsigned long rr, mpfn_t mpfn, unsigned int prot)
{
	if (rr == cpu.rr[va >> 61])
	{
		vhpt_chmod(va, mpfn, prot);
		ia64_ptc_l(va, PAGE_SHIFT<<2);
	}
	else
	{
		unsigned long shadow_va = ((va << 3) >> 3) | (1UL << 61);

		asm volatile("mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((rr & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
		vhpt_chmod(shadow_va, mpfn, prot);
		ia64_ptc_l(shadow_va, PAGE_SHIFT<<2);
		asm volatile(";; srlz.d ;; mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((cpu.rr[1] & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
	}
}

#ifdef ENABLE_REMAP
static void
remap_va(unsigned long va, unsigned long rr, mpfn_t mpfn, pfn_t pfn)
{
	if (rr == cpu.rr[va >> 61])
	{
		vhpt_remap(va, mpfn, pfn);
		ia64_ptc_l(va, PAGE_SHIFT<<2);
	}
	else
	{
		unsigned long shadow_va = ((va << 3) >> 3) | (1UL << 61);

		asm volatile("mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((rr & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
		vhpt_remap(shadow_va, mpfn, pfn);
		ia64_ptc_l(shadow_va, PAGE_SHIFT<<2);
		asm volatile(";; srlz.d ;; mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((cpu.rr[1] & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));
	}
}
#endif

/* L2 PT */
// should be CPU local

struct mapping {
	uint64_t va;
	uint64_t mask;
	uint64_t pte;
	struct mapping *next;
};

struct mapping *mappings[8];

#if 0
static inline struct mapping *
mapping_lookup(uint64_t va)
{
	struct mapping *mapping = mappings[va>>61];
	if (likely(!mapping))
		return NULL;

	do {
		if ((va & mapping->mask) == mapping->va)
			return mapping;
		mapping = mapping->next;
	} while (mapping);

	return NULL;
}
#endif

static void
mapping_insert(uint64_t va, uint64_t mask, uint64_t pte)
{
	struct mapping *mapping = mappings[va>>61];

	if ((va >> 61) == 6)
	{
		printf("mapping_insert %lx %lx %lx\n", va, mask, pte);
		pte |= PTE_VNUMA_PIB;
	}

	for (; mapping != NULL; mapping = mapping->next)
	{
		unsigned long comask = mapping->mask & mask;
		if ((va & comask) == (mapping->va & comask))
		{
			//DEBUG_MM("[%lx] mapping_update %lx %lx %lx\n", read_iip(), va, mask, pte);
			mapping->va = va;
			mapping->mask = mask;
			mapping->pte = pte;
			return;
		}
	}

	//DEBUG_MM("[%lx] mapping_insert %lx %lx %lx\n", read_iip(), va, mask, pte);
	mapping = malloc(sizeof(struct mapping));
	mapping->va = va;
	mapping->mask = mask;
	mapping->pte = pte;
	mapping->next = mappings[va>>61];
	mappings[va>>61] = mapping;
}


// ****************
// MA -> ATTR, {VA}
// ****************
// BEWARE: this is shared between CPUs

struct page {
/*0*/	pfn_t pfn;
/*4*/	uint8_t prot;
/*5*/	uint8_t lock;
/*6*/	uint8_t mode;
/*7*/	uint8_t owner;
/*8*/	uint8_t wcount;
/*9*/	uint8_t unused;
/*10*/	uint8_t sends_pending_this;
/*11*/	uint8_t sends_pending_prev;
/*12*/	uint16_t wu_version;
/*14*/  copyset_t copyset;
/*16*/	struct request_entry *wait_queue;
#ifdef COHERENT_WRITE_UPDATES
/*24*/	struct write_log *write_log;
#endif
};
struct page pages[MAX_MEM_SIZE>>PAGE_SHIFT];

//#define lookup_page(mpa)	({if ((mpa)>=MAX_MEM_SIZE) { printf("[%lx] access to VM space %lx\n", read_iip(), mpa); } &pages[(mpa)>>PAGE_SHIFT];})
#define mpfn_to_page(mpfn)	({assert((mpfn) < (MAX_MEM_SIZE>>PAGE_SHIFT)); &pages[mpfn];})
#define mpfn_to_mpa(mpfn)	((mpfn) << PAGE_SHIFT)
#define mpa_to_mpfn(mpa)	((mpa) >> PAGE_SHIFT)
#define page_to_pa(page)	((uint64_t)(page->pfn) << PAGE_SHIFT)

#ifdef SPECIAL_LOCKS
static void check_locks(struct page *page, char *context);
static void apply_locks(struct page *page, mpfn_t mpfn);
static void drop_locks(struct page *page, mpfn_t mpfn, node_t new_owner);
#endif

static inline void *
map_window(struct page *page, uint64_t mpa)
{
	uint64_t window_address = (1UL << 61) - PAGE_SIZE;
	ia64_itr_d(dtr_index, window_address, PAGE_SHIFT << 2, page_to_pa(page)|PTE_AR_RW|PTE_D|PTE_A|PTE_P);
	return (void *)(window_address + (mpa&PAGE_MASK));
}

static inline void
unmap_window(void)
{
	uint64_t window_address = (1UL << 61) - PAGE_SIZE;
	ia64_ptr_d(window_address, PAGE_SHIFT << 2);
}


#ifdef DSM
// **********
// MA -> {VA}
// **********
// should be CPU local
struct vmap
{
	uint64_t va;
	uint32_t rr;
	uint32_t link_idx;
};
static struct vmap m2v[MAX_MEM_SIZE>>PAGE_SHIFT];
#define M2V_EXT_SIZE (MAX_MEM_SIZE>>PAGE_SHIFT)
static struct vmap m2v_ext[M2V_EXT_SIZE];
/* this value is chosen so that 0 falls outside the next_idx-M2V_EXT_SIZE..next_idx-1 window */
static uint32_t next_idx = M2V_EXT_SIZE+1;

#define mpfn_to_vmap(mpfn)	({assert((mpfn) < (MAX_MEM_SIZE>>PAGE_SHIFT)); &m2v[mpfn];})

static void
register_virtual(mpfn_t mpfn, unsigned long va)
{
	struct vmap *vmap = mpfn_to_vmap(mpfn);
	struct vmap *vmap_ext;
	uint64_t first_va = vmap->va;
	uint32_t first_rr = vmap->rr;
	uint32_t first_link_idx = vmap->link_idx;
	uint32_t rr = (uint32_t)cpu.rr[va >> 61];

	if ((first_va == va) && (first_rr == rr))
		return;

	vmap->va = va;
	vmap->rr = rr;
	if (first_rr == 0)
		return;
	vmap->link_idx = next_idx;

	vmap_ext = &m2v_ext[next_idx % M2V_EXT_SIZE];
	if (vmap_ext->rr != 0)
		flush_va(vmap_ext->va, vmap_ext->rr);
	vmap_ext->va = first_va;
	vmap_ext->rr = first_rr;
	vmap_ext->link_idx = first_link_idx;
	next_idx++;
#ifndef NDEBUG
	if (next_idx == 0)
		printf("next_idx wraparound\n");
#endif
}

static void
chmod_virtual(mpfn_t mpfn, unsigned int prot)
{
	//printf("chmod_virtual %lx %x\n", pa, prot);
	struct vmap *vmap = mpfn_to_vmap(mpfn);
	while (1)
	{
		/* maybe flush is better? */
		if (vmap->rr != 0)
			chmod_va(vmap->va, vmap->rr, mpfn, prot);
		/* CHECKME */
		if ((vmap->link_idx + M2V_EXT_SIZE) < next_idx)
			return;
		vmap = &m2v_ext[vmap->link_idx % M2V_EXT_SIZE];
	}
}

#ifdef ENABLE_REMAP
static void
remap_virtual(mpfn_t mpfn, pfn_t pfn)
{
	//printf("remap_virtual %lx %x\n", pa, pfn);
	struct vmap *vmap = mpfn_to_vmap(mpfn);
	while (1)
	{
		/* maybe flush is better? */
		if (vmap->rr != 0)
			remap_va(vmap->va, vmap->rr, mpfn, pfn);
		/* CHECKME */
		if ((vmap->link_idx + M2V_EXT_SIZE) < next_idx)
			return;
		vmap = &m2v_ext[vmap->link_idx % M2V_EXT_SIZE];
	}
}

pfn_t next_free_frame;

// Frame allocator
static void *
frame_alloc_and_map(pfn_t *ret_pfn)
{
	pfn_t pfn = next_free_frame;
	uint64_t pa = (uint64_t)pfn << PAGE_SHIFT;
	uint64_t window_address = (1UL << 61) - PAGE_SIZE;

	if (pfn == (pfn_t)-1)
	{
		printf("frame_alloc_and_map: out of memory!\n");
		while(1);
	}

	ia64_itr_d(dtr_index, window_address, PAGE_SHIFT << 2, pa|PTE_AR_RW|PTE_D|PTE_A|PTE_P);
	next_free_frame = *(pfn_t *)window_address;
	*ret_pfn = pfn;
	return (void *)window_address;
}

static void
frame_free(pfn_t pfn)
{
	uint64_t window_address = (1UL << 61) - PAGE_SIZE;
	uint64_t pa = (uint64_t)pfn << PAGE_SHIFT;

	ia64_itr_d(dtr_index, window_address, PAGE_SHIFT << 2, pa|PTE_AR_RW|PTE_D|PTE_A|PTE_P);
	*(pfn_t *)window_address = next_free_frame;
	ia64_ptr_d(window_address, PAGE_SHIFT << 2);
	next_free_frame = pfn;
}
#endif

// **************
// Request queues
// **************

struct request_entry {
	struct request_entry *next;
	struct request request;
} *progress_wait_queue;

static void
enqueue_request(struct request_entry **queue, struct request *request)
{
	struct request_entry *entry = malloc(sizeof(struct request_entry));

	entry->request = *request;
	entry->next = *queue;
	*queue = entry;
}

static void
process_request_queue_reverse(struct request_entry *queue)
{
	if (queue == NULL)
		return;
	process_request_queue_reverse(queue->next);
	ipc_handle_async(&queue->request, NULL, 1);
	free(queue);
}

int processing_queue; /* debug */
static void
process_request_queue(struct request_entry **queue)
{
	/* XXX what order should we process? */
	struct request_entry *old_queue = *queue;
	*queue = NULL;
	processing_queue = 1;
	process_request_queue_reverse(old_queue);
	processing_queue = 0;
}

#ifdef THRASH_PROTECTION
void
handle_single_step(void)
{
	struct page *page;

	write_ipsr(read_ipsr() & ~PSR_SS);
#if THRASH_PROTECTION == 1
	last_fault_ip = 0;
#else
	assert(last_data_fault_mpfn != 0);
#endif
	if (last_data_fault_mpfn != 0)
	{
		page = mpfn_to_page(last_data_fault_mpfn);
#ifdef DEBUG_DELAY
		if (page->wait_queue)
			printf("R:%x ", last_data_fault_mpfn);
#endif
		last_data_fault_mpfn = 0;
		process_request_queue(&page->wait_queue);
	}
}

static inline void
thrash_protection_start(mpfn_t mpfn, access_type_t access_type)
{
#if THRASH_PROTECTION == 2
	write_pmd(5, 0);
#else
	last_fault_ip = read_iip() | BITS(read_ipsr(), PSR_RI_BITS, 2);
#endif
	if (access_type == ACCESS_DATA)
	{
		if (last_data_fault_mpfn != mpfn)
		{
			if (last_data_fault_mpfn != 0)
			{
				/* can happen due to mandatory RSE accesses */
				/* not anymore? */
				struct page *page = mpfn_to_page(last_data_fault_mpfn);
				assert(!page->wait_queue);
				if (page->wait_queue)
					printf("lmpfn=%x mpfn=%x\n", last_data_fault_mpfn, mpfn);
				last_data_fault_mpfn = 0;
				process_request_queue(&page->wait_queue);
			}

			last_data_fault_mpfn = mpfn;
		}
	}
#if THRASH_PROTECTION == 1
	write_ipsr(read_ipsr() | PSR_SS);
#endif
}
#else
#define thrash_protection_start(mpfn, access_type)
#endif

#if 0
static void
dsm_dump_page_info(mpfn_t mpfn)
{
	struct page *page = mpfn_to_page(mpfn);

	printf("mpfn=%lx prot=%x lock=%x owner=%x copyset=%lx\n", mpfn, page->prot, page->lock, page->owner, page->copyset);
#if 0
	owner = mpfn_to_manager(pa);
	if (owner == node_id)
		printf(" owner=%x copyset=%lx", page->owner, page->copyset);
	printf("\n");
#endif
}

static void
dump_vmap(mpfn_t mpfn)
{
	struct vmap *vmap = mpfn_to_vmap(mpfn);
	while (1)
	{
		if (vmap->rr != 0)
			printf("  va = %lx, rr = %lx\n", vmap->va, vmap->rr);
		/* CHECKME */
		if ((vmap->link_idx + M2V_EXT_SIZE) < next_idx)
			break;
		vmap = &m2v_ext[vmap->link_idx % M2V_EXT_SIZE];
	}
}

static int
vhpt_consistency_check(unsigned long pa, char *label)
{
	struct vhpt_entry *mapping;
	struct page *page;
	unsigned long va, rid;
	unsigned int da, realda, i;
	mpfn_t mpfn;
	extern void _vhpt;
	int ok = 1;

	for (i = 0; i < ((1UL << 20) / sizeof(struct vhpt_entry)); i++)
	{
		mapping = &((struct vhpt_entry *)(&_vhpt))[i];
		mpfn = mapping->ig;

		page = mpfn_to_page(mpfn);
		if (BITS(mapping->pte, 12, 38) != page->pfn)
		{
			printf("%s: mpfn %lx pfn %lx tpfn %lx\n", mpfn, page->pfn, BITS(mapping->pte, 12, 38));
			ok = 0;
		}

		da = (mapping->pte >> PTE_A_BIT) & 3;
		realda = (mapping->pte >> 53) & 3;
		if (da != (realda & page->prot))
		{
			rid = ((mapping->tag >> 49) << 10) | ((i ^ mapping->tag) & 0x3ff);
			va = (mapping->tag ^ (rid << 39)) << PAGE_SHIFT;
			printf("%s: mpfn %lx %x&%x=%x va=%lx rid=%lx\n", label, mpfn, page->prot, realda, da, va, rid);
			dump_vmap(mpfn);
			ok = 0;
		}
	}
	return ok;
}

static void
dsm_consistency_check(mpfn_t mpfn, char *label)
{
#if 0
	struct page *page;

	if (mpfn_to_manager(mpa) != node_id)
		return;

	page = mpfn_to_page(mpfn);

	// if RWX, must be owner and only copy holder
	if (page->prot == PROT_RWX)
	{
		if ((page->owner != node_id) || (page->copyset != (1UL << node_id)))
			goto bad;
	}

	// if at least RX, must be in copyset; else must not be in copyset
	if (page->prot & PROT_RX)
	{
		if ((page->copyset & (1UL << node_id)) == 0)
			goto bad;
	}
	else
	{
		if ((page->copyset & (1UL << node_id)) != 0)
			goto bad;
	}

	// if owner, must have at least read; if not owner, must have no access
	if (page->owner == node_id)
	{
		if (page->prot == PROT_NONE)
			goto bad;
	}
	else
	{
		if (page->prot != PROT_NONE)
			goto bad;
	}

	return;
 bad:
	printf("%s: ", label);
	dsm_dump_page_info(pa);
#endif
}
#endif

static void mm_drop_store_history(struct page *page);
static void mm_send_store_history(struct page *page, mpfn_t mpfn, int node);

void
mm_handle_fetch(node_t node, mpfn_t mpfn, boolean is_write, access_type_t access_type, node_t for_node, struct request *orig_req, boolean synchronous)
{
	struct request request;
	struct page *page;
	node_t manager;

#if 0
	if (is_mpfn_monitored(mpfn))
		logf("rfetch %lx %s\n", mpfn, is_write ? "write" : "read");
#endif
#ifdef WRITE_UPDATES
#ifdef WRITE_UPDATES_USE_RC
	emu_fwb_if_hit(mpfn);
#else
	emu_fwb();
#endif
#endif

	page = mpfn_to_page(mpfn);
#ifdef WRITE_UPDATE_THRESHOLD 
	page->wcount = 0;
#endif

#ifdef THRASH_PROTECTION
	assert(synchronous);
	if (synchronous && (/*(last_exec_fault_mpfn == mpfn) ||*/ (last_data_fault_mpfn == mpfn)) && THRASH_PROTECTION_NO_PROGRESS())
	{
		DEBUG_DSM(mpfn, "%d:p%lx ", for_node, mpfn);
		//assert (page->wait_queue == NULL); /* otherwise we will end up servicing requests out of order */
		//enqueue_request(&progress_wait_queue, orig_req);
		enqueue_request(&page->wait_queue, orig_req);
#if THRASH_PROTECTION == 2
		write_ipsr(read_ipsr() | PSR_SS);
#endif
		return;
	}
#endif

	if (synchronous && page->lock)
	{
		DEBUG_LOCK(mpfn, "%d:l%lx(%d) ", for_node, mpfn, page->lock);
		enqueue_request(&page->wait_queue, orig_req);
		return;
	}

	manager = mpfn_to_manager(mpfn);
	//assert(manager == 0);

	if (page->owner == node_id)
	{
		DEBUG_DSM(mpfn, "%d:%c%lx ", for_node, is_write ? 'w' : 'r', mpfn);

#if defined(PAGE_VERSIONING_COPYSET)
		if (!(page->copyset & (1 << for_node)))
#endif
			page->sends_pending_this++;

#if 0
		if (is_mpfn_monitored(mpfn)
#if defined(PAGE_VERSIONING_COPYSET)
				&& !(page->copyset & (1 << for_node))
#endif
		   )
		{
			uint8_t *p = map_window(page, 0);
			logf("^%x", *(uint64_t *)(p + (MONITOR_ADDRESS&PAGE_MASK)));
			unmap_window();
		}
#endif

#ifdef COHERENT_WRITE_UPDATES
		if (!is_write && page->write_log)
			mm_send_store_history(page, mpfn, for_node);
#endif

		ipc_send_fetch_reply(for_node, mpfn, page_to_pa(page), is_write,
#if defined(PAGE_VERSIONING_COPYSET)
				     !(page->copyset & (1 << for_node))
#else
				     1
#endif
				     , page->wu_version, page->copyset);
		mm_chmod(mpfn, is_write ? PROT_NONE : PROT_RX);
		if (is_write
#ifdef MANAGER_IS_PREFERRED_OWNER
			|| (for_node == manager)
#endif
			)
		{
			page->owner = for_node;
#ifdef SPECIAL_LOCKS
			if (page->locks)
				drop_locks(page, mpfn, for_node);
#endif
		}
		else
			page->copyset |= (1 << for_node);

#ifdef PARALLEL_INVALIDATES
		if (is_write)
		{
			copyset_t copyset = page->copyset &~ (1UL << for_node);
			if (copyset != 0)
			{
				request.type = IPC_INVALIDATE;
				request.invalidate.mpfn = mpfn;
				request.invalidate.reply_to = for_node;
				ipc_broadcast_sequenced(copyset, &request);
			}
		}
#endif
		return;
	}

	assert(manager == node_id);
	if (manager != node_id)
	{
		assert(processing_queue);
		printf("%x: fetch %lx for node %x (is_write=%x access_type=%x)\n", node_id, mpfn, for_node, is_write, access_type);
		//dsm_dump_page_info(mpfn);
	}

	DEBUG_DSM(mpfn, "%d:f%lx(%d) ", for_node, mpfn, page->owner);

	/* forward request to owner */
	request.type = IPC_FETCH;
	request.fetch.reply_to = for_node;
	request.fetch.mpfn = mpfn;
	request.fetch.is_write = is_write;
	request.fetch.access_type = access_type;
	ipc_send_sequenced(page->owner, &request);

	if (is_write)
	{
		page->owner = for_node;
#ifdef SPECIAL_LOCKS
		if (page->locks)
			drop_locks(page, mpfn, for_node);
#endif
	}

	//dsm_consistency_check(mpfn, "mm_handle_fetch");
}

void
mm_fetch(mpfn_t mpfn, boolean is_write, access_type_t access_type, unsigned long va, boolean wait)
{
	struct request request;
	struct page *page;
	node_t manager, forward_to;
	copyset_t copyset;
	unsigned long start_time;
	unsigned int pending_tx;

#if 0
	logf("[%lx] fetch %x %s %s\n", read_iip(), mpfn,
		is_write ? "write" : "read",
		(access_type == ACCESS_DATA) ? "data" : (access_type == ACCESS_EXECUTE) ? "exec" : "rse");
#endif

	page = mpfn_to_page(mpfn);

#ifndef NDEBUG
	{
		assert((page->prot & PROT_RWX) != PROT_RWX);
		if (!is_write)
			assert((page->prot & PROT_RWX) == PROT_NONE);
	}
#endif

	if (page->lock)
	{
		//printf("hmm, page already value, what to do now?\n");
		return;
	}

	page->lock = 1;

	if (page->owner == node_id)
	{
		copyset = page->copyset;

		assert(is_write);
		assert(copyset);
		DEBUG_DSM(mpfn, "w%lx ", mpfn);
		/* send out invalidates */
		request.type = IPC_INVALIDATE;
		request.invalidate.mpfn = mpfn;
		request.invalidate.reply_to = node_id;
		ipc_broadcast_sequenced(copyset, &request);

		fetch_access_type = access_type;

		start_time = read_itc();
		pending_tx = get_pending_tx();
		ipc_wait(-1, &request, &page->lock);
		if (debug_flags & DEBUG_FLAG_PROFILE)
			profile_invalidate(start_time, pending_tx, mpfn, copyset);
		return;
	}

	/* forward to owner (or manager if we are not the manager) */
	manager = mpfn_to_manager(mpfn);
	//assert(manager == 0);
	forward_to = (manager == node_id) ? page->owner : manager;

	DEBUG_DSM(mpfn, "f%lx(%d) ", mpfn, forward_to);

	request.type = IPC_FETCH;
	request.fetch.reply_to = node_id;
	request.fetch.mpfn = mpfn;
	request.fetch.is_write = is_write;
	request.fetch.access_type = access_type;
	ipc_send_sequenced(forward_to, &request);
	if (wait) {
		fetch_waiting_for_mpfn = mpfn;
		fetch_access_type = access_type;
		fetch_wu_version = page->wu_version;
		//memset(updates_from, 0, nodes);

		start_time = read_itc();
		pending_tx = get_pending_tx();
		ipc_wait(forward_to, &request, &page->lock);
		if (debug_flags & DEBUG_FLAG_PROFILE)
			profile_fetch(start_time, pending_tx, mpfn, is_write, access_type, forward_to, fetch_rcvd_from, fetch_rcvd_data);
	}

	//dsm_consistency_check(mpfn, "mm_fetch");
}

void
mm_handle_fetch_data(mpfn_t mpfn, void *data)
{
	struct page *page = mpfn_to_page(mpfn);
	unsigned long mpa = mpfn_to_mpa(mpfn);
	void *p;

	p = map_window(page, mpa);
	//memcpy(p, data, PAGE_SIZE);

	copy_page(p, data);
#if 0
	if (is_mpfn_monitored(mpfn))
		logf("*%x", *(uint64_t *)(p + (MONITOR_ADDRESS&PAGE_MASK)));
#endif
	unmap_window();
}

void
mm_handle_fetch_data_partial(mpfn_t mpfn, void *data, uint32_t offset, uint32_t length)
{
	struct page *page = mpfn_to_page(mpfn);
	unsigned long mpa = mpfn_to_mpa(mpfn) + offset;
	void *p;

	p = map_window(page, mpa);
	//memcpy(p, data, length);
	copy_area(p, data, length);
#if 0
	if (is_mpfn_monitored(mpfn) && !offset)
		logf("*%x", *(uint64_t *)(p + (MONITOR_ADDRESS&PAGE_MASK)));
#endif
	unmap_window();
}

void
mm_handle_fetch_reply(mpfn_t mpfn, boolean is_write, copyset_t copyset, uint16_t wu_version, node_t from, boolean rcvd_data)
{
	struct request request;
	struct page *page;
	node_t manager;

	DEBUG_DSM(mpfn, "%c%lx ", is_write ? 'W' : 'R', mpfn);

#ifdef READ_CRITICAL_WORD_FIRST
	if (mpfn == fetch_waiting_for_mpfn)
#else
	assert (mpfn == fetch_waiting_for_mpfn);
#endif
	{
		mm_pretend_chmod(mpfn, is_write ? PROT_RWX : PROT_RX);
		fetch_waiting_for_mpfn = 0;
		fetch_rcvd_from = from;
		fetch_rcvd_data = rcvd_data;
		if (rcvd_data)
			process_store_queue_partial(from, wu_version);
		else
			process_store_queue();
		thrash_protection_start(mpfn, fetch_access_type);
	}

	page = mpfn_to_page(mpfn);
	manager = mpfn_to_manager(mpfn);
	//assert(manager == 0);

	if (is_write)
	{
		/* remove ourselves from copyset */
		copyset &= ~(1UL << node_id);

		page->copyset = copyset;
		if (page->owner != node_id)
		{
			page->owner = node_id;
#ifdef SPECIAL_LOCKS
			if (page->locks)
				apply_locks(page, mpfn);
#endif
		}

		if (copyset != 0) /* waiting for invalidations */
#ifdef PARALLEL_INVALIDATES
			return;
#else
		{
			DEBUG_DSM(mpfn, "I%lx(%lx) ", mpfn, copyset);
			request.type = IPC_INVALIDATE;
			request.invalidate.mpfn = mpfn;
			request.invalidate.reply_to = node_id;
			ipc_broadcast_sequenced(copyset, &request);
			return;
		}
#endif
	}
#ifdef MANAGER_IS_PREFERRED_OWNER
	else if (node_id == manager)
	{
		assert(page->owner != node_id);
		assert((copyset & (1UL << node_id)) == 0);

		/* we make ourselves the owner */
		page->copyset = copyset | (1UL << page->owner);
		page->owner = node_id;
#ifdef SPECIAL_LOCKS
		if (page->locks)
			apply_locks(page, mpfn);
#endif
	}
#endif

	mm_chmod(mpfn, is_write ? PROT_RWX : PROT_RX);
	page->lock = 0;
	process_request_queue(&page->wait_queue);

	//dsm_consistency_check(mpfn, "mm_handle_fetch_reply");
}

void
mm_handle_invalidate(node_t for_node, mpfn_t mpfn, struct request *orig_req)
{
	struct page *page;
	struct request request;

	//DEBUG_DSM(mpfn, "mm_handle_invalidate: %lx %s\n", mpfn, is_write ? "write" : "read");

	page = mpfn_to_page(mpfn);
#ifdef WRITE_UPDATE_THRESHOLD
	page->wcount = 0;
#endif

#ifdef THRASH_PROTECTION
	if ((/*(last_exec_fault_mpfn == mpfn) ||*/ (last_data_fault_mpfn == mpfn)) && THRASH_PROTECTION_NO_PROGRESS())
	{
		DEBUG_DSM(mpfn, "Y%lx(%d) ", mpfn, for_node);
		enqueue_request(&page->wait_queue, orig_req);
#if THRASH_PROTECTION == 2
		write_ipsr(read_ipsr() | PSR_SS);
#endif
		return;
	}
#endif

	DEBUG_DSM(mpfn, "X%lx ", mpfn);

	request.type = IPC_INVALIDATE_REPLY;
	request.invalidate.mpfn = mpfn;
	ipc_send(for_node, &request);

	assert(page->prot & PROT_RX);
	mm_chmod(mpfn, PROT_NONE);
}

void
mm_handle_invalidate_reply(node_t node, mpfn_t mpfn)
{
	struct page *page;

	page = mpfn_to_page(mpfn);
	page->copyset &= ~(1UL << node);
	if (page->copyset == 0)
	{
#warning does not handle asynchronous invalidates
		thrash_protection_start(mpfn, fetch_access_type);

		mm_chmod(mpfn, PROT_RWX);
		page->lock = 0;
		process_request_queue(&page->wait_queue);
	}

	//dsm_consistency_check(mpfn, "mm_handle_invalidate_reply");
}

void
mm_handle_invalidate_whole(mpfn_t mpfn, unsigned int npages, node_t node)
{
	struct page *page = mpfn_to_page(mpfn);

	//logf("invwhole %x %x %lx\n", mpfn, npages, node);
	for (; npages--; mpfn++, page++)
	{
		page->owner = node;
		page->copyset = 0;

		if (page->prot != PROT_NONE)
			mm_chmod(mpfn, PROT_NONE);
	}
}

void
mm_invalidate_whole(mpfn_t mpfn, unsigned int npages)
{
	struct page *page = mpfn_to_page(mpfn);
	struct request request;
	void *p;

	//logf("invwhole %x %x\n", mpfn, npages);

	emu_fwb_if_hit(mpfn);

	request.type = IPC_INVALIDATE_WHOLE;
	request.invwhole.mpfn = mpfn;
	request.invwhole.npages = npages;
	ipc_send(63, &request);

	for (; npages--; mpfn++, page++)
	{
		page->owner = node_id;
		page->copyset = 0;

		if (page->prot != PROT_RWX)
			mm_chmod(mpfn, PROT_RWX);

		p = map_window(page, 0);
		clear_page(p);
		unmap_window();
	}
}

void
mm_pretend_chmod(mpfn_t mpfn, unsigned int prot)
{
	struct page *page = mpfn_to_page(mpfn);
	page->prot = prot;
}

void
mm_chmod(mpfn_t mpfn, unsigned int prot)
{
	struct page *page = mpfn_to_page(mpfn);
	unsigned long mpa = mpfn_to_mpa(mpfn);

	asm volatile("mov ar.csd=%0" :: "r"(mpfn));

	//DEBUG_DSM(mpfn, "mm_chmod %lx %x %x\n", mpfn, page->prot, prot);
	chmod_virtual(mpfn, prot);
	vhpt_chmod(mpa, mpfn, prot);
	ia64_ptc_l(mpa, PAGE_SHIFT<<2);
	page->prot = prot;

#ifdef COHERENT_WRITE_UPDATES
	if (prot != PROT_RX)
		mm_drop_store_history(page);
#endif
}

#ifdef ENABLE_REMAP
void
mm_remap(mpfn_t mpfn, pfn_t pfn)
{
	struct page *page = mpfn_to_page(mpfn);
	unsigned long mpa = mpfn_to_mpa(mpfn);

	//DEBUG_DSM(mpfn, "mm_remap %lx %x %x\n", mpfn, page->pfn, pfn);
	remap_virtual(mpfn, pfn);
	vhpt_remap(mpa, mpfn, pfn);
	ia64_ptc_l(mpa, PAGE_SHIFT<<2);
	page->pfn = pfn;
}
#endif

void
mm_handle_tlb_purge(node_t node, unsigned long rr, unsigned long orig_va, unsigned long sz)
{
	unsigned long shadow_va = ((orig_va << 3) >> 3) | (1UL << 61);
	unsigned long size = 1UL << BITS(sz, 2, 6);
	unsigned long va, end;

	asm volatile("mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr(rr)));

	for (va = shadow_va, end = shadow_va + size; va < end; va += PAGE_SIZE)
		vhpt_purge(va);

#if 0
	if (global)
		ia64_ptc_ga(shadow_va, sz);
	else
#endif
		ia64_ptc_l(shadow_va, sz);


	asm volatile(";; srlz.d ;; mov rr[%0] = %1 ;; srlz.d ;;" : : "r"(shadow_va), "r"(translate_rr((cpu.rr[1] & 0xffffff00UL) | (PAGE_SHIFT << 2) | 1)));

#if 0
	ipc_send_reply(node, IPC_PURGE_REPLY);
#endif
}

#if 0
static uint8_t purge_replies_outstanding;

void
mm_handle_purge_reply(void)
{
	assert(purge_replies_outstanding > 0);
	purge_replies_outstanding--;
}
#endif
#endif

void
tlb_insert(unsigned long ifa, unsigned long itir, unsigned long pte, boolean pinned, boolean is_exec)
{
	unsigned long sz = itir & 0xfc;
	unsigned long size = 1UL << (sz >> 2);
	unsigned long mask = ~(size - 1);
	unsigned long va = ifa & mask;
	mpfn_t mpfn = BITS(pte, 12, 38) & (mask >> PAGE_SHIFT);
	unsigned long da = BITS(pte, PTE_A_BIT, 2);
	unsigned long end, realpte, realda;
	struct page *page;

	//DEBUG_MM("[%lx] Mapping from %lx to %lx, size %lx\n", read_iip(), va, mpfn_to_mpa(mpfn), size);

#if 0
	if (mpfn > (MAX_MEM_SIZE>>PAGE_SHIFT))
	{
		printf("tlb_insert ifa=%lx itir=%lx pte=%lx pinned=%x is_exec=%x\n", ifa, itir, pte, pinned, is_exec);
		debug();
	}

	if ((va >> 61) == 0)
	{
		printf("Mapping into region 0 not allowed (%lx)\n", va);
		return;
	}
#endif

	if (unlikely(pinned || (size > 65536)))
	{
#ifndef LINUX
		tlb_purge(va, sz, 0);		/* flush TLB and VHPT */
#endif
		//pte |= (pinned << PTE_VNUMA_PINNED_BIT);
		mapping_insert(va, mask, pte);
	}
	else
	{
		pte &= PTE_P|PTE_AR|PTE_ED;	/* preserve P, AR, ED fields from guest */
		pte |= PTE_PL_3 | (da << PTE_VNUMA_A_BIT);	/* PL 3, original D|A in ignored field */
		for (end = va + size; va < end; va += PAGE_SIZE, mpfn++)
		{
			page = mpfn_to_page(mpfn);
			realda = (page->prot & da) << PTE_A_BIT;
			realpte = pte | realda | page_to_pa(page);
			vhpt_insert(va, realpte, mpfn, is_exec);
#ifdef DSM
			register_virtual(mpfn, va);
#endif
		}

#ifdef LONG_VHPT
#ifndef PRE_INSERT
		ia64_ptc_l(va, sz);		/* only flush TLB - VHPT entries overwritten above */
#endif
#endif
	}
}

void
tlb_purge(unsigned long address, unsigned long sz, boolean global)
{
	unsigned long size = 1UL << BITS(sz, 2, 6);
	unsigned long mask = ~(size - 1);
	unsigned long va = address & mask;
	unsigned long end;

	//DEBUG_MM("tlb_purge %lx %lx\n", address, sz);

	if ((va >> 61) == 0)
	{
		printf("Purging region 0 not allowed (%lx)\n", va);
		return;
	}

#ifdef DSM
	if (global)
	{
		/* FIXME for more than one other node */
		struct request request;
		request.type = IPC_PURGE;
		request.purge.rr = cpu.rr[address >> 61];
		request.purge.va = va;
		request.purge.itir = sz;
		/* purge_replies_outstanding = */ ipc_broadcast((unsigned long)-1, &request);
#if 0
		/* It's my belief that we don't need to wait on a purge broadcast - matthewc */
		ipc_wait(-1, &request, &purge_replies_outstanding);
#endif
	}
#endif

	for (end = va + size; va < end; va += PAGE_SIZE)
		vhpt_purge(va);

#if 0
	if (global)
		ia64_ptc_ga(address, sz);
	else
#endif
		ia64_ptc_l(address, sz);
}

void
tlb_purge_all(void)
{
#ifdef LONG_VHPT
	extern void _vhpt;
	unsigned long offset;
	for (offset = 0; offset < (1UL << 24); offset += PAGE_SIZE)
		clear_page(&_vhpt + offset);
	/* memset(&_vhpt, 0, 1UL << 24); */
#else
	memset(pgd_hash_table, 0, sizeof(pgd_hash_table));
#endif
	ia64_ptc_e(0);
}

extern unsigned long thash(unsigned long address);

#if 0
static void
sleep_usec(void)
{
	unsigned long start = read_itc();
	unsigned long end = start + 900;

	while (read_itc() < end)
		;
}
#endif

/* McKinley note: when the RSB is enabled, the VHPT walker sometimes
   aborts on the first ITLB miss after a context switch in schedule(),
   leading us here.  DTLB misses don't ever seem to occur. */
void
handle_itlb_miss(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("itlb %lx %lx\n", read_iip(), read_ifa());
	DEBUG_SHORT('i');
	STATS_INC(itlb_miss);
	{
	STATS_IF(unsigned long start_time = read_itc());
	mm_tlb_miss(address, isr);
	STATS_ADD(tlb_miss_time, read_itc()-start_time);
	}
}

#ifdef LONG_VHPT
void
handle_dtlb_miss(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("dtlb %lx %lx\n", read_iip(), read_ifa());
	DEBUG_SHORT('d');
	STATS_INC(dtlb_miss);
	{
	STATS_IF(unsigned long start_time = read_itc());
	mm_tlb_miss(address, isr);
	STATS_ADD(tlb_miss_time, read_itc()-start_time);
	}
}
#else
void
handle_vhpt_miss(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("vhpt %lx %lx\n", read_iip(), read_ifa());
	DEBUG_SHORT('v');
	STATS_INC(vhpt_miss);
	{
	STATS_IF(unsigned long start_time = read_itc());
	mm_tlb_miss(address, isr);
	STATS_ADD(tlb_miss_time, read_itc()-start_time);
	}
}
#endif

void
handle_dirty_bit(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("dirty %lx %lx %lx\n", read_iip(), address, isr);
	DEBUG_SHORT('W');
	STATS_INC(dirty_bit);
	mm_fault(address, isr);
}

void
handle_iaccess_bit(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("iaccess %lx %lx %lx\n", read_iip(), address, isr);
	DEBUG_SHORT('I');
	STATS_INC(access_bit);
	mm_fault(address, isr);
}

void
handle_daccess_bit(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("daccess %lx %lx %lx\n", read_iip(), address, isr);
	DEBUG_SHORT('D');
	STATS_INC(access_bit);
	mm_fault(address, isr);
}

void
handle_page_not_present(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("notpresent %lx %lx\n", read_iip(), address);
	DEBUG_SHORT('N');
	STATS_INC(not_present);
#ifdef LONG_VHPT
	cpu.cr[CR_IFA] = address;
	cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
	deliver_exception(0x5000, isr);
#else
	{
	STATS_IF(unsigned long start_time = read_itc());
	mm_tlb_miss(address, isr);
	STATS_ADD(tlb_miss_time, read_itc()-start_time);
	}
#endif
}

void
handle_iaccess_rights(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("irights %lx %lx\n", read_iip(), address);
	DEBUG_SHORT('x');
	STATS_INC(access_rights);
	cpu.cr[CR_IFA] = address;
	cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
	deliver_exception(0x5200, isr);
}

void
handle_daccess_rights(void)
{
	unsigned long address = read_ifa();
	unsigned long isr = read_isr();

	DEBUG_INTERRUPTION("drights %lx %lx\n", read_iip(), address);
	DEBUG_SHORT('w');
	STATS_INC(access_rights);
	cpu.cr[CR_IFA] = address;
	cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
	deliver_exception(0x5300, isr);
}

void
mm_tlb_miss(unsigned long address, unsigned long isr)
{
	unsigned int is_exec = (isr & (1UL << 32)) ? 1 : 0;
	struct mapping *mapping;
	struct page *page;
	unsigned long pte;
	mpfn_t mpfn;

	//DEBUG_MM("%lx %lx\n", address, cpu.rr[address >> 61]);
	//printf("[%lx] mm_tlb_miss %lx\n", read_iip(), address);

	if (unlikely(address < MAX_MEM_SIZE))
	{
		if (unlikely(address < (4095*PAGE_SIZE)))
		{
			printf("null pointer dereference at %lx\n", read_iip());

			/* NaT Page Consumption */
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[address >> 61];
			deliver_exception(0x5600, isr);
			return;
		}
		mpfn = address >> PAGE_SHIFT;
		page = mpfn_to_page(mpfn);
		pte = (3UL << PTE_VNUMA_A_BIT)|PTE_AR_RWX|PTE_PL_3|(page->prot<<PTE_A_BIT)|PTE_P|page_to_pa(page);
		vhpt_insert(address, pte, mpfn, is_exec);
		STATS_INC(tlb_miss_phys);
		return;
	}

	mapping = mappings[address>>61];
	if (unlikely(mapping != NULL))
	{
		do {
		if ((address & mapping->mask) == mapping->va)
		{
			unsigned long va = (address & ~(unsigned long)(PAGE_SIZE-1));
			unsigned long pte = mapping->pte;
			unsigned long mpa = (BITS(pte, 12, 38) << 12) & mapping->mask;
			unsigned long da = BITS(pte, PTE_A_BIT, 2);
			unsigned long realpte, realda;

			if (unlikely(pte & PTE_VNUMA_PIB))
			{
				emu_store_PIB(address);
				return;
			}

			pte &= PTE_P|PTE_AR|PTE_ED;	/* preserve P, AR, ED fields from guest */
			pte |= PTE_PL_3 | (da << PTE_VNUMA_A_BIT);	/* PL 3, original D|A in ignored field */

			if (unlikely(mpa > MAX_MEM_SIZE))
			{
				printf("ip %lx va %lx mva %lx mpte %lx mmask %lx\n",
						read_iip(), address, mapping->va, mapping->pte, mapping->mask);
			}

			mpa += va - mapping->va;
			mpfn = mpa_to_mpfn(mpa);
			page = mpfn_to_page(mpfn);
			realda = (page->prot & da) << PTE_A_BIT;
			realpte = pte | realda | page_to_pa(page);
			//printf("vhpt_insert %lx %lx %d\n", va, realpte, is_exec);
			vhpt_insert(va, realpte, mpfn, is_exec);
#ifdef DSM
			register_virtual(mpfn, va);
#endif
			STATS_INC(tlb_miss_tr);
			return;
		}
		} while ((mapping = mapping->next) != NULL);
	}

	//printf("mapping lookup done\n");

	/* possibly handing another exception, precover incomplete register frame */
	if (unlikely((isr & (1UL << 38)) && !(cpu.psr & PSR_IC)))
	{
		DEBUG_MM("precover\n");
		emu_precover();
		return;
	}

#ifdef VHPT_WALKER
	if (likely(cpu.psr & PSR_IC))
	{
		if (likely(cpu.rr[address >> 61] & 1)) /* VHPT walker enabled */
		{
			/* assumes short format VHPT */
			/* doesn't check permissions on virtual mapping */
			struct vhpt_entry *entry;
			unsigned long pte_addr, pte_pa;

			pte_addr = thash(address);
#if 1
			entry = vhpt_lookup(pte_addr);
			if (!vhpt_entry_valid(entry, pte_addr) || ((entry->pte & (PTE_P|PTE_A)) != (PTE_P|PTE_A)))
#else
			pte_pa = tpa(pte_addr);
			if (unlikely(pte_pa == 0))
#endif
			{
				/* VHPT miss */
				DEBUG_MM("VHPT miss\n");
				cpu.cr[CR_IFA] = address;
				cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
				cpu.cr[CR_IHA] = thash(address);
				deliver_exception(0x0000, isr);
				STATS_INC(tlb_miss_vhpt_miss);
				return;
			}
#if 0
			page = mpfn_to_page(mpa_to_mpfn(pte_pa));
			if (unlikely(!(page->prot & PROT_RX)))
			{
				/* not on this node - VHPT walker aborts */
				DEBUG_MM("not on this node\n");
				cpu.cr[CR_IFA] = address;
				cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
				cpu.cr[CR_IHA] = thash(address);
				deliver_exception(is_exec ? 0x0400 : 0x0800, isr);
				STATS_INC(tlb_miss_vhpt_miss);
				return;
			}
#endif
#if 0
			if (unlikely(tpa(pte_pa) == 0))
			{
				STATS_INC(internal_pa_miss);
				mm_tlb_miss(pte_pa, 0);
			}
			pte = *(unsigned long *)pte_pa;
#elif 0
#define read8_physical(address) ({ unsigned long value; asm("rsm psr.dt ;; srlz.d ;; ld8 %0=[%1] ;; ssm psr.dt ;; srlz.d" : "=r"(value) : "r"(address)); value; })
			pte = read8_physical(pte_pa);
#else
			pte = *(unsigned long *)pte_addr;
#endif
#if 0
			if (((pte >> 12) & 0x3fffffffffUL) > (MAX_MEM_SIZE>>PAGE_SHIFT))
				printf("bad pte: %lx %lx %lx\n", pte, pte_pa, tpa(pte_pa));
#endif

			if (unlikely(!(pte & PTE_P)))
			{
				/* Optimisation: abort VHPT walker if page not present, rather than inserting not present translation */
				DEBUG_MM("not present\n");
				cpu.cr[CR_IFA] = address;
				cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
				cpu.cr[CR_IHA] = thash(address);
				deliver_exception(is_exec ? 0x0400 : 0x0800, isr);
				STATS_INC(tlb_miss_vhpt_miss_np);
				return;
			}
			tlb_insert(address, cpu.rr[address >> 61], pte, 0, is_exec);
			STATS_INC(tlb_miss_vhpt_hit);
		}
		else
		{
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
			cpu.cr[CR_IHA] = thash(address);
			deliver_exception(is_exec ? 0x0c00 : 0x1000, isr);
			STATS_INC(tlb_miss_vhpt_off);
			PROFILE_TLB(address, PROFILE_TLB_VHPT_OFF);
		}
	}
	else
	{
		//printf("iip=%lx address=%lx isr=%lx\n", read_iip(), address, isr);
		deliver_exception(0x1400, isr);
	}
#else
	DEBUG_MM("TLB miss @ %lx\n", address);

	/* deliver TLB miss */
	if (cpu.psr & PSR_IC)
	{
		cpu.cr[CR_IFA] = address;
		cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
		cpu.cr[CR_IHA] = thash(address);
		if (cpu.rr[address >> 61] & 1)
			deliver_exception(is_exec ? 0x0400 : 0x0800, isr);
		else
			deliver_exception(is_exec ? 0x0c00 : 0x1000, isr);
	}
	else
	{
		deliver_exception(0x1400, isr);
	}
#endif
}

void
mm_fault(unsigned long address, unsigned long isr)
{
	struct vhpt_entry *mapping;
	boolean is_exec = (isr >> 32) & 1;
	boolean is_write = (isr >> 33) & 1;
	boolean is_rse = (isr >> 37) & 1;
	access_type_t access_type;
	boolean critical_word_first = 0;
#ifdef DSM
	mpfn_t mpfn;
	struct page *page;
#endif

	//printf("mm_fault %lx %lx\n", address, isr);

	mapping = vhpt_lookup(address);
	if (unlikely(!vhpt_entry_valid(mapping, address)))
	{
		mm_tlb_miss(address, isr);
		return;
	}

	if (!BIT(mapping->pte, 53))
	{
		if ((isr & (1UL << 38)) && !(cpu.psr & PSR_IC))
			emu_precover();
		else
		{
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
			deliver_exception(is_exec ? 0x2400 : 0x2800, isr);
		}
		return;
	}
	if (is_write && !BIT(mapping->pte, 54))
	{
		if ((isr & (1UL << 38)) && !(cpu.psr & PSR_IC))
			emu_precover();
		else
		{
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[address >> 61] & ~1UL;
			deliver_exception(0x2000, isr);
		}
		return;
	}

	//printf("[%lx] unrecoverable fault at %lx\n", read_iip(), address);

#ifdef DSM
#ifdef LONG_VHPT
	mpfn = mapping->ig;
#else
#warning Assumes 1:1 mpfn
	mpfn = (mapping->pte >> PAGE_SHIFT) & 0x3fffffffffUL;
#endif
	page = mpfn_to_page(mpfn);

#if 0 /* ndef NDEBUG */
	assert(page->pfn == BITS(mapping->pte, 12, 38));
	if (is_write)
	{
		if ((mapping->pte & PTE_D) != 0)
		{
			printf("[%lx] PTE_D is set (%lx isr=%lx)\n", read_iip(), address, isr);
			ia64_ptc_l(address, PAGE_SHIFT<<2);
			return;
		}
	}
	else
	{
		if ((mapping->pte & PTE_A) != 0)
		{
			printf("[%lx] PTE_A is set (%lx isr=%lx)\n", read_iip(), address, isr);
			ia64_ptc_l(address, PAGE_SHIFT<<2);
			return;
		}
	}

	if (is_write)
	{
		if (page->prot == PROT_RWX)
		{
			printf("[%lx] write fault on RWX page (%lx isr=%lx)\n", read_iip(), address, isr);
			printf("VHPT a: %lx, itir: %lx, tag: %lx\n", mapping->pte, mapping->ps, mapping->tag);
			//printf("page->va: %lx, page->rr: %lx\n", page->va, page->rr);
			mm_chmod(mpfn, PROT_RWX);
			printf("VHPT b: %lx\n", mapping->pte);
			return;
		}
	}
	else
	{
		if (page->prot & PROT_RX)
		{
			printf("[%lx] read fault on RX page (%lx isr=%lx)\n", read_iip(), address, isr);
			printf("VHPT a: %lx, itir: %lx, tag: %lx\n", mapping->pte, mapping->ps, mapping->tag);
			//printf("page->va: %lx, page->rr: %lx\n", page->va, page->rr);
			mm_chmod(mpfn, page->prot);
			printf("VHPT b: %lx\n", mapping->pte);
			return;
		}
	}
#endif

	if (page->lock) /* waiting on page data already */
		return;

	if (is_rse)
	{
		access_type = ACCESS_RSE;
		/* RSE references cannot be emulated */
	}
	else if (is_write)
	{
		access_type = ACCESS_DATA;
#ifdef WRITE_UPDATES
#ifdef WRITE_UPDATE_THRESHOLD
		if (page->wcount >= write_update_threshold)
		{
			//printf("write threshold reached\n");
			goto invalidate;
		}
#endif

		if (page->sends_pending_this && page->prot)
		{
#ifndef TIGON3
			extern void *ether_drv;
			dp83820_reap_tx_descs(ether_drv);
			if (page->sends_pending_this)
#endif
			{
#ifdef ENABLE_REMAP
				uint64_t mpa;
				pfn_t new_pfn;
				void *p;

				if (unlikely(page->sends_pending_prev))
				{
					//printf("woah, sends_pending_prev set\n");
					STATS_INC(inv_sends_pending);
					goto invalidate;
				}

				mpa = mpfn_to_mpa(mpfn);
				p = frame_alloc_and_map(&new_pfn);
				//logf("remap %x %x %x %x %x\n", mpfn, page->pfn, new_pfn, page->sends_pending_this, page->sends_pending_prev);

				if (tpa(mpa) == 0)
					mm_tlb_miss(mpa, 0);

				/* we know it is readable from the page->prot check */
				/* memcpy(p, (void *)mpa, PAGE_SIZE); */
				copy_page(p, (void *)mpa);
				unmap_window();

				mm_remap(mpfn, new_pfn);

				page->sends_pending_prev = page->sends_pending_this;
				page->sends_pending_this = 0;
#else
				static unsigned long fetch_due_to_pending_send;
				fetch_due_to_pending_send++;
				if ((fetch_due_to_pending_send % 65536) == 0)
					printf("fdtps: %x\n", fetch_due_to_pending_send);
				//printf("pending send\n");
				DEBUG_DSM(mpfn, "&");
				goto invalidate;
#endif
			}
		}
		if (emu_store(address, mpfn_to_mpa(mpfn) | (address & PAGE_MASK)))
		{
			//printf("[%lx] store %lx\n", read_iip(), address);
			page->mode = PAGE_MODE_WRITE_UPDATE;
			page->wcount++;
			return;
		}
#endif
	}
	else if (!is_exec)
	{
		access_type = ACCESS_DATA;
		critical_word_first = 1;
	}
	else
	{
		access_type = ACCESS_EXECUTE;
	}

 invalidate:
#ifdef WRITE_UPDATES
	/* make sure write buffers are flushed when switching to write invalidate */
	if (page->mode == PAGE_MODE_WRITE_UPDATE)
	{
		STATS_INC(fwb_mode_change);
		emu_fwb();
		if (is_write)
			page->mode = PAGE_MODE_WRITE_INVALIDATE;
	}
#endif

#ifdef READ_CRITICAL_WORD_FIRST
	if (critical_word_first)
	{
		if (emu_load(address, mpfn_to_mpa(mpfn) | (address & PAGE_MASK)))
		{
			//printf("[%lx] load %lx\n", read_iip(), address);
			return;
		}
	}
#endif

	//printf("[%lx] fetch %lx\n", read_iip(), address);
	mm_fetch(mpfn, is_write, access_type, address, 1);
#else
	assert(0);
#endif
}

#ifdef SPECIAL_CMPXCHG_FETCHADD
uint64_t
mm_handle_cmpxchg(uint64_t mpa, uint64_t value, uint64_t cmpvalue, uint8_t size, uint64_t *newvalue)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	uint64_t origvalue;
	void *p;

	assert(!page->lock);
	assert(page->owner == node_id);

	//printf("handle_cmpxchg %lx %lx %lx %x\n", mpa, value, cmpvalue, size);
	emu_fwb();

#warning FIXME for SMP
	p = map_window(page, mpa);
	switch (size)
	{
	case 0: origvalue = *(uint8_t *)p; break;
	case 1: origvalue = *(uint16_t *)p; break;
	case 2: origvalue = *(uint32_t *)p; break;
	case 3: origvalue = *(uint64_t *)p; break;
	default: printf("Bad size in cmpxchg\n");
	}
	if (origvalue == cmpvalue)
	{
		switch (size)
		{
		case 0: *(uint8_t *)p = value; break;
		case 1: *(uint16_t *)p = value; break;
		case 2: *(uint32_t *)p = value; break;
		case 3: *(uint64_t *)p = value; break;
		}
		*newvalue = value;
#ifdef MONITOR
		if (is_mpa_monitored(mpa))
			logf("@%lx", value);
#endif
	}
	else
	{
		*newvalue = origvalue;
#ifdef MONITOR
		if (is_mpa_monitored(mpa))
			logf("$%lx", origvalue);
#endif
	}
	unmap_window();

#warning Need to propagate this write to other nodes

	return origvalue;
}

uint64_t
mm_handle_fetchadd(uint64_t mpa, uint8_t size, uint8_t inc, uint64_t *newvalue)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	uint64_t origvalue;
	void *p;
	int8_t inc_map[] = { 16, 8, 4, 1, -16, -8, -4, -1 };
	int8_t increment = inc_map[inc & 7];

	assert(!page->lock);
	assert(page->owner == node_id);

	//printf("handle_fetchadd %lx %x %x %x\n", mpa, size, inc, increment);
	emu_fwb();

#warning FIXME for SMP
	p = map_window(page, mpa);
	switch (size)
	{
	case 2: origvalue = *(uint32_t *)p; *(uint32_t *)p = origvalue+increment; break;
	case 3: origvalue = *(uint64_t *)p; *(uint64_t *)p = origvalue+increment; break;
	default: printf("Bad size in fetchadd\n");
	}
	unmap_window();

#warning Need to propagate this write to other nodes

	*newvalue = origvalue+increment;
#ifdef MONITOR
	if (is_mpa_monitored(mpa))
		logf("%%%lx", origvalue+increment);
#endif
	return origvalue;
}

/* This is really nasty, FIXME */
static struct {
	uint64_t mpa;
	uint64_t cmpvalue;
	uint64_t oldvalue;
	uint8_t size;
	uint8_t isfetchadd;
	uint8_t oldvalue_invalid;
} semaphore;

void
mm_handle_semaphore_reply(uint64_t oldvalue, uint64_t newvalue)
{
	if (semaphore.isfetchadd || (oldvalue == semaphore.cmpvalue))
		mm_handle_store(semaphore.mpa, newvalue, semaphore.size, 0);

	semaphore.oldvalue = oldvalue;
	semaphore.oldvalue_invalid = 0;
}

uint64_t
mm_cmpxchg(uint64_t mpa, uint64_t value, uint64_t cmpvalue, uint8_t size, int release, uint64_t va)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	node_t owner = page->owner;
	struct request request;
	uint64_t oldvalue, newvalue;

#if 0
	//printf("cmpxchg %lx %lx %lx %x\n", mpa, value, cmpvalue, size);
	return 0x12345678UL;
#endif

	assert(!page->lock);

	if (owner == node_id)
	{
		oldvalue = mm_handle_cmpxchg(mpa, value, cmpvalue, size, &newvalue);
		if (oldvalue == cmpvalue)
			emu_store_remote(mpa, newvalue, size, 1 /* release */);
		return oldvalue;
	}

	emu_fwb();
	semaphore.mpa = mpa;
	semaphore.cmpvalue = cmpvalue;
	semaphore.size = size;
	semaphore.isfetchadd = 0;
	semaphore.oldvalue_invalid = 1;
	request.type = IPC_CMPXCHG;
	request.cmpxchg.mpa = mpa;
	request.cmpxchg.value = value;
	request.cmpxchg.cmpvalue = cmpvalue;
	request.cmpxchg.size = size;
	ipc_send(owner, &request);
	ipc_wait(owner, &request, &semaphore.oldvalue_invalid);
	return semaphore.oldvalue;
}

/* BEWARE delayed write-updates overwriting the value we fetch */
uint64_t
mm_fetchadd(uint64_t mpa, uint8_t size, uint8_t inc, int release)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	node_t owner = page->owner;
	struct request request;
	uint64_t newvalue, oldvalue;

#if 0
	//printf("fetchadd %lx %x %x\n", mpa, size, inc);
	return 0x12345678UL;
#endif

	assert(!page->lock);

	if (owner == node_id)
	{
		assert(!((page->prot & PROT_RWX) == PROT_RWX));
		assert(page->prot & PROT_RX);

		oldvalue = mm_handle_fetchadd(mpa, size, inc, &newvalue);
		emu_store_remote(mpa, newvalue, size, 1 /* release */);
		return oldvalue;
	}

	emu_fwb();
	semaphore.mpa = mpa;
	semaphore.size = size;
	semaphore.isfetchadd = 1;
	semaphore.oldvalue_invalid = 1;
	request.type = IPC_FETCHADD;
	request.fetchadd.mpa = mpa;
	request.fetchadd.size = size;
	request.fetchadd.inc = inc;
	ipc_send(owner, &request);
	ipc_wait(owner, &request, &semaphore.oldvalue_invalid);
	return semaphore.oldvalue;
}
#endif

static uint64_t load_value;
static uint8_t load_value_invalid;

boolean
mm_load(uint64_t mpa, uint8_t size, uint64_t *value)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	node_t manager, forward_to;
	struct request request;

	DEBUG_DSM(mpfn, "r%lx ", mpfn);

	if (page->lock)
	{
		//printf("hmm, page already value, what to do now?\n");
		return 0;
	}

	page->lock = 1;

	assert(page->owner != node_id);

	manager = mpfn_to_manager(mpfn);
	//assert(manager == 0);
	forward_to = (manager == node_id) ? page->owner : manager;

	load_value_invalid = 1;
	request.type = IPC_LOAD;
	request.load.mpa = mpa;
	request.load.size = size;
	request.load.reply_to = node_id;
	ipc_send(forward_to, &request);
	ipc_wait(forward_to, &request, &load_value_invalid);
	*value = load_value;
	return 1;
}

void
mm_handle_load(node_t node, uint64_t mpa, uint8_t size, node_t for_node, struct request *orig_req)
{
	unsigned long ip = read_iip() | BITS(read_ipsr(), PSR_RI_BITS, 2);
	struct request reply;
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	uint64_t value;
	void *p;

	DEBUG_DSM(mpfn, "L%lx ", mpfn);

#ifdef WRITE_UPDATES
	emu_fwb_if_hit(mpfn);
#endif

#ifdef THRASH_PROTECTION
	if ((/*(last_exec_fault_mpfn == mpfn) ||*/ (last_data_fault_mpfn == mpfn)) && THRASH_PROTECTION_NO_PROGRESS())
	{
		DEBUG_DSM(mpfn, "%d:p%lx ", node, mpfn);
		enqueue_request(&page->wait_queue, orig_req);
#if THRASH_PROTECTION == 2
		write_ipsr(read_ipsr() | PSR_SS);
#endif
		return;
	}
#endif

	if (page->lock)
	{
		DEBUG_LOCK(mpfn, "%d:l%lx(%d) ", node, mpfn, page->lock);
		enqueue_request(&page->wait_queue, orig_req);
		return;
	}

	p = map_window(page, mpa);
	switch (size)
	{
	case 0: value = *(uint8_t *)p; break;
	case 1: value = *(uint16_t *)p; break;
	case 2: value = *(uint32_t *)p; break;
	case 3: value = *(uint64_t *)p; break;
	default: printf("Bad size in load\n");
	}
	unmap_window();

	//printf("load %lx %x -> %lx\n", mpa, size, value);
	reply.type = IPC_LOAD_REPLY;
	reply.loadreply.value = value;
	ipc_send(for_node, &reply);

	mm_handle_fetch(node, mpfn, 0, ACCESS_DATA, for_node, orig_req, 0);
}

void
mm_handle_load_reply(uint64_t value)
{
	load_value = value;
	load_value_invalid = 0;
}

struct queued_store {
	struct queued_store *next;
	struct store store;
	uint8_t from;
	//uint8_t updates_from[MAX_NODES];
} *store_queue, **store_queue_next = &store_queue;

#ifdef COHERENT_WRITE_UPDATES
#ifdef COHERENT_WRITE_UPDATES_COMPRESSED
struct write_log_revision {
	uint16_t version;
	uint8_t node;
	uint8_t mask;
};
struct write_log_entry {
	uint16_t offset;
	uint16_t entries;
};
#else
struct write_log_revision {
	uint16_t version;
#if 0
	uint8_t node;
#endif
};
struct write_log_entry {
	uint16_t offset;
	uint8_t mask;
	struct write_log_revision byte[8];
};
#endif

struct write_log {
	struct write_log *next;
	struct write_log **pthis;
	struct write_log *cleanup_next;
	struct write_log_entry x;
};

struct write_log *cleanup_list[4];
struct write_log *free_logs;
unsigned int total_write_logs, free_write_logs;

void
mm_dump_write_log_stats(void)
{
	struct write_log *log;
	struct page *page;
	unsigned int i, count, half_free;

	printf("total_write_logs = %d\nfree_write_logs = %d\n",
		total_write_logs, free_write_logs);

	for (i = 0; i < 4; i++)
	{
		count = half_free = 0;
		for (log = cleanup_list[i]; log; log = log->cleanup_next)
		{
			count++;
			if (!log->x.mask)
				half_free++;
		}
		printf("cleanups[%x] = %x (%x)\n", i, count, half_free);
	}

	for (i = 0; i < (MAX_MEM_SIZE >> PAGE_SHIFT); i++)
	{
		page = &pages[i];
		if (!page->write_log)
			continue;

		count = half_free = 0;
		for (log = page->write_log; log; log = log->next)
		{
			count++;
			if (!log->x.mask)
				half_free++;
		}
		if (count > 16)
			printf("page[%x] = %x (%x)\n", i, count, half_free);
		if (count > 512)
		{
			for (log = page->write_log, count = 0; log && (count < 512); log = log->next, count++)
				printf(" %x", log->x.offset);
			printf("\n");
		}
	}
}

static inline struct write_log *alloc_write_log()
{
	struct write_log *log;

	if (free_logs)
	{
		log = free_logs;
		free_logs = log->next;
		free_write_logs--;
		return log;
	}

	total_write_logs++;
	return malloc(sizeof(struct write_log));
}

static inline void free_write_log(struct write_log *log)
{
	log->next = free_logs;
	free_logs = log;
	free_write_logs++;
}

static void
mm_drop_store_history(struct page *page)
{
	struct write_log *log;

	for (log = page->write_log; log != NULL; log = log->next)
	{
		//nextlog = log->next;
		//free_write_log(log);
		log->x.mask = 0;
	}

	//page->write_log = NULL;
}

static void
mm_send_store_history(struct page *page, mpfn_t mpfn, int node)
{
	struct request request;
	struct write_log *log;
	struct write_log_entry *rlogbuf, *rlog;
	unsigned int size = 0;
	uint16_t count = 0;

	for (log = page->write_log; log != NULL; log = log->next)
	{
		if (!log->x.mask)
			continue;

		size += sizeof(struct write_log_entry);
		count++;
	}

	if (!count)
		return;

	if (count >= (1500-64)/sizeof(struct write_log_entry))
	{
		printf("mpfn: %x count: %x\n", mpfn, count);
		for (log = page->write_log; log != NULL; log = log->next)
		{
			if (!log->x.mask)
				continue;

			printf(" %x", log->x.offset);
		}
		printf("\n");
	}
	assert(count < (1500-64)/sizeof(struct write_log_entry));
	assert(size < (1500-64));
	rlog = rlogbuf = malloc(size);
	for (log = page->write_log; log != NULL; log = log->next)
	{
		if (!log->x.mask)
			continue;

		memcpy(rlog, &log->x, sizeof(struct write_log_entry));
		rlog++;
	}

	request.type = IPC_STORE_HISTORY;
	request.storehistory.mpfn = mpfn;
	request.storehistory.count = count;
	ipc_send_page(node, &request, __pa(rlogbuf), size, DATA_AREA_MALLOC);
}

void
mm_handle_store_history(mpfn_t mpfn, uint16_t count, void *data)
{
	struct page *page = mpfn_to_page(mpfn);
	struct write_log_entry *rlog = data;
	struct write_log *log, **prevlog;
	unsigned int min_generation;
	int rel_generation, min_rel_generation;
	unsigned int i, j;

	if (page->write_log)
		page->write_log->pthis = NULL;

	prevlog = &page->write_log;

	assert(count < (1500-64)/sizeof(struct write_log_entry));
	for (j = 0; j < count; j++)
	{
		*prevlog = log = alloc_write_log();
		memcpy(&log->x, rlog, sizeof(struct write_log_entry));
		log->pthis = prevlog;

		min_rel_generation = 2;
		for (i = 0; i < 8; i++)
		{
			rel_generation = ((unsigned int)((log->x.byte[i].version>>14)-current_generation+1))%4-1;
			//printf("%x[%x] ", log->x.byte[i].version>>14, rel_generation);
			if (rel_generation == 2)
				log->x.mask &= ~(1<<i);
			else if (rel_generation < min_rel_generation)
				min_rel_generation = rel_generation;
		}
		min_generation = (unsigned int)(current_generation+min_rel_generation)%4;
		//printf("current_generation=%x min_rel_generation=%x min_generation=%x\n", current_generation, min_rel_generation, min_generation);
		//assert_not_in_list(cleanup_list[min_generation], log, "calc");
		log->cleanup_next = cleanup_list[min_generation];
		cleanup_list[min_generation] = log;

		rlog++;
		prevlog = &log->next;
	}

	*prevlog = NULL;
}

static const uint8_t size_to_mask[4] = { 0x01, 0x03, 0x0f, 0xff };
static const uint32_t mask_to_halfmask[16] = { 0x00000000, 0x000000ff, 0x0000ff00, 0x0000ffff, 0x00ff0000, 0x00ff00ff, 0x00ffff00, 0x00ffffff, 0xff000000, 0xff0000ff, 0xff00ff00, 0xff00ffff, 0xffff0000, 0xffff00ff, 0xffffff00, 0xffffffff };
#define expand_mask(mask) (mask_to_halfmask[mask & 0xf] | ((uint64_t)mask_to_halfmask[mask>>4] << 32))

#ifdef COHERENT_WRITE_UPDATES_COMPRESSED
static uint64_t
mm_calculate_store_mask(struct page *page, uint64_t mpa, uint8_t size, node_t node, uint16_t version)
{
	struct write_log *log;
	uint8_t mask, next_mask, grant_mask, residual_mask;
	node_t next_node;
	uint16_t next_version, offset;
	unsigned int i, j;

	offset = mpa & (PAGE_MASK &~ 7UL);
	mask = size_to_mask[size] << (mpa & 7);
	residual_mask = 0xff;

	for (log = page->write_log; (log != NULL) && (log->offset != offset);
		log = log->next);

	if (!log)
	{
		log = alloc_write_log();
		log->next = page->write_log;
		page->write_log = log;
		log->offset = offset;
		log->entries = 0;
	}

	for (i = 0; i < log->entries; i++)
	{
		if (((int16_t)(version - log->entry[i].version) > 0)
			|| ((version == log->entry[i].version) && (node >= log->entry[i].node)))
			goto entryfound;

		mask &= ~log->entry[i].mask;
		residual_mask &= ~log->entry[i].mask;
		if (!mask) /* write does not show through newer writes */
		{
			//printf("drop %lx\n", mpa);
			return 0;
		}
	}

	/* write older than all writes in log, but still shows through */
	// if (i == 0) printf("first %lx\n", mpa); else printf("showx %lx %x\n", mpa, mask);
	grant_mask = mask;
	goto addentry;

entryfound:
	/* newer than some but not all existing writes, recalculate log */
	//printf("show %lx %x\n", mpa, mask);
	grant_mask = mask;
	j = i;
	do
	{
		next_version = log->entry[j].version;
		next_node = log->entry[j].node;
		next_mask = log->entry[j].mask;
		j++;
		if (mask) {
			log->entry[i].version = version;
			log->entry[i].node = node;
			log->entry[i].mask = mask;
			i++;
			residual_mask &= ~mask;
		}
		version = next_version;
		node = next_node;
		mask = next_mask & residual_mask;
	} while ((j < log->entries) && residual_mask);

addentry:
	if (mask) {
		log->entry[i].version = version;
		log->entry[i].node = node;
		log->entry[i].mask = mask;
		i++;
	}
	assert(i <= 8);
	if (i > 8)
	{
		for (j = 0; j < i; j++)
		{
			printf("%x %x %x\n", log->entry[j].version,
				log->entry[j].node, log->entry[j].mask);
		}
	}
	else
		log->entries = i;
	return expand_mask(grant_mask);
}
#else /* !COHERENT_WRITE_UPDATES_COMPRESSED */
#if 0
void
assert_not_in_list(struct write_log *list, struct write_log *log, char *where)
{
	for (; list != NULL; list = list->cleanup_next)
	{
		if (list == log)
			printf("assert_not_in_list failed: %s!!\n", where);
	}
}
#endif

static uint64_t
mm_calculate_store_mask(struct page *page, uint64_t mpa, uint8_t size, node_t node, uint16_t version)
{
	struct write_log *log;
	unsigned long grant_mask = 0;
	uint16_t offset;
	uint8_t store_mask, valid_mask;
	unsigned int i;

	offset = mpa & (PAGE_MASK &~ 7UL);
	store_mask = size_to_mask[size] << (mpa & 7);

#if 0
	return expand_mask(store_mask);
#endif

	for (log = page->write_log; (log != NULL) && (log->x.offset != offset);
		log = log->next);

	if (!log)
	{
		log = alloc_write_log();
		log->next = page->write_log;
		page->write_log = log;

		log->pthis = &page->write_log;
		if (log->next)
			(log->next)->pthis = &(log->next);

		//assert_not_in_list(cleanup_list[version>>14], log, "calc");
		log->cleanup_next = cleanup_list[version>>14];
		cleanup_list[version>>14] = log;

		log->x.offset = offset;
		log->x.mask = 0;
	}

	valid_mask = log->x.mask;
	for (i = 0; i < 8; i++)
	{
		grant_mask >>= 8;
		if ((store_mask & 1)
			&& (!(valid_mask & (1<<i))
				|| ((int16_t)(version - log->x.byte[i].version) >= 0)))
#if 0
						|| ((((int16_t)(version - log->x.byte[i].version) > 0)
						   || ((version == log->x.byte[i].version) && (node >= log->x.byte[i].node))
#endif
		{
			grant_mask |= 0xff00000000000000UL;
			valid_mask |= 1<<i;
			log->x.byte[i].version = version;
#if 0
			log->x.byte[i].node = node;
#endif
		}
		store_mask >>= 1;
	}
	log->x.mask = valid_mask;
	return grant_mask;
}

void
mm_new_generation(void)
{
	struct write_log *log, *nextlog;
	struct write_log **next_cleanup_list;
	unsigned int i;
	uint32_t mask;
	uint16_t cleanup_generation;

	printf("wu_applied = %x\n", wu_applied);
	cleanup_generation = (current_generation+3)%4;

	for (i = 0; i < nodes; i++)
		if ((i != node_id) && ((remote_wu_version[i]>>14) == cleanup_generation))
			printf("OH SHIT!!! current_generation=%x cleanup_generation=%x remote[%x]=%x\n", current_generation, cleanup_generation, i, remote_wu_version[i]);

	next_cleanup_list = &cleanup_list[(cleanup_generation+1)%4];
	for (log = cleanup_list[cleanup_generation]; log != NULL; log = nextlog)
	{
		nextlog = log->cleanup_next;
		mask = log->x.mask;
		for (i = 0; i < 8; i++)
		{
			if ((log->x.byte[i].version >> 14) == cleanup_generation)
				mask &= ~(1 << i);
		}
		if (mask)
		{
			log->x.mask = mask;
			//assert_not_in_list(*next_cleanup_list, log, "cleanup");
			log->cleanup_next = *next_cleanup_list;
			*next_cleanup_list = log;
		}
		else
		{
			if (log->next)
				(log->next)->pthis = log->pthis;
			if (log->pthis)
				*(log->pthis) = log->next;
#if 0
			for (i = 0; i < 4; i++)
				if (i != cleanup_generation)
					assert_not_in_list(cleanup_list[i], log, "free");
#endif
			free_write_log(log);
		}
	}
	cleanup_list[cleanup_generation] = NULL;

	current_generation = (current_generation+1)%4;
}
#endif
#else /* !COHERENT_WRITE_UPDATES */
void
mm_dump_write_log_stats(void)
{
}

void
mm_handle_store_history(mpfn_t mpfn, uint16_t count, void *data)
{
}

void
mm_new_generation(void)
{
}
#endif

void
mm_handle_store(uint64_t mpa, uint64_t value, uint8_t size, node_t node, uint16_t version, int replay)
{
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);
	uint64_t origvalue, mask;
	void *p;

	if (!replay)
		page->wu_version++;

	if (node != node_id)
	{
#ifdef WRITE_UPDATE_THRESHOLD
		page->wcount = 0;
#endif

		if (mpfn == fetch_waiting_for_mpfn)
		{
			struct queued_store *store;

#ifdef MONITOR
			if (is_mpa_monitored(mpa))
				logf("Q%c%lx", (mpfn == 0x4001) ? '^' : (mpfn == 0x4013) ? '$' : '!', value);
#endif

			store = malloc(sizeof(struct queued_store));
			*store_queue_next = store;
			store_queue_next = &store->next;
			store->next = NULL;
			store->store.mpa = mpa;
			store->store.value = value;
			store->store.size = size;
			store->store.wu_version = version;
			store->from = node;
			//memcpy(store->updates_from, updates_from, nodes);
			return;
		}
#if 0
		if (page->lock)
		{
			struct request request;
			if (page->lock != 1)
				printf("warning: page->lock = %d\n", page->lock);
			//printf("queue store %lx:%x = %lx\n", mpa, size, value);
#ifdef MONITOR
			if (is_mpa_monitored(mpa))
				logf("Q%c%lx", (mpfn == 0x4001) ? '^' : (mpfn == 0x4013) ? '$' : '!', value);
#endif
			request.type = IPC_STORE_SINGLE;
			request.from = node;
			request.store1.mpa = mpa;
			request.store1.value = value;
			request.store1.size = size;
			request.store1.wu_version = version;
			request.store1.arrival_time = wu_applied;
			enqueue_request(&page->wait_queue, &request);
			return;
		}
#endif
	}

#ifdef MONITOR
	if (is_mpa_monitored(mpa))
	{
		if (mpfn == 0x4001)
			logf("%c%lx", (node == node_id) ? '%' : '^', value);
		else if (mpfn == 0x4013)
			logf("%c%lx", (node == node_id) ? '@' : '$', value);
		else
			logf("%c%lx", (node == node_id) ? '#' : '!', value);
	}
#endif

	if (!page->prot)
		return;

#ifdef COHERENT_WRITE_UPDATES
	if (page->prot == PROT_RX)
	{
		mask = mm_calculate_store_mask(page, mpa, size, node, version);
		if (!mask && is_mpa_monitored(mpa))
			logf("*");
		p = map_window(page, mpa &~ 7);
		origvalue = *(uint64_t *)p;
		value <<= (mpa & 7)*8;
		*(uint64_t *)p = (value & mask) | (origvalue & ~mask);
		unmap_window();
		return;
	}
#endif

	//printf("store %lx:%x = %lx\n", mpa, size, value);
	//ia64_itr_d(dtr_index, mpa, PAGE_SHIFT << 2, page_to_pa(page)|PTE_AR_RW|PTE_D|PTE_A|PTE_P);
	p = map_window(page, mpa);
	switch (size) {
	case 0: *(uint8_t *)p = value; break;
	case 1: *(uint16_t *)p = value; break;
	case 2: *(uint32_t *)p = value; break;
	case 3: *(uint64_t *)p = value; break;
	default: printf("Bad size in store\n");
	}
	unmap_window();
	//ia64_ptr_d(mpa, PAGE_SHIFT << 2);
}

void
process_store_queue(void)
{
	struct queued_store *store, *next;

	store = store_queue;
	if (!store)
		return;

	do
	{
		mm_handle_store(store->store.mpa, store->store.value, store->store.size, store->from, store->store.wu_version, 1);
		next = store->next;
		free(store);
		store = next;
	} while (store);

	store_queue = NULL;
	store_queue_next = &store_queue;
}

void
process_store_queue_partial(node_t from, uint16_t page_wu_version)
{
	struct queued_store *store, *next;
	unsigned int applied_updates, applied_source_updates, applied_other_updates, other_updates;

	store = store_queue;
	if (!store)
		return;

	applied_updates = page_wu_version - fetch_wu_version;
	applied_source_updates = 0;
	do {
		if (store->from == from)
			applied_source_updates++;
		store = store->next;
	} while (store);
	assert(applied_updates >= applied_source_updates);
	applied_other_updates = applied_updates - applied_source_updates;

	store = store_queue;
	other_updates = 0;
	do
	{
		if (store->from != from)
		{
			if (other_updates >= applied_other_updates)
				break;
			other_updates++;
		}

		next = store->next;
		free(store);
		store = next;
	} while (store);

	while (store)
	{
		if (store->from != from)
		{
			mm_handle_store(store->store.mpa, store->store.value, store->store.size, store->from, store->store.wu_version, 1);
		}
		next = store->next;
		free(store);
		store = next;
	}

	store_queue = NULL;
	store_queue_next = &store_queue;
}

void
mm_check_mpa(unsigned long mpa, unsigned int prot)
{
#ifdef DSM
	mpfn_t mpfn = mpa_to_mpfn(mpa);
	struct page *page = mpfn_to_page(mpfn);

	if (prot & ~(page->prot))
	{
		assert(page->prot != PROT_RWX);
		mm_fetch(mpfn, (prot == PROT_RWX), ACCESS_DATA, mpa, 1);
	}
#endif

	if (tpa(mpa) == 0)
		mm_tlb_miss(mpa, 0);
}

void
mm_copy_to_user(unsigned long dest_mpa, const void *src, unsigned int n)
{
	while (n > 0)
	{
		unsigned int remsize = PAGE_SIZE - (dest_mpa&PAGE_MASK);
		unsigned int size = (n > remsize) ? remsize : n;
		mm_check_mpa(dest_mpa, PROT_RWX);
		memcpy((void *)dest_mpa, src, size);
		dest_mpa += size;
		src += size;
		n -= size;
	}
}

void
mm_copy_from_user(void *dest, unsigned long src_mpa, unsigned int n)
{
	while (n > 0)
	{
		unsigned int remsize = PAGE_SIZE - (src_mpa&PAGE_MASK);
		unsigned int size = (n > remsize) ? remsize : n;
		mm_check_mpa(src_mpa, PROT_RX);
		memcpy(dest, (void *)src_mpa, size);
		dest += size;
		src_mpa += size;
		n -= size;
	}
}

void
mm_release_page(mpfn_t mpfn, unsigned long pa)
{
	struct page *page;

	assert(mpfn <= (MAX_MEM_SIZE>>PAGE_SHIFT));
	page = mpfn_to_page(mpfn);
#ifdef ENABLE_REMAP
	if (page->sends_pending_prev)
	{
		page->sends_pending_prev--;
		if (page->sends_pending_prev == 0)
			frame_free(pa >> PAGE_SHIFT);
	}
	else if (page->sends_pending_this)
	{
		page->sends_pending_this--;
#if 0
		if (page->sends_pending_this == 0)
			frame_free(pa >> PAGE_SHIFT);
#endif
	}
#else
	if (likely(page->sends_pending_this))
		page->sends_pending_this--;
#endif
	else
		printf("mm_release_page: no sends pending\n");
}

unsigned int outstanding_io_requests, outstanding_dsm_requests;
unsigned int last_request_key, last_quiescent_request_key;
extern int guest_scsi_irq;

#ifdef GUESTIO_DISK
static void
mm_disk_io_done(void *cbdata, uint64_t key)
{
	struct scsi_sglist *sgl = (struct scsi_sglist *)cbdata;
	struct scsi_sge *sge;
	struct page *page;
	unsigned int i, npages;
	mpfn_t mpfn;

	//printf("mm_disk_io_done\n");

	for (i = 0; i < sgl->nsge; i++)
	{
		sge = &sgl->sge[i];

		mpfn = sge->mpfn;
		npages = ((sge->addr & PAGE_MASK) + sge->len + PAGE_SIZE-1) >> PAGE_SHIFT;
		for (; npages; npages--, mpfn++)
		{
			page = mpfn_to_page(mpfn);

			//assert(page->lock == 2);
			assert(page->lock != 0);
			assert(page->lock != 1);
			page->lock = 0;
			process_request_queue(&page->wait_queue);
		}
	}

	free(sgl);

	outstanding_io_requests--;
	if ((outstanding_io_requests == 0) && (outstanding_dsm_requests == 0) && guest_scsi_irq)
	{
		//printf("!%d", last_request_key);
		last_quiescent_request_key = last_request_key;
		deliver_interrupt(guest_scsi_irq);
	}
	if ((outstanding_io_requests == 0) && (outstanding_dsm_requests == 0) && !guest_scsi_irq)
		printf("hmm, guest_scsi_irq not set\n");
}

#define DISPATCH_REQUEST() do { \
	outstanding_io_requests++; \
	sgl->nsge = nsge; \
	disk_xfer(sgl, req_offset, req_length, is_write, mm_disk_io_done); \
	nsge = 0; \
} while(0);

unsigned int
mm_handle_disk_io(struct scsi_sge *sge_list_in, unsigned int sge_count_in, unsigned long offset, int is_write, int key_valid, unsigned int request_key)
{
	struct scsi_sglist *sgl;
	struct scsi_sge *sge, *sge_in;
	struct page *page;
	unsigned long req_offset, mpa, mpa_end, pa, start_offset = offset;
	unsigned int req_length, length_in, length_out, i, nsge = 0, needfetch;
	mpfn_t mpfn;

	assert(node_id == 0);
	if (request_key == last_quiescent_request_key)
		printf("XXX request key has wrapped around XXX\n");

	//printf("mm_handle_disk_io\n");
	for (i = 0; i < sge_count_in; i++)
	{
		sge_in = &sge_list_in[i];

		for (mpa = sge_in->addr, length_in = sge_in->len;
			length_in > 0;
			length_in -= length_out, mpa += length_out, offset += length_out)
		{
			length_out = length_in;
			mpa_end = (mpa & ~PAGE_MASK) + PAGE_SIZE;
			if (mpa + length_out > mpa_end)
				length_out = mpa_end - mpa;

			mpfn = mpa_to_mpfn(mpa);
			page = mpfn_to_page(mpfn);
			needfetch = 0;

			if (!is_write && (page->prot != PROT_RWX))
			{
#if 0
			if (pa_end == pa + PAGE_SIZE)
			{
				struct request request;

				/* reading in whole page, just invalidate */
				request.type = IPC_INVALIDATE_WHOLE;
				request.invalidate.mpfn = mpfn;

				assert(!page->lock); /* what to do? */
				if (page->owner == node_id)
				{
					assert(page->copyset);
					ipc_broadcast(page->copyset, &request);
				}
				else
				{
					ipc_send(MPA_TO_MANAGER(pa), &request);
				}
				page->owner = node_id;
				page->copyset = 0;
				mm_chmod(mpfn, PROT_RWX);
			}
			else
			{
				needfetch = 1;
			}
#else
			needfetch = 1;
#endif
			}
			else if (is_write && (page->prot == PROT_NONE))
			{
				/* writing out page, need to fetch page data */
				needfetch = 1;
			}

			if (needfetch)
			{
				/* first send off any waiting disk requests, the next one won't be contiguous */
				if (nsge > 0)
					DISPATCH_REQUEST();

				mm_fetch(mpfn, !is_write, ACCESS_DATA, mpa, 0);
			}
			if (page->lock)
			{
				struct request request;
				request.type = IPC_DISK_IO_SIMPLE;
				request.diskio.mpa = mpa;
				request.diskio.nsge = 0;
				request.diskio.is_write = is_write;
				request.diskio.key_valid = 0;
				request.diskio.len = length_out;
				request.diskio.offset = offset;
				enqueue_request(&page->wait_queue, &request);
				outstanding_dsm_requests++;
				continue;
			}

			pa = page_to_pa(page) | (mpa & PAGE_MASK);

			/* try to coalesce disk requests */
			if ((nsge > 0) && (pa == sge[nsge-1].addr + sge[nsge-1].len))
			{
#warning Small possibility of not metaphysically continuous coalesces if memory not 1:1
				sge[nsge-1].len += length_out;
			}
			else
			{
				/* prevent overflowing SGL... mpt supports up to 49 entries */
				if (nsge >= 49)
					DISPATCH_REQUEST();

				/* allocate a new SGL if necessary */
				if (!nsge)
				{
					sgl = malloc(sizeof(struct scsi_sglist));
					sge = sgl->sge;
					nsge = 0;
					req_offset = offset;
					req_length = 0;
				}

				sge[nsge].addr = pa;
				sge[nsge].len = length_out;
				sge[nsge].mpfn = mpfn;
				nsge++;
			}

			req_length += length_out;
			assert(!page->lock);
			page->lock = 2;
		}
	}

	/* send any outstanding contiguous disk requests */
	if (nsge > 0)
		DISPATCH_REQUEST();

	if (key_valid)
	{
		//printf("*%d", request_key);
		last_request_key = request_key;
	}
	else
		outstanding_dsm_requests--;
	return offset - start_offset;
}
#endif

void
mm_init(void)
{
	unsigned int i, prot;

	prot = (node_id == 0) ? PROT_RWX : PROT_NONE;
	printf("mm_init on node %u, prot = %u\n", node_id, prot);
	for (i = 0; i < 4095; i++)
	{
		pages[i].pfn = 0;
		pages[i].prot = 0;
		pages[i].mode = PAGE_MODE_WRITE_INVALIDATE;
		pages[i].sends_pending_this = pages[i].sends_pending_prev = 0;
#ifdef DSM
		pages[i].wait_queue = NULL;
		pages[i].wu_version = 0;
		pages[i].copyset = 0;
		pages[i].owner = 0;
#endif
	}

	for (i = 4095; i < (MAX_MEM_SIZE>>PAGE_SHIFT); i++)
	{
		pages[i].pfn = i << (PAGE_SHIFT-12);
		pages[i].prot = prot;
		pages[i].mode = PAGE_MODE_WRITE_INVALIDATE;
		pages[i].sends_pending_this = pages[i].sends_pending_prev = 0;
#ifdef DSM
		pages[i].wait_queue = NULL;
		pages[i].wu_version = 0;
		pages[i].copyset = 0;
		pages[i].owner = 0;
#endif
	}

	for (i = 0; i < (MAX_MEM_SIZE>>PAGE_SHIFT); i++)
	{
		m2v[i].rr = 0;
		m2v[i].link_idx = 0;
	}

	emu_wbinit();
#ifdef WRITE_UPDATES
	printf("*** write updates enabled ***\n");
#ifdef WRITE_UPDATE_THRESHOLD
	write_update_threshold = WRITE_UPDATE_THRESHOLD;
	printf("  (threshold = %d)\n", write_update_threshold);
#endif
#endif
}

void
vhpt_init(void)
{
	unsigned long pta, rr0;
	unsigned int i;
	extern void _vhpt;
	if ((unsigned long)&_vhpt & ((1UL << 24)-1))
	{
		assert(0);
		while(1);
	}
	memset(&_vhpt, 0, 1UL<<24);

#ifdef LONG_VHPT
	printf("Using long format VHPT\n");
	pta = (unsigned long)&_vhpt | (1 << 8) | (24 << 2) | 1;
#else
	printf("Using short format VHPT\n");
	pta = (1UL << 60) | (52 << 2) | 1;
#endif

	rr0 = (PAGE_SHIFT << 2) | 1;
	asm volatile("mov cr.pta = %0 ;; mov rr[r0] = %1 ;; srlz.d ;;" : : "r"(pta), "r"(rr0));

	memset(mappings, 0, sizeof(mappings));

#ifdef ENABLE_REMAP
	printf("Initialising free frame list...\n");
	/* frame 0 unused, frame 1 is used for CPU structure */
	for (i = 2; i < 16; i++)
		*(pfn_t *)((uint64_t)i << PAGE_SHIFT) = (i == 15) ? (pfn_t)-1 : (pfn_t)(i+1);
	next_free_frame = 2;
#endif
}

