#include "vm.h"
#include "config.h"
#include "ssc.h"

extern node_t node_id, nodes;
unsigned long smp_rendezvous_ip = 0x4007f20, smp_rendezvous_gp;
int guest_serial_irq, guest_network_irq, guest_scsi_irq;
struct disk_stat stat;
extern unsigned int last_quiescent_request_key, outstanding_io_requests;
extern unsigned char my_ethaddr[6];
extern char *last_interruption; /* debug */
unsigned long last_netdev_send_addr, last_netdev_send_size;

void
ssc_connect_irq(unsigned int type, unsigned char irq)
{
	printf("ssc_connect_irq %x = %x\n", type, irq);
	switch (type)
	{
		case KEYBOARD_INTR:	guest_serial_irq = irq; break;
		case NETWORK_INTR:	guest_network_irq = irq; break;
		case SIMSCSI_INTR:	guest_scsi_irq = irq; break;
	}
}

unsigned long
do_ssc(unsigned long arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3, unsigned long nr)
{
	unsigned int ret;
	switch (nr)
	{
		case SSC_CONSOLE_INIT:
			return 0;
		case SSC_OPEN:
		{
#ifdef GUESTIO_DISK
			char *name = (char *)arg0;
			mm_check_mpa(arg0, PROT_RX);

			assert(node_id == 0);
			if (!memcmp(name, "sda", 4))
			{
				printf("ssc_open sda\n");
				return 3;
			}
			printf("ssc_open fail %s\n", name);
#endif
			return -1;
		}
#ifdef GUESTIO_DISK
		case SSC_READ:
		case SSC_WRITE:
		{
			unsigned int request_key = arg0;
			unsigned int sge_count = arg1;
			struct scsi_sge *sglist = (struct scsi_sge *)arg2;
			unsigned long offset = arg3;
			int is_write = (nr == SSC_WRITE);
			struct request request;

			mm_check_mpa((unsigned long)(sglist), PROT_RX);
			mm_check_mpa((unsigned long)(sglist+(sge_count-1)), PROT_RX);

			if (node_id != 0)
			{
				printf("remote disk IO\n");
				if (sge_count == 1)
				{
					request.type = IPC_DISK_IO_SIMPLE;
					request.diskio.mpa = sglist[0].addr;
					request.diskio.len = sglist[0].len;
					request.diskio.is_write = is_write;
					request.diskio.key_valid = 1;
					request.diskio.offset = offset;
					request.diskio.key = request_key;
					ipc_send(0, &request);
				}
				else
				{
					request.type = IPC_DISK_IO_SGL;
					request.diskio.nsge = sge_count;
					request.diskio.is_write = is_write;
					request.diskio.key_valid = 1;
					request.diskio.offset = offset;
					request.diskio.key = request_key;
#warning Potentially dangerous use of DATA_AREA_STATIC
					ipc_send_page(0, &request, __pa(sglist), sge_count*sizeof(struct scsi_sge), DATA_AREA_STATIC);
				}
			}
			else
			{
				stat.count = mm_handle_disk_io(sglist, sge_count, offset, is_write, 1, request_key);
			}
			return 0;
		}
#endif
		case SSC_NULL:
			assert(node_id == 0);
			/* note: even with SCSI disabled, this is needed so that the guest can detect the lack of disks */
			last_quiescent_request_key = arg0;
			deliver_interrupt(guest_scsi_irq);
			return 0;
		case SSC_GET_COMPLETION:
			assert(node_id == 0);
			return last_quiescent_request_key;
#ifdef GUESTIO_DISK
		case SSC_WAIT_COMPLETION:
			assert(node_id == 0);
			printf("warning: legacy SSC_WAIT_COMPLETION called\n");
			if (outstanding_io_requests > 0)
			{
				rewind_ip();
				return -1;
			}

			mm_check_mpa(arg0, PROT_RWX);
			*(struct disk_stat *)arg0 = stat;
			return 0;
#endif
#ifdef GUESTIO_NET
		case SSC_NETDEV_PROBE:
			mm_copy_to_user(arg1, my_ethaddr, sizeof(my_ethaddr));
			return 4;
		case SSC_NETDEV_SEND:
			ipc_send_raw(arg1, arg2);
			return 0;
		case SSC_NETDEV_RECV:
			return dequeue_packet(arg1, arg2);
#endif
		case SSC_CONNECT_INTERRUPT:
		{
			struct request request;
			unsigned int type = arg0;
			unsigned char irq = arg1;

			ssc_connect_irq(type, irq);

			request.type = IPC_CONNECT_IRQ;
			request.connectirq.type = type;
			request.connectirq.irq = irq;
			ipc_send(63 /* !node_id */, &request); /* FIXME should broadcast */
			return 0;
		}
		case SSC_PUTCHAR:
			putchar(arg0);
			return 0;
		case SSC_GETCHAR:
			return getc();
		case SSC_SET_RENDEZVOUS:
			smp_rendezvous_ip = arg0;
			smp_rendezvous_gp = arg1;
			return 0;
		case SSC_NCPUS:
			ret = nodes;
			while (ret & (ret-1))
				ret++;
			printf("vNUMA: reporting %d nodes\n", ret);
			return ret;
		default:
			printf("unimplemented SSC %ld\n", nr);
			return -1;
	}
}
